 
UNINITIALIZED_VERSION = -1

class CFG

  class Node
    def initialize(no)
      @no = no
    end

    def to_s
      @no.to_s
    end

    def inspect
      @no.to_s
    end
  end

  attr_reader :nodes
  attr_accessor :start

  def initialize()
    @start = nil
    @nodes = []
    @succ = {}
    @pred = {}
    @vars = {}
  end

  def add_node(node)
    @nodes << node
    @start = node if @nodes.size == 1
    node
  end

  def add_edge(from, to)
    if @pred.has_key?(to)
      @pred[to] << from
    else
      @pred[to] = [from]
    end
    if @succ.has_key?(from)
      @succ[from] << to
    else
      @succ[from] = [to]
    end
  end

  def predecessors(node)
    unless @pred.has_key?(node)
      []
    else
      @pred[node].clone()
    end
  end

  def successors(node)
    if @succ.has_key?(node)
      @succ[node].clone()
    else
      []
    end
  end

  def postorder(node = @start, visited = {})
    if visited.has_key?(node)
      return []
    else
      res = []
      successors(node).each do |succ|
        res = res + postorder(succ, visited.merge!({node => true}))
      end
      res << node
    end
    res
  end

  def traverse_reverse_postorder(&block)
    postorder().reverse().each do |node|
      block.call(node)
    end
  end

  def preorder(node = @start, visited = {})
    if visited.has_key?(node)
      return []
    else
      res = [node]
      successors(node).each do |succ|
        res = res + preorder(succ, visited.merge!({node => true}))
      end
    end
    res
  end

  def traverse_preorder(&block)
    preorder().each do |node|
      block.call(node)
    end
  end

	def calculate_dominators()
	  # simple, iterative algorithm
	  dom = {}
	  nodes.each do |node|
	    dom[node] = nodes.clone()
	  end

	  changed = true
	  while changed do
	    changed = false
	    traverse_reverse_postorder do |node|
	      new_set = nodes.clone()
	      predecessors(node).each do |pred|
		new_set = (new_set & dom[pred]) | [node]
	      end
	      new_set = [node] if predecessors(node).empty?
	      if new_set != dom[node]
		dom[node] = new_set
		changed = true
	      end
	    end
	  end
	  dom

	  # also calculate immediate dominators
	  idom = {}
	  nodes.each do |node|
	    sdom = dom[node] - [node]
	    res = sdom
	    sdom.each do |sdnode|
	      res -= dom[sdnode] - [sdnode]
	    end
	    raise RuntimeError.new("immediate dominator for node #{node} is not unique.") if res.size > 1
	    idom[node] = res[0]
	  end

	  return dom, idom

	  # TODO: implement optimizations from
	  # Keith D. Cooper, Timothy J. Harvey, Key Kennedy - "A Simple, Fast Dominance Algorithm"
	end

	def calculate_dominance_frontiers()
	  dom, idom = calculate_dominators()

	  res = {}
	  nodes.each do |node|
	    res[node] = []
	  end

	  nodes.each do |node|
	    pred = predecessors(node)
	    next unless pred.size > 1
	    pred.each do |pnode|
	      runner = pnode
	      while runner != idom[node] do
		res[runner] << node
		runner = idom[runner]
	      end
	    end
	  end
	  res
	end

	def to_dot()
          dom, idom = calculate_dominators()
	  res = "digraph CFG {\n"
          @nodes.each do |node|
            mark = ""
	    res += "Node" + node.to_s + "[shape=record,label=\"#{node.contents()}\"#{mark}]\n"
          end
          @succ.each do |k,v|
            v.each do |target|
	      res += "Node" + k.to_s + " -> " + "Node" + target.to_s + "\n"
            end
	  end
          idom.each do |k,v|
            next if v == nil
            res += "Node" + k.to_s + " -> " + "Node" + v.to_s + "[color=red]\n"
          end
          res += "}\n"
	end

  def empty_state()
    vars = get_vars()
    res = {}
    vars.each do |var|
      res[var] = []
    end
    res
  end

  def start_state(params)
    vars = get_vars()
    res = {}
    vars.each do |var|
      if params.include?(var) then
        res[var] = [0]
      else
        res[var] = [UNINITIALIZED_VERSION]
      end
    end
    res
  end

  def join_states(a,b)
    vars = a.keys | b.keys
    res = {}
    vars.each do |var|
      aval = a[var] || []
      bval = b[var] || []
      res[var] = aval | bval
    end
    res
  end

  def analyse_values(params)
    # init
    res = {}
    nodes.each do |node|
      res[node] = empty_state()
    end
    res[@start] = start_state(params)

    # fixpoint iteration
    change = true
    while change do
      change = false
      traverse_preorder do |node|
        new = res[node]
        # propagate
        predecessors(node).each do |pred|
          new = join_states(new, res[pred])
        end
        # simulate basic block execution
        new = node.simulate_execution(new)
        # store
        change ||= new != res[node]
        res[node] = new
      end
    end
    res
  end

  def find_read_with_multiple_versions()
    nodes.each do |node|
      node.ops.each do |op|
        if op[0] == :read and op[2].size > 1 then
          var = op[1]
          versions = op[2]
          return node, var, versions
        end
      end
    end
    return nil
  end

  def insert_psi_assignments(params)
    #dom, idom = calculate_dominators()
    values = analyse_values(params)
    node, var, versions = find_read_with_multiple_versions()
    while node != nil do
      # a read with multiple possible versions is not allowed in SSA => we have to insert a psi function
      # possibility A: walk up the dominator tree to find the best location
      #runner = node
      #last = node
      #while values[runner][var] == versions
      #  last = runner
      #  runner = idom[runner]
      #end

      # possibility B: follow the versions to merge back in the predecessor relation
      # NOTE: does not require any dominance information!
      runner = node
      last = node
      visited = [last]
      while runner = predecessors(node).find { |pred| !visited.include?(pred) && values[pred][var] == versions }
        last = runner
        visited |= [last]
      end

      last.insert_psi_assignment(var, versions)
      values = analyse_values(params)
      node, var, versions = find_read_with_multiple_versions()
    end
  end

  def check_for_uninitialized_var_access(params, method)
    values = analyse_values(params)
    nodes.each do |node|
      node.ops.each do |op|
        if op[0] == :read and op[2].include?(UNINITIALIZED_VERSION) then
          typeerror("detected potential uninitialized access to local variable #{op[1].inspect} in method #{method}")
        end
      end
    end
  end

  def typeerror(msg)
    raise DynTypeError.new("TYPEERROR: #{msg}")
  end

  def substitute_variables()
    nodes.each do |node|
      node.substitute_variables()
    end
  end

  def get_asmt_no(var)
      if @vars.has_key?(var)
        @vars[var] += 1
      else
        @vars[var] = 1
      end
  end

  def add_var(var)
    if @vars.has_key?(var)
      # do nothing
    else
      @vars[var] = 0
    end
  end

  def get_vars()
      @vars.keys
  end

end

