#!/usr/bin/ruby

require_relative 'dyn_parsing'
require_relative 'utils'
require_relative 'dyn_types'

CONSTRUCTOR_NAME = "init"
NULLVALUE_LOCATION = 0
NULLCLASS_LOCATION = 1
NULLCLASS_NAME = "nullclass"
INITIAL_PROGRAM = [:method_call, [:method_call, [:constructor_call, "main", [], CodeLoc.generated()], "run", [], CodeLoc.generated()], "toString", [], CodeLoc.generated()]

class DynConfiguration

    include Working_with_types

    attr_reader :error_text, :prog
    attr_reader :class_by_name, :store

    def initialize(parse_tree)
      initial_configuration()
      parse_program(parse_tree)
      @update_counter = 0
      @utd = {}
    end

    def eval()
      eval_inner(@prog, @prog)
    end

    def result()
      if @prog[0] == :loc
        obj = @store[@prog[1]]
        result = @prog.inspect + ": " + obj.inspect
        if obj.clazz == @class_by_name["string"]
          result += "\n#{decode_string_at(@prog[1])}"
        end
        result
      else
        @prog.inspect
      end
    end

    def output_store()
      @next_location.times do |i|
        puts "#{i.to_s}\t=> #{@store[i].inspect}"
      end
    end

    def has_error?()
      @error
    end

    def pretty_print_code(code)
      pretty_print_code_internal(code)
    end

private

    def eval_inner(tree, subtree)
      return false if @error

      begin
      while !@error
        case subtree[0]
          when :variable
            name = subtree[1]
            unless @env.has_key?(name)
              typeerror("accessed unknown local variable #{name.inspect}")
            end
            substitute(tree, subtree, [:loc, @env[name]])
            return !@error
  
          when :instance_variable
            name = subtree[1]
            bindings = @store[@self].bindings
            unless bindings.has_key?(name)
              typeerror("accessed unknown instance variable #{name.inspect} on object @ loc(#{@self.inspect})")
            end
            substitute(tree, subtree, [:loc, bindings[name]])
            return !@error
  
          when :self
            substitute(tree, subtree, [:loc, @self])
            return !@error

          when :assignment
            name = subtree[1]
            exp = subtree[2]
            if eval_inner(tree, exp)
              # TODO: names that should not be assigned should have been filtered out in the parser
              @env[name] = exp[1]
	      @utd[name] = @update_counter += 1
              substitute(tree, subtree, exp)
            end
            return !@error
  
          when :instance_assignment
            name = subtree[1]
            exp = subtree[2]
            if eval_inner(tree, exp)
              bindings = @store[@self].bindings
              bindings[name] = exp[1]
              substitute(tree, subtree, exp)
            end
            return !@error
  
          when :method_call
            exp = subtree[1]
            name = subtree[2]
            typeerror("constructors cannot be called after instantiation") if name == CONSTRUCTOR_NAME
            params = subtree[3]
	    loc = subtree[4]
            if eval_inner(tree, exp)
              if eval_inner_params(tree, params)
                substitute(tree, subtree, [:call, exp[1], name, params, loc])
              end
            end
  
          when :constructor_call
            classname = subtree[1]
            typeerror("unknown class #{classname.inspect}") unless @class_by_name.has_key?(classname)
            params = subtree[2]
	    loc = subtree[3]
            if eval_inner_params(tree, params)
              l = fresh_location()
              @store[l] = DynObject.new(@class_by_name[classname])
              substitute(tree, subtree, [:call, l, CONSTRUCTOR_NAME, params, loc])
            end
  
          when :call
            loc = subtree[1]
            name = subtree[2]
            values = subtree[3]

            obj = @store[loc]
            methods = @store[obj.clazz()].dynmethods()
            unless methods.has_key?(name)
              if loc == NULLVALUE_LOCATION
                error("unknown method #{name.inspect} called on null value")
                return !@error
              else
                typeerror("unknown method #{name.inspect} called on object @ loc(#{loc.inspect}) [of class #{find_class_name(obj.clazz())}]")
              end
            end
            method = methods[name]
            typeerror("wrong number of parameters passed to method #{find_class_name(obj.clazz())}.#{name}: expected #{method.params.size.inspect}, passed #{values.size.inspect}") unless method.params.size == values.size

            # NOTE: name and obj.clazz() are stored just for documentation purposes. This way it is much easier to understand a callstack.
            @stack.push([@env, @self, name, obj.clazz()])
            @self = loc
            @env = {}
            
            new_tree = [:statement] + method.params.zip(values).map{ |param, value| [:assignment, param, value] } + [method.code(), CodeLoc.generated()]
            if eval_inner(new_tree, new_tree)
              @env, @self, _, _ = @stack.pop()
              substitute(tree, subtree, new_tree)
            end
            return !@error
  
          when :loc
            return !@error
  
          when :if
            cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]

            if eval_inner(tree, cond)
              new_tree = [:method_call, cond, "inner", []]
              if eval_inner(new_tree, new_tree)
                syserror("eval did not return a final state") unless new_tree[0] == :loc
                if new_tree[1] == 0
                  substitute(tree, subtree, false_branch)
                else
                  substitute(tree, subtree, true_branch)
                end
              else
                return !@error
              end
            else
              return !@error
            end
            
          when :ifPS
            cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]

            if eval_inner(tree, cond)
              new_tree = [:method_call, cond, "inner", []]
              if eval_inner(new_tree, new_tree)
                syserror("eval did not return a final state") unless new_tree[0] == :loc
                if new_tree[1] == 0
                  substitute(tree, subtree, false_branch)
                else
                  substitute(tree, subtree, true_branch)
                end
              else
                return !@error
              end
            else
              return !@error
            end

          when :while
            cond = subtree[1]
            loop = subtree[2]
	    loc = subtree[3]
            substitute(tree, subtree, [:if, cond.dclone(), [:statement, loop.dclone(), [:while, cond, loop, loc], loc], NULL_CONST, loc])

          when :statement
            first = subtree[1]
            if eval_inner(tree, first)
              if subtree.size > 3
                subtree.delete_at(1)
              else
                substitute(tree, subtree, first)
              end
            else  
              return !@error
            end

          when :error
            error(subtree[1])
            return !@error

          # --- builtin functionality ---

          when :hook_loc_identical
            a = subtree[1]
            b = subtree[2]
            if eval_inner(tree, a)
              if eval_inner(tree, b)
                if a[1] == b[1]
                  substitute(tree, subtree, TRUE_CONST)
                else
                  substitute(tree, subtree, FALSE_CONST)
                end
              end
            end

          when :hook_get_loc
            exp = subtree[1]
            loc = subtree[2]
            if eval_inner(tree, exp)
              substitute(tree, subtree, generate_numeric(exp[1], loc))
            end

          when :hook_is_a
            exp = subtree[1]
            #TODO: clazzname must be a string
            clazzname = subtree[2]
            if eval_inner(tree, exp)
              object = @store[exp[1]]
              typeerror("unknown class #{clazzname.inspect}") unless @class_by_name.has_key?(clazzname)
              clazz = @class_by_name[clazzname]
              if object.clazz == clazz || object.clazz == NULLCLASS_LOCATION
                result = TRUE_CONST
              else
                result = FALSE_CONST
              end
              substitute(tree, subtree, result)
            end

          when :hook_psi
            exps = subtree[1..subtree.size()]
	    i = 0
	    max = 0
	    selected = nil
	    while i < exps.size()
	      exp = exps[i]
	      raise RuntimeError.new("expected only variables inside a psi function") unless exp[0] == :variable
	      var = exp[1]
	      if @utd.has_key?(var) and @utd[var] > max
		selected = i
		max = @utd[var]
	      end
	      i += 1
	    end
	    substitute(tree, subtree, exps[selected])

          when :hook_cast
	    type = subtree[1]
            exp = subtree[2]
	    unless res = eval_inner(subtree, exp) then
	      return res
	    end
	    raise RuntimeError.new("expression did not evaluate to a location") unless exp[0] == :loc
	    raise RuntimeError.new("expected uniontype to cast to") unless type[0] == :uniontype
	    classes = type[1].map do |classname|
	      val = @class_by_name[classname]
	      raise RuntimeError.new("unknown class #{classname}") unless val
	      val
	    end
	    clazz = @store[exp[1]].clazz
	    typeerror("cannot cast #{find_class_name(clazz)} to #{type.inspect}") unless classes.include?(clazz)
	    substitute(tree, subtree, exp)
            return !@error

          when :assertion
	    # TODO: implement (optional) runtime assertion checking?
            return !@error

          else
            syserror("found unknown node in parsetree: #{subtree.inspect}")
        end
      end

      rescue DynTypeError => e
        # do nothing

      rescue SystemStackError => e
        puts "ERROR: stack level too deep:"
        puts e.backtrace.join("\n")
        puts "STACK:"
        puts @stack.map{|x| x.inspect}.join("\n")
        p @class_by_name
        exit(-1)
      end

      return !@error
    end

    def find_class_name(loc)
      @class_by_name.each do |key, value|
        return key if value == loc
      end
      return nil
    end

    def eval_inner_params(tree, params)
      params.each do |param|
        return !@error unless eval_inner(tree, param)
      end
    end

    def substitute(tree, subtree, replacement)
      subtree.size.times do
        subtree.pop()
      end
      replacement.each do |item|
        subtree << item
      end
    end

    def decode_numeric_at(loc)
      n = -1
      while loc > 0
        prog = [:method_call, [:loc, loc], "pred", []]
        unless eval_inner(prog, prog)
          syserror("could not decode numeric at loc #{loc.inspect}")
        end
        n += 1
        loc = prog[1]
      end
      n
    end

    def decode_char_at(loc)
        decode_numeric_at(loc).chr
    end

    def decode_char_array_at(loc)
      res = ""
      while loc > 0
        prog = [:method_call, [:loc, loc], "head", []]
        unless eval_inner(prog, prog)
          syserror("could not decode char array entry at loc #{loc.inspect}")
        end
        res += decode_char_at(prog[1])

        prog = [:method_call, [:loc, loc], "tail", []]
        unless eval_inner(prog, prog)
          syserror("could not decode char array at loc #{loc.inspect}")
        end
        loc = prog[1]
      end
      res
    end
 
    def decode_string_at(loc)
        prog = [:method_call, [:loc, loc], "chars", []]
        unless eval_inner(prog, prog)
          syserror("could not decode string at loc #{loc.inspect}")
        end
        decode_char_array_at(prog[1])
    end

    def initial_configuration()
      @class_by_name = {}
      @stack = []
      @self = nil
      @store = { 0 => DynObject.new(NULLCLASS_LOCATION) }
      @env = {}
      @prog = INITIAL_PROGRAM
      @error = false
      @error_text = nil
      @next_location = 1
    end

    def final?()
      @prog.size == 2 && @prog[0] == :loc
    end

    def fresh_location()
      old = @next_location
      @next_location += 1
      old
    end

    def parse_program(parse_tree)
      syserror("not a program") unless parse_tree[0] == :program

      classmap = {}
      parse_tree[1..parse_tree.size()-2].each do |classdef|
        initialize_classdef(classdef)
        classmap[classdef[1]] = classdef
      end
      parse_classmap(classmap)
    end

    def initialize_classdef(classdef)
      syserror("not a classdef") unless classdef[0] == :classdef

      name = classdef[1]
      codeloc = classdef[classdef.size()-1]

      # initialize class
      l = fresh_location()
      @store[l] = DynClass.new(codeloc)
      error("duplicate class #{name}") if @class_by_name.has_key?(name)
      @class_by_name[name] = l
    end

    def parse_classmap(classmap)
      while !classmap.empty?
        parse_classdef_step(classmap)
      end
    end

    def parse_classdef_step(classmap)
      # draw a classdef
      name, classdef = classmap.first

      parse_classdef(classdef, classmap, [])
    end

    def parse_classdef(classdef, classmap, in_progress)
      syserror("not a classdef") unless classdef[0] == :classdef

      name = classdef[1]
      in_progress << name
      if in_progress.uniq.size < in_progress.size
        raise BootstrapError.new("circular dependencies: #{in_progress.join(',')}")
      end

      # add inherited methods
      parent = classdef[2]
      if classmap.has_key?(parent) && name != "object"
        # parent is not finished -> parse it first
        parse_classdef(classmap[parent], classmap, in_progress)
      end
      clazzloc = @class_by_name[name]
      unless clazzloc
        raise BootstrapError.new("unknown class #{name}")
      end
      clazz = @store[clazzloc]
      parentloc = @class_by_name[parent]
      unless parentloc
        raise BootstrapError.new("unknown class #{parent}")
      end
      clazz.inherit(@store[parentloc])

      # add declared class components
      classdef[3].each do |component|
        case component[0]
          when :method
            clazz.add_method(parse_method(component))
          when :rename
            clazz.rename(component[1], component[2])
          else
            syserror("found invalid member in classdef: #{component[0].inspect}")
        end
      end

      # remove finished classdef from classmap
      classmap.delete(name)
    end

    def parse_method(methoddef)
      l = fresh_location()
      @store[l] = DynMethod.new(methoddef)
    end

    def syserror(text)
      raise RuntimeError.new("SYSERROR: " + text)
    end

    def error(text)
      @error = true
      @error_text = "ERROR: " + text
    end

    def typeerror(text)
      @error = true
      @error_text = "TYPEERROR: " + text
      raise DynTypeError.new(@error_text)
    end

    def pretty_print_code_internal(subtree)
        case subtree[0]
          when :variable
            name = subtree[1]
            return name
  
          when :instance_variable
            name = subtree[1]
            return "@#{name}"
  
          when :self
            return SELF_NAME

          when :assignment
            name = subtree[1]
            exp = subtree[2]
            return "#{name} = #{pretty_print_code_internal(exp)}"
  
          when :instance_assignment
            name = subtree[1]
            exp = subtree[2]
            return "@#{name} = #{pretty_print_code_internal(exp)}"
  
          when :method_call
            exp = subtree[1]
            name = subtree[2]
            params = subtree[3]
            return "#{pretty_print_code_internal(exp)}.#{name}(#{params.map{|p| pretty_print_code_internal(p)}.join(', ')})"
  
          when :constructor_call
            classname = subtree[1]
            params = subtree[2]
            return "new #{classname}(#{params.map{|p| pretty_print_code_internal(p)}.join(', ')})"
  
          when :call
            raise RuntimeError.new("cannot pretty-print internal datastructures")
  
          when :loc
            return "null" if subtree[1] == 0
            raise RuntimeError.new("cannot pretty-print internal datastructures")
  
          when :if
            cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]
            return "if #{pretty_print_code_internal(cond)} then #{pretty_print_code_internal(true_branch)} else #{pretty_print_code_internal(false_branch)} end"
            
          when :while
            cond = subtree[1]
            loop = subtree[2]
            return "while #{pretty_print_code_internal(cond)} do #{pretty_print_code_internal(loop)} done"

          when :statement
            return subtree[1..subtree.size].map { |e| pretty_print_code_internal(e) }.join("; ")

          when :error
            return "error(#{pretty_print_code_internal(subtree[1])})"

          # --- builtin functionality ---

          when :hook_loc_identical
            a = subtree[1]
            b = subtree[2]
            return "hook_loc_identical(#{pretty_print_code_internal(a)}, #{pretty_print_code_internal(b)})"

          when :hook_get_loc
            exp = subtree[1]
            return "hook_get_loc(#{pretty_print_code_internal(exp)})"

          when :hook_is_a
            exp = subtree[1]
            #TODO: clazzname must be a string
            clazzname = subtree[2]
            
            #TODO: Not sure if this should be hook_is_a or hook-is-a . Find out!
            return "hook_is_a(#{pretty_print_code_internal(exp)}, #{pretty_print_code_internal(clazzname)})"

          else
            syserror("found unknown node in parsetree: #{subtree.inspect}")
        end
    end

end

class DynClass

  attr_reader :dynmethods, :codeloc
  # for type inference
  attr_accessor :type

  def initialize(codeloc)
    @dynmethods = {}
    @type = nil
    @codeloc = codeloc
  end

  def add_method(method)
    @dynmethods[method.name] = method
  end

  def rename(old_name, new_name)
    error("cannot rename unknown method #{old_name}") unless @dynmethods.has_key?(old_name)
    @dynmethods[new_name] = @dynmethods[old_name]
    @dynmethods.delete(old_name)
  end

  def inherit(parent)
    @dynmethods = {}
    parent.dynmethods.each do |key, value|
      @dynmethods[key.clone()] = value.clone()
    end
  end

  def finished_type_inference()
    @type != nil
  end

end

class DynMethod

  attr_reader :name, :params, :codeloc

  def initialize(methoddef)
    syserror("not a method definition") unless methoddef[0] == :method
    @name = methoddef[1]
    @params = methoddef[2]
    @code = methoddef[3]
    @codeloc = methoddef[4]
  end

  def code()
    @code.dclone()
  end

end

class DynObject

  attr_reader :clazz, :bindings

  def initialize(clazz)
    @clazz = clazz
    @bindings = {}
  end

end

class DynError < RuntimeError
end

class DynTypeError < DynError
end

class BootstrapError < RuntimeError
end

