#!/usr/bin/ruby

module MonomorphicTypeInference

	BOOL_TYPE = [:object, [:method, "inner", 0]]

	def monomorphic_type_inference_program(options = {})
		constraint_system = generate_constraints(@configuration)
		if options[:output_constraints]
			constraint_system.print()
		end
		begin
			resolver = MethodResolver.new(@configuration)
			solution = solve_constraints(constraint_system, resolver, options)
		rescue ConstraintSatisfactionError => e
			raise e.to_DynTypeError(@configuration)
		end
	end

	def has_error?()
		false
	end

	def error_text()
		""
	end

private

	def generate_constraints(configuration)
		#TODO: apply container class transformation + method transformation to allow polymorphism
		#TODO: apply dead code removal to speed things up

		cs = ConstraintSystem.new()
		@configuration.class_by_name().each do |cname, loc|
			cs.add_value(cname)
			clazz = @configuration.store[loc]
			clazz.dynmethods.each do |mname, method|
				method.param_types = method.params.map { |p| cs.create_variable(p, method.codeloc) }
				method.return_type = cs.create_variable("result-#{cname}.#{mname}", method.codeloc)
			end
		end
		
		@configuration.class_by_name().each do |cname, loc|
			clazz = @configuration.store[loc]
			generate_constraints_for_class(cs, cname, clazz)
		end
		generate_constraints_for_initial_program(cs)
		
		return cs
	end

	def generate_constraints_for_initial_program(cs)
                method = DynMethod.new([:method, "root", [], INITIAL_PROGRAM])
		method.param_types = []
		method.return_type = cs.create_variable("result", method.codeloc)
		class_context = ["ROOT", nil, {}, nil]
		generate_constraints_for_method(cs, method, class_context)
		cs.add_savety_constraint(method.return_type, [cs.get_value("string")])
	end

	def generate_constraints_for_class(cs, cname, clazz)
		# set up a self type for this class
		self_type = cs.create_variable("self-#{cname}", clazz.codeloc)
		val = value_for_class(cs, cname)
		cs.add_constraint([val], self_type)    # {C} \subseteq [self-C]
		cs.add_savety_constraint(self_type, [val])    # [self-C] \subseteq {C}

		instance_env = {}
		constructor_env = nil
		self_access_callback = Proc.new() do
			# capture the constructor's instance environment on the first access to self
			constructor_env = constructor_env || instance_env.clone()
		end
		constructor = clazz.dynmethods()["init"]
		if constructor
			generate_constraints_for_method(cs, constructor, [cname, self_type, instance_env, self_access_callback])
		end
		constructor_env = constructor_env || {}

		clazz.dynmethods().each do |mname, method|
			next if mname == "init"
			generate_constraints_for_method(cs, method, [cname, self_type, instance_env, nil])
		end

		rest_env = instance_env.keys - constructor_env.keys
		unless rest_env.size == 0
			typeerror_fatal("class #{cname}: instance variables #{rest_env.join(", ")} are not initialized in constructor prior to the first access to self.")
		end
	end

	def generate_constraints_for_method(cs, method, class_context)
		code = method.code()

		cname = class_context[0]
		self_type = class_context[1]
		self_access_callback = class_context[3]
		instance_env = class_context[2]
		start_state = AbstractState.new(cs, @configuration, self_type, self_access_callback, cname, instance_env, method)
		local_env = {}
		method.params.zip(method.param_types).each do |param, param_type|
			start_state.add_parameter([param,0], param_type)
		end

		end_states = abstract_interpretation_for_expression(start_state, code, method, class_context)

		end_states.each do |end_state|
		  cs.add_constraint(end_state.result, method.return_type)
                end
	end

	def new_scope()
		if @last_scope
			@last_scope += 1
		else
			@last_scope = 0
		end
	end

# abstract interpretation
	def abstract_interpretation_for_expression(state, tree, method_context, class_context)
		# TODO: most of this should be superfluous as the state can handle this
		# TODO: remove method_context parameter
		method = method_context
		cname, self_type, instance_env, self_access_callback = class_context
		case tree[0]
		  when :variable
		    name = tree[1]
		    return state.access_local_var(name, CodeLocation.from(tree))
	  
		  when :instance_variable
		    name = tree[1]
		    return state.access_instance_var(name, CodeLocation.from(tree))
	  
		  when :self
		    return state.self_access(CodeLocation.from(tree))

		  when :assignment
		    name = tree[1]
		    exp = tree[2]
		    codeloc = tree[3]
		    exp_states = abstract_interpretation_for_expression(state, exp, method_context, class_context)
		    states = exp_states.map do |exp_state|
		      exp_state.local_assignment(name, exp_state.result, codeloc)
		    end.inject(:+)
		    return states
	  
		  when :instance_assignment
		    name = tree[1]
		    exp = tree[2]
		    codeloc = tree[3]
		    exp_states = abstract_interpretation_for_expression(state, exp, method_context, class_context)
		    states = exp_states.map do |exp_state|
		      exp_state.instance_assignment(name, exp_state.result, codeloc)
		    end.inject(:+)
		    return states
	  
		  when :method_call
		    recv = tree[1]
		    name = tree[2]
		    params = tree[3]
		    codeloc = tree[4]
		    typeerror_fatal("constructors cannot be called after instantiation") if name == CONSTRUCTOR_NAME

		    recv_states = abstract_interpretation_for_expression(state, recv, method_context, class_context)
		    raise RuntimeError.new("recv is not allowed to branch") unless recv_states.size == 1
		    param_states, params_types = abstract_interpretation_for_params(recv_states[0], params, method_context, class_context)
		    raise RuntimeError.new("params are not allowed to branch") unless param_states.size == 1
		    return param_states[0].method_call(recv_states[0].result, name, params_types, tree, codeloc)
	  
		  when :constructor_call
		    classname = tree[1]
		    params = tree[2]
		    codeloc = tree[3]
		    param_states, params_types = abstract_interpretation_for_params(state, params, method_context, class_context)
		    return param_states[0].constructor_call(classname, params_types, tree, codeloc)

		  when :loc
		    loc = tree[1]
		    codeloc = tree[tree.size()-1]
		    return state.constant_loc(loc, codeloc)
	  
		  when :if
		    cond = tree[1]
		    true_branch = tree[2]
		    false_branch = tree[3]

		    cond_states = abstract_interpretation_for_expression(state, cond, method_context, class_context)
		    raise RuntimeError.new("condition is not allowed to branch") unless cond_states.size == 1
		    true_state, false_state = cond_states[0].conditional()
		    true_after_states = abstract_interpretation_for_expression(true_state, true_branch, method_context, class_context)
		    false_after_states = abstract_interpretation_for_expression(false_state, false_branch, method_context, class_context)
		    raise RuntimeError.new("condition is not allowed to branch") unless cond_states.size == 1
                    # DEBUG
                    if true_after_states == nil then
                      p true_state
                      p true_branch
                    end
		    return [AbstractState.merge(true_after_states + false_after_states)]

		  when :ifPS
		    cond = tree[1]
		    true_branch = tree[2]
		    false_branch = tree[3]

		    cond_states = abstract_interpretation_for_expression(state, cond, method_context, class_context)
		    raise RuntimeError.new("condition is not allowed to branch") unless cond_states.size == 1

		    true_state, false_state = cond_states[0].conditional()
		    true_after_states = abstract_interpretation_for_expression(true_state, true_branch, method_context, class_context)
		    false_after_states = abstract_interpretation_for_expression(false_state, false_branch, method_context, class_context)

		    return true_after_states + false_after_states

		  when :while
		    cond = tree[1]
		    loop = tree[2]

		    cond_states = abstract_interpretation_for_expression(state, cond, method_context, class_context)
		    cond_states.each do |cond_state|
		      cond_state.result_is_condition()
		      loop_states = abstract_interpretation_for_expression(cond_state, loop, method_context, class_context)
		      loop_states.each { |loop_state| loop_state.merge_into(cond_state) }
		    end
		    return cond_states

		  when :statement
		    first = tree[1]
		    result_type = nil
		    states = [state]
		    tree[1..tree.size()-2].each do |item|
			    states = states.map do |curr_state|
			      abstract_interpretation_for_expression(curr_state, item, method_context, class_context)
			    end.inject(:+) || []
		    end
		    return states

		  when :error
		    return []

		  # --- builtin functionality ---

		  when :hook_loc_identical
		    a = tree[1]
		    b = tree[2]
		    codeloc = tree[3]
		    next_states = abstract_interpretation_for_expression(state, a, method_context, class_context)
		    raise RuntimeError.new("arguments are not allowed to branch") unless next_states.size == 1
		    next_states = abstract_interpretation_for_expression(next_states[0], b, method_context, class_context)
		    raise RuntimeError.new("arguments are not allowed to branch") unless next_states.size == 1
		    return next_states[0].hook_loc_identical(tree, codeloc)

		  when :hook_get_loc
		    exp = tree[1]
		    codeloc = tree[2]
		    next_states = abstract_interpretation_for_expression(state, exp, method_context, class_context)
		    raise RuntimeError.new("arguments are not allowed to branch") unless next_states.size == 1
		    return next_states[0].hook_get_loc(tree, codeloc)

		  when :hook_is_a
		    exp = tree[1]
		    #TODO: clazzname must be a string
		    clazzname = tree[2]
		    codeloc = tree[3]
		    typeerror_fatal("unknown class #{clazzname.inspect}") unless @configuration.class_by_name().has_key?(clazzname)
		    next_states = abstract_interpretation_for_expression(state, exp, method_context, class_context)
		    raise RuntimeError.new("arguments are not allowed to branch") unless next_states.size == 1
		    return next_states[0].hook_is_a(clazzname, tree, codeloc)

# assertions
		  when :hook_psi
		    # TODO: a psi function should not be neccessary when the analysis is already flow-sensitive
		    arg = []
		    next_state = state
		    codeloc = tree[tree.size()-1]
		    tree[1..tree.size()-2].each do |exp|
		      raise RuntimError.new("expected variable inside psi function") unless exp[0] == :variable
		      name = exp[1]
		      unless !next_state.local_env.has_key?(name)
		        next_states = abstract_interpretation_for_expression(next_state, exp, method_context, class_context)
			raise RuntimeError.new("psi arguments should not branch. This is not supposed to happen.") if next_states.size != 1
			next_state = next_states[0]
		        arg << next_state.result
		      end
		    end
		    return next_state.hook_psi(arg, tree, codeloc)

		  when :hook_cast
		    type = tree[1]
		    exp = tree[2]
		    codeloc = tree[3]
		    raise RuntimeError.new("expected union type in cast") unless type[0] == :uniontype

	 	    next_states = abstract_interpretation_for_expression(state, exp, method_context, class_context)
		    raise RuntimeError.new("cast expressions should not branch.") if next_states.size != 1
		    return next_states[0].hook_cast(type, tree, codeloc)

		  when :assertion
		    # exp = tree[1]
		    return [state]

		  else
		    syserror("found unknown node in parsetree: #{tree.inspect}")
		end
	end

        def abstract_interpretation_for_params(state, params, method_context, class_context)
		next_state = state
		types = []
		params.each do |param|
		  next_states = abstract_interpretation_for_expression(next_state, param, method_context, class_context)
		  raise RuntimeError.new("params are not allowed to branch") unless next_states.size == 1
		  next_state = next_states[0]
		  types << next_state.result
		end

		return [next_state], types
	end

	def each_class_with_method(name, &block)
		@configuration.class_by_name().each do |cname, loc|
			clazz = @configuration.store[loc]
			next unless clazz.dynmethods.has_key?(name)
			block.call(cname, clazz, clazz.dynmethods[name])
		end
	end

	def solve_constraints(constaints, resolver, options = {})
		solver = ConstraintSolver.new()
		return solver.solve(constaints, resolver, options)
	end


	def value_for_class(cs, cname)
		unless cs.has_value?(cname)
			syserror("unknown class #{cname}")
		end
		cs.get_value(cname)
	end

	def syserror(text)
		raise RuntimeError.new("SYSERROR:" + text)
	end

	def typeerror_fatal(text)
		raise DynTypeError.new("TYPEERROR: " + text)
	end
end

class AbstractState

	attr_accessor :result, :local_env, :instance_env
	attr_reader :cs, :self_type, :self_access_callback, :method, :cname

	def initialize(cs, configuration, self_type, self_access_callback, cname, instance_env, method)
		@cs = cs
		@configuration = configuration
		@local_env = {}
# TODO: BIG PROBLEM: instance_env needs to be the same across all methods of a class!
		@instance_env = instance_env
		@result = nil
		@self_type = self_type
		@self_access_callback = self_access_callback
		@method = method
		@cname = cname
	end

	def self.merge(state_list)
		if state_list.size > 1
			state_list[0].merge(AbstractState.merge(state_list[1..state_list.size()]))
		else
			state_list[0]
		end
	end

	# produces a new state that has both self and other merged into
	def merge(other)
		new = AbstractState.new(@cs, @configuration, @self_type, @self_access_callback, @cname, @instance_env, @method)
		self.merge_into(new)
		other.merge_into(new)
		return new
	end

	# merges the current state into the given one
	def merge_into(state)
		raise RuntimeError.new("cannot merge states with different constraint systems.") unless state.cs == @cs

		# merge @self_type		
		raise RuntimeError.new("cannot merge states with different self types") unless @self_type == state.self_type

		# merge @self_access_callback
		raise RuntimeError.new("cannot merge states with different self access callbacks") unless @self_access_callback == state.self_access_callback

		# merge @method
		raise RuntimeError.new("cannot merge states with different methods") unless @method == state.method

		# merge @local_env
		@local_env.each do |key, value|
			if state.local_env.has_key?(key)
				@cs.add_constraint(@local_env[key][:var], state.local_env[key][:var])
			else
				state.local_env[key] = { :var => @local_env[key][:var] }
			end
		end
		# merge @instance_env
		raise RuntimeError.new("cannot merge states with different instance environments") unless state.instance_env == @instance_env
=begin
		@instance_env.each do |key, value|
			if state.instance_env.has_key?(key)
				@cs.add_constraint(@instance_env[key], state.instance_env[key])
			else
				state.instance_env[key] = @instance_env[key]
			end
		end
=end
		# merge @result
		if result != nil
		  if state.result == nil then
		    state.result = result
		  else
		    own_type = @cs.create_variable("merge_result", CodeLoc.generated())
		    @cs.add_constraint(state.result, own_type)
		    @cs.add_constraint(result, own_type)
		    state.result = own_type
		  end
		end

	end

	def clone()
		res = AbstractState.new(@cs, @configuration, @self_type, @self_access_callback, @cname, @instance_env, @method)
		res.result = @result
		res.local_env = @local_env.clone()
		return res
	end

	def split()
		a = clone()
		b = clone()
		return [a,b]
	end

# setup

	def add_parameter(name, typevar)
		@local_env[name] = {:var => typevar}
	end

# mutators

	def access_local_var(name, codeloc)
	  unless @local_env.has_key?(name)
	    typeerror_fatal("accessed unknown local variable #{name.inspect} in #{codeloc.to_s(self)}")
	  end
	  res = clone()
	  res.result = @local_env[name][:var]
	  return [res]
	end

	def access_instance_var(name, codeloc)
	  unless @instance_env.has_key?(name)
	    @instance_env[name] = @cs.create_variable("#{@cname}.@#{name}", codeloc)
	  end
	  res = clone()
	  res.result = @instance_env[name]
	  return [res]
	end

	def self_access(codeloc)
	  if @self_access_callback
	    @self_access_callback.call()
	  end
	  res = clone()
	  res.result =  @self_type
	  return [res]
	end

	def local_assignment(name, valtype, codeloc)
	  unless @local_env.has_key?(name)
	    @local_env[name] = {:var => @cs.create_variable(name, codeloc)}
	  end
	  var = @local_env[name][:var]

	  @cs.add_constraint(valtype, var)   # [exp] \subseteq [var]
	  res = clone()
	  res.result = valtype               # the result of an assignment it the result of its expression
	  return [res]
	end

	def instance_assignment(name, valtype, codeloc)
	  unless @instance_env.has_key?(name)
	    @instance_env[name] = @cs.create_variable(name, codeloc)
	  end
	  var = @instance_env[name]

	  @cs.add_constraint(valtype, var)   # [exp] \subseteq [var]
	  res = clone()
	  res.result = valtype               # the result of an assignment it the result of its expression
	  return [res]
	end

	def constant_loc(loc, codeloc)
	  unless loc == 0
	    # no idea how this should happen -> if it does, think about a way to type this
	    syserror("this should never happen")
	  end
	  # null has type {}
	  res = clone()
	  res.result = @cs.create_variable("null", codeloc)
	  @cs.add_savety_constraint(res.result, [])
	  return [res]
	end

	def conditional()
	  result_is_condition()
	  return split()
	end

	def result_is_condition()
	  #TODO: what sense does this constraint make?
	  #cs.add_constraint(["boolean"], cond_type)
	  @cs.add_savety_constraint(result, MonomorphicTypeInference::BOOL_TYPE)
	end

	def hook_loc_identical(tree, codeloc)
	    res = clone()
	    res.result = @cs.create_variable(tree, codeloc)
	    @cs.add_constraint(["boolean"], res.result)
	    return [res]
	end

	def hook_get_loc(tree, codeloc)
	    res = clone()
	    res.result = @cs.create_variable(tree, codeloc)
	    @cs.add_constraint(["numeric"], res.result)
	    return [res]
	end

	def hook_is_a(clazzname, tree, codeloc)
	    res = clone()
	    res.result = @cs.create_variable(tree, codeloc)
	    @cs.add_constraint(["boolean"], res.result)
	    return [res]
	end

	def hook_psi(arg, tree, codeloc)
	    res = clone()
	    res.result = @cs.create_variable(tree, codeloc)
	    arg.each do |typevar|
	      @cs.add_constraint(typevar, res.result)
	    end
	    return [res]
	end

	def method_call(recv, name, params_types, tree, codeloc)
	    res = clone()
	    res.result = @cs.create_variable(tree, codeloc)
	    generate_constraints_for_method_call([recv, name, params_types, codeloc], res.result)
	    return [res]
	end

	def constructor_call(classname, params_types, tree, codeloc)
	    value = @cs.get_value(classname)
	    typeerror_fatal("unknown class #{classname.inspect}") unless value
	    res = clone()
	    res.result = @cs.create_variable(tree, codeloc)
	    generate_constraints_for_method_call([[value], CONSTRUCTOR_NAME, params_types, codeloc], res.result)
	    return [res]
	end

	def hook_cast(type, tree, codeloc)
		raise RuntimeError.new("expected uniontype to cast to") unless type[0] == :uniontype
		type = type[1].map do |classname|
			val = @cs.get_value(classname)
			typeerror_fatal("unknown class #{classname.inspect}") unless val
			val
		end
		res = clone()
		old = res.result
		res.result = @cs.create_variable(tree, codeloc)
		@cs.add_filter_constraint(type, old, res.result)
		return [res]
	end

	def typeerror_fatal(text)
		raise DynTypeError.new("TYPEERROR: " + text)
	end

# private

	def generate_constraints_for_method_call(call, own_type)
		type, name, param_types, codeloc = call

		if @cs.is_variable?(type)
			# [exp] \subseteq {all legal receivers}
			@cs.add_savety_constraint(type, [:object, [:method, name, param_types.size]])
                        @cs.add_method_call_obligation(type, name, param_types, own_type, codeloc)
		else
			unless @cs.is_value_set?(type)
				syserror("illegal argument: receiver must be either a variable or a value set")
			end
			@cs.each_value(type) do |value|
				loc = @configuration.class_by_name[value]
				clazz = @configuration.store[loc]
				if clazz.dynmethods.has_key?(name)
					method = clazz.dynmethods[name]
                                        if method.param_types.size == param_types.size then
  					  method.param_types.zip(param_types).each do |formal, actual|
						@cs.add_constraint(actual, formal)
					  end
					  @cs.add_constraint(method.return_type, own_type)
                                        else
					  # TODO: add code location
					  @cs.typeerror("method #{name} in class #{value} does not match the arity of the call (expected: #{method.param_types.size}, actual: #{param_types.size})")

                                        end
				else
					# TODO: add code location
					@cs.typeerror("class #{value} does not have a method #{name}")
				end
			end
		end

		
	end


end

class CodeLocation

	def initialize(cname)
		@cname = cname
	end

	def self.from(tree)
		# TODO: some magic here to retrieve file:line:from_column-to_column-style information
		CodeLocation.new("unknown")
	end

	def to_s(state)
	  "method #{state.cname}.#{state.method.name}"
	end

end

class ConstraintSystem

	attr_reader :values

	def initialize()
		@values = {}
		@var_desc = {}
		@var_codeloc = {}
		@constraints = []
		@safety_constraints = []
		@next_id = 0
		@next_path = 0
		@typeerrors = []
	end

	def add_value(name)
		if @values.has_key?(name)
			raise RuntimeError.new("duplicate value: #{name.inspect}")
		end
		@values[name] = name
	end

	def has_value?(name)
		@values.has_key?(name)
	end

	def get_value(name)
		@values[name]
	end

	def create_path()
		path = @next_path
		@next_path += 1
		return path
	end

	def create_variable(desc, codeloc)
		var_id = @next_id
		@next_id += 1
		@var_desc[var_id] = desc
		@var_codeloc[var_id] = codeloc
		return var_id
	end

	# means "a \subseteq b" or
	#       "{x,y,z} \subseteq b"
	def add_constraint(a, b)
		assert_variable(b, "b")
		raise RuntimeError.new("expected non-nil value") if a == nil
		@constraints << [:subseteq, a, b]
	end

	def add_filter_constraint(type, a, b)
		assert_variable(a, "a")
		assert_variable(b, "b")
		@constraints << [:filter, type, a, b]
	end

	# means "a \subseteq {x,y,z}"
	def add_savety_constraint(a, b)
		assert_object_type_or_value_set(b, "b")
		@safety_constraints << [:subseteq, a, b]
	end

	# means "if x \in y => a \subseteq b"
	def add_conditional_constraint(x, y, a, b)
		assert_value(x, "x")
		assert_variable(y, "y")
		assert_variable(a, "a")
		assert_variable(b, "b")
		@constraints << [:cond, x, y, [:subseteq, a, b]]
	end

	def add_method_call_obligation(var, name, param_types, return_type, codeloc)
		assert_variable(var, "var")
		param_types.each { |x| assert_variable(x, "element of param_types") }
		assert_variable(return_type, "return_type")
		@constraints << [:obligation, var, [:methodcall, name, param_types, return_type], codeloc]
	end

	def each_constraint(&block)
		@constraints.each(&block)
	end

	def each_safety_constraint(&block)
		@safety_constraints.each(&block)
	end

	def is_variable?(x)
		return x.is_a?(Fixnum) && x >= 0 && x < @next_id
	end

	def is_value_set?(x)
		return x.is_a?(Array) && x.all? { |y| has_value?(y) }
	end

	def is_object_type?(x)
		return x.is_a?(Array) && x[0] == :object
	end

	def each_value(value_set, &block)
		value_set.each(&block)
	end

	def print()
		puts "Constraints:"
		puts
		@constraints.each do |c|
			print_constraint(c)
		end
		puts
		puts "Safety Constraints:"
		puts
		@safety_constraints.each do |c|
			print_constraint(c)
		end
	end

	def print_constraint(c)
		if c[0] == :subseteq
			puts "#{print_val(c[1])} < #{print_val(c[2])}"
		elsif c[0] == :filter
			puts "#{print_val(c[2])} = #{print_type(c[1])} => #{print_val(c[3])}"
		elsif c[0] == :obligation
			codeloc = c[c.size()-1]
			methods = c[2..c.size()-2].map do |x|
				unless x.class.name == "Array" && x[0] == :methodcall
					raise RuntimeError.new("unknown constraint type found: #{c.inspect}")
				end
				"#{x[1]}(#{x[2].map{ |y| print_val(y) }.join(", ")})->#{print_val(x[3])} @#{codeloc.to_s}"
			end.join(", ")
			puts "#{print_val(c[1])} < [#{methods}]"
		else
			raise RuntimeError.new("unknown constraint type found: #{c.inspect}")
		end
	end

	def print_type(type)
		type.inspect
	end

	def print_val(val)
		if is_variable?(val)
			desc = @var_desc[val]
			desc = "<CODE>" unless desc.is_a?(String)
			"#{desc}_#{val.to_s}"
		else
			val.inspect
		end
	end

	def description_for_var(var)
		@var_desc[var]
	end

	def codeloc_for_var(var)
		@var_codeloc[var]
	end

	def typeerror(msg)
		@typeerrors << msg
	end

	def each_typeerror(&block)
		@typeerrors.each(&block)
	end

private

	def assert_value(x, name)
		unless has_value?(x)
			raise RuntimeError.new("wrong type of constraint. #{name} is supposed to be a value.")
		end
	end

	def assert_variable(x, name)
		unless is_variable?(x)
			raise RuntimeError.new("wrong type of constraint. #{name} is supposed to be a variable.")
		end
	end

	def assert_object_type_or_value_set(x, name)
		unless is_object_type?(x) || is_value_set?(x)
			raise RuntimeError.new("wrong type of constraint. #{name} is supposed to be an object type or a value set.")
		end
	end

end

class ConstraintSolver

	def solve(constraints, resolver, options = {})
		graph = ConstraintGraph.new(constraints.values, resolver)
		constraints.each_constraint do |c|
			graph.add_constraint(c)
		end
		constraints.each_typeerror do |msg|
			graph.typeerror(msg)
		end

		if options[:output_solution]
			graph.print_solution(constraints)
		end

		constraints.each_safety_constraint do |c|
			graph.check_safety_constraint(constraints, c)
		end
		graph.print_violations(constraints)
		graph
	end

end

class ConstraintGraph

	NULLCLASS_VALUE = -1

	class Node

		attr_reader :bits, :edges, :filter_edges, :var

		def initialize(var)
			@var = var
			# the implicit bit for nullclass is not represented
			@bits = {}
			@edges = []
			@filter_edges = {}
		end

		def obligations()
			return [] unless defined?(@obligations)
			@obligations
		end

		def add_obligation(name, param_types, return_type, codeloc)
			@obligations = [] unless defined?(@obligations)
			@obligations << [name, param_types, return_type, codeloc]
		end

		def add_filter_edge(type, to)
			raise RuntimeError.new("duplicate filter edge from #{self.inspect} to #{to.inspect}") if @filter_edges.has_key?(to)
			@filter_edges[to] = type
		end
	end

	def initialize(values, resolver)
		@values = {}
		@inverse_values = {}
		val_id = 0
		values.each do |key, value|
			@values[key] = val_id
			@inverse_values[val_id] = key
			val_id += 1
		end
		# nullclass is handled specially
		@values["nullclass"] = NULLCLASS_VALUE
		@inverse_values[NULLCLASS_VALUE] = "nullclass"
		@num_values = val_id
		@nodes = {}
		@resolver = resolver
		@typeerrors = []
	end

	def add_constraint(c)
		if c[0] == :subseteq
			if c[1].is_a?(Fixnum)	# variable
				insert_propagation_constraint(c[1], c[2])
			elsif c[1].is_a?(Array)	# set of values
				insert_start_constraint(c[1], c[2])
			else
				raise RuntimeError.new("unknown constraint type: #{c.inspect}")
			end
		elsif c[0] == :filter
			insert_filter_constraint(c[1], c[2], c[3])
		elsif c[0] == :obligation
			unless c[2][0] == :methodcall
				raise RuntimeError.new("unknown constraint type: #{c.inspect}")
			end
			insert_method_call_obligation(c[1], c[2][1], c[2][2], c[2][3], c[3])
		else
			raise RuntimeError.new("unknown constraint type found: #{c.inspect}")
		end
	end

	def insert_propagation_constraint(a, b)
		from = node_by_variable(a)
		to = node_by_variable(b)
		out_edges = from.edges
		out_edges << to
		out_edges.uniq!
		from.bits.each do |k, v|
			next unless v == true
			propagate(to, k)
		end
	end

	def insert_filter_constraint(type, a, b)
		type = type.map do |classname|
					val = @values[classname]
					raise RuntimeError.new("unknown class #{classname}") unless val
					val
				end
		from = node_by_variable(a)
		to = node_by_variable(b)
		from.add_filter_edge(type, to)
		from.bits.each do |k, v|
			next unless v == true
			next unless type.include?(v)
			propagate(to, k)
		end
	end

	def insert_start_constraint(a, b)
		# union a to the bitvector of b
		node = node_by_variable(b)
		a.each do |classname|
			val = @values[classname]
			next if val == NULLCLASS_VALUE
			unless node.bits.has_key?(val)
				node.bits[val] = true
			end
		end

		# propagate
		node.edges.each do |to|
			node.bits.each do |k, v|
				propagate(to, k)
			end
		end
	end

	def insert_method_call_obligation(var, name, param_types, return_type, codeloc)
		cond_node = node_by_variable(var)
		cond_node.add_obligation(name, param_types, return_type, codeloc)
		
		insert_method_call_constraints(var, NULLCLASS_VALUE, name, param_types, return_type, codeloc)
		cond_node.bits.each do |k, v|
			insert_method_call_constraints(var, k, name, param_types, return_type, codeloc)
		end
	end

	def insert_method_call_constraints(var, value, name, param_types, return_type, codeloc)
		classname = @inverse_values[value]
		method = @resolver.resolve_method(classname, name)
		unless method
			return if value == NULLCLASS_VALUE
			typeerror("var_#{var}: #{classname} does not support method #{name} at #{codeloc.to_s}")
			return
		end
		unless param_types.size == method.param_types.size
			return if value == NULLCLASS_VALUE
			typeerror("var_#{var}: in call to #{classname}.#{name}: expected #{method.param_types.size} parameters, found #{param_types.size}.")
			return
		end

		# parameters
		method.param_types.zip(param_types).each do |formal, actual|
			# c \in [exp] => [exp_i] \subseteq [param_i]
			add_constraint([:subseteq, actual, formal])
		end

		# return value
		# c \in [exp] => [TYPE] \subseteq [exp.name(exp_1,...,exp_n)]
		add_constraint([:subseteq, method.return_type, return_type])
	end

	def node_by_variable(var)
		if @nodes.has_key?(var)
			@nodes[var]
		else
			node = Node.new(var)
			@nodes[var] = node
			node
		end
	end

	def propagate(node, bit)
		if node.bits.has_key?(bit)
			return
		end
		node.bits[bit] = true
		
		node.edges.each do |to|
			propagate(to, bit)
		end

		node.filter_edges.each do |to, type_filter|
			next unless type_filter.include?(bit)
			propagate(to, bit)
		end

		node.obligations.each do |name, param_types, return_type, codeloc|
			insert_method_call_constraints(node.var, bit, name, param_types, return_type, codeloc)
		end
	end

	def check_safety_constraint(cs, c)
		unless c[0] == :subseteq
			raise RuntimeError.new("invalid constraint type: #{c.inspect}")
		end
		
		unless c[1].is_a?(Fixnum)
			raise RuntimeError.new("invalid constraint type: #{c.inspect}")
		end

		unless c[2].is_a?(Array)
			raise RuntimeError.new("invalid constraint type: #{c.inspect}")
		end

		node = node_by_variable(c[1])
		if c[2][0] == :object
			# object type
			ot = c[2]
			ot[1..ot.size].each do |m|
				unless m[0] == :method
					raise RuntimeError.new("invalid constraint type: #{c.inspect}")
				end
				name = m[1]
				num_params = m[2]
				node.bits.each do |k, v|
					classname = @inverse_values[k]
					method = @resolver.resolve_method(classname, name)
					unless method
						var = c[1]
						code = cs.description_for_var(var)
						type = pretty_print_type(c[2])
						raise ConstraintSatisfactionError.new(code, type, var) { |c| "method #{name} is not defined for class #{classname}, but is called on expression #{c}"}
					end
					unless method.param_types.size == num_params
						var = c[1]
						code = cs.description_for_var(var)
						type = pretty_print_type(c[2])
						raise ConstraintSatisfactionError.new(code, type, var) { |c| "method #{classname}.#{name} is expecting #{method.param_types.size}, but is called with #{num_params} on expression #{c}." }
					end
				end
			end
		else
			# set of values
			b_vals = c[2].map {|value| @values[value] }
			node.bits.each do |k, v|
				unless b_vals.include?(k)
					var = c[1]
					code = cs.description_for_var(var)
					type = pretty_print_type(c[2])
					raise ConstraintSatisfactionError.new(code, type, var) { |c| "expression #{c} can yield a result not in #{type}" }
				end
			end
		end

	end

	def pretty_print_type(t)
		unless t.is_a?(Array)
			raise RuntimeError.new("expected object type or value set #{t.inspect}")
		end
		if t[0] == :object
			"[#{t[1..t.size].map{ |m| "#{m[1]}(#{m[2]} args)"}.join(', ')}]"
		else
			"{#{t.join(", ")}}"
		end
	end

	def print_solution(cs)
		puts
		puts "Solution:"
		puts
		@nodes.each do |var, node|
			desc = cs.description_for_var(var)
			desc = "<CODE>" unless desc.is_a?(String)
			desc = "#{desc}_#{var.to_s}"
			puts "#{desc} : #{print_type(node, @inverse_values)}        [#{cs.codeloc_for_var(var).to_s}]"
		end
	end

	def print_violations(cs)
		return if @typeerrors.empty?
		puts
		puts "Type Safety Violations:"
		puts
		@typeerrors.each do |text|
			puts "TYPEERROR: #{text}"
		end
	end

	def typeerror(text)
		@typeerrors << text
	end

private

	def print_type(node, inv_values)
		classlist = node.bits.to_a.map { |k,v| inv_values[k] }
		classlist << "nullclass"
		classlist.uniq!
		type = "{#{classlist.join(", ")}}"
	end

	def print_exp_for_var(cs, var)
		desc = cs.description_for_var(var)
		desc = pretty_print(desc) unless desc.is_a?(String)
		desc = "#{desc}_#{var.to_s}"
	end

end

class ConstraintSatisfactionError < RuntimeError

	attr_reader :code, :type, :var

	def initialize(code, type, var, &block)
		super("")
		@code = code
		@type = type
		@var = var
		@format_message = block
	end

	def to_DynTypeError(configuration)
		c = nil
		if @code.is_a?(String)
			c = "#{@code} (variable #{@code}_#{@var})"
		else
			# configuration.pretty_print_code(@code)
			c = @code.inspect + " (variable <CODE>_#{@var})"
		end
		msg = @format_message.call(c)
		raise DynTypeError.new("TYPEERROR: #{msg}")
	end

end

class MethodResolver

	def initialize(configuration)
		@configuration = configuration
	end

	def resolve_method(classname, methodname)
		loc = @configuration.class_by_name[classname]
		return nil unless loc
		clazz = @configuration.store[loc]
		return clazz.dynmethods[methodname]
	end

end

# extend class DynMethod from dyn_semantics.rb
class DynMethod

  # for type inference
  attr_accessor :return_type, :param_types

end

