#!/usr/bin/ruby

NULL_CONST = [:loc, 0]
FALSE_CONST = [:constructor_call, "boolean", [NULL_CONST]]
TRUE_CONST = [:constructor_call, "boolean", [FALSE_CONST]]

OP_NAME_MAPPING = {
                    "*" => "multiply",
                    "/" => "divide",
                    "+" => "add",
                    "-" => "subtract",
                    "<" => "lt",
                    ">" => "gt",
                    "<=" => "le",
                    ">=" => "ge",
                    "==" => "equal",
                    "===" => "identical"
                  }

class BinaryExpressionNode < Treetop::Runtime::SyntaxNode

  def construct()
    result = a.construct
    if b
      b.elements.each do |e|
        right = e.op.handle_right_hand_side(e.exp)
        result = e.op.construct(result, right)
      end
    end
    result
  end

end

class OperationNode < Treetop::Runtime::SyntaxNode

  def construct(left, right)
    name = method_name()
    if name
      [:method_call, left, name, [right], CodeLoc.from_node(self)]
    else
      [function_symbol(), left, right, CodeLoc.from_node(self)]
    end
  end

  def handle_right_hand_side(exp)
    exp.construct
  end

end

def generate_numeric(n, codeloc)
  if n <= 0
    [:constructor_call, "numeric", [NULL_CONST], codeloc]
  else
    [:method_call, generate_numeric(n-1, codeloc), "succ", [], codeloc]
  end
end

def generate_string(str, codeloc)
  if str.size() <= 0
    [:constructor_call, "string", [NULL_CONST], codeloc]
  else
    [:method_call, generate_string(str[0, str.size()-1], codeloc), "addchar", [generate_numeric(str[str.size()-1].ord, codeloc)], codeloc]
  end
end

class CodeLoc

  def initialize(file, interval, generated = false)
    @int = interval
    @file = file
    @generated = generated
  end

  def self.from_node(node)
    root = node
    while !root.has_file?() do
      root = root.parent
    end
    CodeLoc.new(root.file, node.interval)
  end

  def self.generated()
    CodeLoc.new(nil, nil, true)
  end

  def to_s()
    if @generated then
      return "<generated code>"
    end
    start_line = @file.line_of(@int.min)
    end_line = @file.line_of(@int.max)
    if start_line == end_line then
      "line #{start_line} columns #{@file.index_in_line(@int.min)}-#{@file.index_in_line(@int.max)} in #{@file.filename}"
    else
      # DEBUG
      #$stderr.puts "DEBUG-BEGIN"
      #$stderr.puts @file.inspect
      #$stderr.puts @int.inspect
      "line #{start_line} column #{@file.index_in_line(@int.min)} to line #{end_line} column #{@file.index_in_line(@int.max)} in #{@file.filename}"
    end
  end

  def inspect()
    "CodeLoc:" + to_s()
  end

end

class ParsedFile

  attr_reader :filename, :contents, :parse_tree

  def initialize(filename)
    @filename = filename
    @contents = File.read(filename)
  end

  def line_of(index)
    @contents.line_of(index)
  end

  def index_in_line(index)
    @contents.index_in_line(index)
  end

  def parse(parser)
    result = parser.parse(@contents)
    unless result
      puts parser.failure_reason
      puts parser.failure_line
      puts parser.failure_column
      exit(-1)   # TODO: raise an exception, instead?
    end
    result.file = self
    @parse_tree = result.construct()
  end

  def join_parse_trees(other)
    raise RuntimeError.new("expected a parse tree") unless other.parse_tree[0] == :program
    result = @parse_tree.clone()
    other.parse_tree[1..other.parse_tree.size()-2].each do |elem|
      result.insert(result.size()-1, elem)
    end
    result
  end

end

class String

	def index_in_line(index)
		index - (self[0..index].rindex("\n") || 0)
	end

end

class Treetop::Runtime::SyntaxNode

  def file
    @file
  end

  def file=(val)
    @file = val
  end

  def has_file?
    if defined?(@file) then true else false end
  end

end

