#!/usr/bin/ruby

# NOTES to implementation:
#       - the union types are usually very sparse. Representing them as a hashset should thus save
#         some space over a bitvector
# TODO: Ideas for optimization:
#       - many trace-typevars will usually hold the same values (those propagated through
#         only local constraints). One could thus represent them using only one bitvector
#         and keep track only of those that differ.
#       - if a method is often called with the same parameter types, one should collapse these traces
#         into one. In effect, a method's traces might then be partitioned into equivalence classes
#         of traces (temporarily) calling the method with identical parameter types.
#         # NOTE: this might also work for container classes
#       - When a trace "leaves" an equivalence class EC1 for EC2, then EC1.param_types must be a
#         subset of EC2.param_types. Monotonicity thus demands that all union types in EC1 are
#         subset of those in EC2. Leverage this by only representing their differences.
#         (might be a generalization of the first point)
module PolymorphicTypeInference

	BOOL_TYPE = [:object, [:method, "inner", 0]]

	# indizes in method_context
	NEXT_CALL_ID_INDEX = 0
	METHOD_INDEX = 2
	SCOPE_INDEX = 3

	#TODO: make container classes (fully) parametric polymorphic
        #      - make set of values dynamic
        #      - keep track of the set of non-cyclic call traces for each constructor
        #      -> distinguish between classes created by constructor calls with different call traces
        #      -> BUT: detect cycles to stay decidable
	def polymorphic_type_inference_program(options = {})
		constraint_system, root_method = generate_local_constraints(@configuration)
		if options[:output_constraints]
			constraint_system.print()
		end
		begin
			resolver = MethodResolver.new(@configuration, constraint_system, root_method)
			solution = solve_constraints(constraint_system, resolver, options)
		rescue ConstraintSatisfactionError => e
			raise e.to_DynTypeError(@configuration)
		end
	end

	def has_error?()
		false
	end

	def error_text()
		""
	end

private

	def generate_local_constraints(configuration)
		cs = ConstraintSystem.new()
		@configuration.class_by_name().each do |cname, loc|
			cs.add_value(cname)
			clazz = @configuration.store[loc]
			clazz.dynmethods.each do |mname, method|
				method.param_types = method.params.map { |p| cs.create_variable(p, method) }
				method.return_type = cs.create_variable("result-#{cname}.#{mname}", method)
			end
		end
		
		@configuration.class_by_name().each do |cname, loc|
			clazz = @configuration.store[loc]
			generate_constraints_for_class(cs, cname, clazz)
		end
		root_method = generate_constraints_for_initial_program(cs)

		return cs, root_method
	end

	def generate_constraints_for_initial_program(cs)
		method = DynMethod.new([:method, "root", [], INITIAL_PROGRAM])
		method.param_types = []
		method.return_type = cs.create_variable("result", method)
		class_context = ["ROOT", nil, {}, nil]
		generate_constraints_for_method(cs, method, class_context)
		cs.add_savety_constraint(method.return_type, [cs.get_value("string")])
		return method
	end

	def generate_constraints_for_class(cs, cname, clazz)
		# set up a self value for this class
		self_val = value_for_class(cs, cname)
		# set up a self var for this class
		self_var = cs.create_variable("self-#{cname}", nil, :class_based)
		cs.add_constraint([self_val], self_var)    # {C} \subseteq [self-C]
		cs.add_savety_constraint(self_var, [self_val])    # [self-C] \subseteq {C}

		instance_env = {}
		constructor_env = nil
		self_access_callback = Proc.new() do
			# capture the constructor's instance environment on the first access to self
			constructor_env = constructor_env || instance_env.clone()
		end
		constructor = clazz.dynmethods()["init"]
		if constructor
			generate_constraints_for_method(cs, constructor, [cname, self_var, instance_env, self_access_callback])
		end
		constructor_env = constructor_env || {}

		clazz.dynmethods().each do |mname, method|
			next if mname == "init"
			generate_constraints_for_method(cs, method, [cname, self_var, instance_env, nil])
		end

		rest_env = instance_env.keys - constructor_env.keys
		unless rest_env.size == 0
			typeerror("class #{cname}: instance variables #{rest_env.join(", ")} are not initialized in constructor prior to the first access to self.")
		end
	end

	def generate_constraints_for_method(cs, method, class_context)
		local_env = {}
		scope = new_scope()
		method.params.zip(method.param_types).each do |param, param_type|
			local_env[param] = {:var => param_type, :scope => scope}
		end
		body_type = generate_constraints_for_expression(cs, method.code(), [0, local_env, method, [scope]], class_context)
		cs.add_constraint(body_type, method.return_type, method)
	end

	def new_scope()
		if @last_scope
			@last_scope += 1
		else
			@last_scope = 0
		end
	end

	def generate_constraints_for_expression(cs, tree, method_context, class_context)
		_, local_env, method, in_scopes = method_context
		cname, self_var, instance_env, self_access_callback = class_context
		case tree[0]
		  when :variable
		    name = tree[1]
		    unless local_env.has_key?(name)
		      typeerror("accessed unknown local variable #{name.inspect} in method #{cname}.#{method.name}")
		    end
		    unless in_scopes.include?(local_env[name][:scope])
		      typeerror("accessed local variable #{name.inspect} where its existence cannot be guaranteed in method #{cname}.#{method.name}")
		    end
		    return local_env[name][:var]
	  
		  when :instance_variable
		    name = tree[1]
		    unless instance_env.has_key?(name)
		      # instance variables do not belong to methods, therefore the synchronization variable
		      # is not associated to one
		      instance_env[name] = cs.create_variable(name, nil, :class_based)
		    end
		    cs.add_method_for_var(instance_env[name], method)
		    return instance_env[name]
	  
		  when :self
		    if self_access_callback
			self_access_callback.call()
		    end
		    cs.add_method_for_var(self_var, method)
		    return self_var

		  when :assignment
		    name = tree[1]
		    exp_type = generate_constraints_for_expression(cs, tree[2], method_context, class_context)
		    unless local_env.has_key?(name)
		      local_env[name] = {:var => cs.create_variable(name, method), :scope => in_scopes[-1]}
		    end
		    var = local_env[name][:var]

		    cs.add_constraint(exp_type, var, method)   # [exp] \subseteq [var]
		    own_type = cs.create_variable(tree, method)
		    cs.add_constraint(exp_type, own_type, method)         # [exp] \subseteq [var = exp]
		    # as no other constraint contributes to own_type, the equality can savely be
		    #   weakened to an inclusion
		    # cs.add_constraint(own_type, exp_type, method)  # [var = exp] \subseteq [exp]
		    return own_type
	  
		  when :instance_assignment
		    name = tree[1]
		    exp_type = generate_constraints_for_expression(cs, tree[2], method_context, class_context)
		    unless instance_env.has_key?(name)
		      # instance variables do not belong to methods, therefore the synchronization variable
		      # is not associated to one
		      instance_env[name] = cs.create_variable(name, nil, :class_based)
		    end
		    cs.add_method_for_var(instance_env[name], method)
		    var = instance_env[name]

		    cs.add_constraint(exp_type, var, method)   # [exp] \subseteq [var]
		    own_type = cs.create_variable(tree, method)
		    cs.add_constraint(exp_type, own_type, method)         # [exp] \subseteq [var = exp]
		    # as no other constraint contributes to own_type, the equality can savely be
		    #   weakened to an inclusion
		    # cs.add_constraint(own_type, exp_type, method)  # [var = exp] \subseteq [exp]
		    return own_type
	  
		  when :method_call
		    name = tree[2]
		    typeerror("constructors cannot be called after instantiation") if name == CONSTRUCTOR_NAME
		    params = tree[3]
		    exp_type = generate_constraints_for_expression(cs, tree[1], method_context, class_context)
		    params_types = generate_constraints_for_params(cs, params, method_context, class_context)
		    own_type = cs.create_variable(tree, method)
		    call_id = method_context[NEXT_CALL_ID_INDEX]
		    method_context[NEXT_CALL_ID_INDEX] += 1
		    generate_constraints_for_method_call(cs, [exp_type, name, call_id, params_types], own_type, method_context, class_context)
		    return own_type
	  
		  when :constructor_call
		    classname = tree[1]
		    value = cs.get_value(classname)
		    typeerror("unknown class #{classname.inspect}") unless value
		    params = tree[2]
		    params_types = generate_constraints_for_params(cs, params, method_context, class_context)
		    own_type = cs.create_variable(tree, method)
		    call_id = method_context[NEXT_CALL_ID_INDEX]
		    method_context[NEXT_CALL_ID_INDEX] += 1
		    generate_constraints_for_method_call(cs, [[value], CONSTRUCTOR_NAME, call_id, params_types], own_type, method_context, class_context)
		    return own_type

		  when :loc
		    loc = tree[1]
		    if loc == 0   # null has type {}
		      own_type = cs.create_variable("null", method)
		      cs.add_savety_constraint(own_type, [])
		      return own_type
		    end
		    # no idea how this should happen -> if it does, think about a way to type this
		    syserror("this should never happen")
	  
		  when :if
		    cond = tree[1]
		    true_branch = tree[2]
		    false_branch = tree[3]

		    cond_type = generate_constraints_for_expression(cs, cond, method_context, class_context)
		    # TODO: what sense does this constraint make?
		    #cs.add_constraint(["boolean"], cond_type)
		    cs.add_savety_constraint(cond_type, BOOL_TYPE)
		    own_type = cs.create_variable(tree, method)
		    branching_method_context = method_context.clone()
		    branching_method_context[SCOPE_INDEX] = method_context[SCOPE_INDEX].clone()
		    branching_method_context[SCOPE_INDEX] << new_scope()
		    true_branch_type = generate_constraints_for_expression(cs, true_branch, branching_method_context, class_context)
		    false_branch_type = generate_constraints_for_expression(cs, false_branch, branching_method_context, class_context)
		    cs.add_constraint(true_branch_type, own_type, method)
		    cs.add_constraint(false_branch_type, own_type, method)
		    # as no other constraint contributes to own_type, the equality can savely be
		    #   weakened to an inclusion
		    # cs.add_constraint(own_type, true_branch_type, method)
		    # as no other constraint contributes to own_type, the equality can savely be
		    #   weakened to an inclusion
		    # cs.add_constraint(own_type, false_branch_type, method)
		    return own_type

		  when :while
		    cond = tree[1]
		    loop = tree[2]
		    cond_type = generate_constraints_for_expression(cs, cond, method_context, class_context)
		    #TODO: what sense does this constraint make?
		    #cs.add_constraint(["boolean"], cond_type)
		    cs.add_savety_constraint(cond_type, BOOL_TYPE)
		    own_type = cs.create_variable(tree, method)
		    branching_method_context = method_context.clone()
		    branching_method_context[SCOPE_INDEX] = method_context[SCOPE_INDEX].clone()
		    branching_method_context[SCOPE_INDEX] << new_scope()
		    loop_type = generate_constraints_for_expression(cs, loop, branching_method_context, class_context)
		    cs.add_constraint(loop_type, own_type, method)
		    # as no other constraint contributes to own_type, the equality can savely be
		    #   weakened to an inclusion
		    # cs.add_constraint(own_type, loop_type, method)
		    return own_type

		  when :statement
		    first = tree[1]
		    result_type = nil
		    tree[1..tree.size].each do |item|
			    result_type = generate_constraints_for_expression(cs, item, method_context, class_context)
		    end
		    own_type = cs.create_variable(tree, method)
		    cs.add_constraint(result_type, own_type, method)         # [exp_2] \subseteq [exp_1; exp_2]
		    # as no other constraint contributes to own_type, the equality can savely be
		    #   weakened to an inclusion
		    # cs.add_constraint(own_type, result_type, method)  # [exp_1; exp_2] \subseteq [exp_2]
		    return own_type

		  when :error
		    return []

		  # --- builtin functionality ---

		  when :hook_loc_identical
		    a = tree[1]
		    b = tree[2]
		    generate_constraints_for_expression(cs, a, method_context, class_context)
		    generate_constraints_for_expression(cs, b, method_context, class_context)
		    own_type = cs.create_variable(tree, method)
		    call_id = method_context[NEXT_CALL_ID_INDEX]
		    method_context[NEXT_CALL_ID_INDEX] += 1
		    cs.add_constraint(["boolean"], own_type)
		    return own_type

		  when :hook_get_loc
		    exp = tree[1]
		    generate_constraints_for_expression(cs, exp, method_context, class_context)
		    own_type = cs.create_variable(tree, method)
		    call_id = method_context[NEXT_CALL_ID_INDEX]
		    method_context[NEXT_CALL_ID_INDEX] += 1
		    cs.add_constraint(["numeric"], own_type)
		    return own_type

		  when :hook_is_a
		    exp = tree[1]
		    #TODO: clazzname must be a string
		    clazzname = tree[2]
		    generate_constraints_for_expression(cs, exp, method_context, class_context)
		    typeerror("unknown class #{clazzname.inspect}") unless @configuration.class_by_name().has_key?(clazzname)
		    own_type = cs.create_variable(tree, method)
		    call_id = method_context[NEXT_CALL_ID_INDEX]
		    method_context[NEXT_CALL_ID_INDEX] += 1
		    cs.add_constraint(["boolean"], own_type)
		    return own_type

		  else
		    syserror("found unknown node in parsetree: #{subtree.inspect}")
		end
	end

	def generate_constraints_for_params(cs, params, method_context, class_context)
		params.map do |param|
			generate_constraints_for_expression(cs, param, method_context, class_context)
		end
	end

	def generate_constraints_for_method_call(cs, call, own_type, method_context, class_context)
		type, name, call_id, param_types = call
		method = method_context[METHOD_INDEX]

		if cs.is_variable?(type)
			# [exp] \subseteq {all legal receivers}
			# TODO: is the call obligation not checking enough?
			cs.add_savety_constraint(type, [:object, [:method, name, param_types.size]])
                        cs.add_method_call_obligation(type, name, call_id, param_types, own_type, method)
		else
			unless cs.is_value_set?(type)
				syserror("illegal argument: receiver must be either a variable or a value set")
			end
			typevar = cs.create_variable("constructor_call_#{call_id}", method)
			# TODO: is the call obligation not checking enough?
			cs.add_constraint(type, typevar, method)
			cs.add_savety_constraint(typevar, [:object, [:method, name, param_types.size]])
                        cs.add_method_call_obligation(typevar, name, call_id, param_types, own_type, method)
			
			#cs.each_value(type) do |value|
			#	loc = @configuration.class_by_name[value]
			#	clazz = @configuration.store[loc]
			#	unless clazz.dynmethods.has_key?(name)
			#		typeerror("class #{value} does not have a method #{name}")
			#	end
			#	method = clazz.dynmethods[name]
			#	method.param_types.zip(param_types).each do |formal, actual|
			#		cs.add_constraint(actual, formal, method)
			#	end
			#	cs.add_constraint(method.return_type, own_type, method)
			#end
		end

		
	end

	def each_class_with_method(name, &block)
		@configuration.class_by_name().each do |cname, loc|
			clazz = @configuration.store[loc]
			next unless clazz.dynmethods.has_key?(name)
			block.call(cname, clazz, clazz.dynmethods[name])
		end
	end

	def solve_constraints(constaints, resolver, options = {})
		solver = ConstraintSolver.new()
		return solver.solve(constaints, resolver, options)
	end


	def value_for_class(cs, cname)
		unless cs.has_value?(cname)
			syserror("unknown class #{cname}")
		end
		cs.get_value(cname)
	end

	def syserror(text)
		raise RuntimeError.new("SYSERROR:" + text)
	end

	def typeerror(text)
		raise DynTypeError.new("TYPEERROR: " + text)
	end
end

class ConstraintSystem

	attr_reader :values

	def initialize()
		@values = {}
		@var_desc = {}
		@var_methods = {}
		@var_types = {}
		@constraints = []
		@safety_constraints = []
		@next_id = 0
	end

        # values

	def add_value(name)
		if @values.has_key?(name)
			raise RuntimeError.new("duplicate value: #{name.inspect}")
		end
		@values[name] = name
	end

	def has_value?(name)
		@values.has_key?(name)
	end

	def get_value(name)
		@values[name]
	end

        # variables

	def create_variable(desc, method, type = nil)
		var_id = @next_id
		@next_id += 1
		@var_desc[var_id] = desc
		@var_methods[var_id] = []
		@var_types[var_id] = type if type
		add_method_for_var(var_id, method) if method
		return var_id
	end

	def description_for_var(var)
		@var_desc[var]
	end

	def add_method_for_var(var_id, method)
		@var_methods[var_id] << method
	end

	def methods_for_var(var_id)
		@var_methods[var_id]
	end

	def type_of_var(var_id)
		@var_types[var_id]
	end

	# constraints

	# means "a \subseteq b" or
	#       "{x,y,z} \subseteq b"
	def add_constraint(a, b, method = nil)
		assert_variable(b, "b")
		if method
			@constraints << [:subseteq, a, b, method]
		else
			@constraints << [:subseteq, a, b]
		end
	end

	# means "a \subseteq {x,y,z}" or
        #       "a \subseteq [m(P1,...,Pn)->R]"
	def add_savety_constraint(a, b)
		assert_object_type_or_value_set(b, "b")
		@safety_constraints << [:subseteq, a, b]
	end

        # means "if C \in var => C must support a method name(P1,...,Pn)->R and
	#                        the constraints
	#                             param_types[0] \subseteq P1
	#                             ...
	#                             param_types[n-1] \subseteq Pn
	#                             R \subseteq return_type
	def add_method_call_obligation(var, name, call_id, param_types, return_type, method)
		assert_variable(var, "var")
		param_types.each { |x| assert_variable(x, "element of param_types") }
		assert_variable(return_type, "return_type")
		@constraints << [:obligation, var, method, [:methodcall, name, call_id, param_types, return_type]]
	end

	def each_constraint(&block)
		@constraints.each(&block)
	end

	def each_safety_constraint(&block)
		@safety_constraints.each(&block)
	end

        # assertions

	def is_variable?(x)
		return x.is_a?(Fixnum) && x >= 0 && x < @next_id
	end

	def is_value_set?(x)
		return x.is_a?(Array) && x.all? { |y| has_value?(y) }
	end

	def is_object_type?(x)
		return x.is_a?(Array) && x[0] == :object
	end

        # others

	def each_value(value_set, &block)
		value_set.each(&block)
	end

	def print()
		puts "Constraints:"
		puts
		@constraints.each do |c|
			print_constraint(c)
		end
		puts
		puts "Safety Constraints:"
		puts
		@safety_constraints.each do |c|
			print_constraint(c)
		end
	end

	def print_constraint(c)
		if c[0] == :subseteq
			puts "#{print_val(c[1])} < #{print_val(c[2])}"
		elsif c[0] == :obligation
			methods = c[3..c.size].map do |x|
				unless x[0] == :methodcall
					raise RuntimeError.new("unknown constraint type found: #{c.inspect}")
				end
				"#{x[1]}_#{x[2]}(#{x[3].map{ |y| print_val(y) }.join(", ")})->#{print_val(x[4])}"
			end.join(", ")
			puts "#{print_val(c[1])} < [#{methods}]"
		elsif c[0] == :sync
			puts "#{print_val(c[1])} <s< #{print_val(c[2])}"
		else
			raise RuntimeError.new("unknown constraint type found: #{c.inspect}")
		end
	end

	def print_val(val)
		if is_variable?(val)
			desc = @var_desc[val]
			desc = "<CODE>" unless desc.is_a?(String)
			"#{desc}_#{val.to_s}"
		else
			val.inspect
		end
	end

private

	def assert_value(x, name)
		unless has_value?(x)
			raise RuntimeError.new("wrong type of constraint. #{name} is supposed to be a value.")
		end
	end

	def assert_variable(x, name)
		unless is_variable?(x)
			raise RuntimeError.new("wrong type of constraint. #{name} is supposed to be a variable.")
		end
	end

	def assert_object_type_or_value_set(x, name)
		unless is_object_type?(x) || is_value_set?(x)
			raise RuntimeError.new("wrong type of constraint. #{name} is supposed to be an object type or a value set.")
		end
	end

end

class ConstraintSolver

	def solve(constraints, resolver, options = {})
		graph = ConstraintGraph.new(constraints.values, resolver)
		constraints.each_constraint do |c|
			graph.add_constraint(c)
		end

		# bootstrap the process by creating the first trace in the root node
		root_method = resolver.resolve_method(nil, nil)
		empty_eq_class = DynMethodCallTraceEquivalenceClass.new([], root_method)
		graph.add_trace_to_method(empty_eq_class, root_method)

		if options[:output_solution]
			graph.print_solution(constraints)
		end

		constraints.each_safety_constraint do |c|
			graph.check_safety_constraint(constraints, c)
		end
		graph
	end

end

# TODO: are the call_ids alone not sufficient?
class DynMethodCall

	attr_reader :classname, :dyn_method, :call_id

	def initialize(classname, dyn_method, call_id)
		@classname = classname
		@dyn_method = dyn_method
		@call_id = call_id
	end

	def eql?(other)
		return false unless other.is_a?(DynMethodCall)
		@classname.eql?(other.classname) &&
		@dyn_method.eql?(other.dyn_method) &&
		@call_id.eql?(other.call_id)
	end

	def hash()
		return @hash if defined?(@hash)
		@hash = ["DynMethodCall", @classname, @dyn_method, @call_id].hash
	end

	def to_s()
		"#{@classname}.#{@dyn_method.name}_#{@call_id.to_s}"
	end

end

class CallSite

	attr_reader :eq_class, :param_nodes, :result_node, :dyn_methodcall
	attr_accessor :result_hook_id

	def initialize(eq_class, param_nodes, result_node, dyn_methodcall)
		@eq_class = eq_class
		@param_nodes = param_nodes
		@result_node = result_node
		@dyn_methodcall = dyn_methodcall
		@result_hook_id = nil
	end

end

class DynMethodCallTraceEquivalenceClass

	attr_reader :leads_to, :callsites, :contains, :param_nodes, :result_node
	attr_accessor :signature

	# leads_to = nil means: leads to root method
	# TODO: also neccessary to add result_node?
	def initialize(signature, param_nodes, leads_to = nil)
		@leads_to = leads_to
		@callsites = {}
		@contains = {self => 1}
		@contains_methods = {}
		@signature = signature
		@param_nodes = param_nodes
		@hooks = {}
		@next_hook_id = 0
		@calling_eq_class_hook_ids = {}
	end

	# callsites

	# inverse of remove_callsite(callsite)
	def add_callsite(callsite)
		if @callsites.has_key?(callsite)
			raise RuntimeError.new("callsite #{callsite.inspect} was added to equivalence class #{self.inspect} twice")
		end
		# TODO: cycle detection?
		@callsites[callsite] = true
		hook_calling_eq_class(callsite.eq_class)
		add_callsite_to_contains_hash(callsite)
		each_hook do |id, hook|
			hook.call(:add, self, callsite)
		end
		self
	end

	# inverse of add_callsite(callsite)
	def remove_callsite(callsite)
		unless @callsites.has_key?(callsite)
			raise RuntimeError.new("cannot remove unknown callsite")
		end
		@callsites.delete(callsite)
		unhook_calling_eq_class(callsite.eq_class)
		remove_callsite_from_contains_hash(callsite)
		each_hook do |id, hook|
			hook.call(:remove, self, callsite)
		end
	end

	def empty?()
		@callsites.empty?()
	end

	def each_callsite(&block)
		@callsites.each do |callsite|
			block.call(callsite)
		end
	end

	# hooks

	def add_hook(&block)
		hook_id = @next_hook_id
		@next_hook_id += 1
		@hooks[hook_id] = block
		hook_id
	end

	def remove_hook(hook_id)
		@hooks.delete(hook_id)
	end

	def each_hook(&block)
		@hooks.each do |id, hook|
			block.call(id, hook)
		end
	end

	# operations

	def merge(other)
		raise RuntimeError.new("cannot merge equivalence classes leading to different methods") unless other.leads_to == @leads_to
                other.each_callsite do |callsite|
			add_callsite(callsite)
		end
		self
	end

	def get_last_method()
		if @leads_to
			@leads_to
		else
			:root_method
		end
	end

	def contains_eq_class(eq_class)
		@contains.has_key?(eq_class)
	end

	def contains_method(dyn_method)
		@contains_methods.has_key?(dyn_method)
	end

	def eq_classes_of_method(dyn_method)
		h = @contains_methods[dynmethod]
		return false unless h
		h.keys
	end

	def split(eq_class)
		return self, nil if eq_class = self
		return nil, self unless @contains[eq_class] > 0
		left_ec = DynMethodCallTraceEquivalenceClass.new(@signature, @param_nodes, @leads_to)
		right_ec = DynMethodCallTraceEquivalenceClass.new(@signature, @param_nodes, @leads_to)
		each_callsite do |callsite|
			ec = callsite.eq_class
			left, right = ec.split(eq_class)
			left_ec.add(left) if left
			right_ec.add(right) if right
		end
		left_ec = nil if left_ec.empty?
		right_ec = nil if right_ec.empty?
		return left_ec, right_ec
	end

	# TODO: think about: is there a canonical example trace?
	def example_trace()
		return [] if empty?()
		callsite = @callsites.keys.first
		call = callsite.dyn_methodcall
		eq = callsite.eq_class
		eq.example_trace() + [call]
	end

	def to_s()
		"<EqC>[#{example_trace().map{|e| e.to_s}.join("->")}]"
	end

private

	# inverse of add_callsite_to_contains_hash(callsite)
	def remove_callsite_from_contains_hash(callsite)
		callsite.eq_class.contains.each do |k,v|
			r = @contains[k] -= v
			if r == 0
				@contains.delete(k)
			elsif r < 0
				p @contains
				raise RuntimeError.new("inconsistent contains hash")
			end
			h = @contains_methods[k.leads_to]
			r = h[k] -= v
			if r == 0
				h.delete(k)
				if h.empty?
					@contains_methods.delete(k.leads_to)
				end
			elsif r < 0
				raise RuntimeError.new("inconsistent contains_methods hash")
			end
		end
	end

	# inverse of remove_callsite_from_contains_hash(callsite)
	def add_callsite_to_contains_hash(callsite)
		callsite.eq_class.contains.each do |k,v|
			if @contains.has_key?(k)
				@contains[k] += v
			else
				@contains[k] = v
			end
			if @contains_methods.has_key?(k.leads_to)
				h = @contains_methods[k.leads_to]
				if h.has_key?(k)
					h[k] += v
				else
					h[k] = v
				end
			else
				@contains_methods[k.leads_to] = {k => v}
			end
		end
	end

	# inverse of unhook_calling_eq_class(eq_class)
	def hook_calling_eq_class(eq_class)
		# hook calling equivalence class
		hook_id = eq_class.add_hook do |op, m_eq_class, m_new_sub_eq_class|
			if op == :add
				add_callsite_to_contains_hash(m_new_sub_eq_class)
			elsif op == :remove
				remove_callsite_from_contains_hash(m_new_sub_eq_class)
			else
				raise RuntimeError.new("unknown operation #{op} used in hook")
			end
			# propagate change further
			self.each_hook do |id, hook|
				hook.call(op, m_eq_class, m_new_sub_eq_class)
			end
		end
		# save hook_id
		if @calling_eq_class_hook_ids.has_key?(eq_class)
			@calling_eq_class_hook_ids[eq_class] << hook_id
		else
			@calling_eq_class_hook_ids[eq_class] = [hook_id]
		end
	end

	# inverse of hook_calling_eq_class(eq_class)
	def unhook_calling_eq_class(eq_class)
		unless @calling_eq_class_hook_ids.has_key?(eq_class)
			raise RuntimeError.new("cannot unhooked eq_class #{eq_class.to_s} because it was never hooked")
		end
		list = @calling_eq_class_hook_ids[eq_class]
		if list.empty?
			raise RuntimeError.new("inconsistent hooks hash")
		end
		hook_id = list.delete_at(0)

		# unhook calling equivalence class
		eq_class.remove_hook(hook_id)
	end

end

class ConstraintGraph

	NULLCLASS_VALUE = -1

	# Note: the nodes are not operating on traces anymore, but on equivalence classes
	class Node

		ID_MAPPING = Proc.new{ |trace| trace }

		attr_reader :var, :methods

		def initialize(var, methods, trace_mapping = nil)
			@var = var
			@methods = {}
			@traces_by_method = {}
			@local_edges_by_method = {}
			methods.each do |method|
				add_method(method)
			end
			@traces = {}
			@buckets = {}
			@trace_mapping = trace_mapping || ID_MAPPING
			@connection_edges = {}
			@sync_edges = []
			@applied_obligations = {}
			@hooks_by_buckets = {}
			@next_hook_id = 0
			@hook_calls = {}
		end

		# methods

		def add_method(method)
			unless @methods.has_key?(method)
				@methods[method] = true
				@traces_by_method[method] = {}
				@local_edges_by_method[method] = []
			end
		end

		def each_method(&block)
			@methods.keys.each(&block)
		end

		# traces

		def add_trace(trace, method)
			# remember the trace
			@traces[trace] = true

			# remember this trace belongs to the given method
			@traces_by_method[method][trace] = true

			# add a bucket for the trace if neccessary
			key = @trace_mapping.call(trace)
			unless @buckets.has_key?(key)
				# the implicit entry for nullclass is not represented
				@buckets[key] = {}
			end
		end

		def add_trace_and_initialize(trace, method)
			add_trace(trace, method)
			each_initial_value do |val|
				add_value_to_trace(val, trace)
			end
		end

		def each_trace_for_method(method, &block)
			return unless @traces_by_method.has_key?(method)
			@traces_by_method[method].keys.each do |t|
				block.call(t)
			end
		end

		def each_trace(&block)
			@traces.keys.each do |t|
				block.call(t)
			end
		end

		def has_trace?(trace)
			@traces.has_key?(trace)
		end

		# values

		def add_value_to_trace(val, trace)
			t = get_values(trace)
			t[val] = true
		end

		def values_for_trace(trace)
			get_values(trace).keys
		end

		def each_value_for_trace(trace, &block)
			get_values(trace).keys.each do |v|
				block.call(v)
			end
		end

		def has_value_for_trace?(val, trace)
			get_values(trace).has_key?(val)
		end

		def state(trace)
			get_values(trace)
		end

		# edges

		def add_local_edge(to, method)
			@local_edges_by_method[method] << to
			@local_edges_by_method[method].uniq!
		end

		def each_local_edge_for_method(method, &block)
			return unless @local_edges_by_method.has_key?(method)
			@local_edges_by_method[method].each do |to|
				block.call(to)
			end
		end

#TODO: remove
=begin
		def add_connection_edge(trace, to, to_trace)
			if @connection_edges.has_key?(trace)
				@connection_edges[trace] << [to, to_trace]
				@connection_edges[trace].uniq!
			else
				@connection_edges[trace] = [[to, to_trace]]
			end
		end

		def each_connection_edge_for_trace(trace, &block)
			list = @connection_edges[trace]
			return unless list
			list.each do |to, to_trace|
				block.call(to, to_trace)
			end
		end

		def each_connection_edge(&block)
			@connection_edges.each do |trace, list|
				list.each do |to, to_trace|
					block.call(trace, to, to_trace)
				end
			end
		end
=end

		# hooks

		def add_hook_for_trace(trace, &block)
			id = "hook-#{@var.to_s}-#{@next_hook_id}".to_sym
			@next_hook_id += 1

			key = @trace_mapping.call(trace)
			if @hooks_by_buckets.has_key?(key)
				@hooks_by_buckets[key][id] = block
			else
				@hooks_by_buckets[key] = {id => block}
			end

			return id
		end

		def remove_hook_from_trace(trace, id)
			key = @trace_mapping.call(trace)
			hash = @hooks_by_buckets[key]
			unless hash
				raise RuntimeError.new("cannot remove hook for #{id.inspect} from trace #{trace.inspect}")
			end
			hash.delete(id)
		end

		def each_hook_for_trace(trace, &block)
			key = @trace_mapping.call(trace)
			if @hooks_by_buckets.has_key?(key)
				@hooks_by_buckets[key].each do |id, hook|
					block.call(id, hook)
				end
			end
		end

		def remember_hook_call(val, trace, id)
			key = [val, trace, id]
			@hook_calls[key] = true
		end

		def hook_already_called?(val, trace, id)
			key = [val, trace, id]
			@hook_calls.has_key?(key)
		end

		# obligations

		def each_obligation_for_method(method, &block)
			return unless defined?(@obligations_by_method)
			return unless @obligations_by_method.has_key?(method)
			@obligations_by_method[method].each(&block)
		end

		def add_obligation(method, name, call_id, param_types, return_type)
			@obligations_by_method = {} unless defined?(@obligations_by_method)
			@obligations_by_method[method] = [] unless @obligations_by_method.has_key?(method)
			@obligations_by_method[method] << [name, call_id, param_types, return_type]
		end

		def remember_obligation_application(val, trace, name, call_id)
			key = [val, trace, name, call_id]
			@applied_obligations[key] = true
		end

		def obligation_already_applied?(val, trace, name, call_id)
			key = [val, trace, name, call_id]
			@applied_obligations.has_key?(key)
		end

		# initial values

		# TODO: in the class-polymorphic setting it makes sense to use functions to generate
                #       initial values

		def add_initial_values(values)
			@initial_values = {} unless defined?(@initial_values)
			values.each do |val|
				next if val == NULLCLASS_VALUE
				@initial_values[val] = true
			end
		end

		def each_initial_value(&block)
			return unless defined?(@initial_values)
			@initial_values.keys.each(&block)
		end

private

		def get_values(trace)
			key = @trace_mapping.call(trace)
			@buckets[key]
		end
	end

	class CallAdapter

		def initialize(dyn_method, constraint_graph)
			@dyn_method = dyn_method
			@acyclic_eq_classes = {}
			@acyclic_eq_class_by_callsite = {}
			@constraint_graph = constraint_graph
		end

		# TODO: this class should hold the entire Recursion-Logic + the creation and
                #       linking of equivalence classes

		def add_callsite(eq_class, param_nodes, result_node, dyn_methodcall)

			# cycle detection
			if eq_class.contains_method(@dyn_method)
				# -> recursive call
				# TODO: implement
				raise RuntimeError.new("not implemented, yet")
			else
				# -> non-recursive call
				callsite = CallSite.new(eq_class, param_nodes, result_node, dyn_methodcall)
				handle_non_recursive_call(callsite)
			end
		end

		#TODO: implement cycle propagation

		# a nodes notify adapters of changes in their state

		def on_signature_change(val, node, callsite, param_index)
			ec = find_eq_class(callsite.eq_class, callsite.dyn_methodcall.call_id)
			new_signature = ec.signature.clone
			new_signature[param_index] = node.state(callsite.eq_class)
			if (ec.callsites.size == 1) && !@acyclic_eq_classes.has_key?(new_signature)
				# singular eq_class -> change signature
				@acyclic_eq_classes.delete(ec.signature)
				@acyclic_eq_classes[new_signature] = ec
				ec.signature = new_signature

				# propagate val to formal parameter with index param_index
				@constraint_graph.propagate(ec.param_nodes[param_index], ec, val)
			else
				# -> migrate callsite
				ec.remove_callsite(callsite)
				# also remove result hook
				result_node = @constraint_graph.node_by_variable(@dyn_method.return_type)
				remove_result_hook(callsite, result_node, ec)
				# and the callsite mapping entry
				remove_entry_in_callsite_mapping(callsite, ec)
				# TODO: what about recursive ones?
				handle_non_recursive_call(callsite)
			end
		end

		def remove_result_hook(callsite, result_node, eq_class)
			hook_id = callsite.result_hook_id
			raise RuntimeError.new("no hook id found") unless hook_id
			result_node.remove_hook_from_trace(eq_class, hook_id)
		end

		def find_eq_class(caller_eq_class, call_id)
			@acyclic_eq_class_by_callsite[[caller_eq_class, call_id]]
		end

		def on_result_change(val, method_return_node, called_eq_class, callsite)
			@constraint_graph.propagate(callsite.result_node, callsite.eq_class, val)
		end

		def get_method_param_nodes()
			method_param_nodes = @dyn_method.param_types.map{ |t|
			    @constraint_graph.node_by_variable(t) }
			return method_param_nodes
		end

		def handle_non_recursive_call(callsite)

			signature = callsite.param_nodes.map{ |n| n.state(callsite.eq_class) }
			method_param_nodes = get_method_param_nodes()

			unless callsite.param_nodes.size == method_param_nodes.size
				raise PotentialNullPointerError.new("var_#{var}: in call to #{classname}.#{name}: expected #{method_param_nodes.size} parameters, found #{param_nodes.size}.")
			end

			# TODO: missing case: connected to cycle entry point
			# TODO: missing case: handle in-cycle branching

			# normal acyclic connections
			if @acyclic_eq_classes.has_key?(signature)
				called_eq_class = @acyclic_eq_classes[signature]
				called_eq_class.add_callsite(callsite)
			else
				called_eq_class = create_equivalence_class(callsite, method_param_nodes, signature)
				@acyclic_eq_classes[signature] = called_eq_class
			end
			update_acyclic_eq_class_by_callsite_mapping(callsite, called_eq_class)

			# propagate all values from the result of called_eq_class back
			method_return_node = @constraint_graph.node_by_variable(@dyn_method.return_type)
			hook_result_node(method_return_node, called_eq_class, callsite)
			propagate_all(method_return_node, called_eq_class, callsite.result_node, callsite.eq_class)
		end

		# inverse of remove_entry_in_callsite_mapping(callsite, called_eq_class)
		def update_acyclic_eq_class_by_callsite_mapping(callsite, called_eq_class)
			key = [callsite.eq_class, callsite.dyn_methodcall.call_id]
			if @acyclic_eq_class_by_callsite.has_key?(key)
				raise RuntimeError.new("equivalence class calls method twice with the same call_id")
			end
			@acyclic_eq_class_by_callsite[key] = called_eq_class
		end

		# inverse of update_acyclic_eq_class_by_callsite_mapping(callsite, called_eq_class)
		def remove_entry_in_callsite_mapping(callsite, called_eq_class)
			key = [callsite.eq_class, callsite.dyn_methodcall.call_id]
			if @acyclic_eq_class_by_callsite.has_key?(key)
				@acyclic_eq_class_by_callsite.delete(key)
			else
				raise RuntimeError.new("cannot remove callsite from mapping")
			end
		end

		def create_equivalence_class(callsite, method_param_nodes, signature)
			result = DynMethodCallTraceEquivalenceClass.new(signature, method_param_nodes, @dyn_method)
			result.add_callsite(callsite)

			@constraint_graph.add_trace_to_method(result, callsite.dyn_methodcall.dyn_method)

			# propagate the values of all parameter nodes into the new equivalence class
			param_index = 0
			method_param_nodes.zip(callsite.param_nodes).each do |formal, actual|
				hook_signature_node(actual, callsite, param_index)
				param_index += 1
			end
			method_param_nodes.zip(callsite.param_nodes).each do |formal, actual|
				propagate_all(actual, callsite.eq_class, formal, result)
			end
			return result
		end

		def hook_signature_node(node, callsite, param_index)
			node.add_hook_for_trace(callsite.eq_class) do |val|
				self.on_signature_change(val, node, callsite, param_index)
			end
		end

		def hook_result_node(method_return_node, called_eq_class, callsite)
			hook_id = method_return_node.add_hook_for_trace(called_eq_class) do |val|
				self.on_result_change(val, method_return_node, called_eq_class, callsite)
			end
			callsite.result_hook_id = hook_id
		end

		def propagate_all(from, from_eq_class, to, to_eq_class)
			from.each_value_for_trace(from_eq_class) do |val|
				@constraint_graph.propagate(to, to_eq_class, val)
			end
		end

	end

	def initialize(values, resolver)
		@values = {}
		@inverse_values = {}
		val_id = 0
		values.each do |key, value|
			@values[key] = val_id
			@inverse_values[val_id] = key
			val_id += 1
		end
		# nullclass is handled specially
		@values["nullclass"] = NULLCLASS_VALUE
		@inverse_values[NULLCLASS_VALUE] = "nullclass"
		@num_values = val_id
		@nodes = {}
		@nodes_by_method = {}
		@resolver = resolver
	end

	def add_constraint(c)
		if c[0] == :subseteq
			if c[1].is_a?(Fixnum)	# variable
				insert_propagation_constraint(c[1], c[2], c[3])
			elsif c[1].is_a?(Array)	# set of values
				insert_start_constraint(c[1], c[2])
			else
				raise RuntimeError.new("unknown constraint type found: #{c.inspect}")
			end
		elsif c[0] == :obligation
			c[3..c.size].each do |x|
				unless x[0] == :methodcall
					raise RuntimeError.new("unknown constraint type found: #{c.inspect}")
				end
				insert_method_call_obligation(c[2], c[1], x[1], x[2], x[3], x[4])
			end
		else
			raise RuntimeError.new("unknown constraint type found: #{c.inspect}")
		end
	end

	def insert_propagation_constraint(a, b, method)
		from = node_by_variable(a)
		to = node_by_variable(b)

		# insert local edge
		from.add_local_edge(to, method)

		# propagate values
		from.each_trace_for_method(method) do |t|
			from.each_value_for_trace(t) do |v|
				propagate(to, t, v)
			end
		end
	end

	def insert_start_constraint(a, b)
		# union a to the bitvector of b
		node = node_by_variable(b)
		vals = a.map{ |classname| @values[classname] }
		node.add_initial_values(vals)

		vals.each do |val|
			next if val == NULLCLASS_VALUE
			node.each_trace do |t|
				propagate(node, t, val)
			end
		end
	end

#TODO: remove
=begin
	def insert_connection_constraint(a, a_trace, b, b_trace)
		from = a #from = node_by_variable(a)
		to = b #to = node_by_variable(b)

		# insert connection edge
		from.add_connection_edge(a_trace, b, b_trace)

		# propagate values
		from.each_value_for_trace(a_trace) do |v|
			propagate(to, b_trace, v)
		end
	end
=end

	def insert_method_call_obligation(method, var, name, call_id, param_types, return_type)
		cond_node = node_by_variable(var)
		param_nodes = param_types.map{|t| node_by_variable(t) }
		return_node = node_by_variable(return_type)
		cond_node.add_obligation(method, name, call_id, param_nodes, return_node)

		cond_node.each_trace_for_method(method) do |t|
			insert_method_call_constraints(var, t, NULLCLASS_VALUE, name, call_id, param_nodes, return_node)
			cond_node.each_value_for_trace(t) do |v|
				insert_method_call_constraints(var, t, v, name, call_id, param_nodes, return_node)
			end
		end
	end

	def insert_method_call_constraints(var, trace, value, name, call_id, param_nodes, return_node)
		classname = @inverse_values[value]
		method = @resolver.resolve_method(classname, name)
		unless method
			return if value == NULLCLASS_VALUE
			typeerror("var_#{var}: #{classname} does not support method #{name}")
		end
		adapter = call_adapter_for_method(method)
		begin
			adapter.add_callsite(trace, param_nodes, return_node, DynMethodCall.new(classname, method, call_id))
		rescue PotentialNullPointerError => e
			return if value == NULLCLASS_VALUE
			typeerror(e.message)
		end
	end

	def call_adapter_for_method(dyn_method)
		result = dyn_method.call_adapter
		return result if result
		result = CallAdapter.new(dyn_method, self)
		dyn_method.call_adapter = result
		result
	end

	def add_trace_to_method(new_trace, method)
		# TODO: remove
		# check if the trace is new
		#new_trace = method.minimalize_trace(trace)
		#return new_trace if method.has_trace?(new_trace)

		# DEBUG
		# puts "NEW TRACE for method #{method.name}: #{new_trace.to_s} (#{new_trace.__id__})"

		# add it
		method.add_trace(new_trace)

		# Note: There are 3 ways values can enter a method:
		# - constants
		# - parameters
		# - instance variables
		# while the first 2 could be handled in a straightforward manner, we
		# need to take special care when handling instance variables as these
		# must propagate their values (which have already been propagated) also
		# into the new trace. We do this by making propagate() idempotent and
		# introducing a repropagate() method ensuring the proper propagation of
		# all values
		each_node_for_method(method) do |node|
			# DEBUG
			# puts "initializing node 4 var #{node.var}"
			node.add_trace_and_initialize(new_trace, method)
		end
		# make sure all values are propagated
		each_node_for_method(method) do |node|
			# DEBUG
			# puts "repropagating node 4 var #{node.var}"
			repropagate(node, new_trace)
		end

		return new_trace
	end

	# idempotent
	def propagate(to, trace, val, override_self_check = false)
		# DEBUG
		# puts "propagate(#{to.var}, #{trace}, #{val} (#{@inverse_values[val]}), #{override_self_check})"
		if to.has_value_for_trace?(val, trace)
			return unless override_self_check
		else
			to.add_value_to_trace(val, trace)
		end
		
		method = trace.get_last_method()
		method = @resolver.resolve_method(nil, nil) if method == :root_method

		# propagate along local edges (for the respective method)
		to.each_local_edge_for_method(method) do |peer|
			propagate(peer, trace, val)
		end

		# TODO: remove
		# propagate along connecting edges
		#to.each_connection_edge_for_trace(trace) do |peer, to_trace|
		#	propagate(peer, to_trace, val)
		#end

		# call hooks
		to.each_hook_for_trace(trace) do |id, hook|
			# TODO: think about: would it be better to require hooks to be idempotent?
			next if to.hook_already_called?(val, trace, id)
			to.remember_hook_call(val, trace, id)
			hook.call(val)
		end

		# check obligations
		to.each_obligation_for_method(method) do |name, call_id, param_types, return_type|
			next if to.obligation_already_applied?(val, trace, name, call_id)
			to.remember_obligation_application(val, trace, name, call_id)
			insert_method_call_constraints(to.var, trace, val, name, call_id, param_types, return_type)
		end
	end

	def repropagate(node, trace)
		node.each_value_for_trace(trace) do |val|
			# due to idempotence, propagate can safely be reapplied
			propagate(node, trace, val, true)
		end
	end

	def each_node_for_method(method, &block)
		list = @nodes_by_method[method]
		return unless list
		list.each(&block)
	end

	def node_by_variable(var)
		if @nodes.has_key?(var)
			@nodes[var]
		else
			methods = @resolver.methods_for_var(var)
			trace_mapping = @resolver.trace_mapping_for_var(var)
			node = Node.new(var, methods, trace_mapping)
			@nodes[var] = node
			methods.each do |method|
				if @nodes_by_method.has_key?(method)
					@nodes_by_method[method] << node
				else
					@nodes_by_method[method] = [node]
				end
			end
			node
		end
	end

	def check_safety_constraint(cs, c)
		unless c[0] == :subseteq
			raise RuntimeError.new("invalid constraint type: #{c.inspect}")
		end
		
		unless c[1].is_a?(Fixnum)
			raise RuntimeError.new("invalid constraint type: #{c.inspect}")
		end

		unless c[2].is_a?(Array)
			raise RuntimeError.new("invalid constraint type: #{c.inspect}")
		end

		node = node_by_variable(c[1])
		if c[2][0] == :object
			# TODO: are object types not automatically checked by their method call obligations?
			# object type
			ot = c[2]
			ot[1..ot.size].each do |m|
				unless m[0] == :method
					raise RuntimeError.new("invalid constraint type: #{c.inspect}")
				end
				name = m[1]
				num_params = m[2]
				node.each_trace do |t|
					node.each_value_for_trace(t) do |val|
						classname = @inverse_values[val]
						method = @resolver.resolve_method(classname, name)
						unless method
							var = c[1]
							code = cs.description_for_var(var)
							type = pretty_print_type(c[2])
							raise ConstraintSatisfactionError.new(code, type, var) { |c| "method #{name} is not defined for class #{classname}, but is called on expression #{c}"}
						end
						unless method.param_types.size == num_params
							var = c[1]
							code = cs.description_for_var(var)
							type = pretty_print_type(c[2])
							raise ConstraintSatisfactionError.new(code, type, var) { |c| "method #{classname}.#{name} is expecting #{method.param_types.size}, but is called with #{num_params} on expression #{c}." }
						end
					end
				end
			end
		else
			# set of values
			b_vals = c[2].map {|value| @values[value] }
			node.each_trace do |t|
				node.each_value_for_trace(t) do |val|
					unless b_vals.include?(val)
						var = c[1]
						code = cs.description_for_var(var)
						type = pretty_print_type(c[2])
						raise ConstraintSatisfactionError.new(code, type, var) { |c| "expression #{c} can yield a result not in #{type}" }
					end
				end
			end
		end

	end

	def pretty_print_type(t)
		unless t.is_a?(Array)
			raise RuntimeError.new("expected object type or value set #{t.inspect}")
		end
		if t[0] == :object
			"[#{t[1..t.size].map{ |m| "#{m[1]}(#{m[2]} args)"}.join(', ')}]"
		else
			"{#{t.join(", ")}}"
		end
	end

	def print_solution(cs)
		puts
		puts "Solution:"
		puts
		@nodes.each do |var, node|
			desc = cs.description_for_var(var)
			desc = "<CODE>" unless desc.is_a?(String)
			desc = "#{desc}_#{var.to_s}"
			node.each_method do |m|
				node.each_trace_for_method(m) do |t|
					puts "#{desc}\##{t.to_s}: #{print_type(node, t, @inverse_values)}"
				end
			end
		end
	end

private

	def print_type(node, trace, inv_values)
		classlist = node.values_for_trace(trace).map { |v| inv_values[v] }
		classlist << "nullclass"
		classlist.uniq!
		"{#{classlist.join(", ")}}"
	end

	def print_exp_for_var(cs, var)
		desc = cs.description_for_var(var)
		desc = pretty_print(desc) unless desc.is_a?(String)
		desc = "#{desc}_#{var.to_s}"
	end

	def typeerror(text)
		raise DynTypeError.new("TYPEERROR: " + text)
	end

end

class ConstraintSatisfactionError < RuntimeError

	attr_reader :code, :type, :var

	def initialize(code, type, var, &block)
		super("")
		@code = code
		@type = type
		@var = var
		@format_message = block
	end

	def to_DynTypeError(configuration)
		c = nil
		if @code.is_a?(String)
			c = "#{@code} (variable #{@code}_#{@var})"
		else
			c = configuration.pretty_print_code(@code) + " (variable <CODE>_#{@var})"
		end
		msg = @format_message.call(c)
		raise DynTypeError.new("TYPEERROR: #{msg}")
	end

end

class PotentialNullPointerError < RuntimeError
end

class MethodResolver

	def initialize(configuration, constraint_system, root_method)
		@configuration = configuration
		@constraint_system = constraint_system
		@root_method = root_method
	end

	def resolve_method(classname, methodname)
		return @root_method unless classname
		loc = @configuration.class_by_name[classname]
		return nil unless loc
		clazz = @configuration.store[loc]
		return clazz.dynmethods[methodname]
	end

	def methods_for_var(var)
		@constraint_system.methods_for_var(var)
	end

	CONSTANT_MAPPING = Proc.new { |trace| nil }

	def trace_mapping_for_var(var)
		if @constraint_system.type_of_var(var) == :class_based
			# only one equivalence class
			CONSTANT_MAPPING
		else
			# map tracewise
			nil
		end
	end

end

# extend class DynMethod from dyn_semantics.rb
class DynMethod

  # for type inference
  attr_accessor :return_type, :param_types, :call_adapter

  # traces

  def add_trace(trace)
    @traces = {} unless defined?(@traces)
    @traces[trace] = true
  end

  def has_trace?(trace)
    return false unless defined?(@traces)
    @traces.has_key?(trace)
  end

  def minimalize_trace(trace)
    # TODO: cycle detection for equivalence classes works different
    return trace unless defined?(@traces)
    trace.each_prefix do |prefix|
      return prefix if @traces.has_key?(prefix)
    end
    return trace
  end

end

