#!/usr/bin/ruby

require_relative 'dominance'

class SSATransformer

  class Node

    # to be useful for an SSA transform each CFG node needs
    # - a list of all variables read within it (with their respective statements)
    # - a list of all variables assigned within it (also with their assignment statements)
    # - a program location at the beginning of the code block where it can insert phi functions if neccessary

    attr_reader :ops

    def initialize(cfg, *args)
      @cfg = cfg
      @subtree = []
      while !args.empty? do
        tree = args.shift
        subtree = args.shift
        mode = args.shift || :start
        next unless subtree != nil
        @subtree << [tree, subtree, mode]
      end

      @@counter = 0 unless defined? @@counter
      @@counter += 1
      @no = @@counter
      @ops = []
    end

    def associate_with_subtree(*args)
      if @subtree == nil || @subtree.empty?
        while !args.empty? do
          tree = args.shift
          subtree = args.shift
          mode = args.shift || :start
          next unless subtree != nil
          @subtree << [tree, subtree, mode]
        end
      end
    end

    def addRead(subtree)
      raise RuntimeError.new("variable access expected") unless subtree[0] == :variable
      @ops << [:read, subtree[1], 0, subtree]
      @cfg.add_var(subtree[1])
    end

    def addInstRead(subtree)
      raise RuntimeError.new("instance variable access expected") unless subtree[0] == :instance_variable
      @ops << [:read, "@#{subtree[1]}", 0, subtree]
      @cfg.add_var("@#{subtree[1]}")
    end

    def addAsmt(subtree)
      raise RuntimeError.new("variable assignment expected") unless subtree[0] == :assignment
      var = subtree[1]
      @ops << [:write, var, @cfg.get_asmt_no(var), subtree]
    end

    def addInstAsmt(subtree)
      raise RuntimeError.new("instance variable assignment expected") unless subtree[0] == :instance_assignment
      var = "@#{subtree[1]}"
      @ops << [:write, var, @cfg.get_asmt_no(var), subtree]
    end

    def to_s()
      @no.to_s
    end

    def op_to_str(op)
      if op[0] == :read or op[0] == :write then
        op[0].to_s + " " + op[1].to_s + "_" + op[2].to_s
      elsif op[0] == :psi then
        op[0].to_s + " " + op[1].to_s + "_" + op[2].to_s + " = " + op[3].map{|v| op[1].to_s + "_" + v.to_s}.join(',')
      else
        raise RuntimeError.new("unexpected op #{op.inspect}")
      end
    end

    def contents()
      "{#{to_s()}| " + @ops.map{|op| op_to_str(op)}.join("|") + "}"
    end

    def simulate_execution(state)
      @ops.each do |op|
        var = op[1]
        if op[0] == :write or op[0] == :psi
          version = op[2]
          state[var] = [version]
        elsif op[0] == :read
          # mark the read with all possible versions
          op[2] = state[var]
        else
          raise RuntimeError.new("unexpected op: #{op.inspect}")
        end
      end
      state
    end

    def insert_psi_assignment(var, versions)
      raise RuntimeError.new("encountered unassociated Node in CFG") if @subtree == nil || @subtree.empty?
      new_version = @cfg.get_asmt_no(var)
      # CFG
      @ops.unshift([:psi, var, new_version, versions])
      # program
      @subtree.each do |tree, subtree, mode|
        if mode == :start

          rep = [:statement, [:assignment, [var, new_version], [:hook_psi] + versions.map{|v| [:variable, [var, v]]}], subtree, CodeLoc.generated()]

          substitute(tree, subtree, rep)
        else

	   rep = [:statement, subtree, [:assignment, [var, new_version], [:hook_psi] + versions.map{|v| [:variable, [var, v]]}], CodeLoc.generated()]

          substitute(tree, subtree, rep)
        end
      end
    end

    def substitute(subtree, element, replacement)
      unless substitute_inner(subtree, element, replacement)
        raise RuntimeError.new("could not find element #{element.inspect} to substitute in #{subtree.inspect}")
      end
    end

    def substitute_inner(subtree, element, replacement)
      index = subtree.find_index(element)
      unless index
        found = false
        subtree.each do |elem|
          next unless elem.class.name == "Array"
          next unless elem[0] == :statement
          found ||= substitute_inner(elem, element, replacement)
        end
        return found
      end
      subtree[index] = replacement
      return true
    end

    def substitute_variables()
      @ops.each do |op|
        subtree = op[3]
        if op[0] == :read then
          unless subtree[0] == :variable then
            raise RuntimeError.new("should not happen")
          end
          name = subtree[1]
          raise RuntimeError.new("multi-version read found in what should be SSA form (#{op.inspect})") unless op[2].size == 1
          version = op[2][0]
          subtree[1] = [name, version]
        elsif op[0] == :write then
          raise RuntimeError.new("should not happen") unless subtree[0] == :assignment
          name = subtree[1]
          version = op[2]
          subtree[1] = [name, version]
        elsif op[0] == :psi
          # do nothing
        else
          raise RuntimeError.new("unexpected op found: #{op.inspect}")
        end
      end
    end

  end

  def initialize()
  end

  def construct_CFG(methoddef)
    params = methoddef[2]
    tree = methoddef[3]
    cfg = CFG.new()
    # register parameters as variables
    params.each { |p| cfg.add_var(p) }
    start_node = cfg.add_node(Node.new(cfg, methoddef, tree))
    construct_CFG_inner(cfg, methoddef, tree, start_node)
    cfg
  end

  def construct_CFG_inner(cfg, parent, subtree, currNode)
        case subtree[0]
          when :variable
            currNode.addRead(subtree)
            return currNode
  
          when :instance_variable
            # TODO: make instance variables flow-sensitive
            #currNode.addReadInst(subtree)
            return currNode
  
          when :self
            return currNode

          when :assignment
            var = subtree[1]
            exp = subtree[2]
            currNode = construct_CFG_inner(cfg, subtree, exp, currNode)
            currNode.associate_with_subtree(parent, subtree)
            currNode.addAsmt(subtree)
            return currNode
  
          when :instance_assignment
            var = subtree[1]
            exp = subtree[2]
            currNode = construct_CFG_inner(cfg, subtree, exp, currNode)
            currNode.associate_with_subtree(parent, subtree)
            # TODO: make instance variables flow-sensitive
            #currNode.addAsmtInst(subtree)
            return currNode
  
          when :method_call
            recv = subtree[1]
            name = subtree[2]
            params = subtree[3]
            currNode = construct_CFG_inner(cfg, subtree, recv, currNode)
            currNode.associate_with_subtree(parent, subtree)
            params.each do |param|
              currNode = construct_CFG_inner(cfg, params, param, currNode)
              currNode.associate_with_subtree(parent, subtree)
            end
            return currNode
  
          when :constructor_call
	    classname = subtree[1]
	    params = subtree[2]
            params.each do |param|
              currNode = construct_CFG_inner(cfg, params, param, currNode)
              currNode.associate_with_subtree(parent, subtree)
            end
            return currNode
  
          when :call
            return currNode
  
          when :loc
            return currNode
  
          when :if
            cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]

            currNode = construct_CFG_inner(cfg, subtree, cond, currNode)
            currNode.associate_with_subtree(parent, subtree)
            oldNode = currNode

            true_branch_node = cfg.add_node(Node.new(cfg, subtree, true_branch))
            true_branch_end_node = construct_CFG_inner(cfg, subtree, true_branch, true_branch_node)
            true_branch_end_node.associate_with_subtree(subtree, true_branch, :end)

            false_branch_node = cfg.add_node(Node.new(cfg, subtree, false_branch))
            false_branch_end_node = construct_CFG_inner(cfg, subtree, false_branch, false_branch_node)
            false_branch_end_node.associate_with_subtree(subtree, false_branch, :end)

            currNode = cfg.add_node(Node.new(cfg, nil))
            cfg.add_edge(oldNode, true_branch_node)
            cfg.add_edge(oldNode, false_branch_node)
            cfg.add_edge(true_branch_end_node, currNode)
            cfg.add_edge(false_branch_end_node, currNode)
            return currNode
            
          when :ifPS
	    cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]

            currNode = construct_CFG_inner(cfg, subtree, cond, currNode)
            currNode.associate_with_subtree(parent, subtree)
            oldNode = currNode

            true_branch_node = cfg.add_node(Node.new(cfg, subtree, true_branch))
            true_branch_end_node = construct_CFG_inner(cfg, subtree, true_branch, true_branch_node)
            true_branch_end_node.associate_with_subtree(subtree, true_branch, :end)

            false_branch_node = cfg.add_node(Node.new(cfg, subtree, false_branch))
            false_branch_end_node = construct_CFG_inner(cfg, subtree, false_branch, false_branch_node)
            false_branch_end_node.associate_with_subtree(subtree, false_branch, :end)

            currNode = cfg.add_node(Node.new(cfg, nil))
            cfg.add_edge(oldNode, true_branch_node)
            cfg.add_edge(oldNode, false_branch_node)
            cfg.add_edge(true_branch_end_node, currNode)
            cfg.add_edge(false_branch_end_node, currNode)
            return currNode

          when :while
            cond = subtree[1]
            loop = subtree[2]

	    # create a loop header
            header_node = cfg.add_node(Node.new(cfg, parent, subtree, :start, subtree, loop, :end))
            header_end_node = construct_CFG_inner(cfg, subtree, cond, header_node)
            header_end_node.associate_with_subtree(subtree, loop)

            body_node = cfg.add_node(Node.new(cfg, subtree, loop))
            body_end_node = construct_CFG_inner(cfg, subtree, loop, body_node)
            body_end_node.associate_with_subtree(subtree, loop, :end)

            old_node = currNode
            currNode = cfg.add_node(Node.new(cfg, nil)) # end of loop
            cfg.add_edge(old_node, header_node)
            cfg.add_edge(header_end_node, currNode)
            cfg.add_edge(header_end_node, body_node)
            cfg.add_edge(body_end_node, header_node)
            return currNode

          when :statement
            subtree[1..subtree.size()-2].each do |stmt|
              currNode.associate_with_subtree(subtree, stmt)
              currNode = construct_CFG_inner(cfg, subtree, stmt, currNode)
            end
            return currNode

          when :error
            return currNode

          # --- builtin functionality ---

          when :hook_loc_identical
            a = subtree[1]
            b = subtree[2]
            currNode = construct_CFG_inner(cfg, subtree, a, currNode)
            currNode = construct_CFG_inner(cfg, subtree, b, currNode)
            return currNode

          when :hook_get_loc
            exp = subtree[1]
            currNode = construct_CFG_inner(cfg, subtree, exp, currNode)
            return currNode

          when :hook_is_a
            exp = subtree[1]
            currNode = construct_CFG_inner(cfg, subtree, exp, currNode)
            return currNode

          when :hook_cast
            type = subtree[1]
            exp = subtree[2]
            currNode = construct_CFG_inner(cfg, subtree, exp, currNode)
            return currNode

          when :assertion
	    return currNode

          else
            syserror("found unknown node in parsetree: #{subtree.inspect}")
        end
  end

  def transform_assertions(subtree)
        case subtree[0]
          when :variable, :instance_variable, :self, :call, :loc,
               :error, :hook_loc_identical, :hook_get_loc, :hook_is_a, :hook_cast
            return

          when :assignment
            exp = subtree[2]
            transform_assertions(exp)
            return
  
          when :instance_assignment
            exp = subtree[2]
            transform_assertions(exp)
            return
  
          when :method_call
            recv = subtree[1]
            params = subtree[3]
            transform_assertions(recv)
            params.each do |param|
              transform_assertions(param)
            end
            return
  
          when :constructor_call
	    params = subtree[2]
            params.each do |param|
              transform_assertions(param)
            end
            return
  
          when :if
            cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]
            transform_assertions(cond)
            transform_assertions(true_branch)
            transform_assertions(false_branch)
            return
            
          when :ifPS
            cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]
            transform_assertions(cond)
            transform_assertions(true_branch)
            transform_assertions(false_branch)
            return

          when :while
            cond = subtree[1]
            loop = subtree[2]
            transform_assertions(cond)
            transform_assertions(loop)
            return

          when :statement
            subtree[1..subtree.size()-2].each do |stmt|
              transform_assertions(stmt)
            end
            return

          # --- builtin functionality ---

          when :hook_loc_identical
            a = subtree[1]
            b = subtree[2]
            transform_assertions(a)
            transform_assertions(b)
            return

          when :hook_get_loc
            exp = subtree[1]
            transform_assertions(exp)
            return

          when :hook_is_a
            exp = subtree[1]
            transform_assertions(exp)
            return

          when :hook_cast
            exp = subtree[2]
            transform_assertions(exp)
            return

          when :assertion
	    # DEBUG-SWITCH
	    #puts "DEBUG: transforming assertion #{subtree.inspect}"
	    assertion = subtree[1]
	    exp = transform_assertion_to_cast(assertion)
	    #puts "DEBUG: to expression #{exp.inspect}"
	    substitute_subexp(subtree, exp)
            return

          else
            syserror("found unknown node in parsetree: #{subtree.inspect}")
        end
  end

  def transform_to_DNF(assertion)
    convert(assertion)
  end

  def conjoin(l, r)
    if l[0] == :conjunction
      l = l[1..l.size()]
    else
      l = [l]
    end
    if r[0] == :conjunction
      r = r[1..r.size()]
    else
      r = [r]
    end
    [:conjunction] + l + r
  end

  def disjoin(*args)
    args = args.map do |arg|
      if arg[0] == :disjunction
        arg[1..arg.size()]
      else
        [arg]
      end
    end
    [:disjunction] + args.inject(:+)
  end

  def convert(formula)
    case formula[0]
    when :type_literal
      return [:disjunction, [:conjunction, formula]]
    when :conjunction
      p = convert(formula[1])  # in DNF
      raise RuntimeError.new("not supposed to happen") unless p[0] == :disjunction
      q = convert(formula[2])  # in DNF
      raise RuntimeError.new("not supposed to happen") unless q[0] == :disjunction
      # (p1 \/ ... \/ pn) /\ (q1 \/ ... \/ qn) = (p1 /\ q1) \/ (p1 /\ q2) \/ ... \/ (pn /\ qn)
      res = [:disjunction]
      p[1..p.size()].each do |pc|
        q[1..q.size()].each do |qc|
          res << conjoin(pc, qc)
        end
      end
      return res
    when :disjunction
      return disjoin(*formula[1..formula.size()-2].map{|x| convert(x)})
    when :negation
      case formula[1][0]
      when :type_literal
        return formula
      when :negation
        # double negation
        return convert(formula[1][1])
      when :conjunction
        # de morgan: !(a /\ b) = !a \/ !b
        return convert([:disjunction] + formula[1][1..formula[1].size].map{|x| [:negation, x]})
      when :disjunction
        # de morgan: !(a \/ b) = !a /\ !b
        return convert([:conjunction] + formula[1][1..formula[1].size].map{|x| [:negation, x]})
      else
        raise RuntimeError.new("encountered unknown element in formula: #{formula[1].inspect}")
      end
    else
      raise RuntimeError.new("encountered unknown element in formula: #{formula.inspect}")
    end
  end

  def transform_assertion_to_cast(assertion)
    dnf = transform_to_DNF(assertion)
    raise RuntimeError.new("should not happen") unless dnf[0] == :disjunction
    dnf[1..dnf.size()].map do |disjunct|
      raise RuntimeError.new("should not happen") unless disjunct[0] == :conjunction
      [:statement] + disjunct[1..disjunct.size()].map do |conjunct|
        raise RuntimeError.new("expected only type literals") unless conjunct[0] == :type_literal
        varname = conjunct[1]
        type = conjunct[2]
        raise RuntimeError.new("expected a union type") unless type[0] == :uniontype
        [:assignment, varname, [:hook_cast, type, [:variable, varname], CodeLoc.generated()], CodeLoc.generated()]
      end + [CodeLoc.generated()]
    end.inject{|x,y| [:ifPS, [:constructor_call, "boolean", [[:loc, 0]]], x, y, CodeLoc.generated()]}
  end

  def substitute_subexp(from, to)
    while !from.empty?() do
      from.delete_at(0)
    end
    to.each do |item|
      from << item
    end
  end

  def syserror(text)
      raise RuntimeError.new("SYSERROR: " + text)
  end

  def transform(parse_tree)
    syserror("given parse tree is not a dyn program") unless parse_tree[0] == :program

    parse_tree[1..parse_tree.size()].each do |classdef|
      next unless classdef.class.name == "Array"
      syserror("not a classdef") unless classdef[0] == :classdef
      
      classdef[3].each do |methoddef|
	next if methoddef[0] == :rename
        syserror("not a methoddef: #{methoddef.inspect}") unless methoddef[0] == :method
	params = methoddef[2]
	# TODO: transform ALL method bodies to SSA form
        #if classdef[1] == "main" && methoddef[1] == "run"
          # first transform all typing assertions into casts
          transform_assertions(methoddef[3])

          # then construct the control flow graph
          cfg = construct_CFG(methoddef)
          cfg.check_for_uninitialized_var_access(params, methoddef[1])

          cfg.insert_psi_assignments(params)

	  # DEBUG-SWITCH
          #File.open("cfg/#{classdef[1]}.#{methoddef[1]}.dot", "w") do |f|
          #   f.puts cfg.to_dot()
          #end
          cfg.substitute_variables()

        #end

        # transform to SSA
        # 0. preparation: number all assignments
        # 1. insert psi-functions
        #    - Idea: take all nodes containing a write to a variable x,
        #      get their dominance frontiers
        #    - in each node in the dominance frontier of an assignment to x_i,
        #      insert a psi-function x_{n+1} = psi(x_i, x_j) where n is the number of versions of x and
        #      x_j is the version x would have without the assignment to x_i (how to find x_j?)
        # 2. substitute variables by their versions

      end
    end
    parse_tree
  end

end
