#!/usr/bin/ruby

def run_program(program, params = "")
  comm = "ruby interpreter/interpreter.rb #{params} programs/#{program} 2> /dev/null"

  result = `#{comm}`
  success = $?.success?
  [result, success]
end

def test(program, expected, params = "--typecheck")
  # DEBUG
  puts "#{Time.now} - testing #{program} with params #{params} ..."

  result, success = run_program(program, params)
  unless success
    puts "Program #{program} failed unexpectedly. Output:"
    puts result
    raise RuntimeError.new("program #{program} failed!")
  end
  actual = result.split("\n")[2]
  unless expected === actual
    puts "expected: #{expected.inspect}"
    puts "actual: #{actual.inspect}"
    raise RuntimeError.new("Test failed")
  end
end

def expect_fatal_error(program, error, params = "")
  puts "#{Time.now} - testing #{program} with params #{params} and expecting fatal error..."
  
  result, success = run_program(program, params)
  if success
    puts "Program #{program} succeeded unexpectedly. Output:"
    puts result
    raise RuntimeError.new("program #{program} succeeded unexpectedly")
  end
  actual = result.split("\n")[0]
  unless error === actual
    puts "expected: #{error.inspect}"
    puts "actual: #{actual.inspect}"
    raise RuntimeError.new("Wrong error")
  end
end

def expect_error(program, error, params = "")
  # DEBUG
  puts "#{Time.now} - testing #{program} with params #{params} and expecting errors..."

  result, success = run_program(program, params)
  if success
    puts "Program #{program} succeeded unexpectedly. Output:"
    puts result
    raise RuntimeError.new("program #{program} succeeded unexpectedly")
  end
  actual = result.split("\n")[3]
  unless error === actual
    puts "expected: #{error.inspect}"
    puts "actual: #{actual.inspect}"
    raise RuntimeError.new("Wrong error")
  end
end

test("test.dyn", "3")
test("strings.dyn", "hi")
test("terms.dyn", "success")
# test("primes.dyn", "[2, 3, 5, 7]")    # TODO: with class polymorphism, it runs until it runs out of memory
test("precedence.dyn", "6")
test("inheritance.dyn", "Hello World")
test("coerce.dyn", "1 mod 3", "") # without typecheck
test("is_a.dyn", "1")
test("minimal.dyn", "null")

expect_fatal_error("errornous/circular_dependencies.dyn", "ERROR: circular dependencies: a,b,a")
expect_fatal_error("errornous/class_defined_twice.dyn", "ERROR: duplicate class a")
expect_fatal_error("errornous/class_extends_itself.dyn", "ERROR: circular dependencies: main,main")
expect_fatal_error("errornous/error.dyn", "ERROR: an especially long and gruesome error")

expect_fatal_error("errornous/new_object.dyn", "TYPEERROR: unknown method \"init\" called on object @ loc(104) [of class object]")
expect_fatal_error("errornous/typeerror.dyn", "TYPEERROR: unknown method \"dosomething\" called on object @ loc(106) [of class numeric]")
expect_fatal_error("errornous/uses_loc.dyn", /^Expected one of/)
# type checking / inference
expect_fatal_error("errornous/typeerror.dyn", /TYPEERROR:\ method\ dosomething\ is\ not\ defined\ for\ class\ numeric,\ but\ is\ called\ on\ expression\ .*/, "--typecheck")
test("type-unknown.dyn", "null")
# NOTE: this is a runtime error, but not a typeerror. errors caused by null values are not considered typeerrors.
expect_fatal_error("errornous/typeerror-minimal.dyn", "ERROR: unknown method \"dosomething\" called on null value", "--typecheck")

expect_fatal_error("errornous/typeerror-minimal2.dyn", /TYPEERROR:\ method\ dosomething\ is\ not\ defined\ for\ class\ B,\ but\ is\ called\ on\ expression\ .*/, "--typecheck")

expect_error("errornous/typeerror-minimal3.dyn", "TYPEERROR: class B does not have a method init", "--typecheck")
test("typing-problem-null.dyn", "true")
test("type_me.dyn", "1")

expect_fatal_error("errornous/typeerror-minimal4.dyn", /TYPEERROR:\ method\ inner\ is\ not\ defined\ for\ class\ numeric,\ but\ is\ called\ on\ expression\ .*/, "--typecheck")
expect_fatal_error("errornous/typeerror-minimal5.dyn", /TYPEERROR:\ method\ inner\ is\ not\ defined\ for\ class\ numeric,\ but\ is\ called\ on\ expression\ .*/, "--typecheck")

# TODO: SSA transform masks this kind of error -> use it as a better detector for uninitialized var access
expect_fatal_error("errornous/typeerror-optional-var.dyn", "TYPEERROR: detected potential uninitialized access to local variable \"a\" in method run", "--typecheck")

test("optional-var-branching.dyn", "2")


expect_fatal_error("errornous/typeerror-optional-instance-var.dyn", "TYPEERROR: class main: instance variables var are not initialized in constructor prior to the first access to self.", "--typecheck")
expect_fatal_error("errornous/typeerror-instance-var-access-before-init.dyn", "TYPEERROR: class main: instance variables obj are not initialized in constructor prior to the first access to self.", "--typecheck")
expect_fatal_error("errornous/typeerror-instance-var-access-before-init2.dyn", "TYPEERROR: class main: instance variables bar are not initialized in constructor prior to the first access to self.", "--typecheck")
expect_fatal_error("errornous/main_without_run.dyn", /^TYPEERROR:\ method\ run\ is\ not\ defined\ for\ class\ main,\ but\ is\ called\ on\ expression\ .*/, "--typecheck")
expect_fatal_error("errornous/toString_does_not_return_a_string.dyn", /^TYPEERROR:\ expression\ result\ \(variable result_\d*\)\ can\ yield\ a\ result\ not\ in\ \{string\}$/, "--typecheck")
#expect_error("errornous/typeerror-instance-var.dyn", "TYPEERROR: var_3435: numeric does not support method append", "--typecheck")
test("type_me_large.dyn", "1")

test("while-loop.dyn", "null")

# polymorphic typechecking
#test("polymorphic/polymorphic_method.dyn", "foobar")
#test("polymorphic/polymorphic_submethod.dyn", "2foofoo4barbar")
#test("polymorphic/exponential.dyn", "256")

expect_fatal_error("errornous/test2.dyn", /TYPEERROR:\ method\ holla\ is\ not\ defined\ for\ class\ numeric,\ but\ is\ called\ on\ expression\ .*/, "--typecheck")

# flow sensitivity
test("flow_sensitive/test.dyn", "4")

# path sensitivity (with assertions)
test("path_sensitive/addition+assertion.dyn", "foobar")
test("path_sensitive/implication+assertion.dyn", "foobar")

# context sensitivity (with assertions)
test("context-sensitivity/addition.dyn", "foobar")

# object sensitivity (with assertions)
test("object-sensitivity/addition.dyn", "foobar")

# case studies
test("paper-example.dyn", "13")

