﻿namespace Dyn.Console
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using Dyn.Console.CommandLine;
    using Dyn.Diagnostics;
    using Dyn.Lexer;

    /// <summary>
    /// Represents the state of the program including <see cref="DiagnosticEventArgs"/> event listeners and interpreted
    /// command line arguments.
    /// </summary>
    internal class Arguments : IDiagnostic
    {
        private readonly List<SyntaxToken> _tokens = new List<SyntaxToken>();
        private readonly List<string> _files = new List<string>();

        /// <summary>
        /// Initializes a new instance of the <see cref="Arguments"/> class.
        /// </summary>
        public Arguments()
        {
            this.SolverPath = string.Empty;
        }

        /// <inheritdoc/>
        public event EventHandler<DiagnosticEventArgs> DiagnosticInfo;

        /// <summary>Gets a list containing <see cref="SyntaxToken"/> of all source files.</summary>
        public List<SyntaxToken> Tokens
        {
            get
            {
                return this._tokens;
            }
        }

        /// <summary>Gets or sets a value indicating whether expecting user input or only source files.</summary>
        [Option(LongName = "interactive", ShortName = 'i')]
        public bool Interactive { get; set; }

        /// <summary>Gets or sets a SMT solver path.</summary>
        [Option(LongName = "solver", ShortName = 'z')]
        public string SolverPath { get; set; }

        /// <summary>Gets or sets a value indicating whether this program writes summaries.</summary>
        [Option(LongName = "show-summary", ShortName = 's')]
        public bool ShowSummary { get; set; }

        /// <summary>Gets or sets a value indicating whether this program writes source code highlighted.</summary>
        [Option(LongName = "show-highlighted", ShortName = 'h')]
        public bool ShowColors { get; set; }

        /// <summary>Gets or sets a value indicating whether this program verifies the input.</summary>
        [Option(LongName = "verify", ShortName = 'v')]
        public bool Verify { get; set; }

        /// <summary>Gets or sets a value indicating whether this program writes conditions.</summary>
        [Option(LongName = "show-conditions", ShortName = 'c')]
        public bool ShowConditions { get; set; }

        /// <summary>Gets or sets a value indicating whether this program writes abstract syntax trees.</summary>
        [Option(LongName = "show-ast", ShortName = 'a')]
        public bool ShowAst { get; set; }

        /// <summary>Gets or sets a value indicating whether this program writes help information.</summary>
        [Option(LongName = "help", ShortName = '?')]
        public bool ShowHelp { get; set; }

        /// <summary>Gets a list of all source files.</summary>
        public ICollection<string> Files
        {
            get
            {
                return this._files;
            }
        }

        /// <summary>Adds an file to the <seealso cref="Files"/> collection.</summary>
        /// <param name="fileName">The file name.</param>
        [Option(IsArgument = true)]
        public void AddFile(string fileName)
        {
            this._files.Add(fileName);
        }

        /// <inheritdoc/>
        public void DiagnosticInfoListener(object sender, DiagnosticEventArgs args)
        {
            var handler = this.DiagnosticInfo;

            if (handler != null)
            {
                handler(sender, args);
            }
        }

        /// <summary>Validates the input</summary>
        /// <param name="err">The <see cref="TextWriter"/> for error messages.</param>
        /// <returns> Currently <code>true</code>./ </returns>
        public bool Validate(TextWriter err)
        {
            return true;
        }
    }
}
