﻿namespace Dyn.Console
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using Dyn.Diagnostics;
    using Dyn.Lexer;

    /// <summary>Prints and colorizes <see cref="SyntaxToken"/> on the console.</summary>
    public static class ColoredFormatter
    {
        /// <summary>Returns a <see cref="Console.ForegroundColor"/> representing <paramref name="kind"/>.</summary>
        /// <param name="kind">The kind.</param>
        /// <param name="foreground">Default foreground color.</param>
        /// <param name="background">Default background color.</param>
        public static void ColorByKind(
            SyntaxKind kind,
            ref ConsoleColor foreground,
            ref ConsoleColor background)
        {
            switch (kind)
            {
                // Whitespace, Operators and Symbols are normal
                case SyntaxKind.AssignmentOperator:
                case SyntaxKind.EqualityOperator:
                case SyntaxKind.IdentityOperator:
                case SyntaxKind.AdditionOperator:
                case SyntaxKind.SubtractionOperator:
                case SyntaxKind.MultiplicationOperator:
                case SyntaxKind.DivisionOperator:
                case SyntaxKind.LessThanOperator:
                case SyntaxKind.GreaterThanOperator:
                case SyntaxKind.LessThanOrEqualOperator:
                case SyntaxKind.GreaterThanOrEqualOperator:
                case SyntaxKind.ModulosOperator:
                case SyntaxKind.AndOperator:
                case SyntaxKind.OrOperator:
                case SyntaxKind.ImpliesOperator:
                case SyntaxKind.ParenthesisLeftSymbol:
                case SyntaxKind.ParenthesisRightSymbol:
                case SyntaxKind.SquareBracketLeftSymbol:
                case SyntaxKind.SquareBracketRightSymbol:
                case SyntaxKind.CurlyBracketLeftSymbol:
                case SyntaxKind.CurlyBracketRightSymbol:
                case SyntaxKind.AtSymbol:
                case SyntaxKind.DotSymbol:
                case SyntaxKind.CommaSymbol:
                case SyntaxKind.SemicolonSymbol:
                case SyntaxKind.QuestionMarkSymbol:
                case SyntaxKind.ExclamationMarkSymbol:
                case SyntaxKind.ColonSymbol:
                case SyntaxKind.ApostropheSymbol:
                case SyntaxKind.EndOfStreamKind:

                case SyntaxKind.WhitespaceKind:
                    // use default settings
                    break;

                case SyntaxKind.CommentKind:
                    foreground = ConsoleColor.Green;
                    break;

                case SyntaxKind.SelfKeyword:
                case SyntaxKind.NewKeyword:
                case SyntaxKind.IfKeyword:
                case SyntaxKind.ThenKeyword:
                case SyntaxKind.ElseKeyword:
                case SyntaxKind.EndKeyword:
                case SyntaxKind.WhileKeyword:
                case SyntaxKind.DoKeyword:
                case SyntaxKind.DoneKeyword:
                case SyntaxKind.MethodKeyword:
                case SyntaxKind.RenameKeyword:
                case SyntaxKind.NullKeyword:
                case SyntaxKind.ClassKeyword:
                case SyntaxKind.IsAKeyword:
                case SyntaxKind.TrueKeyword:
                case SyntaxKind.FalseKeyword:

                case SyntaxKind.RequiresKeyword:
                case SyntaxKind.EnsuresKeyword:
                case SyntaxKind.InvariantKeyword:
                case SyntaxKind.AssertKeyword:
                case SyntaxKind.ByTheWayKeyword:

                case SyntaxKind.ForAllKeyword:
                case SyntaxKind.ExistsKeyword:

                case SyntaxKind.InKeyword:
                    foreground = ConsoleColor.Blue;
                    break;

                case SyntaxKind.NumericKind:
                    foreground = ConsoleColor.Cyan;
                    break;

                case SyntaxKind.StringKind:
                    foreground = ConsoleColor.Magenta;
                    break;

                case SyntaxKind.IdentifierKind:
                    foreground = ConsoleColor.Yellow;
                    break;

                case SyntaxKind.InvalidKind:
                default:
                    foreground = ConsoleColor.White;
                    background = ConsoleColor.DarkRed;
                    break;
            }
        }

        /// <summary>Prints a source code representation of <paramref name="tokens"/> on the console.</summary>
        /// <param name="tokens">The tokens.</param>
        /// <param name="writer">The destination.</param>
        public static void PrintTokenListHighlighted(IEnumerable<SyntaxToken> tokens, TextWriter writer)
        {
            Guard.EnsureNotNull("writer", writer);

            var defaultForground = Console.ForegroundColor;
            var defaultBackground = Console.BackgroundColor;

            foreach (var token in tokens)
            {
                var currentForground = defaultForground;
                var currentBackground = defaultBackground;

                ColorByKind(token.Kind, ref currentForground, ref currentBackground);

                Console.ForegroundColor = currentForground;
                Console.BackgroundColor = defaultBackground;

                writer.Write(token.ToSource());
            }

            Console.ForegroundColor = defaultForground;
            Console.BackgroundColor = defaultBackground;
        }

        /// <summary>Returns a <see cref="ConsoleColor"/> representing <paramref name="level"/> or <paramref name="defaultColor"/>.</summary>
        /// <param name="level">The severity.</param>
        /// <param name="defaultColor">The default color.</param>
        /// <returns>A <see cref="ConsoleColor"/> representing <paramref name="level"/> or <paramref name="defaultColor"/>.</returns>
        public static ConsoleColor ConsoleColorFromSeverity(DiagnosticSeverity level, ConsoleColor defaultColor)
        {
            switch (level)
            {
                case DiagnosticSeverity.Information:
                    return ConsoleColor.DarkCyan;

                case DiagnosticSeverity.Warning:
                    return ConsoleColor.Yellow;

                case DiagnosticSeverity.Error:
                    return ConsoleColor.Red;

                case DiagnosticSeverity.None:
                default:
                    return defaultColor;
            }
        }
    }
}
