﻿namespace Dyn.Console.CommandLine
{
    using System;

    /// <summary>Provides methods for reading options and parameters from an <see cref="string"/> array.</summary>
    internal sealed class CommandLineLexer
    {
        private readonly string[] _args;
        private int argIndex = 0;
        private int charPos = 0;
        private bool argumentMode = false;

        /// <summary>
        /// Initializes a new instance of the <see cref="CommandLineLexer"/> class with command line arguments.
        /// </summary>
        /// <param name="args">The command line arguments.</param>
        public CommandLineLexer(string[] args)
        {
            Guard.EnsureNotNull("args", args);
            Guard.EnsureElementNotNull("args", args);

            this._args = args;
            this.Wrap();
        }

        /// <summary>
        /// Gets a value indicating whether the command line arguments are completely handled.
        /// </summary>
        public bool IsCompleted
        {
            get
            {
                return this.argIndex == this._args.Length;
            }
        }

        private void Wrap()
        {
            // skip if completed
            if (this.argIndex >= this._args.Length)
            {
                return;
            }

            var a = this._args[this.argIndex];

            // wrap
            if (this.charPos > 0 && this.charPos >= a.Length)
            {
                this.argIndex++;
                this.charPos = 0;
                this.Wrap();
                return;
            }

            // switch to argument mode if --
            if (!this.argumentMode &&
                this.charPos == 0 &&
                "--".Equals(a, StringComparison.Ordinal))
            {
                this.argumentMode = true;
                this.argIndex++;
                this.charPos = 0;
            }
        }

        /// <summary>Tries to read the next argument as a parameter.</summary>
        /// <param name="result">The parameter.</param>
        /// <returns><code>true</code>, if successful; otherwise <code>false</code>.</returns>
        public bool TryParameter(out string result)
        {
            result = default(string);

            if (this.IsCompleted)
            {
                return false;
            }

            var p = this.charPos;
            var a = this._args[this.argIndex];

            if (p == 0)
            {
                result = a;
                this.argIndex++;
                this.charPos = 0;
                this.Wrap();
                return true;
            }

            if (a[p] == '=' || a[p] == ':')
            {
                p++;
            }

            result = a.Substring(p);
            this.argIndex++;
            this.charPos = 0;
            this.Wrap();
            return true;
        }

        /// <summary>Tries to read the next argument as an argument.</summary>
        /// <param name="result">The argument.</param>
        /// <returns><code>true</code>, if successful; otherwise <code>false</code>.</returns>
        public bool TryArgument(out string result)
        {
            result = default(string);

            if (this.IsCompleted)
            {
                return false;
            }

            if (this.charPos != 0)
            {
                return false;
            }

            result = this._args[this.argIndex];
            this.argIndex++;
            this.charPos = 0;
            this.Wrap();
            return true;
        }

        /// <summary>Tries to read the next argument as an option in short form.</summary>
        /// <param name="result">The option.</param>
        /// <returns><code>true</code>, if successful; otherwise <code>false</code>.</returns>
        public bool TryShort(out char result)
        {
            result = default(char);

            if (this.argumentMode || this.IsCompleted)
            {
                return false;
            }

            var p = this.charPos;
            var a = this._args[this.argIndex];

            if (p == 0 &&
                a.Length > 1 &&
                a[0] == '-' &&
                a[1] != '-')
            {
                result = a[p + 1];
                this.charPos = p + 2;
                this.Wrap();
                return true;
            }
            else if (p != 0 &&
                p < a.Length &&
                a[p] != ':' &&
                a[p] != '=')
            {
                result = a[p];
                this.charPos = p + 1;
                this.Wrap();
                return true;
            }

            return false;
        }

        /// <summary>Tries to read the next argument as an option in long form.</summary>
        /// <param name="result">The option.</param>
        /// <returns><code>true</code>, if successful; otherwise <code>false</code>.</returns>
        public bool TryLong(out string result)
        {
            result = default(string);

            if (this.argumentMode || this.IsCompleted)
            {
                return false;
            }

            var p = this.charPos;
            var a = this._args[this.argIndex];

            if (p == 0 &&
                a.Length >= 3 &&
                a[0] == '-' &&
                a[1] == '-' &&
                a[2] != '-')
            {
                var par = a.IndexOfAny(new char[] { '=', ':' }, 3);

                result = par >= 0 ? a.Substring(2, par - 2) : a.Substring(2);
                this.charPos = par >= 0 ? par : a.Length;
                this.Wrap();
                return true;
            }

            return false;
        }
    }
}
