﻿namespace Dyn.Console.CommandLine
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Reflection;
    using System.Text;

    using Invoker = System.Action<CommandLineLexer, object>;

    /// <summary>
    /// Provides the methods interacting with the command line.
    /// </summary>
    /// <typeparam name="T">The type of a class with at least one <see cref="OptionAttribute"/>.</typeparam>
    public static class CommandLineParser<T>
    {
        private static readonly IDictionary<OptionAttribute, Invoker> _options = GetOptions(typeof(T));

        /// <summary>
        /// Inspects <paramref name="type"/> for members annotated with <see cref="OptionAttribute"/> and returns their
        /// invocation delegates.
        /// </summary>
        /// <param name="type">The type to inspect.</param>
        /// <returns>
        /// A <see cref="IDictionary{OptionAttribute, Invoker}"/> mapping <see cref="OptionAttribute"/> instances to
        /// their owners normalized invocation delegates.
        /// </returns>
        private static IDictionary<OptionAttribute, Invoker> GetOptions(Type type)
        {
            var result = new Dictionary<OptionAttribute, Invoker>();
            Invoker arguments = null;

            HashSet<char> shorts = new HashSet<char>();
            HashSet<string> longs = new HashSet<string>(StringComparer.Ordinal);

            foreach (var info in type.GetTypeInfo().GetMembers())
            {
                if (!info.IsDefined(typeof(OptionAttribute)))
                {
                    continue;
                }

                Invoker memberInvoker = ReflectionInvoker.Create(info);

                foreach (var attr in info.GetCustomAttributes<OptionAttribute>(true))
                {
                    result.Add(attr, memberInvoker);

                    if (attr.ShortName != '\0' && !shorts.Add(attr.ShortName))
                    {
                        throw new OptionDefinitionException(
                            string.Format(
                                "Option -{0} already exists.",
                                attr.ShortName));
                    }

                    if (!string.IsNullOrWhiteSpace(attr.LongName) && !longs.Add(attr.LongName))
                    {
                        throw new OptionDefinitionException(
                            string.Format(
                                "Option --{0} already exists.",
                                attr.LongName));
                    }

                    if (attr.IsArgument)
                    {
                        if (arguments != null)
                        {
                            throw new OptionDefinitionException("Argument option redefined.");
                        }

                        arguments = memberInvoker;
                    }
                }
            }

            return result;
        }

        private static Invoker GetInvoker(
            IDictionary<OptionAttribute, Invoker> source,
            Func<OptionAttribute, bool> predicate,
            string name)
        {
#if DEBUG
            var result = source.SingleOrDefault(e => predicate(e.Key));

            if (result.Value == null)
            {
                throw new ArgumentParserException("Cannot resolve " + name + ".");
            }

            return result.Value;
#else
            return source.Single(e => predicate(e.Key)).Value;
#endif
        }

        /// <summary>
        /// Interprets command line arguments of <paramref name="args"/> and sets them into <paramref name="instance"/>
        /// </summary>
        /// <param name="args">The source.</param>
        /// <param name="instance">The members of this instance will be assigned to the values from <paramref name="args"/>.</param>
        public static void Parse(string[] args, T instance)
        {
            Guard.EnsureNotNull("args", args);
            Guard.EnsureNotNull("instance", instance);

            var lexer = new CommandLineLexer(args);

            while (!lexer.IsCompleted)
            {
                char ch;
                string str;
                Invoker invoker = null;

                if (lexer.TryShort(out ch))
                {
                    invoker = GetInvoker(
                        _options,
                        e => e.ShortName == ch,
                        "option -" + ch);
                }
                else if (lexer.TryLong(out str))
                {
                    invoker = GetInvoker(
                        _options,
                        e => StringComparer.Ordinal.Equals(e.LongName, str),
                        "option --" + str);
                }
                else
                {
                    invoker = GetInvoker(
                        _options,
                        e => e.IsArgument,
                        "the argument");
                }

                invoker(lexer, instance);
            }
        }

        /*
        /// <summary>
        /// Builds a help text for the options defined in type <typeparamref name="T"/>.
        /// </summary>
        /// <returns>A help text for the options defined in type <typeparamref name="T"/>.</returns>
        public static StringBuilder ShowHelp()
        {
            var builder = new StringBuilder();

            HelpReduce(builder, HelpMap().ToList());

            return builder;
        }

        private static IEnumerable<Tuple<string, string, string>> HelpMap()
        {
            foreach (var element in _options.Keys)
            {
                yield return Tuple.Create(
                    element.ShortName != '\0' ? "-" + element.ShortName : string.Empty,
                    element.LongName != null ? "--" + element.LongName : string.Empty,
                    element.Description ?? string.Empty);
            }
        }

        private static void HelpReduce(
            StringBuilder builder,
            IEnumerable<Tuple<string, string, string>> rows)
        {
            int max1 = 0, max2 = 0;

            int left = max1 + 2 + max2 + 2;

            foreach (var element in rows)
            {
                max1 = Math.Max(max1, element.Item1.Length);
                max2 = Math.Max(max2, element.Item2.Length);
            }

            foreach (var element in rows)
            {
            }
        }*/
    }
}