﻿namespace Dyn.Console.CommandLine
{
    using System;
    using System.Collections.Generic;
    using System.Reflection;

    using Invoker = System.Action<CommandLineLexer, object>;

    /// <summary>
    /// Provides the static method <see cref="Create(MemberInfo)"/>, which
    /// returns a delegate for calling its parameter.
    /// </summary>
    internal static class ReflectionInvoker
    {
        private static readonly Type TypeString = typeof(string);
        private static readonly Type TypeBool = typeof(bool);
        private static readonly Type TypeInt = typeof(int);
        private static readonly Type TypeNullable = typeof(Nullable<>);
        private static readonly Type TypeNullableBool = typeof(bool?);
        private static readonly Type TypeNullableInt = typeof(int?);

        private static string GetString(CommandLineLexer lexer)
        {
            string result;

            if (!lexer.TryParameter(out result))
            {
                throw new ArgumentParserException("Expected parameter");
            }

            return result;
        }

        private static bool ParseBool(string input)
        {
            bool result;

            if (bool.TryParse(input, out result))
            {
                return result;
            }

            switch (input)
            {
                case "1":
                case "y":
                case "yes":
                case "t":
                case "+":
                    return true;
                case "0":
                case "-":
                case "n":
                case "f":
                case "no":
                    return false;

                default:
                    throw new ArgumentParserException("Cannot parse parameter as boolean.");
            }
        }

        private static int ParseInt(string input)
        {
            int result;

            if (int.TryParse(input, out result))
            {
                return result;
            }

            throw new ArgumentParserException("Cannot parse parameter as integer.");
        }

        private static T? ParseNullable<T>(CommandLineLexer lexer, Func<string, T> parser)
            where T : struct
        {
            string input;

            if (lexer.TryParameter(out input))
            {
                return new T?(parser(input));
            }

            return null;
        }

        private static Func<CommandLineLexer, object> LasyParser(Type destType)
        {
            if (destType == TypeString)
            {
                return (lexer) => GetString(lexer);
            }
            else if (destType == TypeBool)
            {
                return (lexer) => ParseBool(GetString(lexer));
            }
            else if (destType == TypeNullableBool)
            {
                return (lexer) => ParseNullable(lexer, ParseBool);
            }
            else if (destType == TypeInt)
            {
                return (lexer) => ParseInt(GetString(lexer));
            }
            else if (destType == TypeNullableInt)
            {
                return (lexer) => ParseNullable(lexer, ParseInt);
            }

            throw new OptionDefinitionException("Cannot write to type " + destType.Name + ".");
        }

        /// <summary>
        /// Returns a delegate for calling <paramref name="info"/>.
        /// </summary>
        /// <param name="info">The <see cref="MemberInfo"/> to be called.</param>
        /// <returns>A delegate for calling <paramref name="info"/>.</returns>
        public static Invoker Create(MemberInfo info)
        {
            Guard.EnsureNotNull("info", info);

            if (info is FieldInfo)
            {
                return Create((FieldInfo)info);
            }
            else if (info is PropertyInfo)
            {
                return Create((PropertyInfo)info);
            }
            else if (info is MethodInfo)
            {
                return Create((MethodInfo)info);
            }

            throw new OptionDefinitionException(
                string.Format("Unsupported member type of {0}.", info.Name),
                new NotSupportedException());
        }

        private static Invoker Create(FieldInfo info)
        {
            Guard.EnsureNotNull("info", info);

            if (info.IsInitOnly)
            {
                throw new OptionDefinitionException(
                    string.Format("Field {0} is read only.", info.Name),
                    new FieldAccessException()); // TODO: Describe
            }

            if (!info.IsPublic)
            {
                throw new OptionDefinitionException(
                    string.Format("{0} is not publicly accessible.", info.Name),
                    new MemberAccessException());
            }

            var fieldType = info.FieldType;

            var ft = info.FieldType;
            if (ft == TypeBool || ft == TypeNullableBool)
            {
                return (lexer, instance) => info.SetValue(instance, true);
            }

            var converter = LasyParser(fieldType);
            return (lexer, instance) => info.SetValue(instance, converter(lexer));
        }

        private static Invoker Create(PropertyInfo info)
        {
            Guard.EnsureNotNull("info", info);

            if (!info.CanWrite)
            {
                throw new OptionDefinitionException(
                    string.Format("Property {0} is read only.", info.Name),
                    new MemberAccessException());
            }

            var pi = info.PropertyType;

            if (info.GetIndexParameters().Length == 0)
            {
                if (pi == TypeNullableBool || pi == TypeBool)
                {
                    return (lexer, instance) => info.SetValue(instance, true);
                }
            }

            return Create(info.SetMethod);
        }

        private static Invoker Create(MethodInfo info)
        {
            Guard.EnsureNotNull("info", info);

            if (!info.IsPublic)
            {
                throw new OptionDefinitionException(
                    string.Format("{0} is not publicly accessible.", info.Name),
                    new MemberAccessException());
            }

            var pt = info.GetParameters();

            if (pt.Length == 0)
            {
                return (low, instance) => info.Invoke(instance, Type.EmptyTypes);
            }

            var funcs = new List<Func<CommandLineLexer, object>>();

            foreach (var t in pt)
            {
                funcs.Add(LasyParser(t.ParameterType));
            }

            int ptl = pt.Length;

            return (lexer, instance) =>
            {
                var result = new object[ptl];

                for (int i = 0; i < ptl; ++i)
                {
                    result[i] = funcs[i](lexer);
                }

                info.Invoke(instance, result);
            };
        }
    }
}
