﻿namespace Dyn.Console
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using Dyn.Diagnostics;
    using Dyn.Lexer;

    /// <summary>
    /// Decorates diagnostics with their error location.
    /// <para>This class is a diagnostic middle ware.</para>
    /// </summary>
    public sealed class ConsoleDiagnosticHandler : IDiagnostic
    {
        private readonly IList<SyntaxToken> _tokens;

        /// <summary>
        /// Initializes a new instance of the <see cref="ConsoleDiagnosticHandler"/> class with source code tokens.
        /// </summary>
        /// <param name="tokens">Source code tokens.</param>
        /// <remarks><paramref name="tokens"/> have to be sorted by <see cref="Span.FileName"/> and <see cref="Span.Position"/></remarks>
        public ConsoleDiagnosticHandler(IList<SyntaxToken> tokens)
        {
            Guard.EnsureNotNull("tokens", tokens);
            this._tokens = tokens;
        }

        /// <inheritdoc/>
        public event EventHandler<DiagnosticEventArgs> DiagnosticInfo;

        /// <inheritdoc/>
        public void DiagnosticInfoListener(object sender, DiagnosticEventArgs args)
        {
            this.WriteDiagnostic(args.Info);

            var handler = this.DiagnosticInfo;

            if (handler != null)
            {
                handler(this, args);
            }
        }

        /// <summary>
        /// Writes a human readable diagnostic into <see cref="Console"/>.
        /// </summary>
        /// <param name="info">The diagnostic info.</param>
        private void WriteDiagnostic(DiagnosticInfo info)
        {
            var location = info.Location;

            var fg = Console.ForegroundColor;

            Console.Error.Write("{0}:{1}:{2}: ", SafeFileName(location.FileName), location.Line, location.Column);
            Console.ForegroundColor = ColoredFormatter.ConsoleColorFromSeverity(info.Level, fg);
            Console.Error.Write("{0}:", info.Level);
            Console.ForegroundColor = fg;
            Console.Error.WriteLine(" {0}", info.Message);

            var relevantTokens = new List<SyntaxToken>(
                this._tokens.ExtractSpan(location));

            if (relevantTokens.Count > 0)
            {
                ColoredFormatter.PrintTokenListHighlighted(
                    relevantTokens, Console.Error);

                Console.Error.WriteLine();
                Console.Error.Write(new string(' ', location.Column));
                Console.Error.WriteLine(new string('^', location.Length));
            }
        }

        /// <summary>
        /// Returns the file name segment from a path.
        /// </summary>
        /// <param name="fileName">The path to inspect</param>
        /// <returns>The file name segment from a path</returns>
        private static string SafeFileName(string fileName)
        {
            try
            {
                return Path.GetFileName(fileName);
            }
            catch (ArgumentException)
            {
                return fileName;
            }
        }
    }
}
