﻿namespace Dyn.Console
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Text;
    using Dyn.Ast;
    using Dyn.Diagnostics;
    using Dyn.Lexer;
    using Dyn.Parser;
    using Dyn.VerficationConditions;

    /// <summary>Main class, which contains the program entry point <see cref="Main"/>.</summary>
    internal static class Program
    {
        /// <summary>Program entry.</summary>
        /// <param name="args">Command line arguments</param>
        /// <returns><code>0</code>on success, <code>1</code> on argument parsing errors or the negation of the number of all <see cref="DiagnosticSeverity.Error"/> diagnostics.</returns>
        internal static int Main(string[] args)
        {
            // ensure consistent color theme
            Console.ResetColor();

            var state = new Arguments();

            bool parseError = false;

            try
            {
                CommandLine.CommandLineParser<Arguments>.Parse(args, state);

                if (state.ShowHelp)
                {
                    ShowHelp();
                    return 0;
                }
            }
            catch (CommandLine.ArgumentParserException)
            {
                parseError = true;
            }

            if (parseError || !state.Validate(Console.Error))
            {
                ShowHelp();
                return -1;
            }

            return -Run(state);
        }

        /// <summary>Helps. Hopefully.</summary>
        private static void ShowHelp()
        {
            Console.WriteLine(@"
libdyn.console [-i] [-achs] [-?] -z path [file...]
libdyn.console formally verifies Dyn programs.

Commands:
-i      --interactive                  reads stdin for user input
                                       (automatically activated if no 
                                       files are provided)
-z      --solver           PATH        executes a SMTLIB2 solver with PATH
-s      --show-summary                 prints a summary
-h      --show-highlighted             prints highlighted source files
-v      --verify                       verifies the method contracts
-c      --show-conditions              prints verification conditions
-a      --show-ast                     prints abstract syntax trees 
-?      --help                         prints this message");
        }

        /// <summary>Parses and optionally verifies based on program state.</summary>
        /// <param name="args">The state.</param>
        /// <returns><code>0</code> on success, otherwise the number of diagnostics negated.</returns>
        public static int Run(Arguments args)
        {
            // set up debug handlers
            var summary = CreateDiagnostics(args);

            var classes = ReadFiles(args);

            // inheritance checks
            classes = InheritanceTransformer.Transform(args.DiagnosticInfoListener, classes).ToList();

            if (classes.Count == 0)
            {
                DiagnosticExtensions.Invoke(
                    args.DiagnosticInfoListener,
                    null,
                    DiagnosticSeverity.Error,
                    Span.Empty,
                    "No classes found.");
            }

            if (args.ShowAst)
            {
                var sourceCode = string.Join("\n\n", classes.Select(e => e.ToSource()));

                ColoredFormatter.PrintTokenListHighlighted(
                    Tokenize(string.Empty, new StringReader(sourceCode)),
                    Console.Out);
            }

            if (args.Verify)
            {
                Verfication(args, classes);
            }

            if (args.ShowSummary)
            {
                Console.Out.WriteLine("Diagnostic Summary:");
                WriteSummaryLine(Console.ForegroundColor, summary[DiagnosticSeverity.None], " other(s)");
                WriteSummaryLine(ConsoleColor.Cyan, summary[DiagnosticSeverity.Information], " info(s)");
                WriteSummaryLine(ConsoleColor.Yellow, summary[DiagnosticSeverity.Warning], " warning(s)");
                WriteSummaryLine(ConsoleColor.Red, summary[DiagnosticSeverity.Error], " error(s)");
            }

            return summary[DiagnosticSeverity.Error];
        }

        private static string CollectPredicates(
            Arguments args)
        {
            StringBuilder sb = new StringBuilder();
            foreach (string file in args.Files)
            {
                string filename = Path.ChangeExtension(file, ".theory.smt");
                try
                {
                    sb.AppendFormat("\n; from {0}:\n", filename);
                    sb.Append(File.ReadAllText(filename));
                    Console.Out.WriteLine("Read Predicates from {0}:", filename);
                }
                catch (IOException)
                {
                    // likely FileNotFoundException

                    // silently ignore
                }
            }

            return sb.ToString();
        }

        /// <summary>Verifies a program.</summary>
        /// <param name="args">The state.</param>
        /// <param name="classes">Classes to verify.</param>
        private static void Verfication(
            Arguments args,
            IList<DynProgramClassNode> classes)
        {
            string predicates = CollectPredicates(args);
            var z3 = new Z3Adapter(args.SolverPath, classes, predicates);
            z3.DiagnosticInfo += args.DiagnosticInfoListener;

            var vcg = new VerficationConditionGenerator(classes);
            vcg.DiagnosticInfo += args.DiagnosticInfoListener;

            var debug = Console.Error;

            int counter = 0;
            int success = 0;
            int skip = 0;
            int failure = 0;

            // verification chain (reverse buildup)
            Func<VerificationRequest, bool> handler =
                ((Func<DynAssertionTypeInference.Result, bool>)(r =>
                    z3.Run(r) == Z3Adapter.Z3Result.Unsatisfiable))
                .Chain((VerificationRequest vc, Func<DynAssertionTypeInference.Result, bool> inner) =>
                    {
                        var result = DynAssertionTypeInference.Run(
                                args.DiagnosticInfoListener,
                                vcg.PredicateLookup,
                                vc.Expression,
                                vc.ClassName);

                        if (!result.IsValid)
                        {
                            debug.WriteLine("\tAssertion type inference failed.", result);
                            return false;
                        }

                        if (args.ShowConditions)
                        {
                            debug.WriteLine("\tTransformed Assertion : {0}", vc.Expression.ToSource());

                            foreach (var element in result)
                            {
                                debug.WriteLine("\t\tVariable '{0}' has type '{1}'", element.Key, element.Value);
                            }

                            return inner(result);
                        }

                        // skip solving
                        if (!File.Exists(args.SolverPath))
                        {
                            return true;
                        }

                        return inner(result);
                    })
                .Chain((VerificationRequest vc, Func<VerificationRequest, bool> inner) =>
                    {
                        counter++;

                        if (args.ShowConditions)
                        {
                            debug.WriteLine("Verification request {0}:", counter);
                            debug.WriteLine("\tAssertion: {0}", vc.Expression.ToSource(), vc.Message);
                            debug.WriteLine("\tDescription: {0}", vc.Message);
                            debug.WriteLine("\tIgnored: {0}", false);
                        }

                        if (false)
                        {
                            // TODO: Handle Ignore
                            debug.WriteLine("\tIgnored verification condition.");

                            skip++;

                            args.Invoke(
                                null,
                                DiagnosticSeverity.Warning,
                                vc.Expression.Location,
                                "Ignored.");

                            return true;
                        }

                        var result = inner(vc);

                        if (!result)
                        {
                            failure++;

                            args.Invoke(
                                null,
                                DiagnosticSeverity.Error,
                                vc.Expression.Location,
                                vc.Message);
                        }
                        else
                        {
                            success++;
                        }

                        return result;
                    });

            var wpc = new WeakestPreconditionCalculus(vcg, e => handler(e));
            wpc.DiagnosticInfo += args.DiagnosticInfoListener;
            wpc.Verify(classes);

            if (args.ShowSummary)
            {
                Console.ResetColor();
                Console.Out.WriteLine("Verification condition summary:");

                WriteSummaryLine(Console.ForegroundColor, counter, " found.");
                WriteSummaryLine(ConsoleColor.Green, success, " successfully verified.");
                WriteSummaryLine(ConsoleColor.Yellow, skip, " skipped.");
                WriteSummaryLine(ConsoleColor.Red, failure, " failed to verify.");
            }
        }

        /// <summary>Write a line to the console.</summary>
        /// <param name="fg">Fore ground color of the number.</param>
        /// <param name="count">The number.</param>
        /// <param name="text">Text following the number.</param>
        private static void WriteSummaryLine(ConsoleColor fg, int count, string text)
        {
            Console.Write('\t');

            var t = Console.ForegroundColor;
            Console.ForegroundColor = fg;
            Console.Write(count.ToString().PadLeft(8));
            Console.ForegroundColor = t;

            Console.WriteLine(text);
        }

        /// <summary>Crates a <see cref="DiagnosticSummary"/> instance.</summary>
        /// <param name="args">The args.</param>
        /// <returns>A <see cref="DiagnosticSummary"/> instance bound to the <see cref="IDiagnostic.DiagnosticInfo"/> of <paramref name="args"/>.</returns>
        private static DiagnosticSummary CreateDiagnostics(Arguments args)
        {
            var diagPrinter = new ConsoleDiagnosticHandler(args.Tokens);

            var result = new DiagnosticSummary();
            result.DiagnosticInfo += diagPrinter.DiagnosticInfoListener;
            args.DiagnosticInfo += result.DiagnosticInfoListener;

            return result;
        }

        /// <summary>Parses all requested files.</summary>
        /// <param name="state">The state.</param>
        /// <returns>A token stream from all source files.</returns>
        private static IList<DynProgramClassNode> ReadFiles(
            Arguments state)
        {
            var result = new List<DynProgramClassNode>();

            if (state.Interactive || state.Files.Any())
            {
                Console.Error.WriteLine("Press ^D or ^Z to finish input.");

                result.AddRange(Parse(state, "$stdin", Console.In));
            }

            foreach (var element in state.Files)
            {
                if (!File.Exists(element))
                {
                    DiagnosticExtensions.Invoke(
                        state.DiagnosticInfoListener,
                        null,
                        DiagnosticSeverity.Error,
                        new Span(element, 0, 0, 0, 0),
                        "File not found.");
                }

                try
                {
                    using (var reader = File.OpenText(element))
                    {
                        result.AddRange(Parse(state, element, reader));
                    }
                }
                catch (IOException e)
                {
                    DiagnosticExtensions.Invoke(
                        state.DiagnosticInfoListener,
                        null,
                        DiagnosticSeverity.Error,
                        new Span(element, 0, 0, 0, 0),
                        "Exception while reading file: " + e.Message);
                }
            }

            return result;
        }

        /// <summary>Parses a Dyn program from a file stream and file name.</summary>
        /// <param name="args">The arguments.</param>
        /// <param name="fileName">The name of the file.</param>
        /// <param name="reader">The file reader</param>
        /// <returns>A sequence of class nodes.</returns>
        private static IEnumerable<DynProgramClassNode> Parse(
            Arguments args,
            string fileName,
            TextReader reader)
        {
            var tokens = Tokenize(fileName, reader);
            args.Tokens.AddRange(tokens);

            var parser = new DynParser(tokens.GetEnumerator());
            parser.DiagnosticInfo += args.DiagnosticInfoListener;

            var x = parser.ParseAll(p => p.ParseClass());
            return x;
        }

        /// <summary>
        /// Returns the content of <paramref name="reader"/> interpreted as <see cref="SyntaxToken"/> with their location
        /// file name set to <paramref name="fileName"/>
        /// </summary>
        /// <param name="fileName">The file name.</param>
        /// <param name="reader"><see cref="TextReader"/> for the tokens.</param>
        /// <returns>A list of <see cref="SyntaxToken"/> representing the content of <paramref name="reader"/>.</returns>
        private static IList<SyntaxToken> Tokenize(string fileName, TextReader reader)
        {
            return Compatibility.TransformCommentToAssertion(
                    new DynLexer(fileName, reader))
                .ToList();
        }

        /// <summary>Allows function chaining.</summary>
        /// <typeparam name="T">The first type.</typeparam>
        /// <typeparam name="U">The second type.</typeparam>
        /// <typeparam name="V">The third type.</typeparam>
        /// <typeparam name="W">Thr forth type.</typeparam>
        /// <param name="inner">The delegate to be wrapped by <paramref name="outer"/>.</param>
        /// <param name="outer">The wrapping delegate.</param>
        /// <returns>A new delegate.</returns>
        private static Func<V, W> Chain<T, U, V, W>(
            this Func<T, U> inner,
            Func<V, Func<T, U>, W> outer)
        {
            return e => outer(e, inner);
        }
    }
}
