﻿using Dyn.Lexer;

namespace Dyn.Ast
{
    public static class AstExtensions
    {
        public const string RootClass = "object";

        public static class Programs
        {
            public static readonly DynProgramVariableNode Global = new DynProgramVariableNode(Span.Empty, DynProgramStatementNode.NodeKind.Variable, "_Global");
            public static readonly DynProgramVariableNode Null = new DynProgramVariableNode(Span.Empty, DynProgramStatementNode.NodeKind.Variable, "null");
            public static readonly DynProgramConstructorCallNode False = new DynProgramConstructorCallNode(Span.Empty, "boolean", Dyn.Utilities.AsArray(Null));
            public static readonly DynProgramConstructorCallNode True = new DynProgramConstructorCallNode(Span.Empty, "boolean", Dyn.Utilities.AsArray(False));
            public static readonly DynProgramConstructorCallNode Zero = new DynProgramConstructorCallNode(Span.Empty, "numeric", Dyn.Utilities.AsArray(Null));
            public static readonly DynProgramConstructorCallNode EmptyString = new DynProgramConstructorCallNode(Span.Empty, "string", Dyn.Utilities.AsArray(Null));
            public static readonly DynProgramVariableNode Undefined = new DynProgramVariableNode(Span.Empty, DynProgramStatementNode.NodeKind.Variable, "undefined");
            public static readonly DynProgramVariableNode r = new DynProgramVariableNode(Span.Empty, DynProgramStatementNode.NodeKind.Variable, "result");
            public static readonly DynProgramVariableNode Self = new DynProgramVariableNode(Span.Empty, DynProgramStatementNode.NodeKind.Variable, "self");
            public static readonly string ConstructorMethodName = "init";
        }

        public static class Assertions
        {
            public static readonly DynAssertionVariableNode Null = new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.ProgramVariable, "null");
            public static readonly DynAssertionVariableNode True = new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "true");
            public static readonly DynAssertionVariableNode False = new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "false");
            public static readonly DynAssertionVariableNode Self = new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.ProgramVariable, "self");
            public static readonly DynAssertionVariableNode r = new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.ProgramVariable, "result");
            public static readonly DynAssertionInstanceVariableNode ClassInstanceGhost = new DynAssertionInstanceVariableNode(Span.Empty, AstExtensions.Assertions.Self, "c");
        }

        public static DynAssertionExpressionNode Negate(this DynAssertionExpressionNode value, Span span)
        {
            return new DynAssertionUnaryNode(span, value);
        }

        public static DynAssertionExpressionNode And(this DynAssertionExpressionNode first, Span span, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(span, DynAssertionExpressionNode.NodeKind.And, first, second);
        }

        //TODO: Harmonize
        public static DynAssertionExpressionNode Implies(this DynAssertionExpressionNode first, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(Span.Empty, DynAssertionExpressionNode.NodeKind.Implies, first, second);
        }

        public static DynAssertionExpressionNode Addition(this DynAssertionExpressionNode first, Span span, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(span, DynAssertionExpressionNode.NodeKind.Addition, first, second);
        }

        public static DynAssertionExpressionNode Multiplication(this DynAssertionExpressionNode first, Span span, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(span, DynAssertionExpressionNode.NodeKind.Multiplication, first, second);
        }

        public static DynAssertionExpressionNode Or(this DynAssertionExpressionNode first, Span span, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(span, DynAssertionExpressionNode.NodeKind.Or, first, second);
        }

        public static DynAssertionExpressionNode Equality(this DynAssertionExpressionNode first, Span span, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(span, DynAssertionExpressionNode.NodeKind.Equality, first, second);
        }

        public static DynAssertionExpressionNode GreaterThan(this DynAssertionExpressionNode first, Span span, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(span, DynAssertionExpressionNode.NodeKind.GreaterThan, first, second);
        }

        public static DynProgramExpressionNode TransformNumber(Span span, int value)
        {
            Guard.EnsureNonnegative("value", value);

            DynProgramExpressionNode result = value == 0 ? new DynProgramConstructorCallNode(span, "numeric", Dyn.Utilities.AsArray(Programs.Null)) : Programs.Zero;

            while (value > 0)
            {
                result = new DynProgramMethodCallNode(value == 1 ? span : Span.Empty, result, "succ", null);
                value--;
            }

            return result;
        }

        public static DynProgramExpressionNode TransformBoolean(Span span, bool value)
        {
            if (value)
            {
                return new DynProgramConstructorCallNode(span, "boolean", Dyn.Utilities.AsArray(Programs.False));
            }
            else
            {
                return new DynProgramConstructorCallNode(span, "boolean", Dyn.Utilities.AsArray(Programs.Null));
            }
        }

        public static DynProgramExpressionNode TransformString(Span span, string value)
        {
            Guard.EnsureNotNull("value", value);

            // TODO: set span single character here:
            DynProgramExpressionNode result = new DynProgramConstructorCallNode(span, "string", Dyn.Utilities.AsArray(Programs.Null));

            foreach (var c in value)
            {
                result = new DynProgramMethodCallNode(span, result, "addchar", Dyn.Utilities.AsArray(TransformNumber(Span.Empty, (int)c)));
            }

            return result;
        }

        public static readonly DynNodeEqualityComparer Normal = new DynNodeEqualityComparer();
        public static readonly DynNodeIgnoreLocationEqualityComparer IgnoreLocation = new DynNodeIgnoreLocationEqualityComparer();
    }
}
