﻿using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Dyn.Ast
{
    /// <summary>
    /// <para>Provides extensions methods transforms abstract syntax tree nodes like
    /// <see cref="DynAssertionExpressionNode"/> or
    /// <see cref="DynProgramClassNode"/> to
    /// <see cref="string"/>.</para>
    /// <para>This is the <see cref="Dyn.Parser.DynParser"/> parser inverse.</para>
    /// </summary>
    public static class DynNodeFormatter
    {
        private static readonly StringComparer Ordinal = StringComparer.Ordinal;

        #region Primitive

        private static StringBuilder VisitEnumerable<TType>(
            IEnumerable<TType> source,
            StringBuilder state,
            Func<StringBuilder, StringBuilder> delimeter,
            Func<StringBuilder, TType, StringBuilder> function)
        {
            bool tail = false;

            foreach (var node in source)
            {
                if (tail)
                {
                    state = delimeter(state);
                }

                state = function(state, node);

                tail = true;
            }

            return state;
        }

        private static StringBuilder VisitEnumerable<TType>(
            IEnumerable<TType> source,
            StringBuilder state,
            int depth,
            bool multiLine,
            Func<StringBuilder, int, bool, StringBuilder> delimeter,
            Func<StringBuilder, TType, int, bool, StringBuilder> function)
        {
            bool tail = false;

            foreach (var node in source)
            {
                if (tail && delimeter != null)
                {
                    state = delimeter(state, depth, multiLine);
                }

                state = function(state, node, depth, multiLine);

                tail = true;
            }

            return state;
        }

        private const string DEFAULT_INDENTATION = "   ";

        private static StringBuilder Indent(this StringBuilder builder, int depth)
        {
            for (int i = 0; i < depth; ++i)
            {
                builder.Append(DEFAULT_INDENTATION);
            }

            return builder;
        }

        #endregion Primitive

        #region Assertions

        /// <summary>Returns a string representing an <see cref="DynAssertionExpressionNode"/> instance.</summary>
        /// <param name="node">The node.</param>
        /// <returns>A string whose value is <paramref name="node"/>.</returns>
        public static string ToSource(this DynAssertionExpressionNode node)
        {
            Guard.EnsureNotNull("node", node);
            return VisitAssertionExpression(new StringBuilder(), node, 0).ToString();
        }

        private static StringBuilder VisitAssertionExpression(
            StringBuilder builder,
            DynAssertionExpressionNode node)
        {
            return VisitAssertionExpression(builder, node, 0);
        }

        private static StringBuilder VisitAssertionExpression(
            StringBuilder builder,
            DynAssertionExpressionNode node,
            int precedence)
        {
            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.String:
                    return builder.Append('"').Append(((DynAssertionStringNode)node).Value).Append('"');

                case DynAssertionExpressionNode.NodeKind.Numeric:
                    return builder.Append(((DynAssertionNumericNode)node).Value);

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    {
                        var element = (DynAssertionIfThenElseNode)node;
                        builder.Append("if ");
                        builder = VisitAssertionExpression(
                            builder,
                            element.Condition,
                            0).Append(" then ");
                        builder = VisitAssertionExpression(
                            builder,
                            element.TrueBranch,
                            0).Append(" else ");
                        builder = VisitAssertionExpression(
                            builder,
                            element.FalseBranch,
                            0).Append(" end");
                    }
                    return builder;

                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    return builder.Append(((DynAssertionVariableNode)node).Name);

                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                    {
                        var expression = (DynAssertionVariableNode)node;

                        var name = expression.Name;

                        if (StringComparer.Ordinal.Equals("true", name) ||
                            StringComparer.Ordinal.Equals("false", name))
                        {
                            return builder.Append(expression.Name);
                        }

                        return builder.Append('\'').Append(expression.Name);
                    }
                case DynAssertionExpressionNode.NodeKind.Predicate:
                    {
                        var element = (DynAssertionPredicateNode)node;
                        builder = builder.Append(element.Name).Append('(');
                        builder = VisitEnumerable(element.Arguments, builder, (b) => b.Append(", "), VisitAssertionExpression);
                        builder = builder.Append(')');
                    }
                    return builder;
            }

            bool parenthesisRequired = OperatorTranslation.AssertionPrecedence(node.Kind) < precedence;

            if (parenthesisRequired) { builder = builder.Append('('); }

            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Addition:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " + ", true); break;
                case DynAssertionExpressionNode.NodeKind.And:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " && ", true); break;
                case DynAssertionExpressionNode.NodeKind.Division:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " / ", true); break;
                case DynAssertionExpressionNode.NodeKind.Equality:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " == ", false); break;
                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " > ", true); break;
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " >= ", true); break;
                case DynAssertionExpressionNode.NodeKind.Implies:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " implies ", true); break;
                case DynAssertionExpressionNode.NodeKind.LessThan:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " < ", true); break;
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " <= ", true); break;
                case DynAssertionExpressionNode.NodeKind.Modulus:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " % ", true); break;
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " * ", true); break;
                case DynAssertionExpressionNode.NodeKind.Negate:
                    builder = VisitAssertionUnary(builder, (DynAssertionUnaryNode)node, "!"); break;
                case DynAssertionExpressionNode.NodeKind.Or:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " || ", true); break;
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                    builder = VisitAssertionBinary(builder, (DynAssertionBinaryNode)node, " - ", true); break;
                case DynAssertionExpressionNode.NodeKind.String:
                    builder = builder.Append('"').Append(((DynAssertionStringNode)node).Value).Append('"'); break;
                case DynAssertionExpressionNode.NodeKind.Numeric:
                    builder = builder.Append(((DynAssertionNumericNode)node).Value); break;
                case DynAssertionExpressionNode.NodeKind.Universal:
                    builder = VisitAssertionQuantifier(builder, (DynAssertionQuantifierNode)node, "forall "); break;
                case DynAssertionExpressionNode.NodeKind.Existential:
                    builder = VisitAssertionQuantifier(builder, (DynAssertionQuantifierNode)node, "exists "); break;
                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    {
                        var element = (DynAssertionTypeCheckNode)node;
                        builder = VisitAssertionExpression(
                            builder,
                            element.Expression,
                            OperatorTranslation.AssertionPrecedence(node.Kind))
                            .Append(" in {").Join(", ", element.TypeSet).Append('}');
                    }
                    break;

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    {
                        var element = (DynAssertionIfThenElseNode)node;
                        builder.Append("if ");
                        builder = VisitAssertionExpression(
                            builder,
                            element.Condition,
                            0).Append(" then ");
                        builder = VisitAssertionExpression(
                            builder,
                            element.TrueBranch,
                            0).Append(" else ");
                        builder = VisitAssertionExpression(
                            builder,
                            element.FalseBranch,
                            0).Append(" end");
                    }
                    break;

                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    {
                        var element = (DynAssertionInstanceVariableNode)node;
                        builder = VisitAssertionExpression(
                            builder,
                            element.Expression,
                            OperatorTranslation.AssertionPrecedence(node.Kind))
                            .Append(".@").Append(element.Name);
                    }
                    break;

                case DynAssertionExpressionNode.NodeKind.Predicate:
                    {
                        var element = (DynAssertionPredicateNode)node;
                        builder = builder.Append(element.Name).Append('(');
                        builder = VisitEnumerable(element.Arguments, builder, (b) => b.Append(", "), VisitAssertionExpression);
                        builder = builder.Append(')');
                    }
                    break;

                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("node.Kind", (int)node.Kind);
            }

            if (parenthesisRequired) { builder.Append(')'); }
            return builder;
        }

        private static StringBuilder VisitAssertionQuantifier(
            StringBuilder builder,
            DynAssertionQuantifierNode node,
            string text)
        {
            builder.Append(text).Append('\'').Append(node.Variable).Append(" : ").Append(node.Type).Append(". ");
            return VisitAssertionExpression(
                            builder,
                            node.Expression,
                            OperatorTranslation.AssertionPrecedence(node.Kind));
        }

        private static StringBuilder VisitAssertionBinary(
            StringBuilder builder,
            DynAssertionBinaryNode node,
            string op,
            bool leftAssoc)
        {
            var precedence = OperatorTranslation.AssertionPrecedence(node.Kind);
            VisitAssertionExpression(builder, node.First, precedence + (leftAssoc ? 0 : 1)).Append(op);
            return VisitAssertionExpression(builder, node.Second, precedence + (!leftAssoc ? 0 : 1));
        }

        private static StringBuilder VisitAssertionUnary(StringBuilder builder, DynAssertionUnaryNode node, string op)
        {
            builder.Append(op);
            return VisitAssertionExpression(builder, node.Expression, OperatorTranslation.AssertionPrecedence(node.Kind) - 1);
        }

        #endregion Assertions

        private static StringBuilder LinebreakDelimeter(StringBuilder builder, int depth, bool multiline)
        {
            if (multiline)
            {
                return builder.AppendLine().Indent(depth);
            }
            else
            {
                return builder.Append(' ');
            }
        }

        private static StringBuilder CommaDelimeter(StringBuilder builder, int depth, bool multiline)
        {
            return builder.Append(", ");
        }

        /// <summary>Returns a string representing an <see cref="DynProgramClassNode"/> instance.</summary>
        /// <param name="node">The node.</param>
        /// <returns>A string whose value is <paramref name="node"/>.</returns>
        public static string ToSource(this DynProgramClassNode node)
        {
            bool multiLine = true;

            Guard.EnsureNotNull("node", node);

            var builder = new StringBuilder();
            builder.Append("class ").Append(node.Name);

            if (!Ordinal.Equals(node.Parent, AstExtensions.RootClass))
            {
                builder.Append(" < ").Append(node.Parent);
            }

            if (node.Contracts.Count > 0)
            {
                builder = LinebreakDelimeter(builder, 1, multiLine);
                VisitEnumerable(node.Contracts, builder, 1, multiLine, LinebreakDelimeter, VisitProgramStatement);
            }

            builder.AppendLine();
            builder.Append("{");

            if (node.Renames.Count > 0)
            {
                builder = LinebreakDelimeter(builder, 1, multiLine);
                VisitEnumerable(node.Renames, builder, 1, multiLine, LinebreakDelimeter, VisitRenameDeclaration);
                builder.AppendLine();
            }

            if (node.Methods.Count > 0)
            {
                builder = LinebreakDelimeter(builder, 1, multiLine);
                VisitEnumerable(node.Methods, builder, 1, multiLine, LinebreakDelimeter, VisitMethodDeclaration);
                builder.AppendLine();
            }

            builder.AppendLine("}");

            return builder.ToString();
        }

        /// <summary>Returns a string representing an <see cref="DynProgramRenameDeclarationNode"/> instance.</summary>
        /// <param name="node">The node.</param>
        /// <returns>A string whose value is <paramref name="node"/>.</returns>
        public static string ToSource(this DynProgramRenameDeclarationNode node)
        {
            Guard.EnsureNotNull("node", node);
            return VisitRenameDeclaration(new StringBuilder(), node, 0, true).ToString();
        }

        /// <summary>Returns a string representing an <see cref="DynProgramMethodDeclarationNode"/> instance.</summary>
        /// <param name="node">The node.</param>
        /// <returns>A string whose value is <paramref name="node"/>.</returns>
        public static string ToSource(this DynProgramMethodDeclarationNode node)
        {
            Guard.EnsureNotNull("node", node);
            return VisitMethodDeclaration(new StringBuilder(), node, 0, true).ToString();
        }

        /// <summary>Returns a string representing an <see cref="DynProgramStatementNode"/> instance.</summary>
        /// <param name="node">The node.</param>
        /// <returns>A string whose value is <paramref name="node"/>.</returns>
        public static string ToSource(this DynProgramStatementNode node)
        {
            Guard.EnsureNotNull("node", node);
            return VisitProgramStatement(new StringBuilder(), node, 0, true).ToString();
        }

        /// <summary>Returns a string representing an <see cref="DynProgramContractDeclarationNode"/> instance.</summary>
        /// <param name="node">The node.</param>
        /// <returns>A string whose value is <paramref name="node"/>.</returns>
        public static string ToSource(this DynProgramContractDeclarationNode node)
        {
            Guard.EnsureNotNull("node", node);
            var builder = new StringBuilder();
            VisitProgramStatement(builder, node, 0, true);
            return builder.ToString();
        }

        private static StringBuilder VisitRenameDeclaration(
            StringBuilder builder,
            DynProgramRenameDeclarationNode node,
            int depth,
            bool multiLine)
        {
            return builder.Append("rename ").Append(node.From).Append(" ").Append(node.To);
        }

        private static StringBuilder VisitMethodDeclaration(
            StringBuilder builder,
            DynProgramMethodDeclarationNode node,
            int depth,
            bool multiLine)
        {
            builder.Append("method ").Append(node.Name);
            builder.Append('(').Join(", ", node.Parameters).Append(')');
            if (node.Contracts.Count > 0)
            {
                builder = LinebreakDelimeter(builder, depth + 1, multiLine);
                VisitEnumerable(node.Contracts, builder, depth + 1, multiLine, LinebreakDelimeter, VisitProgramStatement);
            }
            builder.AppendLine();
            builder.Indent(depth).Append('{');

            if (node.Statements.Any())
            {
                builder = LinebreakDelimeter(builder, depth + 1, multiLine);
                VisitProgramStatements(builder, node.Statements, depth + 1, multiLine);
            }

            builder.AppendLine().Indent(depth).Append('}');

            return builder;
        }

        private static StringBuilder VisitProgramStatement(
            StringBuilder builder,
            DynProgramStatementNode node,
            int depth,
            bool multiLine)
        {
            switch (node.Kind)
            {
                case DynProgramStatementNode.NodeKind.Assignment:
                case DynProgramStatementNode.NodeKind.ConstructorCall:
                case DynProgramStatementNode.NodeKind.If:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                case DynProgramStatementNode.NodeKind.MethodCall:
                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.While:
                    return VisitProgramExpression(builder, (DynProgramExpressionNode)node, depth, multiLine, 0);

                case DynProgramStatementNode.NodeKind.Assert:
                    builder = builder.Append("assert "); break;
                case DynProgramStatementNode.NodeKind.Ensures:
                    builder = builder.Append("ensures "); break;
                case DynProgramStatementNode.NodeKind.Requires:
                    builder = builder.Append("requires "); break;
                case DynProgramStatementNode.NodeKind.Invariant:
                    builder = builder.Append("invariant "); break;
                case DynProgramStatementNode.NodeKind.ByTheWay:
                    builder = builder.Append("bytheway "); break;
                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("node.Kind", (int)node.Kind);
            }

            return VisitAssertionExpression(builder, ((DynProgramContractDeclarationNode)node).Expression);
        }

        private static StringBuilder VisitProgramStatements(
            StringBuilder builder,
            IEnumerable<DynProgramStatementNode> source,
            int depth,
            bool multiLine)
        {
            return VisitEnumerable(
                source,
                builder,
                0,
                false,
                (b, d, m) => LinebreakDelimeter(builder.Append(';'), d, m),
                VisitProgramStatement);
        }

        private static StringBuilder VisitProgramExpression(
            StringBuilder builder,
            DynProgramExpressionNode node,
            int depth,
            bool multiLine,
            int precedence)
        {
            int num;
            string str;

            if (TryExtractNumber(node, out num))
            {
                return builder.Append(num);
            }
            else if (TryExtractString(node, out str))
            {
                return builder.Append('"').Append(str).Append('"');
            }
            else if (TryVisitProgramExpression(builder, node, depth, precedence))
            {
                return builder;
            }

            switch (node.Kind)
            {
                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return VisitConstructorCall(builder, (DynProgramConstructorCallNode)node, depth);

                case DynProgramStatementNode.NodeKind.If:
                    return VisitProgramIf(builder, (DynProgramIfNode)node, depth, multiLine);

                case DynProgramStatementNode.NodeKind.InstanceVariable:
                case DynProgramStatementNode.NodeKind.Variable:
                    return VisitProgramVariable(builder, ((DynProgramVariableNode)node));

                case DynProgramStatementNode.NodeKind.While:
                    return VisitProgramWhile(builder, node, depth, multiLine);
            }

            bool parenthesisRequired = OperatorTranslation.ProgramPrecedence(node.Kind) < precedence;

            if (parenthesisRequired)
            {
                builder.Append('(');
            }

            switch (node.Kind)
            {
                case DynProgramStatementNode.NodeKind.Assignment:
                    {
                        var element = (DynProgramAssignmentNode)node;

                        VisitProgramVariable(builder, element.Variable).Append(" = ");
                        VisitProgramExpression(
                            builder,
                            element.Expression,
                            depth,
                            false,
                            OperatorTranslation.ProgramPrecedence(node.Kind) - 1);
                    }
                    break;

                case DynProgramStatementNode.NodeKind.MethodCall:
                    VisitMethodCall(builder, (DynProgramMethodCallNode)node, depth); break;
                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("node.Kind", (int)node.Kind);
            }

            if (parenthesisRequired)
            {
                builder.Append(')');
            }

            return builder;
        }

        private static StringBuilder VisitProgramWhile(
            StringBuilder builder,
            DynProgramExpressionNode node,
            int depth,
            bool multiLine)
        {
            var element = (DynProgramWhileNode)node;

            builder.Append("while ");
            VisitProgramExpression(
                builder,
                element.Condition,
                depth + 1,
                false,
                0);
            builder.Append(" invariant ");
            VisitAssertionExpression(
                builder,
                element.Invariant);
            builder.Append(" do");

            LinebreakDelimeter(builder, depth + 1, multiLine);
            VisitProgramStatements(builder, element.Body, depth + 1, multiLine);

            LinebreakDelimeter(builder, depth, multiLine);
            return builder.Append("done");
        }

        private static bool TryVisitProgramExpression(
            StringBuilder builder,
            DynProgramExpressionNode node,
            int depth,
            int precedence)
        {
            if (node.Kind != DynProgramStatementNode.NodeKind.MethodCall)
            {
                return false;
            }

            DynProgramMethodCallNode callNode = (DynProgramMethodCallNode)node;

            SyntaxKind kind = OperatorTranslation.MethodToKind(callNode.Name);

            if (kind == SyntaxKind.EndOfStreamKind)
            {
                return false;
            }

            // binary
            if (callNode.Arguments.Count != 1)
            {
                return false;
            }

            VisitBinaryOperation(builder, depth, callNode.Expression, kind, callNode.Arguments[0], precedence);

            return true;
        }

        private static StringBuilder VisitProgramIf(StringBuilder builder, DynProgramIfNode node, int depth, bool multiLine)
        {
            builder.Append("if ");
            VisitProgramExpression(
                builder,
                node.Condition,
                depth + 1,
                false,
                0);
            builder.Append(" then");
            LinebreakDelimeter(builder, depth + 1, multiLine);
            VisitProgramStatements(builder, node.TrueBranch, depth + 1, multiLine);

            if (node.FalseBranch.Count > 0)
            {
                LinebreakDelimeter(builder, depth, multiLine);
                builder.Append(" else");
                LinebreakDelimeter(builder, depth + 1, multiLine);
                VisitProgramStatements(builder, node.FalseBranch, depth + 1, multiLine);
            }

            LinebreakDelimeter(builder, depth, multiLine);
            return builder.Append("end");
        }

        private static bool TryExtractNumber(DynProgramExpressionNode node, out int result)
        {
            result = 0;

            if (node.Kind == DynProgramStatementNode.NodeKind.ConstructorCall)
            {
                var cc = (DynProgramConstructorCallNode)node;

                // 0 is desugared to new numeric(null)!!!
                return AstExtensions.IgnoreLocation.Equals(cc, AstExtensions.Programs.Zero);
            }
            else if (node.Kind != DynProgramStatementNode.NodeKind.MethodCall)
            {
                return false;
            }

            var mc = (DynProgramMethodCallNode)node;

            while (Ordinal.Equals("succ", mc.Name) && mc.Arguments.Count == 0)
            {
                result++;

                if (mc.Expression.Kind == DynProgramStatementNode.NodeKind.ConstructorCall)
                {
                    var zero = (DynProgramConstructorCallNode)mc.Expression;

                    // 0 is desugared to new numeric(null)!!!
                    if (AstExtensions.IgnoreLocation.Equals(mc.Expression, AstExtensions.Programs.Zero))
                    {
                        return true;
                    }
                }

                if (mc.Expression.Kind != DynProgramStatementNode.NodeKind.MethodCall)
                {
                    return false;
                }

                mc = (DynProgramMethodCallNode)mc.Expression;
            }
            return false;
        }

        private static bool TryExtractString(DynProgramExpressionNode node, out string result)
        {
            result = null;

            if (node.Kind != DynProgramStatementNode.NodeKind.MethodCall)
            {
                return false;
            }

            DynProgramMethodCallNode methodCall = (DynProgramMethodCallNode)node;

            var builder = new StringBuilder();

            while (Ordinal.Equals("addchar", methodCall.Name) && methodCall.Arguments.Count == 1)
            {
                var arg0 = methodCall.Arguments[0];

                int ascii;

                if (!TryExtractNumber(methodCall.Arguments[0], out ascii) || ascii < 0 || ascii > 127)
                {
                    return false;
                }

                builder.Insert(0, (char)ascii);

                if (AstExtensions.IgnoreLocation.Equals(methodCall.Expression, AstExtensions.Programs.EmptyString))
                {
                    result = builder.ToString();
                    return true;
                }

                if (methodCall.Expression.Kind != DynProgramStatementNode.NodeKind.MethodCall)
                {
                    return false;
                }

                methodCall = (DynProgramMethodCallNode)methodCall.Expression;
            }
            return false;
        }

        private static void VisitBinaryOperation(
            StringBuilder builder,
            int depth,
            DynProgramExpressionNode first,
            SyntaxKind kind,
            DynProgramExpressionNode second,
            int precedence)
        {
            bool leftAssoc;

            switch (kind)
            {
                case SyntaxKind.AdditionOperator:
                case SyntaxKind.AndOperator:
                case SyntaxKind.MultiplicationOperator:
                case SyntaxKind.OrOperator:
                case SyntaxKind.GreaterThanOperator:
                case SyntaxKind.GreaterThanOrEqualOperator:
                case SyntaxKind.LessThanOperator:
                case SyntaxKind.LessThanOrEqualOperator:
                case SyntaxKind.IsAKeyword:
                case SyntaxKind.SubtractionOperator:
                case SyntaxKind.ModulosOperator:
                case SyntaxKind.DivisionOperator:
                    leftAssoc = true; break;

                case SyntaxKind.EqualityOperator:
                case SyntaxKind.IdentityOperator:
                    leftAssoc = false; break;
                default:
                    throw Guard.EnumOutOfRangeException<SyntaxKind>("kind", (int)kind);
            }

            int innerPrecedence = OperatorTranslation.ProgramPrecedence(kind);

            bool parenthesisRequired = innerPrecedence < precedence;

            if (parenthesisRequired)
            {
                builder.Append('(');
            }

            VisitProgramExpression(
                builder,
                first,
                depth,
                false,
                innerPrecedence + (leftAssoc ? 0 : 1));

            builder.Append(' ').Append(new SyntaxToken(kind, Span.Empty).ToSource());

            // special case for is_a? operator
            if (kind == SyntaxKind.IsAKeyword) { builder.Append('?'); }

            builder.Append(' ');
            VisitProgramExpression(
                builder,
                second,
                depth,
                false,
                innerPrecedence + (!leftAssoc ? 0 : 1));

            if (parenthesisRequired)
            {
                builder.Append(')');
            }
        }

        private static void VisitMethodCall(StringBuilder builder, DynProgramMethodCallNode node, int depth)
        {
            if (!AstExtensions.IgnoreLocation.Equals(node.Expression, AstExtensions.Programs.Global))
            {
                VisitProgramExpression(
                    builder,
                    node.Expression,
                    depth,
                    false,
                    OperatorTranslation.ProgramPrecedence(node.Kind));
                builder.Append('.');
            }

            builder.Append(node.Name).Append('(');
            VisitEnumerable(node.Arguments, builder, depth + 1, false, CommaDelimeter, VisitProgramStatement);
            builder.Append(')');
        }

        private static StringBuilder VisitConstructorCall(StringBuilder builder, DynProgramConstructorCallNode node, int depth)
        {
            // special case: boolean
            if (Ordinal.Equals("boolean", node.Name))
            {
                if (AstExtensions.IgnoreLocation.Equals(node, AstExtensions.Programs.False))
                {
                    return builder.Append("false");
                }
                if (AstExtensions.IgnoreLocation.Equals(node, AstExtensions.Programs.True))
                {
                    return builder.Append("true");
                }
            }
            // special case: numeric
            else if (Ordinal.Equals("numeric", node.Name) && node.Arguments.Count == 0)
            {
                return builder.Append("0");
            }
            // special case: string
            else if (Ordinal.Equals("string", node.Name) && AstExtensions.IgnoreLocation.Equals(node, AstExtensions.Programs.EmptyString))
            {
                return builder.Append("\"\"");
            }

            builder.Append("new ").Append(node.Name).Append('(');
            VisitEnumerable(node.Arguments, builder, depth + 1, false, CommaDelimeter, VisitProgramStatement);
            return builder.Append(')');
        }

        private static StringBuilder VisitProgramVariable(StringBuilder builder, DynProgramVariableNode node)
        {
            if (node.Kind == DynProgramStatementNode.NodeKind.InstanceVariable)
            {
                builder.Append('@');
            }
            builder.Append(node.Name);
            return builder;
        }

        /// <summary>Returns a string representing an <see cref="DynHoareNode"/> instance.</summary>
        /// <param name="node">The node.</param>
        /// <returns>A string whose value is <paramref name="node"/>.</returns>
        public static string ToSource(this DynHoareNode node)
        {
            Guard.EnsureNotNull("node", node);

            var builder = new StringBuilder();

            builder.Append('{');
            VisitAssertionExpression(builder, node.PreCondition);
            builder.Append("} ");
            VisitProgramStatements(builder, node.Statements, 0, false);

            if (node.Statements.Count > 0)
            {
                builder.Append(' ');
            }

            builder.Append('{');
            VisitAssertionExpression(builder, node.PostCondition);
            builder.Append('}');

            return builder.ToString();
        }
    }
}
