﻿using System;
using System.Collections.Generic;

namespace Dyn.Ast
{
	public static class FreeVariableAspect
	{
		/// <summary>Returns a list of free variables of a <see cref="DynAssertionExpressionNode"/> instance.</summary>
		/// <param name="node">The assertion node.</param>
		/// <returns>An immutable list of variable names in <paramref name="node"/>.</returns>
		public static IList<string> FreeVariables(this DynAssertionExpressionNode node)
		{
			Guard.EnsureNotNull("node", node);
			List<string> result = null;
			switch (node.Kind) {
				case DynAssertionExpressionNode.NodeKind.Addition:
				case DynAssertionExpressionNode.NodeKind.And:
				case DynAssertionExpressionNode.NodeKind.Or:
				case DynAssertionExpressionNode.NodeKind.Division:
				case DynAssertionExpressionNode.NodeKind.Modulus:
				case DynAssertionExpressionNode.NodeKind.Equality:
				case DynAssertionExpressionNode.NodeKind.GreaterThan:
				case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
				case DynAssertionExpressionNode.NodeKind.Implies:
				case DynAssertionExpressionNode.NodeKind.Multiplication:
				case DynAssertionExpressionNode.NodeKind.Subtraction:
				case DynAssertionExpressionNode.NodeKind.LessThan:
				case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
					var nodeB = ((DynAssertionBinaryNode)node);
					result = (List<string>)nodeB.First.FreeVariables();
					result.AddRange(nodeB.Second.FreeVariables());
					return result;

				case DynAssertionExpressionNode.NodeKind.Negate:
					return ((DynAssertionUnaryNode)node).Expression.FreeVariables();

				case DynAssertionExpressionNode.NodeKind.AssertionVariable:
					return new List<string>(){ string.Format("'{0}", ((DynAssertionVariableNode)node).Name) };

				case DynAssertionExpressionNode.NodeKind.ProgramVariable:
					return new List<string>(){ ((DynAssertionVariableNode)node).Name };


				case DynAssertionExpressionNode.NodeKind.Existential:
				case DynAssertionExpressionNode.NodeKind.Universal:
					var nodeQ = ((DynAssertionQuantifierNode)node);
					result = (List<string>)nodeQ.Expression.FreeVariables();
					result.RemoveAll(item => item.Equals(nodeQ.Variable));
					return result;

				case DynAssertionExpressionNode.NodeKind.IfThenElse:
					var nodeIf = ((DynAssertionIfThenElseNode)node);
					result = (List<string>)nodeIf.Condition.FreeVariables();
					result.AddRange(nodeIf.TrueBranch.FreeVariables());
					result.AddRange(nodeIf.FalseBranch.FreeVariables());
					return result;


				case DynAssertionExpressionNode.NodeKind.InstanceVariable:
				case DynAssertionExpressionNode.NodeKind.Numeric:
				case DynAssertionExpressionNode.NodeKind.String:
					return (IList<string>)new List<string>();

				case DynAssertionExpressionNode.NodeKind.Predicate:
					result = new List<string>();
					foreach (DynAssertionExpressionNode exp in ((DynAssertionPredicateNode)node).Arguments) {
						result.AddRange(exp.FreeVariables());
					}
					return result;

				case DynAssertionExpressionNode.NodeKind.TypeCheck:
					return ((DynAssertionTypeCheckNode)node).Expression.FreeVariables();

				default:
					throw new ArgumentOutOfRangeException();
			}
		}

	}
}

