﻿using System;
using System.Collections.Generic;

namespace Dyn.Ast
{
#if DEBUG
    internal sealed class Dyn_ListDebugView<T>
    {
        private ImmutableList<T> _collection;

        [System.Diagnostics.DebuggerBrowsable(System.Diagnostics.DebuggerBrowsableState.RootHidden)]
        public T[] Items
        {
            get
            {
                T[] array = new T[this._collection.Count];
                this._collection.CopyTo(array, 0);
                return array;
            }
        }

        public Dyn_ListDebugView(ImmutableList<T> collection)
        {
            if (collection == null) { throw new ArgumentNullException("collection"); }
            this._collection = collection;
        }
    }
#endif

#if DEBUG
    [System.Diagnostics.DebuggerDisplay("Count = {Count}"), System.Diagnostics.DebuggerTypeProxy(typeof(Dyn_ListDebugView<>))]
    [System.Diagnostics.DebuggerStepThrough]
#endif
    public sealed class ImmutableList<T> : IList<T>
    {
        private static void ThrowNotSupportedException()
        {
            throw new NotSupportedException("Cannot change immutable array.");
        }

        private static readonly ImmutableList<T> _EMPTY = new ImmutableList<T>(new T[0]);

        public static ImmutableList<T> Empty { get { return _EMPTY; } }

        public ImmutableList(T[] array)
        {
            Guard.EnsureNotNull("array", array);
            this._data = (T[])array.Clone();
        }

        public ImmutableList(IEnumerable<T> source)
        {
            Guard.EnsureNotNull("source", source);

            var immutable = source as ImmutableList<T>;

            if (immutable != null)
            {
                this._data = immutable._data;
                return;
            }

            var collection = source as ICollection<T>;

            if (collection != null)
            {
                this._data = new T[collection.Count];
                collection.CopyTo(this._data, 0);
                return;
            }

            this._data = new List<T>(source).ToArray();
        }

        private readonly T[] _data;

        #region IList<T> Members

        public T this[int index]
        {
            get
            {
                return this._data[index];
            }
        }

        #endregion IList<T> Members

        #region ICollection<T> Members

        public int IndexOf(T item)
        {
            return ((IList<T>)this._data).IndexOf(item);
        }

        public bool Contains(T item)
        {
            return ((IList<T>)this._data).Contains(item);
        }

        public void CopyTo(T[] array, int arrayIndex)
        {
            this._data.CopyTo(array, arrayIndex);
        }

        public int Count
        {
            get { return this._data.Length; }
        }

        #endregion ICollection<T> Members

        #region IEnumerable<T> Members

        public IEnumerator<T> GetEnumerator()
        {
            return ((IEnumerable<T>)this._data).GetEnumerator();
        }

        #endregion IEnumerable<T> Members

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<T>)this).GetEnumerator();
        }

        #endregion IEnumerable Members

        #region IList<T> Members

        void IList<T>.Insert(int index, T item)
        {
            ThrowNotSupportedException();
        }

        void IList<T>.RemoveAt(int index)
        {
            ThrowNotSupportedException();
        }

        T IList<T>.this[int index]
        {
            get
            {
                return this._data[index];
            }
            set
            {
                ThrowNotSupportedException();
            }
        }

        #endregion IList<T> Members

        #region ICollection<T> Members

        void ICollection<T>.Add(T item)
        {
            ThrowNotSupportedException();
        }

        void ICollection<T>.Clear()
        {
            ThrowNotSupportedException();
        }

        bool ICollection<T>.IsReadOnly
        {
            get { return true; }
        }

        bool ICollection<T>.Remove(T item)
        {
            ThrowNotSupportedException();
            return false;
        }

        #endregion ICollection<T> Members
    }
}
