﻿using Dyn.Lexer;

namespace Dyn.Ast
{
    [System.Diagnostics.DebuggerStepThrough]
    internal static class OperatorTranslation
    {
        private enum AssertionOperatorPrecedence : int
        {
            None = 0,
            Quantifier = 10,
            IfThenElse = 20,
            Boolean = 30,
            Equality = 40,
            Relational = 50,
            Addition = 60,
            Multiplication = 70,
            InstanceVariable = 80,
            TypeCheck = 90,
            Negation = 100,
            Parenthesis = 200,
        }

        public static int AssertionPrecedence(DynAssertionExpressionNode.NodeKind kind)
        {
            switch (kind)
            {
                case DynAssertionExpressionNode.NodeKind.String:
                case DynAssertionExpressionNode.NodeKind.Numeric:
                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                case DynAssertionExpressionNode.NodeKind.Predicate:
                    return 0;

                case DynAssertionExpressionNode.NodeKind.Existential:
                case DynAssertionExpressionNode.NodeKind.Universal:
                    return (int)AssertionOperatorPrecedence.Quantifier;

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    return (int)AssertionOperatorPrecedence.IfThenElse;

                case DynAssertionExpressionNode.NodeKind.And:
                case DynAssertionExpressionNode.NodeKind.Or:
                case DynAssertionExpressionNode.NodeKind.Implies:
                    return (int)AssertionOperatorPrecedence.Boolean;

                case DynAssertionExpressionNode.NodeKind.Equality:
                    return (int)AssertionOperatorPrecedence.Equality;

                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.LessThan:
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                    return (int)AssertionOperatorPrecedence.Relational;

                case DynAssertionExpressionNode.NodeKind.Addition:
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                    return (int)AssertionOperatorPrecedence.Addition;

                case DynAssertionExpressionNode.NodeKind.Multiplication:
                case DynAssertionExpressionNode.NodeKind.Division:
                case DynAssertionExpressionNode.NodeKind.Modulus:
                    return (int)AssertionOperatorPrecedence.Multiplication;

                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    return (int)AssertionOperatorPrecedence.InstanceVariable;

                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    return (int)AssertionOperatorPrecedence.TypeCheck;

                case DynAssertionExpressionNode.NodeKind.Negate:
                    return (int)AssertionOperatorPrecedence.Negation;

                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("kind", (int)kind);
            }
        }

        public static int AssertionPrecedence(SyntaxKind kind)
        {
            switch (kind)
            {
                case SyntaxKind.ExistsKeyword:
                case SyntaxKind.ForAllKeyword:
                    return (int)AssertionOperatorPrecedence.Quantifier;

                case SyntaxKind.IfKeyword:
                    return (int)AssertionOperatorPrecedence.IfThenElse;

                case SyntaxKind.AndOperator:
                case SyntaxKind.OrOperator:
                case SyntaxKind.ImpliesOperator:
                    return (int)AssertionOperatorPrecedence.Boolean;

                case SyntaxKind.EqualityOperator:
                    return (int)AssertionOperatorPrecedence.Equality;

                case SyntaxKind.LessThanOperator:
                case SyntaxKind.LessThanOrEqualOperator:
                case SyntaxKind.GreaterThanOperator:
                case SyntaxKind.GreaterThanOrEqualOperator:
                    return (int)AssertionOperatorPrecedence.Relational;

                case SyntaxKind.AdditionOperator:
                case SyntaxKind.SubtractionOperator:
                    return (int)AssertionOperatorPrecedence.Addition;

                case SyntaxKind.MultiplicationOperator:
                case SyntaxKind.DivisionOperator:
                case SyntaxKind.ModulosOperator:
                    return (int)AssertionOperatorPrecedence.Multiplication;

                case SyntaxKind.DotSymbol:
                    //                case SyntaxKind.AtSymbol:
                    return (int)AssertionOperatorPrecedence.InstanceVariable;

                case SyntaxKind.InKeyword:
                    //                case SyntaxKind.IsAKeyword:
                    return (int)AssertionOperatorPrecedence.TypeCheck;

                case SyntaxKind.ExclamationMarkSymbol:
                    return (int)AssertionOperatorPrecedence.Negation;

                case SyntaxKind.ParenthesisLeftSymbol:
                    return (int)AssertionOperatorPrecedence.Parenthesis;

                case SyntaxKind.IdentifierKind:
                case SyntaxKind.NumericKind:
                case SyntaxKind.StringKind:
                    return 0;
                // invalid
                case SyntaxKind.RequiresKeyword:
                case SyntaxKind.EnsuresKeyword:
                case SyntaxKind.InvariantKeyword:
                case SyntaxKind.AssertKeyword:
                case SyntaxKind.WhileKeyword:
                case SyntaxKind.IdentityOperator:
                case SyntaxKind.AssignmentOperator:
                case SyntaxKind.NewKeyword:
                    return 0;

                default:
                    return 0;
            }
        }

        private enum ProgramOperatorPrecedence : int
        {
            None = 0,
            Assignment = 10,
            Boolean = 20,
            Equality = 30,
            Relational = 40,
            TypeCheck = 50,
            Addition = 60,
            Multiplication = 70,
            MethodCall = 80,
            Parenthesis = 200,
        }

        public static int ProgramPrecedence(DynProgramStatementNode.NodeKind kind)
        {
            switch (kind)
            {
                case DynProgramStatementNode.NodeKind.Assert:
                case DynProgramStatementNode.NodeKind.Ensures:
                case DynProgramStatementNode.NodeKind.Requires:
                case DynProgramStatementNode.NodeKind.Invariant:
                case DynProgramStatementNode.NodeKind.ByTheWay:
                    return 0;

                case DynProgramStatementNode.NodeKind.InstanceVariable:
                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.If:
                case DynProgramStatementNode.NodeKind.While:
                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return 0;

                case DynProgramStatementNode.NodeKind.Assignment:
                    return (int)ProgramOperatorPrecedence.Assignment;

                case DynProgramStatementNode.NodeKind.MethodCall:
                    return (int)ProgramOperatorPrecedence.MethodCall;

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("kind", (int)kind);
            }
        }

        public static int ProgramPrecedence(SyntaxKind kind)
        {
            switch (kind)
            {
                case SyntaxKind.RequiresKeyword:
                case SyntaxKind.EnsuresKeyword:
                case SyntaxKind.InvariantKeyword:
                case SyntaxKind.AssertKeyword:
                case SyntaxKind.ExclamationMarkSymbol:
                case SyntaxKind.InKeyword:
                    return 0;

                case SyntaxKind.IfKeyword:
                case SyntaxKind.WhileKeyword:
                case SyntaxKind.NumericKind:
                case SyntaxKind.StringKind:
                case SyntaxKind.NewKeyword:
                    return 0;

                case SyntaxKind.AtSymbol:
                case SyntaxKind.IdentifierKind:
                    return 0;

                case SyntaxKind.DotSymbol:
                    return (int)ProgramOperatorPrecedence.MethodCall;

                case SyntaxKind.ParenthesisLeftSymbol:
                    return (int)ProgramOperatorPrecedence.Parenthesis;

                case SyntaxKind.AssignmentOperator:
                    return (int)ProgramOperatorPrecedence.Assignment;

                case SyntaxKind.AndOperator:
                case SyntaxKind.OrOperator:
                case SyntaxKind.ImpliesOperator:
                    return (int)ProgramOperatorPrecedence.Boolean;

                case SyntaxKind.EqualityOperator:
                case SyntaxKind.IdentityOperator:
                    return (int)ProgramOperatorPrecedence.Equality;

                case SyntaxKind.LessThanOperator:
                case SyntaxKind.LessThanOrEqualOperator:
                case SyntaxKind.GreaterThanOperator:
                case SyntaxKind.GreaterThanOrEqualOperator:
                    return (int)ProgramOperatorPrecedence.Relational;

                case SyntaxKind.AdditionOperator:
                case SyntaxKind.SubtractionOperator:
                    return (int)ProgramOperatorPrecedence.Addition;

                case SyntaxKind.MultiplicationOperator:
                case SyntaxKind.DivisionOperator:
                case SyntaxKind.ModulosOperator:
                    return (int)ProgramOperatorPrecedence.Multiplication;

                case SyntaxKind.IsAKeyword:
                    return (int)ProgramOperatorPrecedence.TypeCheck;

                default:
                    return 0;
            }
        }

        public static SyntaxKind MethodToKind(string methodName)
        {
            switch (methodName)
            {
                case "op_add": return SyntaxKind.AdditionOperator;
                case "op_and": return SyntaxKind.AndOperator;
                case "op_div": return SyntaxKind.DivisionOperator;
                case "op_equal": return SyntaxKind.EqualityOperator;
                case "op_gt": return SyntaxKind.GreaterThanOperator;
                case "op_ge": return SyntaxKind.GreaterThanOrEqualOperator;
                case "op_identical": return SyntaxKind.IdentityOperator;
                case "op_isa": return SyntaxKind.IsAKeyword;
                case "op_lt": return SyntaxKind.LessThanOperator;
                case "op_le": return SyntaxKind.LessThanOrEqualOperator;
                case "op_multiply": return SyntaxKind.MultiplicationOperator;
                case "op_or": return SyntaxKind.OrOperator;
                case "op_substract": return SyntaxKind.SubtractionOperator;
                case "op_mod": return SyntaxKind.ModulosOperator;
                default:
                    return SyntaxKind.EndOfStreamKind;
            }
        }

        public static string KindToMethod(SyntaxKind kind)
        {
            switch (kind)
            {
                case SyntaxKind.AdditionOperator:
                    return "op_add";

                case SyntaxKind.AndOperator:
                    return "op_and";

                case SyntaxKind.DivisionOperator:
                    return "op_div";

                case SyntaxKind.EqualityOperator:
                    return "op_equal";

                case SyntaxKind.GreaterThanOperator:
                    return "op_gt";

                case SyntaxKind.GreaterThanOrEqualOperator:
                    return "op_ge";

                case SyntaxKind.IdentityOperator:
                    return "op_identical";

                case SyntaxKind.IsAKeyword:
                    return "op_isa";

                case SyntaxKind.LessThanOperator:
                    return "op_lt";

                case SyntaxKind.LessThanOrEqualOperator:
                    return "op_le";

                case SyntaxKind.MultiplicationOperator:
                    return "op_multiply";

                case SyntaxKind.OrOperator:
                    return "op_or";

                case SyntaxKind.SubtractionOperator:
                    return "op_substract";

                case SyntaxKind.ModulosOperator:
                    return "op_mod";

                default:
                    throw Guard.EnumOutOfRangeException<SyntaxKind>("op", (int)kind);
            }
        }
    }
}
