﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Dyn.Ast
{
    public static class Utilities
    {
        public static ImmutableList<T> AsImmutableList<T>(IEnumerable<T> source) where T : class
        {
            if (source == null)
            {
                return ImmutableList<T>.Empty;
            }

            var immutable = source as ImmutableList<T>;

            if (immutable != null)
            {
                return immutable;
            }
            var array = source as T[];

            if (array != null)
            {
                return new ImmutableList<T>(array);
            }

            return new ImmutableList<T>(source);
        }

        public static StringBuilder Join(this StringBuilder builder, string delimeter, IEnumerable<string> source)
        {
            bool tail = false;

            foreach (var element in source)
            {
                if (tail)
                {
                    builder.Append(delimeter);
                }

                tail = true;
                builder.Append(element);
            }

            return builder;
        }

        internal const int HASH_PRIME = 127;
        internal const int HASH_SEED = 0x6d68c3d1;

        public static int EnumerableDeepGetHashCode<T>(IEnumerable<T> source, Func<T, int> func)
        {
            Guard.EnsureNotNull("source", source);
            Guard.EnsureNotNull("func", func);

            int result = 0;

            foreach (var element in source)
            {
                result = result * HASH_PRIME + func(element);
            }

            return result;
        }

        internal static bool EnumerableDeepEquals<T>(IEnumerable<T> first, IEnumerable<T> second, Func<T, T, bool> func)
        {
            Guard.EnsureNotNull("func", func);

            if (object.ReferenceEquals(first, second)) { return true; };
            if (object.ReferenceEquals(first, null) || object.ReferenceEquals(second, null)) { return false; };

            bool mn1, mn2;

            using (IEnumerator<T> e1 = first.GetEnumerator(), e2 = second.GetEnumerator())
            {
                while ((mn1 = e1.MoveNext()) & (mn2 = e2.MoveNext()))
                {
                    if (!func(e1.Current, e2.Current))
                    {
                        return false;
                    }
                }
            }

            return mn1 == false && mn2 == false;
        }

        internal static IList<T> Transform<T>(IList<T> source, Func<T, T> transform)
        {
            return source.Select(transform).ToList();
        }
    }
}
