﻿using Dyn.Lexer;
using System.Text;

namespace Dyn.Diagnostics
{

    /// <summary>
    /// Represents a diagnostic annotation of a location where it occurred.
    /// </summary>
    public class DiagnosticInfo
    {
        private readonly Span _span;
        private readonly string _message;
        private readonly DiagnosticSeverity _severity;

        /// <summary>
        /// Creates a <see cref="DiagnosticInfo"/> instance.
        /// </summary>
        /// <param name="severity">The severity of the diagnostic.</param>
        /// <param name="span">The location of the diagnostic.</param>
        /// <param name="message"></param>
        public DiagnosticInfo(DiagnosticSeverity severity, Span span, string message)
        {
            Guard.EnsureNotNull("message", message);
            Guard.EnsureEnumDefined<DiagnosticSeverity>("level", (int)severity);

            this._span = span;
            this._message = message;
            this._severity = severity;
        }

        /// <summary>
        /// Gets the <see cref="Location"/> of the diagnostic.
        /// </summary>
        public Span Location { get { return this._span; } }

        /// <summary>
        /// Gets the diagnostic description.
        /// </summary>
        public string Message { get { return this._message; } }

        /// <summary>
        /// Get the diagnostic's severity.
        /// </summary>
        public DiagnosticSeverity Level { get { return this._severity; } }

        public override string ToString()
        {
            var builder = new StringBuilder();

            switch (this.Level)
            {
                case DiagnosticSeverity.None:
                    builder.Append("[None]\t"); break;
                case DiagnosticSeverity.Information:
                    builder.Append("[Info]\t"); break;
                case DiagnosticSeverity.Warning:
                    builder.Append("[Warn]\t"); break;
                case DiagnosticSeverity.Error:
                    builder.Append("[Fail]\t"); break;
            }

            builder.Append(this.Location).Append('\t').Append(this.Message);

            return builder.ToString();
        }
    }
}
