﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Dyn.Lexer
{
    /// <summary>
    /// Extracts single spans and its type from an input stream or string.
    /// </summary>
    /// <remarks>
    /// <see cref="DynLexer"/> does not implement <see cref="IDisposable"/> nor disposes the reader given to a constructor.
    /// The caller is responsible for resource management.
    /// </remarks>
    public sealed class DynLexer : IEnumerable<SyntaxToken>
    {
        private readonly SourceReaderHelper _source;

        /// <summary>Instantiates a new instance of a <see cref="Span"/> structure with the source as a <see cref="string"/>.</summary>
        /// <param name="source">The source.</param>
        public DynLexer(string source)
            : this(string.Empty, new StringReader(source))
        { }

        /// <summary>Instantiates a new instance of a <see cref="Span"/> structure with the source as a <see cref="TextReader"/>.</summary>
        /// <param name="reader">The source.</param>
        public DynLexer(TextReader reader)
            : this(string.Empty, reader)
        { }

        /// <summary>Instantiates a new instance of a <see cref="Span"/> structure with a file name and the source as a <see cref="string"/>.</summary>
        /// <param name="fileName">The file name.</param>
        /// <param name="source">The source.</param>
        public DynLexer(string fileName, string source)
            : this(fileName, new StringReader(source))
        { }

        /// <summary>Instantiates a new instance of a <see cref="Span"/> structure with a file name and the source as a <see cref="TextReader"/>.</summary>
        /// <param name="fileName">The file name.</param>
        /// <param name="reader">The source.</param>
        public DynLexer(string fileName, TextReader reader)
        {
            Guard.EnsureNotNull("fileName", fileName);
            Guard.EnsureNotNull("reader", reader);
            this._source = new SourceReaderHelper(fileName, reader);
        }

        /// <summary>
        /// Gets the filename associated with lexer.
        /// </summary>
        public string FileName
        {
            get
            {
                return this._source.FileName;
            }
        }

        // input equivalence groups:
        private const byte IEQ_COUNT = 13;

        private static readonly byte[] ieqt = new byte[128]
        {
            0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x0c, 0x03, 0x01, 0x01, 0x0c, 0x01, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
            0x0c, 0x07, 0x06, 0x02, 0x01, 0x07, 0x09, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07,
            0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x07, 0x07, 0x0b, 0x08, 0x0b, 0x07,
            0x07, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x07, 0x01, 0x07, 0x01, 0x04,
            0x01, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,
            0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x07, 0x0a, 0x07, 0x01, 0x01
        };

        private const byte TRANSITION_TOKEN_COMPLETE_FLAG = 0x80;
        private const byte STATE_INITIAL = 0x00;
        private const byte STATE_INVALID = 0x01;
        private const byte TRANSITION_EOS = 0x00;
        private const byte TRANSITION_INVALID = 0x01;

        private const byte STT_STATES = 14;

        private static readonly byte[] stt = new byte[IEQ_COUNT * STT_STATES]
        {
            0x80, 0x01, 0x03, 0x0d, 0x04, 0x05, 0x07, 0x0a, 0x09, 0x0c, 0x0b, 0x08, 0x0d,
            0x80, 0x01, 0x83, 0x8d, 0x84, 0x85, 0x87, 0x8a, 0x89, 0x8c, 0x8b, 0x88, 0x8d,
            0x80, 0x81, 0x83, 0x8d, 0x84, 0x85, 0x87, 0x8a, 0x89, 0x8c, 0x8b, 0x88, 0x8d,
            0x80, 0x03, 0x03, 0x02, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03,
            0x80, 0x81, 0x83, 0x8d, 0x04, 0x04, 0x87, 0x8a, 0x89, 0x8c, 0x8b, 0x88, 0x8d,
            0x80, 0x81, 0x83, 0x8d, 0x84, 0x05, 0x87, 0x8a, 0x89, 0x8c, 0x8b, 0x88, 0x8d,
            0x80, 0x81, 0x83, 0x8d, 0x84, 0x85, 0x87, 0x8a, 0x89, 0x8c, 0x8b, 0x88, 0x8d,
            0x80, 0x07, 0x07, 0x07, 0x07, 0x07, 0x06, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07,
            0x80, 0x81, 0x83, 0x8d, 0x84, 0x85, 0x87, 0x8a, 0x0a, 0x8c, 0x8b, 0x88, 0x8d,
            0x80, 0x81, 0x83, 0x8d, 0x84, 0x85, 0x87, 0x8a, 0x08, 0x8c, 0x8b, 0x88, 0x8d,
            0x80, 0x81, 0x83, 0x8d, 0x84, 0x85, 0x87, 0x8a, 0x89, 0x8c, 0x8b, 0x88, 0x8d,
            0x80, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x0a, 0x01, 0x01,
            0x80, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x0a, 0x01, 0x01, 0x01,
            0x80, 0x81, 0x83, 0x0d, 0x84, 0x85, 0x87, 0x8a, 0x89, 0x8c, 0x8b, 0x88, 0x0d
        };

        private IEnumerator<SyntaxToken> GetTokens()
        {
            if (_source.BeforeContent)
            {
                _source.TryNext();
            }

            if (_source.AfterContent)
            {
                yield return new SyntaxToken(SyntaxKind.EndOfStreamKind, _source.CreateSpan());
                yield break;
            }

            Span start = _source.CreateSpan();

            short state = STATE_INITIAL;

            StringBuilder builder = new StringBuilder();

            while (true)
            {
                char c = _source.Current;

                int it = ieqt[(c < 0 || c > ieqt.Length) ? TRANSITION_INVALID : (byte)c];

                short nextState = stt[state * IEQ_COUNT + it];

                if ((nextState & TRANSITION_TOKEN_COMPLETE_FLAG) == TRANSITION_TOKEN_COMPLETE_FLAG)
                {
                    Span span = new Span(start.FileName, start.Position, start.Line, start.Column, builder.Length);

                    switch (state & ~TRANSITION_TOKEN_COMPLETE_FLAG)
                    {
                        /// end of stream
                        case STATE_INITIAL:
                            yield return new SyntaxToken(SyntaxKind.EndOfStreamKind, span);
                            yield break;
                        case 2:
                            yield return ParseComments(span, builder);
                            break;

                        case 4:
                            yield return ParseIdentifiers(span, builder);
                            break;

                        case 5:
                            yield return ParseNumeric(span, builder);
                            break;

                        case 6:
                            yield return ParseStrings(span, builder);
                            break;

                        case 8:
                        case 9:
                        case 10:
                            yield return ParseSymbols(span, builder);
                            break;

                        case 13:
                            yield return ParseWhitespaces(span, builder);
                            break;

                        case 1:
                        case 3:
                        case 7:
                        case 11:
                        case 12:
                            yield return ParseInvalid(span, builder);
                            break;

                        default:
                            throw new NotImplementedException("Ooops");
                    }

                    builder.Length = 0;
                    start = _source.CreateSpan();
                }
                if (c != '\0')
                {
                    builder.Append(c);
                }

                state = (short)(nextState & ~TRANSITION_TOKEN_COMPLETE_FLAG);

                _source.TryNext();
            }
        }

        #region Parsing

        private static SyntaxToken ParseInvalid(Span span, StringBuilder value)
        {
            return new SyntaxToken(SyntaxKind.InvalidKind, span, value.ToString());
        }

        private static SyntaxToken ParseComments(Span span, StringBuilder value)
        {
            return new SyntaxToken(SyntaxKind.CommentKind, span, value.ToString());
        }

        private static SyntaxToken ParseIdentifiers(Span span, StringBuilder value)
        {
            var identifier = value.ToString();

            //TODO: Make it fast
            switch (identifier)
            {
                case "class":
                    return new SyntaxToken(SyntaxKind.ClassKeyword, span);

                case "do":
                    return new SyntaxToken(SyntaxKind.DoKeyword, span);

                case "done":
                    return new SyntaxToken(SyntaxKind.DoneKeyword, span);

                case "else":
                    return new SyntaxToken(SyntaxKind.ElseKeyword, span);

                case "end":
                    return new SyntaxToken(SyntaxKind.EndKeyword, span);

                case "if":
                    return new SyntaxToken(SyntaxKind.IfKeyword, span);

                case "in":
                    return new SyntaxToken(SyntaxKind.InKeyword, span);

                case "is_a":
                    return new SyntaxToken(SyntaxKind.IsAKeyword, span);

                case "method":
                    return new SyntaxToken(SyntaxKind.MethodKeyword, span);

                case "new":
                    return new SyntaxToken(SyntaxKind.NewKeyword, span);

                case "null":
                    return new SyntaxToken(SyntaxKind.NullKeyword, span);

                case "rename":
                    return new SyntaxToken(SyntaxKind.RenameKeyword, span);

                case "self":
                    return new SyntaxToken(SyntaxKind.SelfKeyword, span);

                case "then":
                    return new SyntaxToken(SyntaxKind.ThenKeyword, span);

                case "while":
                    return new SyntaxToken(SyntaxKind.WhileKeyword, span);

                case "false":
                    return new SyntaxToken(SyntaxKind.FalseKeyword, span);

                case "true":
                    return new SyntaxToken(SyntaxKind.TrueKeyword, span);

                case "requires":
                    return new SyntaxToken(SyntaxKind.RequiresKeyword, span);

                case "ensures":
                    return new SyntaxToken(SyntaxKind.EnsuresKeyword, span);

                case "exists":
                    return new SyntaxToken(SyntaxKind.ExistsKeyword, span);

                case "forall":
                    return new SyntaxToken(SyntaxKind.ForAllKeyword, span);

                case "invariant":
                    return new SyntaxToken(SyntaxKind.InvariantKeyword, span);

                case "assert":
                    return new SyntaxToken(SyntaxKind.AssertKeyword, span);

                case "implies":
                    return new SyntaxToken(SyntaxKind.ImpliesOperator, span);

                case "btw":
                    return new SyntaxToken(SyntaxKind.ByTheWayKeyword, span);

                default:
                    return new SyntaxToken(SyntaxKind.IdentifierKind, span, identifier);
            }
        }

        private static SyntaxToken ParseNumeric(Span span, StringBuilder value)
        {
            var number = value.ToString();

            int result;

            if (int.TryParse(number, out result))
            {
                return new SyntaxToken(SyntaxKind.NumericKind, span, result);
            }
            return new SyntaxToken(SyntaxKind.InvalidKind, span, number);
        }

        private static SyntaxToken ParseStrings(Span span, StringBuilder value)
        {
            return new SyntaxToken(SyntaxKind.StringKind, span, value.ToString());
        }

        private static SyntaxToken ParseSymbols(Span span, StringBuilder value)
        {
            int len = value.Length;

            if (len == 1)
            {
                switch (value[0])
                {
                    case '+':
                        return new SyntaxToken(SyntaxKind.AdditionOperator, span);

                    case '-':
                        return new SyntaxToken(SyntaxKind.SubtractionOperator, span);

                    case '*':
                        return new SyntaxToken(SyntaxKind.MultiplicationOperator, span);

                    case '/':
                        return new SyntaxToken(SyntaxKind.DivisionOperator, span);

                    case '<':
                        return new SyntaxToken(SyntaxKind.LessThanOperator, span);

                    case '>':
                        return new SyntaxToken(SyntaxKind.GreaterThanOperator, span);

                    case '=':
                        return new SyntaxToken(SyntaxKind.AssignmentOperator, span);

                    case '?':
                        return new SyntaxToken(SyntaxKind.QuestionMarkSymbol, span);

                    case '(':
                        return new SyntaxToken(SyntaxKind.ParenthesisLeftSymbol, span);

                    case ')':
                        return new SyntaxToken(SyntaxKind.ParenthesisRightSymbol, span);

                    case '[':
                        return new SyntaxToken(SyntaxKind.SquareBracketLeftSymbol, span);

                    case ']':
                        return new SyntaxToken(SyntaxKind.SquareBracketRightSymbol, span);

                    case '{':
                        return new SyntaxToken(SyntaxKind.CurlyBracketLeftSymbol, span);

                    case '}':
                        return new SyntaxToken(SyntaxKind.CurlyBracketRightSymbol, span);

                    case '@':
                        return new SyntaxToken(SyntaxKind.AtSymbol, span);

                    case '.':
                        return new SyntaxToken(SyntaxKind.DotSymbol, span);

                    case ',':
                        return new SyntaxToken(SyntaxKind.CommaSymbol, span);

                    case ';':
                        return new SyntaxToken(SyntaxKind.SemicolonSymbol, span);

                    case '!':
                        return new SyntaxToken(SyntaxKind.ExclamationMarkSymbol, span);

                    case ':':
                        return new SyntaxToken(SyntaxKind.ColonSymbol, span);

                    case '%':
                        return new SyntaxToken(SyntaxKind.ModulosOperator, span);

                    case '\'':
                        return new SyntaxToken(SyntaxKind.ApostropheSymbol, span);

                    default:
                        return new SyntaxToken(SyntaxKind.InvalidKind, span, value.ToString());
                }
            }
            else if (len == 2)
            {
                switch (value[0] << 8 | value[1])
                {
                    case '&' << 8 | '&':
                        return new SyntaxToken(SyntaxKind.AndOperator, span);

                    case '|' << 8 | '|':
                        return new SyntaxToken(SyntaxKind.OrOperator, span);

                    case '<' << 8 | '=':
                        return new SyntaxToken(SyntaxKind.LessThanOrEqualOperator, span);

                    case '>' << 8 | '=':
                        return new SyntaxToken(SyntaxKind.GreaterThanOrEqualOperator, span);

                    case '=' << 8 | '=':
                        return new SyntaxToken(SyntaxKind.EqualityOperator, span);

                    default:
                        break;
                }
            }
            else if (len == 3 && value[0] == '=' && value[1] == '=' && value[2] == '=')
            {
                return new SyntaxToken(SyntaxKind.IdentityOperator, span);
            }

            return new SyntaxToken(SyntaxKind.InvalidKind, span, value.ToString());
        }

        private static SyntaxToken ParseWhitespaces(Span span, StringBuilder value)
        {
            return new SyntaxToken(SyntaxKind.WhitespaceKind, span, value.ToString());
        }

        #endregion Parsing

        /// <summary>Returns an enumerator that iterates through lexed <see cref="SyntaxToken"/>.</summary>
        /// <returns>a <see cref="IEnumerator{SyntaxToken}"/> that can be used to iterate through lexed.</returns>
        public IEnumerator<SyntaxToken> GetEnumerator()
        {
            return GetTokens();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<SyntaxToken>)this).GetEnumerator();
        }
    }
}
