using System;
using System.Diagnostics;
using System.Text;

namespace Dyn.Lexer
{
    /// <summary>Represents a location of a source file.</summary>
    /// <remarks>This structure is immutable.</remarks>
    [DebuggerDisplay("{ToString()}")]
    public struct Span : IEquatable<Span>
    {
        private readonly string _fileName;
        private readonly int _position;
        private readonly int _line;
        private readonly int _column;
        private readonly int length;

        /// <summary>Instantiates a new instance of a <see cref="Span"/> structure.</summary>
        /// <param name="fileName">The file name.</param>
        /// <param name="position">The position relative to the start of the file.</param>
        /// <param name="line">The line starting with zero.</param>
        /// <param name="column">The column starting with zero.</param>
        /// <param name="length">The span length.</param>
        public Span(string fileName, int position, int line, int column, int length)
        {
            Guard.EnsureNotNull("fileName", fileName);
            Guard.EnsureNonnegative("position", position);
            Guard.EnsureNonnegative("line", line);
            Guard.EnsureNonnegative("column", column);
            Guard.EnsureNonnegative("length", length);

            this._fileName = fileName;
            this._position = position;
            this._line = line;
            this._column = column;
            this.length = length;
        }

        /// <summary>Gets the file name.</summary>
        public string FileName { get { return this._fileName ?? string.Empty; } }

        /// <summary>Gets the position relative to the start of the file.</summary>
        public int Position { get { return this._position; } }

        /// <summary>Gets the file name.</summary>
        public int Line { get { return this._line; } }

        /// <summary>Gets the line starting with zero.</summary>
        public int Column { get { return this._column; } }

        /// <summary>Gets the column starting with zero.</summary>
        public int Length { get { return this.length; } }

        public override string ToString()
        {
            StringBuilder builder = new StringBuilder();

            if (!string.IsNullOrEmpty(FileName))
            {
                builder.Append(FileName).Append(':');
            }

            builder.Append(Line).Append(':').Append(Column);

            if (Length > 0)
            {
                builder.Append("+").Append(Length);
            }

            return builder.ToString();
        }

        public bool Equals(Span other)
        {
            return StringComparer.Ordinal.Equals(this.FileName, other.FileName) &&
                this.Position == other.Position &&
                this.Line == other.Line &&
                this.Column == other.Column &&
                this.Length == other.Length;
        }

        public override bool Equals(object obj)
        {
            return obj is Span && Equals((Span)obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                const int PRIME = 37;
                int result = 1;
                result = result * PRIME ^ FileName.GetHashCode();
                result = result * PRIME ^ Position.GetHashCode();
                result = result * PRIME ^ Line.GetHashCode();
                result = result * PRIME ^ Column.GetHashCode();
                result = result * PRIME ^ Length.GetHashCode();
                return result;
            }
        }

        /// <summary>Determines whether two specified instances of <see cref="Span"/> are equal.</summary>
        /// <param name="first">A <see cref="Span"/>.</param>
        /// <param name="second">A <see cref="Span"/>.</param>
        /// <returns>
        ///     <code>true</code> if <paramref name="left"/> and <paramref name="right"/>
        ///     represent the same location; otherwise <code>false</code>.
        /// </returns>
        public static bool operator ==(Span first, Span second)
        {
            return first.Equals(second);
        }

        /// <summary>Determines whether two specified instances of <see cref="Span"/> ar note equal.</summary>
        /// <param name="first">A <see cref="Span"/>.</param>
        /// <param name="second">A <see cref="Span"/>.</param>
        /// <returns>
        ///     <code>false</code> if <paramref name="left"/> and <paramref name="right"/>
        ///     represent the same location; otherwise <code>true</code>.
        /// </returns>
        public static bool operator !=(Span first, Span second)
        {
            return !first.Equals(second);
        }

        /// <summary>Represents an empty, default or invalid span.</summary>
        public readonly static Span Empty = new Span(string.Empty, 0, 0, 0, 0);

        /// <summary>
        /// Gets a <see cref="Span"/> structure that bounds the union of the two <see cref="Span"/> structures.
        /// </summary>
        /// <param name="first">A <see cref="Span"/>.</param>
        /// <param name="second">A <see cref="Span"/>.</param>
        /// <returns>A <see cref="Span"/> that bounds the union of the two <see cref="Span"/> structures.
        /// <para>Note: The <see cref="Span"/> structure with the smaller  <see cref="Position"/> determines <see cref="FileName"/>, <see cref="Line"/> and <see cref="Column"/>.</para>
        /// </returns>
        public static Span Union(Span first, Span second)
        {
            Span min;
            if (first.Position < second.Position)
            {
                min = first;
            }
            else
            {
                min = second;
            }

            int maxEnd = Math.Max(first.Position + first.length, second.Position + second.length);

            return new Span(
                min.FileName,
                min.Position,
                min.Line,
                min.Column,
                maxEnd - min.Position);
        }
    }
}
