﻿using System;
using System.Text;

namespace Dyn.Lexer
{
    public enum SyntaxKind
    {
        /// <summary>Represents the end of the stream.</summary>
        EndOfStreamKind = 0,

        InvalidKind,
        WhitespaceKind,
        IdentifierKind,
        StringKind,
        CommentKind,

        // Kind with numeric value attached
        NumericKind,

        // Operators
        AssignmentOperator,

        EqualityOperator,
        IdentityOperator,
        AdditionOperator,
        SubtractionOperator,
        MultiplicationOperator,
        DivisionOperator,
        LessThanOperator,
        GreaterThanOperator,
        LessThanOrEqualOperator,
        GreaterThanOrEqualOperator,
        ModulosOperator,
        AndOperator,
        OrOperator,
        ImpliesOperator,

        // Symbols
        ParenthesisLeftSymbol,

        ParenthesisRightSymbol,
        SquareBracketLeftSymbol,
        SquareBracketRightSymbol,
        CurlyBracketLeftSymbol,
        CurlyBracketRightSymbol,
        AtSymbol,
        DotSymbol,
        CommaSymbol,
        SemicolonSymbol,
        QuestionMarkSymbol,
        ExclamationMarkSymbol,
        ColonSymbol,
        ApostropheSymbol,

        // Keywords
        SelfKeyword,

        NewKeyword,
        IfKeyword,
        ThenKeyword,
        ElseKeyword,
        EndKeyword,
        WhileKeyword,
        DoKeyword,
        DoneKeyword,
        MethodKeyword,
        RenameKeyword,
        NullKeyword,
        ClassKeyword,
        IsAKeyword,
        TrueKeyword,
        FalseKeyword,

        // Assertion keywords
        RequiresKeyword,

        EnsuresKeyword,
        InvariantKeyword,
        AssertKeyword,
        ByTheWayKeyword,

        // Quantifier
        ForAllKeyword,

        ExistsKeyword,

        InKeyword
    }

    /// <summary>
    /// Represents a single token read by <see cref="DynLexer"/>
    /// </summary>
    /// <remarks>Immutable</remarks>
    [System.Diagnostics.DebuggerDisplay("{ToString()}")]
    public struct SyntaxToken : IEquatable<SyntaxToken>
    {
        private readonly string _stringValue;
        private readonly int _numericValue;
        private readonly SyntaxKind _kind;
        private readonly Span _span;

        private static void EnsureState(SyntaxKind kind, bool numericValue, bool stringValue)
        {
            switch (kind)
            {
                case SyntaxKind.NumericKind:
                    if (!numericValue || stringValue)
                    {
                        throw new InvalidOperationException("Expected numeric value for numeric constructor.");
                    }
                    return;

                case SyntaxKind.InvalidKind:
                case SyntaxKind.WhitespaceKind:
                case SyntaxKind.StringKind:
                case SyntaxKind.CommentKind:
                case SyntaxKind.IdentifierKind:
                    if (!stringValue || numericValue)
                    {
                        throw new InvalidOperationException("Expected string value for string constructor.");
                    }
                    return;

                default:
                    if (numericValue || stringValue)
                    {
                        throw new InvalidOperationException("Expected no value for empty constructor.");
                    }
                    return;
            }
        }

        public SyntaxToken(SyntaxKind kind, Span span, int value)
        {
            EnsureState(kind, true, false);

            this._kind = kind;
            this._span = span;

            this._stringValue = default(string);
            this._numericValue = value;
        }

        public SyntaxToken(SyntaxKind kind, Span span, string value)
        {
            Guard.EnsureNotNull("value", value);
            EnsureState(kind, false, true);

            this._kind = kind;
            this._span = span;

            this._stringValue = value;
            this._numericValue = default(int);
        }

        public SyntaxToken(SyntaxKind kind, Span span)
        {
            EnsureState(kind, false, false);

            this._kind = kind;
            this._span = span;

            this._stringValue = default(string);
            this._numericValue = default(int);
        }

        public Span Span
        {
            get
            {
                return this._span;
            }
        }

        public SyntaxKind Kind
        {
            get
            {
                return this._kind;
            }
        }

        public int Number
        {
            get
            {
                EnsureState(this._kind, true, false);
                return this._numericValue;
            }
        }

        public string String
        {
            get
            {
                EnsureState(this._kind, false, true);
                return this._stringValue;
            }
        }

        public override bool Equals(object obj)
        {
            return obj is SyntaxToken && Equals((SyntaxToken)obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                const int SEED = (int)2166136261;
                const int PRIME = 16777619;

                int result = SEED;

                result = (result * PRIME ^ (int)this._kind);
                result = (result * PRIME ^ this.Span.GetHashCode());

                switch (this._kind)
                {
                    case SyntaxKind.NumericKind:
                        return result * PRIME ^ this._numericValue;

                    case SyntaxKind.InvalidKind:
                    case SyntaxKind.WhitespaceKind:
                    case SyntaxKind.StringKind:
                    case SyntaxKind.CommentKind:
                    case SyntaxKind.IdentifierKind:
                        return result * PRIME ^ this._stringValue.GetHashCode();

                    default:
                        return result;
                }
            };
        }

        public string ToSource()
        {
            switch (this._kind)
            {
                case SyntaxKind.EndOfStreamKind:
                    return string.Empty;

                case SyntaxKind.InvalidKind:
                case SyntaxKind.WhitespaceKind:
                case SyntaxKind.CommentKind:
                case SyntaxKind.IdentifierKind:
                case SyntaxKind.StringKind:
                    return this._stringValue;

                case SyntaxKind.NumericKind:
                    return this._numericValue.ToString(System.Globalization.CultureInfo.InvariantCulture);

                case SyntaxKind.AssignmentOperator:
                    return "=";

                case SyntaxKind.EqualityOperator:
                    return "==";

                case SyntaxKind.IdentityOperator:
                    return "===";

                case SyntaxKind.AdditionOperator:
                    return "+";

                case SyntaxKind.SubtractionOperator:
                    return "-";

                case SyntaxKind.MultiplicationOperator:
                    return "*";

                case SyntaxKind.DivisionOperator:
                    return "/";

                case SyntaxKind.LessThanOperator:
                    return "<";

                case SyntaxKind.GreaterThanOperator:
                    return ">";

                case SyntaxKind.LessThanOrEqualOperator:
                    return "<=";

                case SyntaxKind.GreaterThanOrEqualOperator:
                    return ">=";

                case SyntaxKind.AndOperator:
                    return "&&";

                case SyntaxKind.OrOperator:
                    return "||";

                case SyntaxKind.ImpliesOperator:
                    return "implies";

                case SyntaxKind.ParenthesisLeftSymbol:
                    return "(";

                case SyntaxKind.ParenthesisRightSymbol:
                    return ")";

                case SyntaxKind.SquareBracketLeftSymbol:
                    return "[";

                case SyntaxKind.SquareBracketRightSymbol:
                    return "]";

                case SyntaxKind.CurlyBracketLeftSymbol:
                    return "{";

                case SyntaxKind.CurlyBracketRightSymbol:
                    return "}";

                case SyntaxKind.AtSymbol:
                    return "@";

                case SyntaxKind.DotSymbol:
                    return ".";

                case SyntaxKind.CommaSymbol:
                    return ",";

                case SyntaxKind.SemicolonSymbol:
                    return ";";
                case SyntaxKind.QuestionMarkSymbol:
                    return "?";

                case SyntaxKind.ExclamationMarkSymbol:
                    return "!";

                case SyntaxKind.ColonSymbol:
                    return ":";

                case SyntaxKind.ApostropheSymbol:
                    return "'";

                case SyntaxKind.ModulosOperator:
                    return "%";

                case SyntaxKind.SelfKeyword:
                    return "self";

                case SyntaxKind.NewKeyword:
                    return "new";

                case SyntaxKind.IfKeyword:
                    return "if";

                case SyntaxKind.ThenKeyword:
                    return "then";

                case SyntaxKind.ElseKeyword:
                    return "else";

                case SyntaxKind.EndKeyword:
                    return "end";

                case SyntaxKind.WhileKeyword:
                    return "while";

                case SyntaxKind.DoKeyword:
                    return "do";

                case SyntaxKind.DoneKeyword:
                    return "done";

                case SyntaxKind.MethodKeyword:
                    return "method";

                case SyntaxKind.RenameKeyword:
                    return "rename";

                case SyntaxKind.NullKeyword:
                    return "null";

                case SyntaxKind.ClassKeyword:
                    return "class";

                case SyntaxKind.RequiresKeyword:
                    return "requires";

                case SyntaxKind.EnsuresKeyword:
                    return "ensures";

                case SyntaxKind.InvariantKeyword:
                    return "invariant";

                case SyntaxKind.AssertKeyword:
                    return "assert";

                case SyntaxKind.ForAllKeyword:
                    return "forall";

                case SyntaxKind.ExistsKeyword:
                    return "exists";

                case SyntaxKind.InKeyword:
                    return "in";

                case SyntaxKind.IsAKeyword:
                    return "is_a";

                case SyntaxKind.TrueKeyword:
                    return "true";

                case SyntaxKind.FalseKeyword:
                    return "false";

                case SyntaxKind.ByTheWayKeyword:
                    return "btw";

                default:
                    throw Guard.EnumOutOfRangeException<SyntaxKind>("value", (int)this._kind);
            }
        }

        public override string ToString()
        {
            StringBuilder builder = new StringBuilder();

            if (this._stringValue != null)
            {
                builder.Append(this.Kind).Append(": ");
            }

            builder.Append(this.ToSource());
            builder.Append(" (").Append(this.Span).Append(")");

            return builder.ToString();
        }

        #region IEquatable<LexerToken> Members

        public bool Equals(SyntaxToken other)
        {
            return
               this._kind == other._kind &&
               this._span == other._span &&
               // a implies b => !a or b
               ((this._kind != SyntaxKind.NumericKind) || this._numericValue == other._numericValue) &&
               ((this._stringValue == null) || StringComparer.OrdinalIgnoreCase.Equals(this._stringValue, other._stringValue));
        }

        public bool EqualsIgnoreSpan(SyntaxToken other)
        {
            return
               this._kind == other._kind &&
               // a implies b => !a or b
               ((this._kind != SyntaxKind.NumericKind) || this._numericValue == other._numericValue) &&
               ((this._stringValue == null) || StringComparer.OrdinalIgnoreCase.Equals(this._stringValue, other._stringValue));
        }

        #endregion IEquatable<LexerToken> Members

        public static bool operator ==(SyntaxToken left, SyntaxToken right)
        {
            return left.Equals(right);
        }

        public static bool operator !=(SyntaxToken left, SyntaxToken right)
        {
            return !left.Equals(right);
        }
    }
}
