﻿using Dyn.Ast;
using Dyn.Lexer;
using System;
using System.Collections.Generic;

namespace Dyn.Parser
{
    public sealed partial class DynParser
    {
        public DynAssertionExpressionNode ParseAssertionExpression()
        {
            return ParseAssertionExpressionPratt(0);
        }

        private DynAssertionExpressionNode ParseAssertionExpressionPratt(int rbp)
        {
            DynAssertionExpressionNode result = AssertionNud(Current), cache = null;

            AssertionDesugarToken();

            while (result != null && rbp < AssertionLbp(Current))
            {
                result = AssertionLed(cache = result, Current);
                AssertionDesugarToken();
            }

            return result ?? cache;
        }

        private void AssertionDesugarToken()
        {
            if (Current.Kind == SyntaxKind.IdentifierKind)
            {
                if (StringComparer.Ordinal.Equals("and", Current.String))
                { _current = new SyntaxToken(SyntaxKind.AndOperator, Current.Span); }
                else if (StringComparer.Ordinal.Equals("or", Current.String))
                { _current = new SyntaxToken(SyntaxKind.OrOperator, Current.Span); }
                else if (StringComparer.Ordinal.Equals("implies", Current.String))
                { _current = new SyntaxToken(SyntaxKind.ImpliesOperator, Current.Span); }
                else if (StringComparer.Ordinal.Equals("mod", Current.String))
                { _current = new SyntaxToken(SyntaxKind.ModulosOperator, Current.Span); }
            }
        }

        private DynAssertionExpressionNode AssertionNud(SyntaxToken token)
        {
            var kind = token.Kind;

            Span first = token.Span;

            switch (kind)
            {
                case SyntaxKind.AtSymbol:
                    {
                        Accept(SyntaxKind.AtSymbol);

                        string name;

                        if ((name = ParseIdentifier()) == null)
                        {
                            OnError("Parsing instance variable");
                            return null;
                        }

                        return new DynAssertionInstanceVariableNode(
                            UnionUpToCurrent(first),
                            AstExtensions.Assertions.Self,
                            name);
                    }

                case SyntaxKind.ApostropheSymbol:
                    {
                        Accept(SyntaxKind.ApostropheSymbol);

                        var name = ParseIdentifier();

                        if (name == null)
                        {
                            return null;
                        }

                        return new DynAssertionVariableNode(UnionUpToCurrent(first),
                            DynAssertionExpressionNode.NodeKind.AssertionVariable,
                            name);
                    }

                case SyntaxKind.IdentifierKind:
                    {
                        var name = ParseIdentifier();

                        if (name == null)
                        {
                            return null;
                        }

                        if (Accept(SyntaxKind.ParenthesisLeftSymbol))
                        {
                            var arguments = new List<DynAssertionExpressionNode>();

                            if (!ParseSequenceWithDelimiter(SyntaxKind.CommaSymbol, arguments, ParseAssertionExpression) ||
                                !Expect(SyntaxKind.ParenthesisRightSymbol))
                            {
                                OnError("Parsing built in method call");
                                return null;
                            }

                            return new DynAssertionPredicateNode(UnionUpToCurrent(first), name, arguments);
                        }

                        return new DynAssertionVariableNode(UnionUpToCurrent(first),
                            DynAssertionExpressionNode.NodeKind.ProgramVariable,
                            name);
                    }
                case SyntaxKind.StringKind:
                    {
                        string value = null;

                        if (Current.Kind != SyntaxKind.StringKind ||
                            (value = Current.String) == null ||
                            value.Length < 2 ||
                            value[0] != '"' ||
                            value[value.Length - 1] != '"')
                        {
                            OnError("Invalid string");
                            return null;
                        }

                        Accept(SyntaxKind.StringKind);

                        return new DynAssertionStringNode(first, value.Substring(1, value.Length - 2));
                    }
                case SyntaxKind.NumericKind:
                    {
                        int value = Current.Number;

                        Accept(SyntaxKind.NumericKind);

                        return new DynAssertionNumericNode(first, value);
                    }
                case SyntaxKind.ParenthesisLeftSymbol:
                    {
                        Accept(SyntaxKind.ParenthesisLeftSymbol);

                        var result = ParseAssertionExpressionPratt(0);

                        if (result == null || !Expect(SyntaxKind.ParenthesisRightSymbol))
                        {
                            OnError(UnionUpToCurrent(first), "Parsing parenthesis");
                            return null;
                        }

                        return result;
                    }
                case SyntaxKind.ExclamationMarkSymbol:
                    {
                        Accept(SyntaxKind.ExclamationMarkSymbol);

                        var result = ParseAssertionExpressionPratt(
                            OperatorTranslation.AssertionPrecedence(
                                DynAssertionExpressionNode.NodeKind.Negate));

                        if (result == null)
                        {
                            OnError(UnionUpToCurrent(first), "Parsing unary operation");
                            return null;
                        }

                        return new DynAssertionUnaryNode(
                            UnionUpToCurrent(first),
                            result);
                    }
                case SyntaxKind.SelfKeyword:
                    Accept(SyntaxKind.SelfKeyword);
                    return new DynAssertionVariableNode(first, DynAssertionExpressionNode.NodeKind.ProgramVariable, "self");

                case SyntaxKind.NullKeyword:
                    Accept(SyntaxKind.NullKeyword);
                    return new DynAssertionVariableNode(first, DynAssertionExpressionNode.NodeKind.ProgramVariable, "null");

                case SyntaxKind.TrueKeyword:
                    Accept(SyntaxKind.TrueKeyword);
                    return new DynAssertionVariableNode(first, DynAssertionExpressionNode.NodeKind.AssertionVariable, "true");

                case SyntaxKind.FalseKeyword:
                    Accept(SyntaxKind.FalseKeyword);
                    return new DynAssertionVariableNode(first, DynAssertionExpressionNode.NodeKind.AssertionVariable, "false");

                case SyntaxKind.IfKeyword:
                    {
                        Accept(SyntaxKind.IfKeyword);

                        DynAssertionExpressionNode condition, trueBranch, falseBranch;

                        if ((condition = ParseAssertionExpressionPratt(0)) == null ||
                            !Expect(SyntaxKind.ThenKeyword) ||
                            (trueBranch = ParseAssertionExpressionPratt(0)) == null ||
                            !Expect(SyntaxKind.ElseKeyword) ||
                            (falseBranch = ParseAssertionExpressionPratt(0)) == null ||
                            !Expect(SyntaxKind.EndKeyword))
                        {
                            OnError(UnionUpToCurrent(first), "Parsing assertion if then else");
                            return null;
                        }

                        return new DynAssertionIfThenElseNode(UnionUpToCurrent(first), condition, trueBranch, falseBranch);
                    }
                case SyntaxKind.ForAllKeyword:
                    return ParseQuantifier(first, DynAssertionExpressionNode.NodeKind.Universal);

                case SyntaxKind.ExistsKeyword:
                    return ParseQuantifier(first, DynAssertionExpressionNode.NodeKind.Existential);

                default:
                    Guard.EnsureEnumDefined<SyntaxKind>("current.Kind", (int)kind);
                    return null;
            }
        }

        private DynAssertionExpressionNode ParseQuantifier(Span first, DynAssertionExpressionNode.NodeKind kind)
        {
            Accept(Current.Kind);

            string name, type;

            Span second = Current.Span;

            DynAssertionExpressionNode expression;

            if (!Expect(SyntaxKind.ApostropheSymbol) ||
                (name = ParseIdentifier()) == null ||
                (!Accept(SyntaxKind.InKeyword) && !Expect(SyntaxKind.ColonSymbol)) ||
                (type = ParseType()) == null ||
                (!Accept(SyntaxKind.ColonSymbol) && !Expect(SyntaxKind.DotSymbol)) ||
                (expression = ParseAssertionExpressionPratt(0)) == null)
            {
                OnError(UnionUpToCurrent(first), "Parsing quantifier");
                return null;
            }

            return new DynAssertionQuantifierNode(
                UnionUpToCurrent(first),
                kind,
                name,
                type,
                expression);
        }

        private string ParseType()
        {
            return ParseIdentifier();
        }

        private int AssertionLbp(SyntaxToken current)
        {
            return OperatorTranslation.AssertionPrecedence(current.Kind);
        }

        private DynAssertionExpressionNode AssertionLed(DynAssertionExpressionNode left, SyntaxToken token)
        {
            switch (token.Kind)
            {
                // short circuit
                case SyntaxKind.CurlyBracketRightSymbol:
                case SyntaxKind.ParenthesisRightSymbol:
                case SyntaxKind.SquareBracketRightSymbol:
                case SyntaxKind.SemicolonSymbol:
                    return null;

                case SyntaxKind.IdentifierKind:
                    return null;
                // boolean
                case SyntaxKind.AndOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.And, true);

                case SyntaxKind.OrOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.Or, true);

                case SyntaxKind.ImpliesOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.Implies, true);

                case SyntaxKind.AdditionOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.Addition, true);

                case SyntaxKind.SubtractionOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.Subtraction, true);

                case SyntaxKind.MultiplicationOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.Multiplication, true);

                case SyntaxKind.DivisionOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.Division, true);

                case SyntaxKind.ModulosOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.Modulus, true);

                case SyntaxKind.EqualityOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.Equality, false);

                case SyntaxKind.GreaterThanOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.GreaterThan, true);

                case SyntaxKind.GreaterThanOrEqualOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual, true);

                case SyntaxKind.LessThanOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.LessThan, true);

                case SyntaxKind.LessThanOrEqualOperator:
                    return ParseAssertionBinaryExpression(left, DynAssertionExpressionNode.NodeKind.LessThanOrEqual, true);

                case SyntaxKind.DotSymbol:
                    {
                        Accept(SyntaxKind.DotSymbol);

                        string name;

                        if (!Expect(SyntaxKind.AtSymbol) ||
                            (name = ParseIdentifier()) == null)
                        {
                            OnError("Parsing instance variable");
                            return null;
                        }

                        return new DynAssertionInstanceVariableNode(UnionUpToCurrent(left.Location), left, name);
                    }

                case SyntaxKind.InKeyword:
                    return ParseTypeCheck(left);

                default:
                    Guard.EnsureEnumDefined<SyntaxKind>("token.Kind", (int)token.Kind);
                    return null;
            }
        }

        private DynAssertionExpressionNode ParseAssertionBinaryExpression(
            DynAssertionExpressionNode left,
            DynAssertionExpressionNode.NodeKind nodeKind,
            bool isLeftAssoc)
        {
            var first = Current.Span;

            Accept(Current.Kind);

            var right = ParseAssertionExpressionPratt(
                OperatorTranslation.AssertionPrecedence(nodeKind) -
                (isLeftAssoc ? 0 : 1));

            if (right == null)
            {
                return null;
            }

            return new DynAssertionBinaryNode(UnionUpToCurrent(left.Location), nodeKind, left, right);
        }

        private DynAssertionTypeCheckNode ParseTypeCheck(DynAssertionExpressionNode left)
        {
            List<string> types = new List<string>();

            if (!Accept(SyntaxKind.InKeyword) ||
                !Expect(SyntaxKind.CurlyBracketLeftSymbol) ||
                !ParseSequenceWithDelimiter(SyntaxKind.CommaSymbol, types, ParseIdentifier) ||
                !Expect(SyntaxKind.CurlyBracketRightSymbol))
            {
                return null;
            }

            return new DynAssertionTypeCheckNode(UnionUpToCurrent(left.Location), left, types);
        }
    }
}
