﻿using Dyn.Diagnostics;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace Dyn.Parser
{
    //TODO: Handle invalid stuff, like reading Class as Expression etc by using a bitfield with expected operations
    public sealed partial class DynParser
    {
        private readonly IEnumerator<SyntaxToken> _source;

        public DynParser(IEnumerator<SyntaxToken> source)
        {
            Guard.EnsureNotNull("source", source);
            this._source = source;

            MoveNext();
        }

        #region Diagnostic

        public event EventHandler<DiagnosticEventArgs> DiagnosticInfo;

        public void OnDiagnosticInfo(DiagnosticSeverity level, Span span, string message)
        {
            var handler = DiagnosticInfo;

            if (handler != null)
            {
                handler(this, new DiagnosticEventArgs(new DiagnosticInfo(level, span, message)));
            }
        }

        #endregion Diagnostic

        #region Primitives

        private SyntaxToken _current;

        public bool IsEndOfStream
        {
            get
            {
                return this._current.Kind == SyntaxKind.EndOfStreamKind;
            }
        }

        private SyntaxToken Current
        {
            get
            {
                return this._current;
            }
        }

        [DebuggerStepThrough]
        private bool MoveNext()
        {
            while (this._source.MoveNext())
            {
                this._current = this._source.Current;

                if (!SkipToken(this._current))
                {
                    return true;
                }
            }

            this._current = default(SyntaxToken);

            return false;
        }

        private static bool SkipToken(SyntaxToken token)
        {
            switch (token.Kind)
            {
                case SyntaxKind.EndOfStreamKind:
                    return true;

                case SyntaxKind.WhitespaceKind:
                case SyntaxKind.CommentKind:
                    return true;

                case SyntaxKind.InvalidKind:
                    // return false to handle invalid token later down the line
                    return false;

                case SyntaxKind.IdentifierKind:
                case SyntaxKind.NumericKind:
                case SyntaxKind.StringKind:
                    return false;

                default:
                    return false;
            }
        }

        [DebuggerStepThrough]
        public bool Accept(SyntaxKind value)
        {
            if (this.Current.Kind == value)
            {
                MoveNext();
                return true;
            }

            return false;
        }

        [DebuggerStepThrough]
        private bool Expect(SyntaxKind value)
        {
            if (this.Current.Kind == value)
            {
                MoveNext();
                return true;
            }

            OnDiagnosticInfo(DiagnosticSeverity.Error, this.Current.Span, "Expected " + value + ", but found " + this._current.ToSource());
            return false;
        }

        #endregion Primitives

        #region Parse Primitives

        [DebuggerStepThrough]
        private string ParseIdentifier()
        {
            string result = default(string);
            if (this._current.Kind == SyntaxKind.IdentifierKind)
            {
                result = this._current.String;
                MoveNext();
            }

            return result;
        }

        //[DebuggerHidden]
        private bool ParseSequenceWithDelimiter<T>(SyntaxKind delimeter, ICollection<T> result, Func<T> parser) where T : class
        {
            Guard.EnsureNotNull("result", result);
            Guard.EnsureNotNull("parser", parser);

            bool head = true;

            T element;

            while ((head || Accept(delimeter)) && ((element = parser()) != null))
            {
                result.Add(element);
                head = delimeter == SyntaxKind.EndOfStreamKind;
            }

            return true;
        }

        #endregion Parse Primitives

        private void OnError(string message)
        {
            OnDiagnosticInfo(DiagnosticSeverity.Error, Current.Span, message);
        }

        private void OnError(Span span, string message)
        {
            OnDiagnosticInfo(DiagnosticSeverity.Error, span, message);
        }

        private Span UnionUpToCurrent(Span first)
        {
            var other = Current.Span;

            if (other.Position == 0 && other.Length == 0)
            {
                return first;
            }

            if (first.Position == 0 && first.Length == 0)
            {
                return other;
            }

            if (first.FileName == other.FileName) {
              return new Span(first.FileName, first.Position, first.Line, first.Column, other.Position - first.Position);
            } else {
              return new Span(first.FileName, first.Position, first.Line, first.Column, 0);
            }
        }
    }
}
