﻿using Dyn.Ast;
using Dyn.Lexer;
using System;
using System.Collections.Generic;

namespace Dyn.Parser
{
    //TODO: Handle invalid stuff, like reading Class as Expression etc by using a bitfield with expected operations
    public sealed partial class DynParser
    {
        private bool ParseConstractDeclarations(ICollection<DynProgramContractDeclarationNode> result, bool ensures, bool requires, bool invariant, bool assert, bool bytheway)
        {
            Guard.EnsureNotNull("result", result);

            DynProgramContractDeclarationNode element;

            while (((element = ParseContractDeclaration(ensures, requires, invariant, assert, bytheway)) != null))
            {
                result.Add(element);
            }

            return true;
        }

        public DynProgramClassNode ParseClass()
        {
            if (!Accept(SyntaxKind.ClassKeyword))
            {
                return null;
            }

            Span first = Current.Span;

            string name;
            string parent = AstExtensions.RootClass;

            if ((name = ParseIdentifier()) == null)
            {
                OnError("Parsing class name");
                return null;
            }

            if (Accept(SyntaxKind.LessThanOperator) && (parent = ParseIdentifier()) == null)
            {
                OnError("parsing class parent");
                return null;
            }

            List<DynProgramContractDeclarationNode> assertions = new List<DynProgramContractDeclarationNode>();
            List<DynProgramRenameDeclarationNode> renames = new List<DynProgramRenameDeclarationNode>();
            List<DynProgramMethodDeclarationNode> methods = new List<DynProgramMethodDeclarationNode>();

            ParseConstractDeclarations(assertions, true, true, true, false, false);

            if (!Expect(SyntaxKind.CurlyBracketLeftSymbol))
            {
                return null;
            }

            {
                DynProgramContractDeclarationNode sa;

                DynProgramMethodDeclarationNode md;
                DynProgramRenameDeclarationNode rd;

                while (!Accept(SyntaxKind.CurlyBracketRightSymbol))
                {
                    if ((sa = ParseContractDeclaration(true, true, true, false, false)) != null)
                    {
                        assertions.Add(sa);
                    }
                    else if ((rd = ParseRenameDeclaration()) != null)
                    {
                        renames.Add(rd);
                    }
                    else if ((md = ParseMethodDeclaration()) != null)
                    {
                        methods.Add(md);
                    }
                    else
                    {
                        OnError("Parsing class member");
                        return null;
                    }
                }
            }

            return new DynProgramClassNode(UnionUpToCurrent(first), name, parent, renames, methods, assertions);
        }

        public DynProgramContractDeclarationNode ParseContractDeclaration(bool ensures, bool requires, bool invariant, bool assert, bool bytheway)
        {
            DynProgramStatementNode.NodeKind quantifier;

            Span first = Current.Span;

            bool valid = false;

            if (Accept(SyntaxKind.AssertKeyword))
            {
                quantifier = DynProgramStatementNode.NodeKind.Assert; valid = assert;
            }
            else if (Accept(SyntaxKind.EnsuresKeyword))
            {
                quantifier = DynProgramStatementNode.NodeKind.Ensures; valid = ensures;
            }
            else if (Accept(SyntaxKind.InvariantKeyword))
            {
                quantifier = DynProgramStatementNode.NodeKind.Invariant; valid = invariant;
            }
            else if (Accept(SyntaxKind.RequiresKeyword))
            {
                quantifier = DynProgramStatementNode.NodeKind.Requires; valid = requires;
            }
            else if (Accept(SyntaxKind.ByTheWayKeyword))
            {
                quantifier = DynProgramStatementNode.NodeKind.ByTheWay; valid = bytheway;
            }
            else
            {
                return null;
            }

            if (!valid)
            {
                OnError("Invalid contract modifier");
                return null;
            }

            DynAssertionExpressionNode expression = ParseAssertionExpression();

            if (expression == null)
            {
                OnError("Parsing assertion expression");
                return null;
            }

            return new DynProgramContractDeclarationNode(UnionUpToCurrent(first), quantifier, expression);
        }

        public DynProgramRenameDeclarationNode ParseRenameDeclaration()
        {
            string from, to;

            Span first = Current.Span;

            if (!Accept(SyntaxKind.RenameKeyword))
            {
                return null;
            }

            if (((from = ParseIdentifier()) == null) ||
                ((to = ParseIdentifier()) == null))
            {
                OnError("Parsing rename");
                return null;
            }

            return new DynProgramRenameDeclarationNode(UnionUpToCurrent(first), from, to);
        }

        public DynProgramMethodDeclarationNode ParseMethodDeclaration()
        {
            Span first = Current.Span;

            if (!Accept(SyntaxKind.MethodKeyword))
            {
                return null;
            }

            string name = ParseIdentifier();

            if (name == null)
            {
                OnError("Parsing method name");
                return null;
            }

            List<string> parameters = new List<string>();
            List<DynProgramContractDeclarationNode> assertions = new List<DynProgramContractDeclarationNode>();
            List<DynProgramStatementNode> statements = new List<DynProgramStatementNode>();

            if (!Expect(SyntaxKind.ParenthesisLeftSymbol) ||
            !ParseSequenceWithDelimiter(SyntaxKind.CommaSymbol, parameters, ParseIdentifier) ||
            !Expect(SyntaxKind.ParenthesisRightSymbol) ||
            !ParseConstractDeclarations(assertions, true, true, true, false, false) ||
            !Expect(SyntaxKind.CurlyBracketLeftSymbol) ||
            !ParseProgramStatements(statements) ||
            !Expect(SyntaxKind.CurlyBracketRightSymbol))
            {
                OnError("Parsing method");
                return null;
            }

            return new DynProgramMethodDeclarationNode(UnionUpToCurrent(first), name, parameters, statements, assertions);
        }

        public bool ParseProgramStatements(IList<DynProgramStatementNode> dest)
        {
            return ParseSequenceWithDelimiter(SyntaxKind.SemicolonSymbol, dest, ParseProgramStatement);
        }

        public DynProgramStatementNode ParseProgramStatement()
        {
            return ParseContractDeclaration(false, false, false, true, true) ??
                (DynProgramStatementNode)ParseProgramExpressionPratt(0);
        }

        public DynProgramExpressionNode ParseProgramExpression()
        {
            return ParseProgramExpressionPratt(0);
        }

        private DynProgramExpressionNode ParseProgramExpressionPratt(int rbp)
        {
            DynProgramExpressionNode result = ProgramNud(Current), cache = null;
            ProgramDesugarToken();
            while (result != null && rbp < ProgramLbp(Current))
            {
                result = ProgramLed(cache = result, Current);
                ProgramDesugarToken();
            }

            return result ?? cache;
        }

        private void ProgramDesugarToken()
        {
            if (Current.Kind == SyntaxKind.IdentifierKind)
            {
                if (StringComparer.Ordinal.Equals("and", Current.String))
                { _current = new SyntaxToken(SyntaxKind.AndOperator, Current.Span); }
                else if (StringComparer.Ordinal.Equals("or", Current.String))
                { _current = new SyntaxToken(SyntaxKind.OrOperator, Current.Span); }
                else if (StringComparer.Ordinal.Equals("implies", Current.String))
                { _current = new SyntaxToken(SyntaxKind.ImpliesOperator, Current.Span); }
            }
        }

        private DynProgramExpressionNode ProgramNud(SyntaxToken current)
        {
            var kind = current.Kind;

            Span first = current.Span;

            switch (kind)
            {
                case SyntaxKind.StringKind:
                    return ParseProgramStringNode();

                case SyntaxKind.NumericKind:
                    return ParseProgramNumericNode();

                case SyntaxKind.IdentifierKind:
                    {
                        var name = ParseIdentifier();

                        // handle method
                        if (Accept(SyntaxKind.ParenthesisLeftSymbol))
                        {
                            List<DynProgramExpressionNode> arguments = new List<DynProgramExpressionNode>();

                            if (!ParseSequenceWithDelimiter(SyntaxKind.CommaSymbol, arguments, ParseProgramExpression) ||
                                !Expect(SyntaxKind.ParenthesisRightSymbol))
                            {
                                OnError("Parsing built in method call");
                                return null;
                            }

                            return new DynProgramMethodCallNode(UnionUpToCurrent(first), AstExtensions.Programs.Global, name, arguments);
                        }

                        return new DynProgramVariableNode(UnionUpToCurrent(first), DynProgramStatementNode.NodeKind.Variable, name);
                    }
                case SyntaxKind.AtSymbol:
                    {
                        Accept(SyntaxKind.AtSymbol);
                        string name;

                        if ((name = ParseIdentifier()) == null)
                        {
                            OnError("Parsing variable");
                            return null;
                        }

                        return new DynProgramVariableNode(UnionUpToCurrent(first), DynProgramStatementNode.NodeKind.InstanceVariable, name);
                    }
                case SyntaxKind.SelfKeyword:
                    Accept(SyntaxKind.SelfKeyword);
                    return new DynProgramVariableNode(first, DynProgramStatementNode.NodeKind.Variable, "self");

                case SyntaxKind.NullKeyword:
                    Accept(SyntaxKind.NullKeyword);
                    return new DynProgramVariableNode(first, DynProgramStatementNode.NodeKind.Variable, "null");

                case SyntaxKind.ParenthesisLeftSymbol:
                    {
                        Accept(SyntaxKind.ParenthesisLeftSymbol);

                        DynProgramExpressionNode result = ParseProgramExpressionPratt(0);

                        if (result == null || !Expect(SyntaxKind.ParenthesisRightSymbol))
                        {
                            OnError("Parsing parenthesis");
                            return null;
                        }

                        return result;
                    }
                case SyntaxKind.TrueKeyword:
                    Accept(SyntaxKind.TrueKeyword);
                    return AstExtensions.TransformBoolean(first, true);

                case SyntaxKind.FalseKeyword:
                    Accept(SyntaxKind.FalseKeyword);
                    return AstExtensions.TransformBoolean(first, false);

                case SyntaxKind.IfKeyword:
                    return ParseIfStatement();

                case SyntaxKind.WhileKeyword:
                    return ParseWhile();

                case SyntaxKind.NewKeyword:
                    return ParseConstructorCall();

                default:
                    Guard.EnsureEnumDefined<SyntaxKind>("current.Kind", (int)kind);
                    return null;
            }
        }

        private DynProgramConstructorCallNode ParseConstructorCall()
        {
            string name;
            List<DynProgramExpressionNode> arguments = new List<DynProgramExpressionNode>();

            Span first = Current.Span;

            if (!Accept(SyntaxKind.NewKeyword) ||
                (name = ParseIdentifier()) == null ||
                !Expect(SyntaxKind.ParenthesisLeftSymbol) ||
                !ParseSequenceWithDelimiter(SyntaxKind.CommaSymbol, arguments, ParseProgramExpression) ||
                !Expect(SyntaxKind.ParenthesisRightSymbol))
            {
                OnError("Parsing constructor call");
                return null;
            }

            return new DynProgramConstructorCallNode(UnionUpToCurrent(first), name, arguments);
        }

        private DynProgramWhileNode ParseWhile()
        {
            DynProgramExpressionNode condition;
            List<DynProgramStatementNode> statements = new List<DynProgramStatementNode>();
            DynAssertionExpressionNode invariant = null;

            Span first = Current.Span;

            if (!Accept(SyntaxKind.WhileKeyword) ||
                (condition = ParseProgramExpressionPratt(0)) == null)
            {
                OnError("Parsing while loop");
                return null;
            }

            if (!Accept(SyntaxKind.InvariantKeyword) ||
                (invariant = ParseAssertionExpression()) == null)
            {
                invariant = Ast.AstExtensions.Assertions.True;
            }

            if (!Accept(SyntaxKind.DoKeyword) ||
                !ParseProgramStatements(statements) ||
                !Expect(SyntaxKind.DoneKeyword))
            {
                OnError("Parsing while loop");
                return null;
            }

            return new DynProgramWhileNode(UnionUpToCurrent(first), condition, invariant, statements);
        }

        private DynProgramIfNode ParseIfStatement()
        {
            DynProgramExpressionNode condition;
            List<DynProgramStatementNode> trueBranch = new List<DynProgramStatementNode>();
            List<DynProgramStatementNode> falseBranch = new List<DynProgramStatementNode>();

            Span first = Current.Span;

            if (!Accept(SyntaxKind.IfKeyword) ||
                (condition = ParseProgramExpressionPratt(0)) == null ||
                !Expect(SyntaxKind.ThenKeyword) ||
                !ParseProgramStatements(trueBranch))
            {
                OnError("Paring if statement");
                return null;
            }

            if (Accept(SyntaxKind.ElseKeyword) && !ParseProgramStatements(falseBranch))
            {
                OnError("Parsing else branch");
                return null;
            }

            if (!Expect(SyntaxKind.EndKeyword))
            {
                return null;
            }

            return new DynProgramIfNode(UnionUpToCurrent(first), condition, trueBranch, falseBranch);
        }

        private DynProgramExpressionNode ParseProgramNumericNode()
        {
            Span first = Current.Span;

            if (Current.Kind != SyntaxKind.NumericKind)
            {
                return null;
            }

            int value = Current.Number;

            Accept(SyntaxKind.NumericKind);

            return AstExtensions.TransformNumber(first, value);
        }

        private DynProgramExpressionNode ParseProgramStringNode()
        {
            string value;

            if (Current.Kind != SyntaxKind.StringKind ||
                (value = Current.String) == null ||
                value.Length < 2 ||
                value[0] != '"' ||
                value[value.Length - 1] != '"')
            {
                OnError("Invalid string");
                return null;
            }

            Span first = Current.Span;

            Accept(SyntaxKind.StringKind);

            return AstExtensions.TransformString(first, value.Substring(1, value.Length - 2));
        }

        private int ProgramLbp(SyntaxToken current)
        {
            return OperatorTranslation.ProgramPrecedence(current.Kind);
        }

        private DynProgramExpressionNode ProgramLed(DynProgramExpressionNode left, SyntaxToken current)
        {
            var kind = current.Kind;

            switch (kind)
            {
                case SyntaxKind.AssignmentOperator:
                    {
                        if (left.Kind != DynProgramStatementNode.NodeKind.Variable &&
                                left.Kind != DynProgramStatementNode.NodeKind.InstanceVariable)
                        {
                            OnError("Assigning to not a variable");
                            return null;
                        }

                        var variable = (DynProgramVariableNode)left;

                        DynProgramExpressionNode expression;
                        if (!Accept(SyntaxKind.AssignmentOperator) ||
                            (expression = ParseProgramExpressionPratt(
                                OperatorTranslation.ProgramPrecedence(SyntaxKind.AssignmentOperator) - 1)) == null)
                        {
                            return null;
                        }

                        return new DynProgramAssignmentNode(UnionUpToCurrent(left.Location), variable, expression);
                    }

                case SyntaxKind.SquareBracketLeftSymbol:
                    throw new NotImplementedException("List not implemented");

                // Operators
                case SyntaxKind.EqualityOperator:
                case SyntaxKind.IdentityOperator:
                    return ParseProgramBinary(left, OperatorTranslation.KindToMethod(current.Kind), true);

                case SyntaxKind.SubtractionOperator:
                case SyntaxKind.DivisionOperator:
                case SyntaxKind.ModulosOperator:
                case SyntaxKind.LessThanOperator:
                case SyntaxKind.GreaterThanOperator:
                case SyntaxKind.LessThanOrEqualOperator:
                case SyntaxKind.GreaterThanOrEqualOperator:
                case SyntaxKind.AndOperator:
                case SyntaxKind.OrOperator:
                case SyntaxKind.AdditionOperator:
                case SyntaxKind.MultiplicationOperator:
                    return ParseProgramBinary(left, OperatorTranslation.KindToMethod(current.Kind), false);
                // Symbols
                case SyntaxKind.DotSymbol:
                    {
                        Accept(SyntaxKind.DotSymbol);

                        string name = ParseIdentifier();

                        List<DynProgramExpressionNode> arguments = new List<DynProgramExpressionNode>();

                        if (!Expect(SyntaxKind.ParenthesisLeftSymbol) ||
                            !ParseSequenceWithDelimiter(SyntaxKind.CommaSymbol, arguments, ParseProgramExpression) ||
                            !Expect(SyntaxKind.ParenthesisRightSymbol))
                        {
                            OnError("Parsing method call");
                            return null;
                        }

                        return new DynProgramMethodCallNode(UnionUpToCurrent(left.Location), left, name, arguments);
                    }
                case SyntaxKind.IsAKeyword:
                    return ParseProgramBinary(left, OperatorTranslation.KindToMethod(current.Kind), false);

                default:
                    Guard.EnsureEnumDefined<SyntaxKind>("current.Kind", (int)kind);
                    OnError("Parsing expression: invalid token " + kind);
                    return null;
            }
        }

        private DynProgramExpressionNode ParseProgramBinary(DynProgramExpressionNode left, string opName, bool rightAssoc)
        {
            var current = Current;

            MoveNext();

            if (current.Kind == SyntaxKind.IsAKeyword && !Accept(SyntaxKind.QuestionMarkSymbol))
            {
                OnError("Expected question mark after isa keyword");
                return null;
            }

            DynProgramExpressionNode right = ParseProgramExpressionPratt(
                OperatorTranslation.ProgramPrecedence(current.Kind) - (rightAssoc ? 1 : 0));

            if (right == null)
            {
                return null;
            }

            return new DynProgramMethodCallNode(UnionUpToCurrent(current.Span), left, opName, Utilities.AsArray(right));
        }

        public DynHoareNode ParseHoare()
        {
            if (!Accept(SyntaxKind.CurlyBracketLeftSymbol))
            {
                return null;
            }

            Span first = Current.Span;
            DynAssertionExpressionNode pre = ParseAssertionExpression();

            if (!Expect(SyntaxKind.CurlyBracketRightSymbol))
            {
                OnError("Parsing pre-condition");
                return null;
            }

            List<DynProgramStatementNode> statements = new List<DynProgramStatementNode>();

            ParseProgramStatements(statements);

            if (!Expect(SyntaxKind.CurlyBracketLeftSymbol))
            {
                OnError("Parsing post-condition");
                return null;
            }

            DynAssertionExpressionNode post = ParseAssertionExpression();
            if (!Expect(SyntaxKind.CurlyBracketRightSymbol))
            {
                OnError("Parsing post-condition");
                return null;
            }

            return new DynHoareNode(UnionUpToCurrent(first), pre, statements, post);
        }
    }
}
