﻿using Dyn.Ast;
using Dyn.Diagnostics;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Linq;

namespace Dyn.Parser
{
    /// <summary>
    /// Capsules a transformation for resolving class inheritance.
    /// </summary>
    public static class InheritanceTransformer
    {
        private static readonly DynProgramClassNode EMPTY_ROOT = new DynProgramClassNode(
            Span.Empty,
            AstExtensions.RootClass,
            AstExtensions.RootClass,
            null,
            null,
            null);

        private static void OnRaiseDiagnostic(
                        EventHandler<DiagnosticEventArgs> diagnostic,
                        DiagnosticSeverity severity,
                        Span location,
                        string message)
        {
            if (diagnostic != null)
            {
                diagnostic(null, new DiagnosticEventArgs(
                    new DiagnosticInfo(
                        severity,
                        location,
                        message)));
            }
        }

        /// <summary>
        /// Transforms <paramref name="classes"/> into a sequence of
        /// <see cref="DynProgramClassNode"/> without inheritance and
        /// raising diagnostics on errors.
        /// </summary>
        /// <param name="diagnostic">An event handler callback in case
        /// of diagnostics, or <code>null</code>.</param>
        /// <param name="classes">A sequence of
        /// <see cref="DynProgramClassNode"/>.</param>
        /// <returns>A sequence of <see cref="DynProgramClassNode"/>
        /// similar to <paramref name="classes"/>, but all inheritance
        /// resolved.</returns>
        public static IList<DynProgramClassNode> Transform(
            EventHandler<DiagnosticEventArgs> diagnostic,
            IEnumerable<DynProgramClassNode> classes)
        {
            var names = new Dictionary<string, DynProgramClassNode>();
            var visited = new HashSet<string>();

            // repeated class names
            foreach (var element in classes)
            {
                if (names.ContainsKey(element.Name))
                {
                    OnRaiseDiagnostic(diagnostic, DiagnosticSeverity.Error, element.Location,
                        string.Format("Program already contains name {0}.", element.Name));
                }

                if (StringComparer.Ordinal.Equals(AstExtensions.RootClass, element.Name))
                {
                    names[element.Name] = HandleClass(diagnostic, element, EMPTY_ROOT);
                    visited.Add(element.Name);
                }
                else
                {
                    names[element.Name] = element;
                }
            }
            var cachedClasses = names.Values.ToList();

            // remove inheritance
            {
                bool loop;
                do
                {
                    loop = false;

                    foreach (var element in cachedClasses)
                    {
                        DynProgramClassNode parent;

                        if (visited.Contains(element.Name) || !visited.Contains(element.Parent))
                        {
                            continue;
                        }

                        if (!names.TryGetValue(element.Parent, out parent))
                        {
                            OnRaiseDiagnostic(diagnostic, DiagnosticSeverity.Error, element.Location,
                                string.Format("Parent {0} of class {1} not found.", element.Parent, element.Name));

                            parent = EMPTY_ROOT;
                        }

                        var modifiedClass = HandleClass(diagnostic, element, parent);
                        names[element.Name] = modifiedClass;
                        visited.Add(element.Name);
                        loop = true;
                    }
                } while (loop);

                // all classes left are in a circular dependency
                // it does not matter which, just all affected classes is ok
                foreach (var element in cachedClasses)
                {
                    if (!visited.Contains(element.Name))
                    {
                        OnRaiseDiagnostic(
                            diagnostic,
                            DiagnosticSeverity.Error,
                            element.Location,
                            string.Format(
                                "Class {0} < {1} part of circular inheritance.",
                                element.Name,
                                element.Parent));

                        names[element.Name] = HandleClass(diagnostic, element, EMPTY_ROOT);
                    }
                }
            }

            var result = new DynProgramClassNode[names.Count];
            names.Values.CopyTo(result, 0);
            return result;
        }

        /// <summary>
        /// Transforms a single class by combining <paramref name="currentClass"/>
        /// with <paramref name="parent"/> and handling <code>rename from to</code>.
        /// </summary>
        /// <param name="diagnostic">An event handler callback in case
        /// of diagnostics, or <code>null</code>.</param>
        /// <param name="currentClass">The class.</param>
        /// <param name="parent">The parent from <paramref name="currentClass"/></param>
        /// <returns></returns>
        private static DynProgramClassNode HandleClass(
            EventHandler<DiagnosticEventArgs> diagnostic,
            DynProgramClassNode currentClass,
            DynProgramClassNode parent)
        {
            // handle methods
            var methods = new Dictionary<string, DynProgramMethodDeclarationNode>();
            var visited = new HashSet<string>();

            {
                // build rename lookup
                var visitedRenames = new HashSet<string>();
                var renames = GenerateRenameMapping(diagnostic, currentClass);

                // parent methods are first
                foreach (var method in parent.Methods)
                {
                    string destName = method.Name;

                    var destMethod = method;

                    if (renames.TryGetValue(method.Name, out destName))
                    {
                        visitedRenames.Add(method.Name);

                        destMethod = new DynProgramMethodDeclarationNode(
                            method.Location,
                            destName,
                            method.Parameters,
                            method.Statements,
                            method.Contracts);
                    }

                    HandleMethod(diagnostic, currentClass, methods, visited, destMethod);
                }

                foreach (var rename in renames.Keys.Except(visitedRenames))
                {
                    OnRaiseDiagnostic(
                        diagnostic,
                        DiagnosticSeverity.Warning,
                        currentClass.Location,
                        string.Format(
                            "Rename directive for {0} in class {1} has not been used.",
                            rename, currentClass.Name));
                }
            }

            visited.Clear();

            foreach (var method in currentClass.Methods)
            {
                HandleMethod(diagnostic, currentClass, methods, visited, method);
            }

            var contracts = new List<DynProgramContractDeclarationNode>();

            contracts.AddRange(parent.Contracts);
            contracts.AddRange(currentClass.Contracts);

            return new DynProgramClassNode(
                    currentClass.Location,
                    currentClass.Name,
                    AstExtensions.RootClass,
                    null,
                    Ast.Utilities.AsImmutableList(methods.Values),
                    contracts);
        }

        /// <summary>
        /// Generates a mapping of renames from a class' list of renames.
        /// </summary>
        /// <param name="diagnostic">An event handler callback in case
        /// of diagnostics, or <code>null</code>.</param>
        /// <param name="currentClass">A <see cref="DynProgramClassNode"/></param>
        /// <returns>A dictionary of all rename statements in <paramref name="currentClass"/>.</returns>
        private static Dictionary<string, string> GenerateRenameMapping(
            EventHandler<DiagnosticEventArgs> diagnostic,
            DynProgramClassNode currentClass)
        {
            var renames = new Dictionary<string, string>();

            foreach (var rename in currentClass.Renames)
            {
                if (renames.ContainsKey(rename.From))
                {
                    OnRaiseDiagnostic(
                        diagnostic,
                        DiagnosticSeverity.Error,
                        currentClass.Location,
                        string.Format(
                            "Rename directive for {0} already found in class {1}.",
                            rename.From, currentClass.Name));
                }

                renames[rename.From] = rename.To;
            }

            return renames;
        }

        /// <summary>
        /// Modifies a method to replace inheritance.
        /// </summary>
        /// <param name="diagnostic"></param>
        /// <param name="element"></param>
        /// <param name="methods"></param>
        /// <param name="visited"></param>
        /// <param name="method"></param>
        private static void HandleMethod(
            EventHandler<DiagnosticEventArgs> diagnostic,
            DynProgramClassNode element,
            Dictionary<string, DynProgramMethodDeclarationNode> methods,
            HashSet<string> visited,
            DynProgramMethodDeclarationNode method)
        {
            if (methods.ContainsKey(method.Name))
            {
                if (visited.Contains(method.Name))
                {
                    OnRaiseDiagnostic(
                        diagnostic,
                        DiagnosticSeverity.Error,
                        element.Location,
                        string.Format("Overriding method {0} in class {1}.",
                        method.Name, element.Name));
                }

                int arity = methods[method.Name].Parameters.Count;

                if (arity != method.Parameters.Count)
                {
                    OnRaiseDiagnostic(
                        diagnostic,
                        DiagnosticSeverity.Error,
                        element.Location,
                        string.Format(
                            "Arity mismatch after overriding method {0} in class {1}.",
                            method.Name, element.Name));
                }

                visited.Add(method.Name);
            }

            methods[method.Name] = method;
        }
    }
}
