﻿using Dyn.Ast;
using Dyn.Lexer;
using System.Collections.Generic;

namespace Dyn.VerficationConditions
{
#if DEBUG

    [System.Diagnostics.DebuggerDisplay("{ToString()}")]
#endif
    public struct Contract
    {
        private readonly DynAssertionExpressionNode _pre, _post;

        public Contract(
            DynAssertionExpressionNode pre,
            DynAssertionExpressionNode post)
        {
            this._pre = pre;
            this._post = post;
        }

        public DynAssertionExpressionNode PreCondition
        {
            get
            {
                return this._pre ?? AstExtensions.Assertions.True;
            }
        }

        public DynAssertionExpressionNode PostCondition
        {
            get
            {
                return this._post ?? AstExtensions.Assertions.True;
            }
        }

        public static string CreateKey(string name, int arity)
        {
            Guard.EnsureNotNull("name", name);
            Guard.EnsureNonnegative("arity", arity);

            return name + "$" + arity;
        }

        public static DynAssertionVariableNode CreateArgumentVariable(int i)
        {
            return new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "_a" + i);
        }

        public static List<string> GetArgumentVariables(int num)
        {
            var result = new List<string>();
            for (int i = 0; i <= num; i++) {
              result.Add("'_a" + i);
            }
            return result;
        }

        public static Contract FromNodes(IEnumerable<DynProgramContractDeclarationNode> source)
        {
            DynAssertionExpressionNode pre = null, post = null;

            foreach (var element in source)
            {
                switch (element.Kind)
                {
                    case DynProgramStatementNode.NodeKind.Requires:
                        pre = NullSafeAnd(pre, element.Expression); break;
                    case DynProgramStatementNode.NodeKind.Ensures:
                        post = NullSafeAnd(post, element.Expression); break;
                    case DynProgramStatementNode.NodeKind.Invariant:
                        pre = NullSafeAnd(pre, element.Expression);
                        post = NullSafeAnd(post, element.Expression); break;
                    default:
                        throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("element.Kind", (int)element.Kind);
                }
            }

            return new Contract(pre, post);
        }

        public Contract Normalize(IList<string> parameters)
        {
            DynAssertionExpressionNode pre = this.PreCondition, post = this.PostCondition;

            int paramIndex = 1;
            foreach (var param in parameters)
            {
                var variable = new DynAssertionVariableNode(
                    Span.Empty,
                    DynAssertionExpressionNode.NodeKind.ProgramVariable,
                    param);
                post = post.Substitute(variable, CreateArgumentVariable(paramIndex));
                pre = pre.Substitute(variable, CreateArgumentVariable(paramIndex));
                ++paramIndex;
            }

            pre = pre.Substitute(AstExtensions.Assertions.Self, CreateArgumentVariable(0));
            pre = pre.Substitute(AstExtensions.Assertions.r, WeakestPreconditionCalculus._r);
            post = post.Substitute(AstExtensions.Assertions.Self, CreateArgumentVariable(0));
            post = post.Substitute(AstExtensions.Assertions.r, WeakestPreconditionCalculus._r);

            return new Contract(pre, post);
        }

        private static DynAssertionExpressionNode NullSafeAnd(DynAssertionExpressionNode first, DynAssertionExpressionNode second)
        {
            if (first == null)
            {
                return second;
            }
            if (second == null)
            {
                return first;
            }
            return first.And(Span.Empty, second);
        }

        public static Contract operator &(Contract first, Contract second)
        {
            return new Contract(
                NullSafeAnd(first._pre, second._pre),
                NullSafeAnd(first._post, second._post));
        }

#if DEBUG

        public override string ToString()
        {
            return string.Format("Pre: {0}, Post: {1}",
                this._pre == null ? "" : this._pre.ToSource(),
                this._post == null ? "" : this._post.ToSource());
        }

#endif
    }
}
