﻿using Dyn.Ast;
using Dyn.SmtLib2;
using System;
using System.IO;
using System.Linq;
using System.Text;

namespace Dyn.VerficationConditions
{
    public static class DynAssertionSmtLibFormatter
    {
        public static string ToSymbol(DynProgramClassNode node)
        {
            return "class_" + node.Name;
        }

        public static string ToSymbolInstanceVariableName(string instanceVariableName)
        {
            return "iv_" + instanceVariableName;
        }

        public static string ToSymbol(string name)
        {
	    if (name.StartsWith("'")) {
              return "lv_" + name.Substring(1);
            } else {
              return "pv_" + name;
            }
        }

        public static string ToSymbolType(string type)
        {
            //TODO: Handle generic types
            switch (type)
            {
                case "B":
                case "bool":
                case "boolean":
                  return "Bool";
                case "N":
                case "num":
                case "numeric":
                  return "Int";
                case "O":
                case "obj":
                case "object":
                  return "Obj";
                default:
                    return type;
            }
        }

        public static string ToSymbolType(DynAssertionTypeInference.InferredType type)
        {
            //TODO: Handle generic types
            switch (type)
            {
                case DynAssertionTypeInference.InferredType.Boolean: return "Bool";
                case DynAssertionTypeInference.InferredType.Numeric: return "Int";
                case DynAssertionTypeInference.InferredType.Object: return "Obj";
                default:
                    return "Obj";
            }
        }

        public static StringBuilder Visit(DynAssertionExpressionNode node)
        {
            Guard.EnsureNotNull("node", node);

            var builder = new StringBuilder();

            using (var writer = new StringWriter(builder))
            {
                var sexpsw = new SexprWriter(writer, false);

                Visit(sexpsw, node);
            }

            return builder;
        }

        private static void Visit(SexprWriter writer, DynAssertionExpressionNode node)
        {
            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Universal:
                case DynAssertionExpressionNode.NodeKind.Existential:
                    Visit(writer, (DynAssertionQuantifierNode)node); return;
                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    Visit(writer, (DynAssertionTypeCheckNode)node); return;
                case DynAssertionExpressionNode.NodeKind.Predicate:
                    Visit(writer, (DynAssertionPredicateNode)node); return;
                case DynAssertionExpressionNode.NodeKind.Negate:
                    Visit(writer, (DynAssertionUnaryNode)node); return;
                case DynAssertionExpressionNode.NodeKind.And:
                case DynAssertionExpressionNode.NodeKind.Or:
                case DynAssertionExpressionNode.NodeKind.Implies:
                case DynAssertionExpressionNode.NodeKind.Addition:
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                case DynAssertionExpressionNode.NodeKind.Division:
                case DynAssertionExpressionNode.NodeKind.Modulus:
                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.LessThan:
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.Equality:
                    Visit(writer, (DynAssertionBinaryNode)node); return;
                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    Visit(writer, (DynAssertionIfThenElseNode)node); return;
                case DynAssertionExpressionNode.NodeKind.Numeric:
                    Visit(writer, (DynAssertionNumericNode)node); return;
                case DynAssertionExpressionNode.NodeKind.String:
                    Visit(writer, (DynAssertionStringNode)node); return;
                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    Visit(writer, (DynAssertionInstanceVariableNode)node); return;
                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    Visit(writer, (DynAssertionVariableNode)node); return;
                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("node.Kind", (int)node.Kind);
            }
        }

        private static void Visit(SexprWriter writer, DynAssertionQuantifierNode node)
        {
            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Universal:
                    writer.Open("forall"); break;
                case DynAssertionExpressionNode.NodeKind.Existential:
                    writer.Open("exists"); break;
                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("node.Kind", (int)node.Kind);
            }
            writer.Open(string.Empty);
            writer.Open(string.Format("lv_{0}", node.Variable));
            writer.Symbol(ToSymbolType(node.Type));
            writer.Close();
            writer.Close();

            Visit(writer, node.Expression);
            writer.Close();
        }

        private static void VisitType(SexprWriter writer, DynAssertionExpressionNode node, string element)
        {
            writer.Open("=");
            {
                writer.Open("class");
                Visit(writer, node);
                writer.Close();
            }
            writer.Symbol("class_" + element);
            writer.Close();
        }

        private static void Visit(SexprWriter writer, DynAssertionTypeCheckNode node)
        {
            if (node.TypeSet.Count == 0)
            {
                writer.Symbol("false");
                return;
            }

            if (node.TypeSet.Count == 1)
            {
                VisitType(writer, node.Expression, node.TypeSet.Single());
                return;
            }

            writer.Open("or");
            foreach (var element in node.TypeSet)
            {
                VisitType(writer, node.Expression, element);
            }
            writer.Close();
        }

        private static void Visit(SexprWriter writer, DynAssertionPredicateNode node)
        {
            writer.Open(node.Name);

            foreach (var arg in node.Arguments)
            {
                Visit(writer, arg);
            }

            writer.Close();
        }

        private static void Visit(SexprWriter writer, DynAssertionUnaryNode node)
        {
            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Negate:
                    writer.Open("not"); break;
                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("node.Kind", (int)node.Kind);
            }

            Visit(writer, node.Expression);
            writer.Close();
        }

        private static void Visit(SexprWriter writer, DynAssertionBinaryNode node)
        {
            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.And:
                    writer.Open("and"); break;
                case DynAssertionExpressionNode.NodeKind.Or:
                    writer.Open("or"); break;
                case DynAssertionExpressionNode.NodeKind.Implies:
                    writer.Open("=>"); break;
                case DynAssertionExpressionNode.NodeKind.Addition:
                    writer.Open("+"); break;
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                    writer.Open("-"); break;
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                    writer.Open("*"); break;
                case DynAssertionExpressionNode.NodeKind.Division:
                    writer.Open("div"); break;
                case DynAssertionExpressionNode.NodeKind.Modulus:
                    writer.Open("mod"); break;
                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                    writer.Open(">"); break;
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                    writer.Open(">="); break;
                case DynAssertionExpressionNode.NodeKind.LessThan:
                    writer.Open("<"); break;
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                    writer.Open("<="); break;
                case DynAssertionExpressionNode.NodeKind.Equality:
                    writer.Open("="); break;
                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("node.Kind", (int)node.Kind);
            }

            Visit(writer, node.First);
            Visit(writer, node.Second);
            writer.Close();
        }

        private static void Visit(SexprWriter writer, DynAssertionIfThenElseNode node)
        {
            writer.Open("ite");
            Visit(writer, node.Condition);
            Visit(writer, node.TrueBranch);
            Visit(writer, node.FalseBranch);
            writer.Close();
        }

        private static void Visit(SexprWriter writer, DynAssertionNumericNode node)
        {
            writer.Symbol(node.Value.ToString());
        }

        private static void Visit(SexprWriter writer, DynAssertionStringNode node)
        {
            throw new NotImplementedException();
        }

        private static void Visit(SexprWriter writer, DynAssertionInstanceVariableNode node)
        {
            writer.Open("inst_ref");
            Visit(writer, node.Expression);
            writer.Symbol(ToSymbolInstanceVariableName(node.Name));
            writer.Close();
        }

        private static void Visit(SexprWriter writer, DynAssertionVariableNode node)
        {
            if(StringComparer.Ordinal.Equals(node.Name, "null"))
            {
                writer.Symbol("Null");
                return;
            }

            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                    writer.Symbol("lv_" + node.Name); break;
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    writer.Symbol("pv_" + node.Name); break;
                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("node.Kind", (int)node.Kind);
            }
        }
    }
}
