﻿using Dyn.Ast;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Collections;

namespace Dyn.VerficationConditions
{
    using Diagnostics;
    using unbound_element = KeyValuePair<string, DynAssertionTypeInference.InferredType>;
    using unbound_type = IEnumerable<KeyValuePair<string, DynAssertionTypeInference.InferredType>>;

    public sealed class DynAssertionTypeInference
    {
        public enum InferredType
        {
            Unknown,
            Boolean,
            String,
            Numeric,
            Object
        }

        public struct Result : unbound_type
        {
            private static unbound_element[] EMPTY = new unbound_element[0];

            private readonly DynAssertionExpressionNode _expr;
            private readonly unbound_type _unbound;
            private readonly string _classname;

	        public static Result INVALID = new Result();

            private readonly bool _valid;

            public Result(DynAssertionExpressionNode expression,
                unbound_type unbound, string classname)
            {
                Guard.EnsureNotNull("expression", expression);
                Guard.EnsureNotNull("freeVariables", unbound);

                this._expr = expression;
                this._unbound = unbound;
                this._classname = classname;
                this._valid = true;
            }

            public bool IsValid
            { get { return this._valid; } }

            public DynAssertionExpressionNode Expression
            { get { return this._expr ?? AstExtensions.Assertions.False; } }

            public string ClassName
            { get { return this._classname; } }

            public IEnumerator<unbound_element> GetEnumerator()
            {
                return (this._unbound ?? EMPTY).GetEnumerator();
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return ((IEnumerable<unbound_element>)this).GetEnumerator();
            }
        }

        private static InferredType FromString(string source)
        {
            switch (source)
            {
                case "B":
                case "bool":
                case "boolean":
                  return InferredType.Boolean;
                case "S":
                case "string":
                  return InferredType.String;
                case "N":
                case "num":
                case "numeric":
                  return InferredType.Numeric;
                case "O":
                case "obj":
                case "object":
                  return InferredType.Object;
                default:
                    return InferredType.Unknown;
            }
        }

        public static string ToString(InferredType type)
        {
            switch (type)
            {
                case InferredType.Boolean: return "bool";
                case InferredType.String: return "string";
                case InferredType.Numeric: return "num";
                case InferredType.Unknown:
                case InferredType.Object: 
                default:
                    return "object";
            }
        }

        private InferredType Combine(Span location, InferredType first, InferredType second, string KindOfType)
        {
            if (first == second)
            {
                return first;
            }

            if (first == InferredType.Unknown)
            {
                return second;
            }

            if (second == InferredType.Unknown)
            {
                return first;
            }

            OnError(location,
                string.Format("Type error. {0} cannot be both {1} and {2}.", KindOfType, first, second));

            return InferredType.Unknown;
        }

        private int counter = 0;
        private const int UNASSIGNED = -1;

        private int GetUniqueCounter()
        {
            // will never reach UNASSIGNED
            return checked(counter++);
        }

        private IEnumerable<KeyValuePair<string, InferredType>> GetMapping()
        {
            foreach (var kvp in _variableSlotMapping)
            {
                yield return new KeyValuePair<string, InferredType>(kvp.Key, _slotTypeMapping[kvp.Value]);
            }
        }

        private InferredType Expect(Span span, string variable, InferredType inferred)
        {
            int slot;

            if (!this._variableSlotMapping.TryGetValue(variable, out slot))
            {
                slot = GetUniqueCounter();

                this._variableSlotMapping[variable] = slot;
                this._slotTypeMapping[slot] = inferred;
                return inferred;
            }

            InferredType previousType = this._slotTypeMapping[slot];

            var result = Combine(span, inferred, previousType, string.Format("Type of Variable {0}", variable));

            this._slotTypeMapping[slot] = result;

            return result;
        }

        private void AssumeEqual(Span span, string first, string second)
        {
            int slotFirst, slotSecond;

            if (!this._variableSlotMapping.TryGetValue(first, out slotFirst))
            {
                slotFirst = UNASSIGNED;
            }
            if (!this._variableSlotMapping.TryGetValue(second, out slotSecond))
            {
                slotSecond = UNASSIGNED;
            }

            if (slotFirst == UNASSIGNED && slotSecond == UNASSIGNED)
            {
                slotFirst = GetUniqueCounter();
                this._variableSlotMapping[first] = slotFirst;
                this._variableSlotMapping[second] = slotFirst;
                this._slotTypeMapping[slotFirst] = InferredType.Unknown;
            }
            else if (slotFirst == UNASSIGNED)
            {
                this._variableSlotMapping[first] = slotSecond;
            }
            else if (slotSecond == UNASSIGNED)
            {
                this._variableSlotMapping[second] = slotFirst;
            }
            else
            {
                InferredType
                    firstType = this._slotTypeMapping[slotFirst],
                    secondType = this._slotTypeMapping[slotSecond];

                InferredType combined = Combine(span, firstType, secondType, string.Format("Type of Variables {0} and {1}", first, second));

                List<string> keysToModify = new List<string>();

                foreach (var kvp in this._variableSlotMapping)
                {
                    if (kvp.Value == slotSecond)
                    {
                        keysToModify.Add(kvp.Key);
                    }
                }

                foreach (var key in keysToModify)
                {
                    this._variableSlotMapping[key] = slotFirst;
                }

                this._slotTypeMapping[slotFirst] = combined;
            }
        }

        private int Assume(string variable, InferredType type)
        {
            Guard.EnsureNotNull("variable", variable);

            int slot = GetUniqueCounter();
            int previousSlot;

            if (!this._variableSlotMapping.TryGetValue(variable, out previousSlot))
            {
                previousSlot = UNASSIGNED;
            }

            this._variableSlotMapping[variable] = slot;
            this._slotTypeMapping[slot] = type;

            return previousSlot;
        }

        private void RevertAssume(string variable, int slot)
        {
            if (slot == UNASSIGNED)
            {
                this._variableSlotMapping.Remove(variable);
            }
            else
            {
                this._variableSlotMapping[variable] = slot;
            }
        }

        private readonly Dictionary<string, int> _variableSlotMapping = new Dictionary<string, int>();
        private readonly Dictionary<int, InferredType> _slotTypeMapping = new Dictionary<int, InferredType>();

        private EventHandler<DiagnosticEventArgs> _diagnostics;
        private readonly Func<string, string[]> _predicateLookup;
        private bool IsValid { get; set; }

        private DynAssertionTypeInference(
            EventHandler<DiagnosticEventArgs> diagnostics,
            Func<string, string[]> predicateLookup)
        {
            Guard.EnsureNotNull("predicateLookup", predicateLookup);

            this._predicateLookup = predicateLookup;
            this._diagnostics = diagnostics;
            IsValid = true;
        }

        private void OnError(Lexer.Span span, string message)
        {
            IsValid = false;
            OnDiagnosticInfo(DiagnosticSeverity.Error, span, message);
        }

        private void OnDiagnosticInfo(DiagnosticSeverity level, Lexer.Span span, string message)
        {
            if (_diagnostics != null)
            {
                _diagnostics(this, new DiagnosticEventArgs(new DiagnosticInfo(level, span, message)));
            }
        }

        public static Result Run(
            EventHandler<DiagnosticEventArgs> diagnostics,
            Func<string, string[]> predicateLookup,
            DynAssertionExpressionNode expression,
            string classname)
        {
            var context = new DynAssertionTypeInference(diagnostics, predicateLookup);

            context.Expect(Span.Empty, "true", InferredType.Boolean);
            context.Expect(Span.Empty, "false", InferredType.Boolean);
            context.Expect(Span.Empty, "null", InferredType.Object);

            context.Visit(InferredType.Boolean, expression);

            context.RevertAssume("true", UNASSIGNED);
            context.RevertAssume("null", UNASSIGNED);
            context.RevertAssume("false", UNASSIGNED);

            if (!context.IsValid)
            {
                return DynAssertionTypeInference.Result.INVALID;
            }

            return new Result(expression, new List<unbound_element>(context.GetMapping()), classname);
        }

        private InferredType Visit(
            InferredType inferred,
            DynAssertionExpressionNode expression)
        {
            switch (expression.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Universal:
                case DynAssertionExpressionNode.NodeKind.Existential:
                    return VisitQuantifier(inferred, (DynAssertionQuantifierNode)expression);

                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    return VisitTypeCheck(inferred, (DynAssertionTypeCheckNode)expression);

                case DynAssertionExpressionNode.NodeKind.Predicate:
                    return VisitPredicate(inferred, (DynAssertionPredicateNode)expression);

                case DynAssertionExpressionNode.NodeKind.Negate:
                    return VisitNegate(inferred, (DynAssertionUnaryNode)expression);

                case DynAssertionExpressionNode.NodeKind.And:
                case DynAssertionExpressionNode.NodeKind.Or:
                case DynAssertionExpressionNode.NodeKind.Implies:
                    return VisitBinary(inferred, InferredType.Boolean, InferredType.Boolean, (DynAssertionBinaryNode)expression);

                case DynAssertionExpressionNode.NodeKind.Addition:
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                case DynAssertionExpressionNode.NodeKind.Division:
                case DynAssertionExpressionNode.NodeKind.Modulus:
                    return VisitBinary(inferred, InferredType.Numeric, InferredType.Numeric, (DynAssertionBinaryNode)expression);

                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.LessThan:
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                    return VisitBinary(inferred, InferredType.Boolean, InferredType.Numeric, (DynAssertionBinaryNode)expression);

                case DynAssertionExpressionNode.NodeKind.Equality:
                    return VisitEquality(inferred, (DynAssertionBinaryNode)expression);

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    return VisitIfThenElse(inferred, (DynAssertionIfThenElseNode)expression);

                case DynAssertionExpressionNode.NodeKind.Numeric:
                    return VisitNumeric(inferred, (DynAssertionNumericNode)expression);

                case DynAssertionExpressionNode.NodeKind.String:
                    return VisitString(inferred, (DynAssertionStringNode)expression);

                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    return VisitInstanceVariable(inferred, (DynAssertionInstanceVariableNode)expression);

                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    return VisitVariable(inferred, (DynAssertionVariableNode)expression);

                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("expression.Kind", (int)expression.Kind);
            }
        }

        private InferredType VisitNumeric(InferredType inferred, DynAssertionNumericNode expression)
        {
            return Combine(expression.Location, InferredType.Numeric, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
        }

        private InferredType VisitQuantifier(
            InferredType inferred,
            DynAssertionQuantifierNode expression)
        {
            var result = Combine(expression.Location, InferredType.Boolean, inferred, string.Format("Type of Expression {0}", expression.ToSource()));

            var variableType = FromString(expression.Type);

            if (variableType == InferredType.Unknown)
            {
                OnError(expression.Location, string.Format("Invalid quantifier variable type {0} for variable {1}.\nMust be one of { bool, string, numeric, object } in {2}", expression.Type, expression.Variable, expression.ToSource()));
            }

            string varName = string.Format("'{0}", expression.Variable);
            var before = Assume(varName, variableType);

            Visit(InferredType.Boolean, expression.Expression);

            RevertAssume(varName, before);

            return result;
        }

        private InferredType VisitTypeCheck(InferredType inferred,
            DynAssertionTypeCheckNode expression)
        {
            var result = Combine(expression.Location, InferredType.Boolean, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
            Visit(InferredType.Object, expression.Expression);
            return result;
        }

        private InferredType VisitPredicate(InferredType inferred,
            DynAssertionPredicateNode expression)
        {
            var result = this._predicateLookup(expression.Name);

            if (result == null)
            {
                OnError(expression.Location, "type information for predicate " + expression.Name + " not found.");
                return inferred;
            }

            if (expression.Arguments.Count != result.Length)
            {
                OnError(expression.Location, string.Format("predicate arity mismatch for predicate {0}.", expression.Name));
                return inferred;
            }

            for (int i = 0; i < result.Length; ++i)
            {
                Visit(FromString(result[i]), expression.Arguments[i]);
            }

            return Combine(expression.Location, InferredType.Boolean, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
        }

        private InferredType VisitNegate(InferredType inferred,
            DynAssertionUnaryNode expression)
        {
            switch (expression.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Negate:
                    {
                        var result = Combine(expression.Location, InferredType.Boolean, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
                        Visit(InferredType.Boolean, expression.Expression);
                        return result;
                    }
                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("expression.Kind", (int)expression.Kind);
            }
        }

        private InferredType VisitEquality(InferredType inferred, DynAssertionBinaryNode expression)
        {
            Combine(expression.Location, InferredType.Boolean, inferred, string.Format("Type of Expression {0}", expression.ToSource()));

            bool firstAssVar = expression.First.Kind == DynAssertionExpressionNode.NodeKind.AssertionVariable;
            bool secondAssVar = expression.Second.Kind == DynAssertionExpressionNode.NodeKind.AssertionVariable;

            if (firstAssVar && secondAssVar)
            {
                DynAssertionVariableNode firstNode = (DynAssertionVariableNode)expression.First;
                DynAssertionVariableNode secondNode = (DynAssertionVariableNode)expression.Second;
                var firstVarName = string.Format("'{0}",firstNode.Name);
                var secondVarName = string.Format("'{0}",secondNode.Name);
                AssumeEqual(
                    expression.Location,
                    firstVarName,
                    secondVarName);
            }
            else if (firstAssVar)
            {
                var otherType = Visit(InferredType.Unknown, expression.Second);
                Visit(otherType, expression.First);
            }
            else
            {
                var otherType = Visit(InferredType.Unknown, expression.First);
                Visit(otherType, expression.Second);
            }

            return InferredType.Boolean;
        }

        private InferredType VisitBinary(InferredType inferred,
            InferredType outerType,
            InferredType innerType,
            DynAssertionBinaryNode expression)
        {
            var result = Combine(expression.Location, outerType, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
            Visit(innerType, expression.First);
            Visit(innerType, expression.Second);
            return result;
        }

        private InferredType VisitIfThenElse(InferredType inferred,
            DynAssertionIfThenElseNode expression)
        {
            Visit(InferredType.Boolean, expression.Condition);
            var first = Visit(inferred, expression.TrueBranch);
            var second = Visit(inferred, expression.FalseBranch);
            return Combine(expression.Location, first, second, string.Format("Type of Expression {0}", expression.ToSource()));
        }

        private InferredType InferType(InferredType inferred,
            DynAssertionNumericNode expression)
        {
            return Combine(expression.Location, InferredType.Numeric, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
        }

        private InferredType VisitString(InferredType inferred,
            DynAssertionStringNode expression)
        {
            return Combine(expression.Location, InferredType.String, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
        }

        private InferredType VisitInstanceVariable(InferredType inferred,
            DynAssertionInstanceVariableNode expression)
        {
            var result = Combine(expression.Location, InferredType.Object, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
            Visit(InferredType.Object, expression.Expression);
            return result;
        }

        private InferredType VisitVariable(InferredType inferred,
            DynAssertionVariableNode expression)
        {
            switch (expression.Name)
            {
                case "true":
                case "false":
                    return Combine(expression.Location, InferredType.Boolean, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
                case "null":
                case "self":
                case "result":
                case "undefined":
                    return Combine(expression.Location, InferredType.Object, inferred, string.Format("Type of Expression {0}", expression.ToSource()));
                default:
                    InferredType type;
                    string varname;

                    if (expression.Kind == DynAssertionExpressionNode.NodeKind.ProgramVariable)
                    {
                        type = InferredType.Object;
                        varname = expression.Name;
                    }
                    else
                    {
                        type = InferredType.Unknown;
                        varname = string.Format("'{0}", expression.Name);
                    }

                    return Expect(expression.Location, varname,
                        Combine(expression.Location, type, inferred, string.Format("Type of Variable {0}", varname)));
            }
        }
    }
}
