﻿using Dyn.Ast;
using System;
using System.Collections.Generic;

namespace Dyn.VerficationConditions
{
    public static class DynProgramVariableVisitor
    {
        public static IList<string> Visit(DynProgramStatementNode node)
        {
            Guard.EnsureNotNull("node", node);

            var result = new List<string>();
            Visit(result, node);
            return result;
        }

        private static void VisitAll<T>(List<string> result, IList<T> source, Action<List<string>, T> visitor)
        {
            foreach (var element in source)
            {
                visitor(result, element);
            }
        }

        private static void Visit(List<string> result, DynProgramStatementNode node)
        {
            switch (node.Kind)
            {
                case DynProgramStatementNode.NodeKind.MethodCall:
                    Visit(result, (DynProgramMethodCallNode)node); return;
                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    Visit(result, (DynProgramConstructorCallNode)node); return;
                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    Visit(result, (DynProgramVariableNode)node); return;
                case DynProgramStatementNode.NodeKind.If:
                    Visit(result, (DynProgramIfNode)node); return;
                case DynProgramStatementNode.NodeKind.While:
                    Visit(result, (DynProgramWhileNode)node); return;
                case DynProgramStatementNode.NodeKind.Assignment:
                    Visit(result, (DynProgramAssignmentNode)node); return;
                case DynProgramStatementNode.NodeKind.Ensures:
                case DynProgramStatementNode.NodeKind.Requires:
                case DynProgramStatementNode.NodeKind.Invariant:
                case DynProgramStatementNode.NodeKind.Assert:
                case DynProgramStatementNode.NodeKind.ByTheWay:
                    return;

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("node.Kind", (int)node.Kind);
            }
        }

        private static void Visit(List<string> result, DynProgramExpressionNode node)
        {
            Guard.EnsureNotNull("node", node);

            switch (node.Kind)
            {
                case DynProgramStatementNode.NodeKind.MethodCall:
                    Visit(result, (DynProgramMethodCallNode)node); return;
                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    Visit(result, (DynProgramConstructorCallNode)node); return;
                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    Visit(result, (DynProgramVariableNode)node); return;
                case DynProgramStatementNode.NodeKind.If:
                    Visit(result, (DynProgramIfNode)node); return;
                case DynProgramStatementNode.NodeKind.While:
                    Visit(result, (DynProgramWhileNode)node); return;
                case DynProgramStatementNode.NodeKind.Assignment:
                    Visit(result, (DynProgramAssignmentNode)node); return;
                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("node.Kind", (int)node.Kind);
            }
        }

        private static void Visit(List<string> result, DynProgramMethodCallNode node)
        {
            if (node.Expression != null)
            {
                Visit(result, node.Expression);
            }

            VisitAll(result, node.Arguments, Visit);
        }

        private static void Visit(List<string> result, DynProgramConstructorCallNode node)
        {
            VisitAll(result, node.Arguments, Visit);
        }

        private static void Visit(List<string> result, DynProgramVariableNode node)
        {
            if (node.Kind == DynProgramStatementNode.NodeKind.Variable)
            {
                switch(node.Name)
                {
                    case "null":
                    case "self":
                        break;
                    default:
                        result.Add(node.Name);
                        break;
                }
            }
        }

        private static void Visit(List<string> result, DynProgramIfNode node)
        {
            Visit(result, node.Condition);
            VisitAll(result, node.TrueBranch, Visit);
            VisitAll(result, node.FalseBranch, Visit);
        }

        private static void Visit(List<string> result, DynProgramWhileNode node)
        {
            Visit(result, node.Condition);
            VisitAll(result, node.Body, Visit);
        }

        private static void Visit(List<string> result, DynProgramAssignmentNode node)
        {
            Visit(result, node.Variable);
            Visit(result, node.Expression);
        }
    }
}
