﻿using Dyn.Ast;
using Dyn.Lexer;
using System;
using System.Linq;

namespace Dyn.VerficationConditions
{
    public static class SubstitutionExtensions
    {
        private static readonly StringComparer CMP = StringComparer.Ordinal;

        private sealed class SubstituteLocalVariable : DynAssertionNodeTransformer
        {
            private readonly DynAssertionVariableNode _element;
            private readonly DynAssertionExpressionNode _replacement;

            public SubstituteLocalVariable(DynAssertionVariableNode element, DynAssertionExpressionNode replacement)
            {
                Guard.EnsureNotNull("element", element);
                Guard.EnsureNotNull("replacement", replacement);

                this._element = element;
                this._replacement = replacement;
            }

            public override DynAssertionExpressionNode Transform(DynAssertionVariableNode node)
            {
                return ((node.Kind == _element.Kind) && CMP.Equals(node.Name, _element.Name)) ? _replacement : node;
            }

            public override DynAssertionExpressionNode Transform(DynAssertionQuantifierNode node)
            {
                if (CMP.Equals(node.Variable, _element.Name))
                {
                    return node;
                }
                else
                {
                    return base.Transform(node);
                }
            }
        }

        private sealed class SubstituteInstanceVariable : DynAssertionNodeTransformer
        {
            private readonly DynAssertionInstanceVariableNode _element;
            private DynAssertionExpressionNode _replacement;

            public SubstituteInstanceVariable(DynAssertionInstanceVariableNode element, DynAssertionExpressionNode replacement)
            {
                Guard.EnsureNotNull("element", element);
                Guard.EnsureNotNull("replacement", replacement);

                this._element = element;
                this._replacement = replacement;
            }

            public override DynAssertionExpressionNode Transform(DynAssertionInstanceVariableNode node)
            {
                if (CMP.Equals(node.Name, _element.Name))
                {
                    return new DynAssertionIfThenElseNode(
                        node.Location,
                        new DynAssertionBinaryNode(
                            node.Location,
                            DynAssertionExpressionNode.NodeKind.Equality,
                            node.Expression,
                            _element.Expression),
                        _replacement,
                        node);
                }
                else
                {
                    return node;
                }
            }
        }

        private sealed class SubstituteObjectCreation : DynAssertionNodeTransformer
        {
            private readonly string _className;

            public SubstituteObjectCreation(string className)
            {
                Guard.EnsureNotNull("className", className);
                this._className = className;
            }

            public override DynAssertionExpressionNode Transform(DynAssertionBinaryNode node)
            {
                switch (node.Kind)
                {
                    // desugaring
                    case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                        return Transform(
                            new DynAssertionBinaryNode(
                                node.Location,
                                DynAssertionExpressionNode.NodeKind.GreaterThan,
                                node.First,
                                node.Second)
                            .Or(
                                node.Location,
                                new DynAssertionBinaryNode(
                                    node.Location,
                                    DynAssertionExpressionNode.NodeKind.Equality,
                                    node.First,
                                    node.Second)));

                    case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                        return Transform(
                            new DynAssertionBinaryNode(
                                node.Location,
                                DynAssertionExpressionNode.NodeKind.LessThan,
                                node.First,
                                node.Second)
                            .Or(
                                node.Location,
                                new DynAssertionBinaryNode(
                                    node.Location,
                                    DynAssertionExpressionNode.NodeKind.Equality,
                                    node.First,
                                    node.Second)));

                    case DynAssertionExpressionNode.NodeKind.Equality:
                        {
                            var r = Ast.AstExtensions.Assertions.r;
                            var rEqualsFirst = AstExtensions.Normal.Equals(r, node.First);
                            var rEqualsSecond = AstExtensions.Normal.Equals(r, node.Second);

                            if (rEqualsFirst && rEqualsSecond)
                            {
                                return AstExtensions.Assertions.True;
                            }
                            else if (rEqualsFirst && !rEqualsSecond && node.Second.Kind != DynAssertionExpressionNode.NodeKind.IfThenElse)
                            {
                                return AstExtensions.Assertions.False;
                            }
                            else if (!rEqualsFirst && rEqualsSecond && node.First.Kind != DynAssertionExpressionNode.NodeKind.IfThenElse)
                            {
                                return AstExtensions.Assertions.False;
                            }
                            else if (node.First.Kind == DynAssertionExpressionNode.NodeKind.IfThenElse)
                            {
                                var element = (DynAssertionIfThenElseNode)node.First;

                                return Transform(new DynAssertionIfThenElseNode(
                                    element.Location,
                                    element.Condition,
                                    element.TrueBranch.Equality(Span.Empty, node.Second),
                                    element.FalseBranch.Equality(Span.Empty, node.Second)));
                            }
                            else if (node.Second.Kind == DynAssertionExpressionNode.NodeKind.IfThenElse)
                            {
                                var element = (DynAssertionIfThenElseNode)node.Second;

                                return Transform(new DynAssertionIfThenElseNode(
                                    element.Location,
                                    element.Condition,
                                    element.TrueBranch.Equality(Span.Empty, node.First),
                                    element.FalseBranch.Equality(Span.Empty, node.First)));
                            }
                            else
                            {
                                return base.Transform(node);
                            }
                        }

                    case DynAssertionExpressionNode.NodeKind.Addition:
                    case DynAssertionExpressionNode.NodeKind.And:
                    case DynAssertionExpressionNode.NodeKind.Division:
                    case DynAssertionExpressionNode.NodeKind.GreaterThan:
                    case DynAssertionExpressionNode.NodeKind.Implies:
                    case DynAssertionExpressionNode.NodeKind.LessThan:
                    case DynAssertionExpressionNode.NodeKind.Modulus:
                    case DynAssertionExpressionNode.NodeKind.Multiplication:
                    case DynAssertionExpressionNode.NodeKind.Or:
                    case DynAssertionExpressionNode.NodeKind.Subtraction:
                        return base.Transform(node);

                    default:
                        throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("node.Kind", (int)node.Kind);
                }
            }

            public override DynAssertionExpressionNode Transform(DynAssertionTypeCheckNode node)
            {
                // since we know the type of 'result', we can in this case decide the typecheck
                if (AstExtensions.IgnoreLocation.Equals(node.Expression, AstExtensions.Assertions.r))
                {
		  return node.TypeSet.Contains(_className, CMP) ? AstExtensions.Assertions.True : AstExtensions.Assertions.False;
                } else {
                  return node;
                }
            }
        }

        public static DynAssertionExpressionNode Substitute(
            this DynAssertionExpressionNode post,
            DynAssertionVariableNode element,
            DynAssertionExpressionNode replacement)
        {
            return new SubstituteLocalVariable(element, replacement).Transform(post);
        }

        public static DynAssertionExpressionNode Substitute(
            this DynAssertionExpressionNode post,
            DynAssertionInstanceVariableNode element,
            DynAssertionExpressionNode replacement)
        {
            return new SubstituteInstanceVariable(element, replacement).Transform(post);
        }

        public static DynAssertionExpressionNode Substitute(
            this DynAssertionExpressionNode post,
            string className)
        {
            return new SubstituteObjectCreation(className).Transform(post);
        }
    }
}
