﻿using Dyn.Ast;
using Dyn.Diagnostics;
using Dyn.Lexer;
using System;
using System.Collections.Generic;

namespace Dyn.VerficationConditions
{
    public sealed class VerficationConditionGenerator
    {
        private readonly IDictionary<string, IDictionary<string, Contract>> _contracts =
            new Dictionary<string, IDictionary<string, Contract>>();

        private readonly IDictionary<string, Contract> _buildins =
            new Dictionary<string, Contract>();

        private readonly IDictionary<string, string[]> _predicates =
            new Dictionary<string, string[]>();

        public VerficationConditionGenerator(IEnumerable<DynProgramClassNode> classes)
        {
            AddContracts(classes);

// hook_loc_identical(a,b)
//    requires true
//    ensures ((map_B(result, true) and a == b) or (map_B(result, false) and !(a == b)))
            var pre = AstExtensions.Assertions.True;
            var leftleft = new DynAssertionPredicateNode(Span.Empty, "map_B", new List<DynAssertionExpressionNode>(){ AstExtensions.Assertions.r, AstExtensions.Assertions.True });
            var leftright = new DynAssertionBinaryNode(Span.Empty, DynAssertionExpressionNode.NodeKind.Equality, new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "a"), new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "b"));
            var left = new DynAssertionBinaryNode(Span.Empty, DynAssertionExpressionNode.NodeKind.And, leftleft, leftright);

            var rightleft = new DynAssertionPredicateNode(Span.Empty, "map_B", new List<DynAssertionExpressionNode>(){ AstExtensions.Assertions.r, AstExtensions.Assertions.False });
            var rightright = new DynAssertionUnaryNode(Span.Empty, 
new DynAssertionBinaryNode(Span.Empty, DynAssertionExpressionNode.NodeKind.Equality, new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "a"), new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "b")));
            var right = new DynAssertionBinaryNode(Span.Empty, DynAssertionExpressionNode.NodeKind.And, rightleft, rightright);
            var post = new DynAssertionBinaryNode(Span.Empty, DynAssertionExpressionNode.NodeKind.Or, left, right);
	    var contract = new Contract(pre, post);
            AddBuildIn("hook_loc_identical", new List<string>(){ "a", "b" }, contract);

	    // boolean
            AddPredicate("is_B", "object");
            AddPredicate("map_B", "object", "bool");

            // numeric
            AddPredicate("is_N", "object");
            AddPredicate("map_N", "object", "num");

            // list
            AddPredicate("is_L", "object");
            AddPredicate("select_L", "object", "num", "object");

            // for example sorting
            AddPredicate("permutation", "object", "object");
            AddPredicate("sorted", "object");
            AddPredicate("sorted_partial", "object", "num", "num");
            AddPredicate("select", "object", "num", "object");
            AddPredicate("le_L", "object", "num", "num", "object");
            AddPredicate("ge_L", "object", "num", "num", "object");
        }

        #region Diagnostic

        public event EventHandler<DiagnosticEventArgs> DiagnosticInfo;

        public void OnDiagnosticInfo(DiagnosticSeverity level, Span span, string message)
        {
            var handler = DiagnosticInfo;

            if (handler != null)
            {
                handler(this, new DiagnosticEventArgs(new DiagnosticInfo(level, span, message)));
            }
        }

        #endregion Diagnostic

        private static readonly string[] EMPTY_STRING = new string[0];

        public bool AddPredicate(string key, params string[] values)
        {
            if (this._predicates.ContainsKey(key))
            {
                return false;
            }

            this._predicates.Add(key, values ?? EMPTY_STRING);
            return true;
        }

        public string[] PredicateLookup(string key)
        {
            string[] result;
            if (!this._predicates.TryGetValue(key, out result) || result == null)
            {
                return null;
            }

            return result;
        }

        public DynAssertionExpressionNode GeneratePreconditionMethodCall(
            string name,
            int arity,
            bool global,
            DynAssertionExpressionNode postCondition)
        {
            Guard.EnsureNonnegative("arity", arity);
            Guard.EnsureNotNull("name", name);
            Guard.EnsureNotNull("postCondition", postCondition);

            DynAssertionExpressionNode result = null;
	    if (global) {
              Contract contract;
              if (!this._buildins.TryGetValue(Contract.CreateKey(name, arity), out contract))
              {
		// NOTE: calling a non-existing buildin causes a Typeerror. Hence the only way for this
                // to be correct is that the call is unreachable...
                // TODO: this of course only applies if we are verifying typesafe partial correctness.
                return AstExtensions.Assertions.False;
              }
              result = GeneratePreconditionInner(arity, postCondition, contract);

            } else {
              IDictionary<string, Contract> outerValue;
              if (!this._contracts.TryGetValue(Contract.CreateKey(name, arity), out outerValue))
              {
		// NOTE: calling a non-existing method causes a Typeerror. Hence the only way for this
                // to be correct is that the call is unreachable...
                // TODO: this of course only applies if we are verifying typesafe partial correctness.
                return AstExtensions.Assertions.False;
              }

              foreach (var kvp in outerValue)
              {
                var current = GeneratePreconditionInner(arity, postCondition, kvp.Value);
                current = new DynAssertionTypeCheckNode(
                    current.Location,
                    Contract.CreateArgumentVariable(0),
                    Utilities.AsArray(kvp.Key)).And(current.Location, current);

                result = result == null ? current : result.Or(result.Location, current);
              }
            }

            return result;
        }

        public DynAssertionExpressionNode GeneratePreconditionContructor(
            string className,
            int arity,
            DynAssertionExpressionNode postCondition)
        {
            Guard.EnsureNonnegative("arity", arity);
            Guard.EnsureNotNull("className", className);
            Guard.EnsureNotNull("postCondition", postCondition);

            IDictionary<string, Contract> outerValue;

            var outerKey = Contract.CreateKey("init", arity);

            if (!this._contracts.TryGetValue(outerKey, out outerValue))
            {
                OnDiagnosticInfo(DiagnosticSeverity.Error, postCondition.Location,
                    string.Format("Method 'init' with arity {0} in class '{1}' not found.", arity, className));

                return Ast.AstExtensions.Assertions.False;
            }

            Contract innerValue;

            if (!outerValue.TryGetValue(className, out innerValue))
            {
                OnDiagnosticInfo(DiagnosticSeverity.Error, postCondition.Location,
                    string.Format("Method 'init' with arity {0} in class '{1}' not found.", arity, className));

                return Ast.AstExtensions.Assertions.False;
            }

            return GeneratePreconditionInner(arity, postCondition, innerValue);
        }

        private DynAssertionExpressionNode GeneratePreconditionInner(
            int arity,
            DynAssertionExpressionNode postCondition,
            Contract contract)
        {
            Guard.EnsureNonnegative("arity", arity);
            Guard.EnsureNotNull("postCondition", postCondition);
            Guard.EnsureNotNull("contract", contract);

            var result = AstExtensions.Implies(contract.PostCondition, postCondition);

            // result is also an argument variable as every method modifies it
            result = new DynAssertionQuantifierNode(
                result.Location,
                DynAssertionExpressionNode.NodeKind.Universal,
                WeakestPreconditionCalculus._r.Name,
                "object",
                result);
            for (int i = 0; i < arity + 1; ++i)
            {
                result = new DynAssertionQuantifierNode(
                    result.Location,
                    DynAssertionExpressionNode.NodeKind.Universal,
                    Contract.CreateArgumentVariable(i).Name,
                    "object",
                    result);
            }

            result = contract.PreCondition.And(result.Location, result);

            foreach(var element in DynAssertionTypeInference.Run(DiagnosticInfo, PredicateLookup, result, string.Empty))
            {
                // program variables are already universally quantified
                if (Contract.GetArgumentVariables(arity).Contains(element.Key)) { continue; }
                // this includes result
                if (String.Equals(element.Key, WeakestPreconditionCalculus._r.Name)) { continue; }
                // also do not quantify variables that occur free in the postCondition
                if (postCondition.FreeVariables().Contains(element.Key)) { continue; }
                // for all others, add a quantifier:
                result = new DynAssertionQuantifierNode(
                    result.Location,
                    DynAssertionExpressionNode.NodeKind.Existential,
                    NameForAssertionVar(element.Key),
                    DynAssertionTypeInference.ToString(element.Value),
                    result);
            }

            return result;
        }

        public string NameForAssertionVar(string varName) {
          if (varName.StartsWith("'")) {
            return varName.Substring(1);
          }
          return varName;
        }

        public void AddBuildIn(string name, List<string> parameters, Contract contract)
        {
          contract = contract.Normalize(parameters);

          var key = Contract.CreateKey(name, parameters.Count);

	  Contract oldContract;
          if (this._buildins.TryGetValue(key, out oldContract))
          {
             throw new ArgumentException(string.Format("Duplicate Buildin {0} with arity {1} registered!"));
          }
          this._buildins.Add(key, contract);
        }

        public void AddContracts(
            IEnumerable<DynProgramClassNode> classes)
        {
            Guard.EnsureNotNull("classes", classes);

            foreach (var c in classes)
            {
                var classContract = Contract.FromNodes(c.Contracts);

                foreach (var m in c.Methods)
                {
                    var contract = Contract.FromNodes(m.Contracts).Normalize(m.Parameters) & classContract;

                    var key = Contract.CreateKey(m.Name, m.Parameters.Count);

                    IDictionary<string, Contract> outerValue;

                    if (!this._contracts.TryGetValue(key, out outerValue))
                    {
                        outerValue = new Dictionary<string, Contract>();
                        this._contracts.Add(key, outerValue);
                    }

                    outerValue.Add(c.Name, contract);
                }
            }
        }
    }
}
