﻿using Dyn.Ast;
using Dyn.Diagnostics;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Linq;

namespace Dyn.VerficationConditions
{


    public sealed class WeakestPreconditionCalculus
    {
        internal readonly static DynAssertionVariableNode _r =
            new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "_result");

        private readonly VerficationConditionGenerator _contractLookup;
        private readonly Action<VerificationRequest> _addVC;
        
        private string _classname = string.Empty;

        public WeakestPreconditionCalculus(
            VerficationConditionGenerator contractLookup,
            Action<VerificationRequest> addVerificationCondition)
        {
            Guard.EnsureNotNull("contractLookup", contractLookup);
            Guard.EnsureNotNull("addVerificationCondition", addVerificationCondition);

            this._contractLookup = contractLookup;
            this._addVC = addVerificationCondition;
        }

        private void RaiseVerificationRequest(DynAssertionExpressionNode expression, string message)
        {
            this._addVC(new VerificationRequest(expression, this._classname, message));
        }

        private int _uniqueCounter = 0;

        private DynAssertionVariableNode GenerateUniqueVariable()
        {
            ++this._uniqueCounter;
            return new DynAssertionVariableNode(Lexer.Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "_u" + _uniqueCounter);
        }

        #region Diagnostic

        public event EventHandler<DiagnosticEventArgs> DiagnosticInfo;

        public void OnDiagnosticInfo(DiagnosticSeverity level, Span span, string message)
        {
            var handler = DiagnosticInfo;

            if (handler != null)
            {
                handler(this, new DiagnosticEventArgs(new DiagnosticInfo(level, span, message)));
            }
        }

        #endregion Diagnostic

        public DynAssertionExpressionNode Wp(IList<DynProgramStatementNode> nodes, DynAssertionExpressionNode post)
        {
            Guard.EnsureNotNull("nodes", nodes);
            Guard.EnsureNotNull("post", post);

            int i = nodes.Count;

            while (i-- > 0)
            {
                post = Wp(nodes[i], post);
            }

            // Bugfix for {} {r == null} => null === null
            //This has to be at the end to be as a -1-th statement
            // post = Wp(AstExtensions.Programs.Null, post);

            return post;
        }

        public DynAssertionExpressionNode Wp(DynProgramStatementNode expression, DynAssertionExpressionNode post)
        {
            Guard.EnsureNotNull("expression", expression);
            Guard.EnsureNotNull("post", post);

            switch (expression.Kind)
            {
                case DynProgramStatementNode.NodeKind.Ensures:
                case DynProgramStatementNode.NodeKind.Requires:
                case DynProgramStatementNode.NodeKind.Invariant:
                    OnDiagnosticInfo(DiagnosticSeverity.Error, expression.Location, "Invalid program statement");
                    return AstExtensions.Assertions.False;

                case DynProgramStatementNode.NodeKind.Assert:
                    {
                        var expr = ((DynProgramContractDeclarationNode)expression).Expression;
                        RaiseVerificationRequest(expr.Implies(post), "Assertion too weak.");
                        return expr;
                    }

                case DynProgramStatementNode.NodeKind.ByTheWay:
                    return post.And(post.Location, ((DynProgramContractDeclarationNode)expression).Expression);

                case DynProgramStatementNode.NodeKind.Assignment:
                    {
                        var assignment = (DynProgramAssignmentNode)expression;
                        var variable = assignment.Variable;
                        var variableKind = variable.Kind;

                        switch (variableKind)
                        {
                            case DynProgramStatementNode.NodeKind.InstanceVariable:
                                return WpAssignment(TransformInstanceVariable(variable), assignment.Expression, post);

                            case DynProgramStatementNode.NodeKind.Variable:
                                return WpAssignment(TransformVariable(variable), assignment.Expression, post);

                            default:
                                throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("Variable.Kind", (int)variableKind);
                        }
                    }
                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return WpVariable((DynProgramVariableNode)expression, post);

                case DynProgramStatementNode.NodeKind.If:
                    return WpIfStatement((DynProgramIfNode)expression, post);

                case DynProgramStatementNode.NodeKind.While:
                    return WpWhileStatement((DynProgramWhileNode)expression, post);

                case DynProgramStatementNode.NodeKind.MethodCall:
                    return WpMethodCall((DynProgramMethodCallNode)expression, post);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return WpConstructorCall((DynProgramConstructorCallNode)expression, post);

                case DynProgramStatementNode.NodeKind.ObjectCreation:
                    return WpObjectCreation((DynProgramObjectCreationNode)expression, post);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("value", (int)expression.Kind);
            }
        }

        private DynAssertionExpressionNode WpWhileStatement(DynProgramWhileNode expression, DynAssertionExpressionNode post)
        {
            var span = expression.Location;

            var i0 = expression.Invariant;
            var i1 = Wp(expression.Condition, expression.Invariant);

            var iLocation = new Span(
                expression.Invariant.Location.FileName,
                expression.Invariant.Location.Position,
                expression.Invariant.Location.Line,
                expression.Invariant.Location.Column,
                0);

            RaiseVerificationRequest(
                i0.And(
                    i0.Location,
                    new DynAssertionPredicateNode(
                        iLocation,
                        "map_B",
                        Utilities.AsArray(
                            AstExtensions.Assertions.r,
                            AstExtensions.Assertions.True))).Implies(Wp(expression.Body, i1)),
                "loop invariant is not inductive");
            RaiseVerificationRequest(
                i0.And(
                    i0.Location,
                    new DynAssertionPredicateNode(
                        iLocation,
                        "map_B",
                        Utilities.AsArray(
                            AstExtensions.Assertions.r,
                            AstExtensions.Assertions.False))).Implies(post),
                "loop invariant too weak");

            return i1;
        }

        private static DynAssertionVariableNode TransformVariable(DynProgramVariableNode expression)
        {
            if (expression.Kind != DynProgramStatementNode.NodeKind.Variable)
            {
                throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("expression.Kind", (int)expression.Kind);
            }

            if (StringComparer.Ordinal.Equals("null", expression.Name) || StringComparer.Ordinal.Equals("self", expression.Name))
            {
                return new DynAssertionVariableNode(expression.Location, DynAssertionExpressionNode.NodeKind.ProgramVariable, expression.Name);
            }

            return new DynAssertionVariableNode(expression.Location, DynAssertionExpressionNode.NodeKind.ProgramVariable, expression.Name);
        }

        private static DynAssertionInstanceVariableNode TransformInstanceVariable(DynProgramVariableNode expression)
        {
            if (expression.Kind != DynProgramStatementNode.NodeKind.InstanceVariable)
            {
                throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("expression.Kind", (int)expression.Kind);
            }

            return new DynAssertionInstanceVariableNode(expression.Location, AstExtensions.Assertions.Self, expression.Name);
        }

        private DynAssertionExpressionNode WpVariable(DynProgramVariableNode expression, DynAssertionExpressionNode post)
        {
            if (expression.Kind == DynProgramStatementNode.NodeKind.Variable)
            {
                return post.Substitute(AstExtensions.Assertions.r, TransformVariable(expression));
            }
            else
            {
                return post.Substitute(AstExtensions.Assertions.r, TransformInstanceVariable(expression));
            }
        }

        private DynAssertionExpressionNode WpAssignment(DynAssertionVariableNode variable, DynProgramExpressionNode expression, DynAssertionExpressionNode post)
        {
            var intermediate = post.Substitute(variable, AstExtensions.Assertions.r);
            return Wp(expression, intermediate);
        }

        private DynAssertionExpressionNode WpAssignment(DynAssertionInstanceVariableNode variable, DynProgramExpressionNode expression, DynAssertionExpressionNode post)
        {
            return Wp(expression, post.Substitute(variable, AstExtensions.Assertions.r));
        }

        private DynAssertionExpressionNode WpIfStatement(DynProgramIfNode expression, DynAssertionExpressionNode post)
        {
            var b = GenerateUniqueVariable();
            var r = AstExtensions.Assertions.r;

            var span = expression.Location;

	    var trueExp = Wp(expression.TrueBranch, post);
            var falseExp = Wp(expression.FalseBranch, post);

            var result = new DynAssertionQuantifierNode(
                span,
                DynAssertionExpressionNode.NodeKind.Existential,
                b.Name,
                "bool",
                new DynAssertionPredicateNode(span, "map_B", Utilities.AsArray(r, b)).And(
                    span,
                    b.And(
                        span,
                        trueExp)
                    .Or(
                        span,
                        b.Negate(span)
                        .And(
                            span,
                            falseExp))));

            return Wp(expression.Condition, result);
        }

        private DynAssertionExpressionNode WpMethodCall(DynProgramMethodCallNode expression, DynAssertionExpressionNode post)
        {
	    // Method Calls may multiply the previous postcondition. Hence this is a good
            // place to use simplification to counteract blow-up.
	    post = Simplifier.Simplify(post);
            post = post.Substitute(AstExtensions.Assertions.r, _r);

            post = this._contractLookup.GeneratePreconditionMethodCall(
                expression.Name,
                expression.Arguments.Count,
                AstExtensions.IgnoreLocation.Equals(expression.Expression, AstExtensions.Programs.Global),
                post);

            int argc = expression.Arguments.Count;

            var _vArguments = new List<DynAssertionVariableNode>(argc);
            var _v0 = GenerateUniqueVariable();

            // need unique variables, because nested method calls are overriding _a1 ...
            for (int i = 0; i < argc; ++i)
            {
                var current = GenerateUniqueVariable();
                _vArguments.Add(current);
                post = post.Substitute(Contract.CreateArgumentVariable(i + 1), current);
            }
            post = post.Substitute(Contract.CreateArgumentVariable(0), _v0);

            // reverse order because of dependencies like a = 0; foo(a = a + 1, a = a + 1, a = a + 1) => foo(1, 2, 3) /\ a == 4
            for (int i = argc - 1; i >= 0; --i)
            {
                post = WpAssignment(_vArguments[i], expression.Arguments[i], post);
            }

            return WpAssignment(_v0, expression.Expression, post);
        }

        private DynAssertionExpressionNode WpConstructorCall(DynProgramConstructorCallNode expression, DynAssertionExpressionNode post)
        {
            post = post.Substitute(AstExtensions.Assertions.r, _r);

            post = this._contractLookup.GeneratePreconditionMethodCall(
                AstExtensions.Programs.ConstructorMethodName,
                expression.Arguments.Count,
                false,
                post);

            int argc = expression.Arguments.Count;

            var _vArguments = new List<DynAssertionVariableNode>(argc);
            var _v0 = GenerateUniqueVariable();

            // need unique variables, because nested method calls are overriding _a1 ...
            for (int i = 0; i < argc; ++i)
            {
                var current = GenerateUniqueVariable();
                _vArguments.Add(current);
                post = post.Substitute(Contract.CreateArgumentVariable(i + 1), current);
            }
            // TODO @Dennis: why could this not be part of the loop?
            post = post.Substitute(Contract.CreateArgumentVariable(0), _v0);

            // reverse order because of dependencies like a = 0; foo(a = a + 1, a = a + 1, a = a + 1) => foo(1, 2, 3) /\ a == 4
            for (int i = argc - 1; i >= 0; --i)
            {
                post = WpAssignment(_vArguments[i], expression.Arguments[i], post);
            }
	    post = WpAssignment(_v0, new DynProgramObjectCreationNode(Span.Empty, expression.Name), post);
            // Constructor Calls often duplicate the previous postcondition. Hence this is a good point
            // to use simplification to counteract blow-up.
            post = Simplifier.Simplify(post);
            return post;
        }

        private DynAssertionExpressionNode WpObjectCreation(DynProgramObjectCreationNode expression, DynAssertionExpressionNode post)
        {
            // substitution for object creation
            return post.Substitute(expression.ClassName);
        }

        public void Verify(IEnumerable<DynProgramClassNode> classes)
        {
            foreach (var c in classes)
            {
                Verify(c);
            }
        }

        public void Verify(DynProgramClassNode node)
        {
            this._classname = node.Name;
            var a = Contract.FromNodes(node.Contracts);

            foreach (var m in node.Methods)
            {
                var b = Contract.FromNodes(m.Contracts);

                var c = a & b;

                Verify(m, c.PreCondition, c.PostCondition);
            }
            this._classname = string.Empty;
        }

        public void Verify(
            DynProgramMethodDeclarationNode node,
            DynAssertionExpressionNode preCondition,
            DynAssertionExpressionNode postCondition)
        {
            var calculatedPreCondition = Wp(node.Statements, postCondition);

            // undefine variables
            foreach (var variable in GetProgramVariables(node.Statements).Distinct())
            {
		// do NOT undefine parameters
		if (node.Parameters.Contains(variable))
                {
                  continue;
                };

                calculatedPreCondition = WpAssignment(
                    new DynAssertionVariableNode(node.Location, DynAssertionExpressionNode.NodeKind.ProgramVariable, variable),
                    AstExtensions.Programs.Undefined,
                    calculatedPreCondition);
            }

            RaiseVerificationRequest(
                preCondition.Implies(calculatedPreCondition),
                string.Format("precondition of method {0}.{1} must imply weakest precondition", this._classname, node.Name));
        }

        private IEnumerable<string> GetProgramVariables(IEnumerable<DynProgramStatementNode> statements)
        {
            return statements.SelectMany(DynProgramVariableVisitor.Visit);
        }
    }
}
