﻿using Dyn.Ast;
using Dyn.Diagnostics;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;

namespace Dyn.VerficationConditions
{
    public sealed class Z3Adapter
    {
        public enum Z3Result
        {
            Invalid,
            Satisfiable,
            Unsatisfiable,
        }

        private readonly string _filePath;
        private readonly string _preamble;
        private int _run_id;

	public int RunID() {
          return this._run_id;
        }

        public Z3Adapter(string filePath, IEnumerable<DynProgramClassNode> context, string predicates)
        {
            Guard.EnsureNotNull("filePath", filePath);
            Guard.EnsureNotNull("context", context);

            this._filePath = filePath;
            this._preamble = BuildPreamble(context, predicates);
            this._run_id = 0;
        }

        private static string BuildPreamble(IEnumerable<DynProgramClassNode> context, string predicates)
        {
            var builder = new StringBuilder();

            builder.AppendLine(@"(set-option :produce-unsat-cores true)
(set-option :produce-models true)
;(set-option :produce-proofs true)

; --------------- Basic Definitions -------------------

(declare-sort Obj 0)

(declare-const Null Obj)

; --------------- Classes -------------------

(declare-sort Class 0)

(declare-fun class(Obj) Class)");

            GenerateClassSection(builder, context);

            builder.AppendLine("(assert (= (class Null) class_nullclass))");
            builder.AppendLine(@"
; --------------- Instance variables -------------------

(declare-sort IVar 0)
");
            GenerateInstanceVariableSection(builder, context);

            builder.AppendLine(@"
(declare-fun inst_ref(Obj IVar) Obj)
(declare-const this Obj)

; --------------- Predicates -------------------------------
");
            builder.AppendLine(predicates);
            builder.AppendLine(@"
; --------------- Contextual Information -------------------

(declare-const lv_true Bool)
(declare-const lv_false Bool)

(assert (= lv_true true))
(assert (= lv_false false))

; TODO: local program variables
");

            return builder.ToString();
        }

        private static void GenerateClassSection(StringBuilder builder, IEnumerable<DynProgramClassNode> context)
        {
            var distinctAssert = new StringBuilder();

            foreach (var c in context)
            {
                var symbol = DynAssertionSmtLibFormatter.ToSymbol(c);

                builder.AppendLine(string.Format("(declare-const {0} Class)", symbol));

                distinctAssert.Append(' ').Append(symbol);
            }

            if (distinctAssert.Length > 0)
            {
                builder.Append("(assert (distinct");
                builder.Append(distinctAssert.ToString());
                builder.AppendLine("))");
            }
        }

        private static void GenerateInstanceVariableSection(StringBuilder builder, IEnumerable<DynProgramClassNode> context)
        {
            var distinctAssert = new StringBuilder();

            foreach (var ivar in context.SelectMany(DynProgramInstanceVariableVisitor.Visit).Distinct())
            {
                var symbol = DynAssertionSmtLibFormatter.ToSymbolInstanceVariableName(ivar);

                builder.AppendLine(string.Format("(declare-const {0} IVar)", symbol));
                distinctAssert.Append(' ').Append(symbol);
            }

            if (distinctAssert.Length > 0)
            {
                builder.Append("(assert (distinct");
                builder.Append(distinctAssert.ToString());
                builder.AppendLine("))");
            }
        }

        #region Diagnostic

        public event EventHandler<DiagnosticEventArgs> DiagnosticInfo;

        private void OnDiagnosticInfo(DiagnosticSeverity level, Span span, string message)
        {
            var handler = DiagnosticInfo;

            if (handler != null)
            {
                handler(this, new DiagnosticEventArgs(new DiagnosticInfo(level, span, message)));
            }
        }

        #endregion Diagnostic

        public Z3Result Run(DynAssertionTypeInference.Result annotatedExpression)
        {
	    this._run_id = this._run_id + 1;
            try
            {
                using (var z3 = new Process())
                {
                    z3.StartInfo.FileName = this._filePath;
                    z3.StartInfo.Arguments = "-smt2 -in";

                    z3.StartInfo.RedirectStandardInput = true;
                    z3.StartInfo.RedirectStandardOutput = true;
                    z3.StartInfo.UseShellExecute = false;
                    z3.StartInfo.CreateNoWindow = true;

                    if (!z3.Start())
                    {
                        OnDiagnosticInfo(DiagnosticSeverity.Error, Span.Empty, "Cannot start z3.");
                        return Z3Result.Invalid;
                    }

                    // DEBUG
                    var tmpFile = new StreamWriter("/tmp/z3-" + this._run_id + ".smt");
                    WriteInput(tmpFile, annotatedExpression);
                    tmpFile.Dispose();

                    WriteInput(z3.StandardInput, annotatedExpression);
                    z3.StandardInput.Dispose();
                    var result = ReadOutput(z3.StandardOutput);

                    z3.WaitForExit();

                    return result;
                }
            }
            catch (Exception e) //TODO: Specialize
            {
                OnDiagnosticInfo(DiagnosticSeverity.Error, Span.Empty, "z3 process exception: " + e.Message);
            }

            return Z3Result.Invalid;
        }

        private void WriteInput(
            StreamWriter writer,
            DynAssertionTypeInference.Result annotatedExpression)
        {
            StringBuilder builder = new StringBuilder(this._preamble);

            builder.AppendLine(@"
; --------------- Contextual Information -------------------");

            builder.AppendLine("(declare-const pv_self Obj)");
            builder.Append("(assert (= (class pv_self) class_");
            builder.Append(annotatedExpression.ClassName);
            builder.AppendLine("))");

            builder.AppendLine(@"
; --------------- Verification Condition -------------------");

            foreach(var element in annotatedExpression)
            {
                builder.AppendFormat("(declare-const {0} {1})\n", 
                    DynAssertionSmtLibFormatter.ToSymbol(element.Key),
                    DynAssertionSmtLibFormatter.ToSymbolType(element.Value));
            }
            builder.AppendLine(@"

");

            builder.AppendFormat("(assert (not {0}))\n", 
                DynAssertionSmtLibFormatter.Visit(annotatedExpression.Expression));

            builder.AppendLine(@"
; --------------- Verify -------------------
(check-sat)");

            OnDiagnosticInfo(DiagnosticSeverity.None, new Span("z3.in", 0, 0, 0, builder.Length), builder.ToString());

            writer.Write(builder.ToString());
        }

        private Z3Result ReadOutput(StreamReader reader)
        {
            var line = reader.ReadLine().TrimEnd();

            var span = new Span("z3.out", 0, 0, 0, 0);

            if (StringComparer.Ordinal.Equals("sat", line))
            {
                OnDiagnosticInfo(DiagnosticSeverity.Information, span, "Verification condition invalid.");
                return Z3Result.Satisfiable;
            }

            if (StringComparer.Ordinal.Equals("unsat", line))
            {
                OnDiagnosticInfo(DiagnosticSeverity.Information, span, "Verification condition successfully verified.");
                return Z3Result.Unsatisfiable;
            }

            var tail = reader.ReadToEnd();

            OnDiagnosticInfo(DiagnosticSeverity.Error, span, "Verification failed: " + line + tail ?? "Nothing");

            return Z3Result.Invalid;
        }
    }
}
