﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace libdyn.generator.Ast
{
    public sealed class CodeGenerator
    {
        private readonly TextWriter _debug, _code;

        public CodeGenerator(TextWriter debug, TextWriter code)
        {
            Guard.EnsureNotNull("debug", debug);
            Guard.EnsureNotNull("code", code);

            this._debug = debug;
            this._code = code;
        }

        public TextWriter Code { get { return this._code; } }
        public TextWriter Debug { get { return this._debug; } }

        public string TypeEnum { get; set; }

        private IEnumerable<AstClass> NonNullAst { get { return Ast ?? Enumerable.Empty<AstClass>(); } }

        public IEnumerable<AstClass> Ast { get; set; }

        public void WriteAstClasses()
        {
            if (Ast == null)
            {
                return;
            }
            bool tail = false;

            foreach (var c in NonNullAst)
            {
                if (tail)
                {
                    Code.WriteLine();
                }

                WriteClass(c);

                tail = true;
            }
        }

        private void WriteClass(AstClass value)
        {
            Debug.WriteLine(value);

            // is abstract if is parent and Defaults in children
            bool isSealed = !NonNullAst.Any(e => value.Name.Equals(e.Parent));
            bool isAbstract = value.IsAbstract;

            if (!string.IsNullOrEmpty(value.Documentation))
            {
                var doc = value.Documentation.TrimEnd();
                Code.Write(Utilities.Indent("<summary>" + doc + "</summary>", "/// "));
                Code.WriteLine("/// <remarks>This class is immutable and therefore thread-safe.</remarks>");
            }

            Code.Write("public{0}{1} class {2}", isAbstract ? " abstract" : "", isSealed ? " sealed" : "", value.Name);

            if (value.Parent != null)
            {
                Code.Write(" : {0}", value.Parent);
            }

            Code.WriteLine();
            Code.WriteLine("{");

            var enumProperty = TypeEnum != null ? value.Properties.SingleOrDefault(e => TypeEnum.Equals(e.Type)) : null;

            if (TypeEnum != null && enumProperty != null)
            {
                WriteTypeEnum(enumProperty.Name);
                Code.WriteLine();
            }

            foreach (var p in value.Properties)
            {
                Code.WriteLine("\tprivate readonly {0} {1};", p.Type, ModelUtilities.ToMember(p.Name));
            }

            WriteLineIfAny(Code, value.Properties);

            GenerateConstructor(value, ModelUtilities.GetConstructorArguments(NonNullAst, value));

            bool tail = false;

            foreach (var p in value.Properties)
            {
                if (tail)
                {
                    Code.WriteLine();
                }

                GenerateProperty(p);

                tail = true;
            }

            Code.WriteLine("}");
        }

        private static void WriteLineIfAny<T>(TextWriter writer, IEnumerable<T> properties)
        {
            if (properties != null && properties.Any())
            {
                writer.WriteLine();
            }
        }

        private void GenerateConstructor(AstClass current, IList<AstProperty> constructorArguments)
        {
            Code.WriteLine("\t/// <summary> Initializes a new instance of the <see cref=\"{0}\" /> class.</summary>", current.Name);
            foreach (var property in constructorArguments.Where(e => !string.IsNullOrEmpty(e.Documentation)))
            {
                Code.WriteLine("\t/// <param name=\"{0}\">{1}</param>", property.Name, property.Documentation.TrimEnd());
            }

            Code.Write("\t{0} {1}({2})", current.IsAbstract ? "internal protected" : "public", current.Name,
                string.Join(", ", constructorArguments.Select(ModelUtilities.ToParameter)));

            Code.WriteLine();

            var validators = new List<KeyValuePair<AstDefault, AstProperty>>();

            if (current.Parent != null)
            {
                Code.Write("\t\t: base(");

                var inheritedDefaults = ModelUtilities.GetInheritedDefaults(NonNullAst, current);
                // replace with defaults if defined

                bool tail = false;

                foreach (var p in ModelUtilities.GetConstructorArguments(NonNullAst, NonNullAst.Single(
                    e => e.Name.Equals(current.Parent))))
                {
                    if (tail)
                    {
                        Code.Write(", ");
                    }

                    var d = inheritedDefaults.SingleOrDefault(e => p.Name.Equals(e.Name));

                    if (d != null && d.Values != null)
                    {
                        var ds = d.Values;

                        if (ds.Count == 1)
                        {
                            Code.Write(p.Type + "." + ds.Single());
                        }
                        else
                        {
                            validators.Add(new KeyValuePair<AstDefault, AstProperty>(d, p));

                            Code.Write("Validate{0}({1})", p.Name, ModelUtilities.FirstLetterLowerCase(p.Name));
                        }
                    }
                    else
                    {
                        Code.Write(ModelUtilities.FirstLetterLowerCase(p.Name));
                    }

                    tail = true;
                }

                Code.WriteLine(")");
            }

            Code.WriteLine("\t{");

            var guards = current.Properties.Where(e => !e.SkipNullCheck);

            foreach (var p in guards)
            {
                if (StringComparer.Ordinal.Equals(TypeEnum, p.Type))
                {
                    Code.WriteLine("\t\tGuard.EnsureEnumDefined<{0}>(\"{1}\", (int) {1});", p.Type, ModelUtilities.FirstLetterLowerCase(p.Name));
                }
                else
                {
                    Code.WriteLine("\t\tGuard.EnsureNotNull(\"{0}\", {0});", ModelUtilities.FirstLetterLowerCase(p.Name));
                }
            }

            WriteLineIfAny(Code, guards);

            foreach (var property in current.Properties)
            {
                if (!string.IsNullOrEmpty(property.Initialization))
                {
                    Code.WriteLine("\t\tthis.{0} = {1};",
                        ModelUtilities.ToMember(property.Name),
                        property.Initialization.Replace("$1", ModelUtilities.FirstLetterLowerCase(property.Name)));
                }
                else
                {
                    Code.WriteLine("\t\tthis.{0} = {1};",
                        ModelUtilities.ToMember(property.Name),
                        ModelUtilities.FirstLetterLowerCase(property.Name));
                }
            }

            Code.WriteLine("\t}");
            Code.WriteLine();

            foreach (var kvp in validators)
            {
                var pcname = ModelUtilities.FirstLetterLowerCase(kvp.Value.Name);
                var tname = kvp.Value.Type;

                Code.WriteLine("\tpublic static {0} Validate{1}({2})", tname, kvp.Key.Name, ModelUtilities.ToParameter(kvp.Value));
                Code.WriteLine("\t{");
                Code.WriteLine("\t\tswitch({0})", pcname);
                Code.WriteLine("\t\t{");

                foreach (var v in kvp.Key.Values)
                {
                    Code.WriteLine("\t\tcase {0}.{1}:", tname, v);
                }

                Code.WriteLine("\t\t\treturn {0};", pcname);
                Code.WriteLine("\t\tdefault:");
                Code.WriteLine("\t\t\tthrow Guard.EnumOutOfRangeException<{0}>(\"{1}\", (int) {1});", tname, pcname);
                Code.WriteLine("\t\t}");
                Code.WriteLine("\t}");
                Code.WriteLine();
            }
        }

        public void WriteEqualityComparerClass(string value, Func<AstProperty, bool> where)
        {
            Code.WriteLine("public sealed class {0}EqualityComparer : ", value);
            Code.WriteLine("\t\t{0}", string.Join(",\n\t\t", NonNullAst.Select(e => string.Format("IEqualityComparer<{0}>", e.Name))));
            Code.WriteLine("{");
            Code.WriteLine("\tprivate static readonly IEqualityComparer<string> STRING_COMPARER = StringComparer.Ordinal;");
            Code.WriteLine("\tprivate const int HASH_PRIME = Utilities.HASH_PRIME;");
            Code.WriteLine("\tprivate const int HASH_SEED = Utilities.HASH_SEED;");
            Code.WriteLine("\tprivate static int GetHashCode(string value)");
            Code.WriteLine("\t{");
            Code.WriteLine("\t\treturn value != null ? value.GetHashCode() : 0;");
            Code.WriteLine("\t}");
            Code.WriteLine();

            bool tail = false;
            foreach (var element in NonNullAst)
            {
                if (tail)
                {
                    Code.WriteLine();
                }

                WriteGetHashCodeMethod(element, where);
                tail = true;
            }
            foreach (var element in NonNullAst)
            {
                if (tail)
                {
                    Code.WriteLine();
                }

                WriteEqualityMethod(element, where);
                tail = true;
            }
            Code.WriteLine("}");
        }

        private void GenerateProperty(AstProperty current)
        {
            if (!string.IsNullOrEmpty(current.Documentation))
            {
                var doc = current.Documentation.TrimEnd();
                Code.Write(Utilities.Indent("<summary> Gets or sets " + ModelUtilities.FirstLetterLowerCase(doc) + "</summary>", "\t/// "));
                Code.WriteLine("\t/// <value>{0}</value>", doc);
                Code.WriteLine("\t/// <returns>{0}</returns>", doc);
            }

            Code.WriteLine("\tpublic {0} {1} {{ get {{ return this.{2}; }} }}", current.Type, current.Name, ModelUtilities.ToMember(current.Name));
        }

        private void WriteTypeEnum(string enumPropertyName)
        {
            Code.WriteLine("\tpublic enum {0}", TypeEnum);
            Code.WriteLine("\t{");
            Code.WriteLine("\t\t{0}", string.Join(",\n\t\t", ModelUtilities.GetEnumValues(NonNullAst, enumPropertyName).OrderBy(e => e)));
            Code.WriteLine("\t}");
        }

        private void WriteGetHashCodeMethod(AstClass current, Func<AstProperty, bool> where)
        {
            bool isSealed = !NonNullAst.Any(e => current.Name.Equals(e.Parent));

            Code.WriteLine("\t/// <summary>Returns a hash code for the specified object.</summary>");
            Code.WriteLine("\t/// <param name=\"obj\">The <see cref=\"{0}\" /> for which a hash code is to be returned.</param>", current.Name);
            Code.WriteLine("\t/// <returns>A hash code for the specified object.</returns>");
            Code.WriteLine("\tpublic int GetHashCode({0} obj)", current.Name);
            Code.WriteLine("\t{");

            var inheritedProperties = ModelUtilities.GetInheritedProperties(NonNullAst, current).ToArray();

            var enumProperty = inheritedProperties.SingleOrDefault(e => e.Type.EndsWith(TypeEnum));

            // assert == isSealed == ep != null
            if (!isSealed && enumProperty != null)
            {
                MultiplexInheritance("obj", current, enumProperty, e => string.Format("return GetHashCode(({0}) obj);", e));
            }
            else
            {
                Code.WriteLine("\t\tunchecked");
                Code.WriteLine("\t\t{");
                Code.WriteLine("\t\t\tint result = HASH_SEED;");

                foreach (var property in inheritedProperties.Where(where).Reverse())
                {
                    Code.Write("\t\t\t");

                    var type = property.Type;
                    var isEnumerable = type.StartsWith("IList<") || type.StartsWith("ICollection<") || type.StartsWith("IEnumerable<");

                    if (isEnumerable)
                    {
                        Code.WriteLine("result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.{0}, GetHashCode);", property.Name);
                    }
                    else if (IsPrimitiveType(type) || StringComparer.Ordinal.Equals(TypeEnum, type))
                    {
                        Code.WriteLine("result = result * HASH_PRIME + obj.{0}.GetHashCode();", property.Name);
                    }
                    else
                    {
                        Code.WriteLine("result = result * HASH_PRIME + GetHashCode(obj.{0});", property.Name);
                    }
                }

                Code.WriteLine("\t\t\treturn result;");
                Code.WriteLine("\t\t}");
            }

            Code.WriteLine("\t}");
        }

        private void WriteEqualityMethod(AstClass current, Func<AstProperty, bool> where)
        {
            bool isSealed = !NonNullAst.Any(e => current.Name.Equals(e.Parent));

            Code.WriteLine("\t/// <summary>Determines whether the specified objects are equal.</summary>");
            Code.WriteLine("\t/// <param name=\"x\">The first object of type <see cref=\"{0}\" /> to compare.</param>", current.Name);
            Code.WriteLine("\t/// <param name=\"y\">The second object of type <see cref=\"{0}\" /> to compare.</param>", current.Name);
            Code.WriteLine("\t/// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>");
            Code.WriteLine("\tpublic bool Equals({0} x, {0} y)", current.Name);
            Code.WriteLine("\t{");
            Code.WriteLine("\t\tif(object.ReferenceEquals(x, y)) { return true; }");
            Code.WriteLine("\t\tif(object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }");

            var inheritedProperties = ModelUtilities.GetInheritedProperties(NonNullAst, current).ToArray();

            var enumProperty = inheritedProperties.SingleOrDefault(e => e.Type.EndsWith(TypeEnum));

            if (enumProperty != null)
            {
                Code.WriteLine("\t\tif(x.{0} != y.{0}) {{ return false; }}", enumProperty.Name);
            }

            Code.WriteLine();

            // assert == isSealed == ep != null
            if (!isSealed && enumProperty != null)
            {
                MultiplexInheritance("x", current, enumProperty, e => string.Format("return Equals(({0}) x, ({0}) y);", e));
            }
            else
            {
                Code.WriteLine("\t\treturn");

                bool tail = false;

                foreach (var property in inheritedProperties.Reverse().Where(where).Where(e => enumProperty == null || !enumProperty.Name.Equals(e.Name)))
                {
                    if (tail)
                    {
                        Code.WriteLine(" &&");
                    }

                    Code.Write("\t\t\t");

                    var type = property.Type;
                    var isEnumerable = type.StartsWith("IList<") || type.StartsWith("ICollection<") || type.StartsWith("IEnumerable<");

                    if (isEnumerable)
                    {
                        Code.Write("Utilities.EnumerableDeepEquals(x.{0}, y.{0}, Equals)", property.Name);
                    }
                    else if (IsPrimitiveType(type))
                    {
                        Code.Write("x.{0} == y.{0}", property.Name);
                    }
                    else if (StringComparer.OrdinalIgnoreCase.Equals("string", type))
                    {
                        Code.Write("STRING_COMPARER.Equals(x.{0}, y.{0})", property.Name);
                    }
                    else
                    {
                        Code.Write("Equals(x.{0}, y.{0})", property.Name);
                    }

                    tail = true;
                }
                Code.WriteLine(";");
            }

            Code.WriteLine("\t}");
        }

        private static bool IsPrimitiveType(string type)
        {
            switch (type)
            {
                case "bool":
                case "Boolean":
                case "int":
                case "Int32":
                case "Int16":
                case "Int64":
                case "Span":
                    return true;

                default:
                    return false;
            }
        }

        private void MultiplexInheritance(string varName, AstClass current, AstProperty ep, Func<string, string> function)
        {
            Code.WriteLine("\t\tswitch({0}.{1})", varName, ep.Name);
            Code.WriteLine("\t\t{");

            var rootClass = ModelUtilities.ClassHierarchyUp(NonNullAst, current).Last();

            var fullEnumTypeName = rootClass.Name + "." + ep.Type;

            var values = ModelUtilities.GetEnumValuesByName(ModelUtilities.SubClasses(NonNullAst, current), ep.Name);

            if (!values.Any())
            {
                throw new ArgumentException();
            }

            foreach (var kvp in values.GroupBy(e => e.Key.Name))
            {
                foreach (var value in kvp.Select(e => e.Value).SelectMany(e => e.Values))
                {
                    Code.WriteLine("\t\tcase {0}.{1}:", fullEnumTypeName, value);
                }

                Code.WriteLine("\t\t\t" + function(kvp.Key));
            }

            Code.WriteLine("\t\tdefault:");
            Code.WriteLine("\t\t\tthrow Guard.EnumOutOfRangeException<{0}>(\"{1}\", (int) {1});", fullEnumTypeName, varName + "." + ep.Name);
            Code.WriteLine("\t\t}");
        }

        #region Visitor

        public void WriteTransformer(string className)
        {
            Code.WriteLine("public class {0}", className);
            Code.WriteLine("{");

            // find all classes used in the ast
            //
            var targetsString = new HashSet<string>(
                NonNullAst.SelectMany(e => e.Properties).SelectMany(
                    e => ModelUtilities.ExtractSimpleTypes(e.Type)).Intersect(
                    NonNullAst.Select(e => e.Name)));

            var publicAccess = NonNullAst.Where(e => e.Parent == null).Concat(
                NonNullAst.Where(e => targetsString.Contains(e.Name))).Distinct().ToList();

            Func<IEnumerable<AstClass>, AstClass, string> transformerReturnType =
                (source, current) => ModelUtilities.ClassHierarchyUp(source, current).Intersect(publicAccess).First().Name;

            bool tail = false;

            // multiplexer
            foreach (var current in NonNullAst) // where class is root
            {
                if (tail)
                {
                    Code.WriteLine();
                }

                bool currentPublicAccess = publicAccess.Contains(current);

                Code.WriteLine("\tpublic virtual {0} Transform({1} node)",
                   transformerReturnType(NonNullAst, current), current.Name);
                Code.WriteLine("\t{");

                if (currentPublicAccess)
                {
                    Code.WriteLine("\t\tGuard.EnsureNotNull(\"node\", node);");
                    Code.WriteLine();
                }

                if (ModelUtilities.SubClasses(NonNullAst, current).Count() > 1)
                {
                    var enumProperty = ModelUtilities.GetInheritedProperties(NonNullAst, current).SingleOrDefault(e => e.Type.EndsWith(TypeEnum));

                    MultiplexInheritance("node", current, enumProperty, e => string.Format("return Transform(({0}) node);", e));
                }
                else
                {
                    Code.Write("\t\treturn new {0}(", current.Name);

                    bool tail2 = false;

                    foreach (var arg in ModelUtilities.GetConstructorArguments(NonNullAst, current))
                    {
                        if (tail2)
                        {
                            Code.Write(',');
                        }
                        Code.WriteLine();

                        Code.Write("\t\t\t");

                        if (!NonNullAst.Any(e => arg.Type.Contains(e.Name)))
                        {
                            Code.Write("node.{0}", arg.Name);
                        }
                        else if (arg.Type.StartsWith("IList") || arg.Type.StartsWith("ICollection") || arg.Type.StartsWith("IEnumerable"))
                        {
                            Code.Write("Utilities.Transform(node.{0}, Transform)", arg.Name);
                        }
                        else
                        {
                            Code.Write("Transform(node.{0})", arg.Name);
                        }

                        tail2 = true;
                    }

                    Code.WriteLine(");");
                }

                // if class is abstract
                tail = true;
                Code.WriteLine("\t}");
            }

            Code.WriteLine("}");
        }

        #endregion Visitor
    }
}
