﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;

namespace libdyn.generator.Ast
{
    public static class DynAst
    {
        public static bool PrintParsedInput
        {
            get; set;
        }

        private static TextWriter _debug = TextWriter.Null, _code = TextWriter.Null;

        public static TextWriter Debug
        {
            get
            {
                return _debug;
            }
            set
            {
                Guard.EnsureNotNull("value", value);
                _debug = value;
            }
        }

        public static TextWriter Code
        {
            get
            {
                return _code;
            }
            set
            {
                Guard.EnsureNotNull("value", value);
                _code = value;
            }
        }

        // Quantifier . Variable == LogicalVariable
        public const string ASSERTION = @"
- Expression abstract
    + Location : Span ? ;
    + Kind : NodeKind ?

- Quantifier : Expression
    ^ Kind = Universal , Existential ;
    + Variable : string ;
    + Type : string ;
    + Expression : Expression

- TypeCheck : Expression
    ^ Kind = TypeCheck ;
    + Expression : Expression ;
    + TypeSet : ICollection<string> ? @ Utilities.AsImmutableList($1)

- Predicate : Expression
    ^ Kind = Predicate ;
    + Name : string ;
    + Arguments : IList<Expression> ? @ Utilities.AsImmutableList($1)

- Unary : Expression
    ^ Kind = Negate ;
    + Expression : Expression

- Binary : Expression
    ^ Kind = And , Or , Implies ,
            Addition , Subtraction , Multiplication , Division , Modulus ,
            GreaterThan , GreaterThanOrEqual , LessThan , LessThanOrEqual , Equality ;
    + First : Expression ;
    + Second : Expression

- IfThenElse : Expression
    ^ Kind = IfThenElse ;
    + Condition : Expression ;
    + TrueBranch : Expression ;
    + FalseBranch : Expression

- Numeric : Expression
    ^ Kind = Numeric ;
    + Value : int ?

- String : Expression
    ^ Kind = String ;
    + Value : string

- InstanceVariable : Expression
    ^ Kind = InstanceVariable ;
    + Expression : Expression ;
    + Name : string

- Variable : Expression
    ^ Kind = AssertionVariable , ProgramVariable ;
    + Name : string
";

        public const string PROGRAM = @"
- Class
    + Location : Span ? ;
    + Name : string ;
    + Parent : string ;
    + Renames : IList<RenameDeclaration> ? @ Utilities.AsImmutableList($1) ;
    + Methods : IList<MethodDeclaration> ? @ Utilities.AsImmutableList($1) ;
    + Contracts : IList<ContractDeclaration> ? @ Utilities.AsImmutableList($1)

- RenameDeclaration
    + Location : Span ? ;
    + From : string ;
    + To : string

- MethodDeclaration
    + Location : Span ? ;
    + Name : string ;
    + Parameters : IList<string> ? @ Utilities.AsImmutableList($1) ;
    + Statements : IList<Statement> ? @ Utilities.AsImmutableList($1) ;
    + Contracts : IList<ContractDeclaration> @ Utilities.AsImmutableList($1)

- Statement abstract
    + Location : Span ? ;
    + Kind : NodeKind

- ContractDeclaration : Statement
    ^ Kind = Ensures , Requires , Invariant , Assert , ByTheWay ;
    + Expression : DynAssertionExpressionNode

- Expression : Statement abstract

- MethodCall : Expression
    ^ Kind = MethodCall ;
    + Expression : Expression ;
    + Name : string ;
    + Arguments : IList<Expression> ? @ Utilities.AsImmutableList($1)

- ConstructorCall : Expression
    ^ Kind = ConstructorCall ;
    + Name : string ;
    + Arguments : IList<Expression> ? @ Utilities.AsImmutableList($1)

- Variable : Expression
    ^ Kind = Variable , InstanceVariable ;
    + Name : string

- If : Expression
    ^ Kind = If ;
    + Condition : Expression ;
    + TrueBranch : IList<Statement> ? @ Utilities.AsImmutableList($1) ;
    + FalseBranch : IList<Statement> ? @ Utilities.AsImmutableList($1)

- While : Expression
    ^ Kind = While ;
    + Condition : Expression ;
    + Invariant : DynAssertionExpressionNode ;
    + Body : IList<Statement> ? @ Utilities.AsImmutableList($1)

- Assignment : Expression
    ^ Kind = Assignment ;
    + Variable : Variable ;
    + Expression : Expression
";

        public const string HOARE = @"
/// A hoare triple containing a precondition, zero or more statements and a postcondition.
- DynHoareNode
    /// The source code position of this node.
    + Location : Span ? ;
    /// The precondition.
    + PreCondition : DynAssertionExpressionNode ? @ $1 ?? AstExtensions.Assertions.True ;
    /// The statements.
    + Statements : IList<DynProgramStatementNode> ? @ Utilities.AsImmutableList($1) ;
    /// The postcondition.
    + PostCondition : DynAssertionExpressionNode ? @ $1 ?? AstExtensions.Assertions.True
";

        private static void WriteHeader(string header)
        {
            // Console.ReadLine();
            // Console.Clear();
            Console.WriteLine("//");
            Console.WriteLine("// " + header);
            Console.WriteLine("//");
            Console.WriteLine("//");
        }

        public static void MainAst()
        {
            //TODO: Ensure no inheritance loops
            //TODO: Ensure all parents are in source
            //TODO: Ensure default types are defined upwards

            var cg = new CodeGenerator(Debug, Code);
            WriteHeader("Assertions");
            var assAst = Parser.Parse(ASSERTION);
            RenameType(assAst, e => "DynAssertion" + e + "Node");
            cg.TypeEnum = "NodeKind";
            cg.Ast = assAst;
            cg.WriteAstClasses();
            cg.WriteTransformer("DynAssertionNodeTransformer");

            WriteHeader("Program");
            var prgAst = Parser.Parse(PROGRAM);
            RenameType(prgAst, e => "DynProgram" + e + "Node");
            cg.TypeEnum = "NodeKind";
            cg.Ast = prgAst;
            cg.WriteAstClasses();
            cg.WriteTransformer("DynProgramNodeTransformer");

            WriteHeader("Hoare");
            var hoareAst = Parser.Parse(HOARE);
            cg.TypeEnum = null;
            cg.Ast = hoareAst;
            cg.WriteAstClasses();

            // visit

            // comparer
            RenameType(assAst, e => StringComparer.Ordinal.Equals(e, "NodeKind") ? "DynAssertionNodeKind.NodeKind" : e);
            RenameType(prgAst, e => StringComparer.Ordinal.Equals(e, "NodeKind") ? "DynProgramNodeKind.NodeKind" : e);

            var allClasses = new List<AstClass>(assAst.Concat(prgAst).Concat(hoareAst));

            WriteHeader("EqualityComparer with Location");

            cg.TypeEnum = "NodeKind";
            cg.Ast = allClasses;
            cg.WriteEqualityComparerClass("DynNode", e => true);

            WriteHeader("EqualityComparer without Location");
            {
                cg.TypeEnum = "NodeKind";
                cg.Ast = allClasses;

                cg.WriteEqualityComparerClass("DynNodeIgnoreLocation", e => !StringComparer.Ordinal.Equals("Span", e.Type));
            }

            Console.ReadLine();
        }

        private static Regex TYPE_MATCH_PATTERN_A = new Regex(@"^(\w+)$");
        private static Regex TYPE_MATCH_PATTERN_B = new Regex(@"[,<](\w+)[,>]");

        private static void RenameType(IEnumerable<AstClass> source, Func<string, string> func)
        {
            ISet<string> sourceNames = new HashSet<string>(source.Select(e => e.Name));

            foreach (var element in source)
            {
                element.Name = func(element.Name);
                element.Parent = element.Parent == null ? null : func(element.Parent);

                foreach (var property in element.Properties)
                {
                    property.Type = TYPE_MATCH_PATTERN_A.Replace(property.Type, m =>
                    {
                        var value = m.Value;
                        return sourceNames.Contains(value) ? func(m.Value) : value;
                    });

                    property.Type = TYPE_MATCH_PATTERN_B.Replace(property.Type, m =>
                    {
                        var cap = m.Groups[1].Value;
                        var value = m.Value;

                        if (sourceNames.Contains(cap))
                        {
                            return value.Replace(cap, func(cap));
                        }

                        return value;
                    });
                }
            }
        }
    }
}
