﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace libdyn.generator.Ast
{
    public sealed class AstClass
    {
        public string Documentation { get; set; }
        public string Name { get; set; }
        public string Parent { get; set; }
        public bool IsAbstract { get; set; }
        public ICollection<AstDefault> Defaults { get; set; }
        public ICollection<AstProperty> Properties { get; set; }

        public override string ToString()
        {
            var builder = new StringBuilder();

            if (Documentation != null)
            {
                Utilities.Indent(Documentation, "/// ", builder);
            }

            builder.Append("- ").Append(Name);

            if (Parent != null)
            {
                builder.Append(" : ").Append(Parent);
            }

            if (IsAbstract)
            {
                builder.Append(" abstract");
            }

            builder.AppendLine();

            bool any = false;

            foreach (var element in Defaults)
            {
                builder.Append(element).AppendLine(" ;");
                any = true;
            }
            foreach (var element in Properties)
            {
                builder.Append(element).AppendLine(" ;");
                any = true;
            }

            if (any)
            {
                builder.Length -= 1 /* space */  /* semicolon */ + System.Environment.NewLine.Length;
            }

            return builder.ToString();
        }
    }

    public sealed class AstProperty
    {
        public string Documentation { get; set; }
        public string Name { get; set; }
        public bool SkipNullCheck { get; set; }
        public string Type { get; set; }
        public string Initialization { get; set; }

        public override string ToString()
        {
            var builder = new StringBuilder();

            if (Documentation != null)
            {
                Utilities.Indent(Documentation, "\t/// ", builder);
            }

            builder.Append("\t+ ").Append(Name).Append(" : ").Append(Type);

            if (SkipNullCheck)
            {
                builder.Append(" ?");
            }

            if (!string.IsNullOrEmpty(Initialization))
            {
                builder.Append(" @ ").Append(Initialization);
            }

            return builder.ToString();
        }
    }

    public sealed class AstDefault
    {
        public string Name { get; set; }
        public ICollection<string> Values { get; set; }

        public override string ToString()
        {
            var builder = new StringBuilder();

            builder.Append("\t^ ").Append(Name).Append(" = ").Append(string.Join(" , ", Values));

            return builder.ToString();
        }
    }

    public static class ModelUtilities
    {
        public static string ToMember(string value)
        {
            return "_" + FirstLetterLowerCase(value);
        }

        public static string FirstLetterLowerCase(string value)
        {
            return char.ToLowerInvariant(value[0]) + value.Substring(1);
        }

        public static string ToParameter(AstProperty property)
        {
            return property.Type + " " + FirstLetterLowerCase(property.Name);
        }

        public static IEnumerable<AstClass> ClassHierarchyUp(IEnumerable<AstClass> source, AstClass current)
        {
            Guard.EnsureElementNotNull("source", source);

            while (current != null)
            {
                yield return current;

                current = Parent(source, current);
            }
        }

        public static AstClass Parent(IEnumerable<AstClass> source, AstClass current)
        {
            return source.Where(e => e.Name.Equals(current.Parent)).SingleOrDefault();
        }

        public static IEnumerable<AstProperty> GetInheritedProperties(IEnumerable<AstClass> source, AstClass current)
        {
            return ClassHierarchyUp(source, current).SelectMany(e => e.Properties);
        }

        public static IEnumerable<AstDefault> GetInheritedDefaults(IEnumerable<AstClass> source, AstClass current)
        {
            return ClassHierarchyUp(source, current).SelectMany(e => e.Defaults);
        }

        public static IEnumerable<string> GetEnumValues(IEnumerable<AstClass> source, string enumName)
        {
            Guard.EnsureElementNotNull("source", source);

            return source.SelectMany(e => e.Defaults).Where(e => enumName.Equals(e.Name)).SelectMany(e => e.Values).Distinct();
        }

        public static IEnumerable<KeyValuePair<AstClass, AstDefault>> GetEnumValuesByName(IEnumerable<AstClass> source, string enumName)
        {
            Guard.EnsureElementNotNull("source", source);

            return source.SelectMany(e => e.Defaults, (e, f) => new KeyValuePair<AstClass, AstDefault>(e, f)).Where(e => enumName.Equals(e.Value.Name));
        }

        public static IEnumerable<AstClass> SubClasses(IEnumerable<AstClass> source, AstClass current)
        {
            Guard.EnsureElementNotNull("source", source);

            var result = new List<AstClass>();

            var args = source.Where(e => current.Name.Equals(e.Parent));

            result.Add(current);
            result.AddRange(args);
            result.AddRange(args.SelectMany(e => SubClasses(source, e)));

            return result.Distinct().ToList();
        }

        public static IList<AstProperty> GetConstructorArguments(IEnumerable<AstClass> source, AstClass current)
        {
            Guard.EnsureElementNotNull("source", source);

            var cup = ClassHierarchyUp(source, current).ToList();

            var result = new List<AstProperty>();

            foreach (var element in cup.Reverse<AstClass>())
            {
                result.AddRange(element.Properties);

                result.RemoveAll(e => element.Defaults.Any(f => e.Name.Equals(f.Name) && f.Values.Count == 1));
            }

            return result;
        }

        public static IEnumerable<string> ExtractSimpleTypes(string typeName)
        {
            return typeName.Split(",<>[] \t\n\r".ToCharArray(), StringSplitOptions.RemoveEmptyEntries);
        }
    }
}
