﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

namespace libdyn.generator.Ast
{
    public static class Parser
    {
        private static IEqualityComparer<string> STR_CMP = StringComparer.Ordinal;

        private static bool Current(IEnumerator<string> source, out string value)
        {
            value = source.Current;

            if (value == null)
            {
                return false;
            }

            source.MoveNext();
            return true;
        }

        private static bool Accept(IEnumerator<string> source, string value, bool expect)
        {
            var current = source.Current;

            if (value == null || !STR_CMP.Equals(current, value))
            {
                if (expect)
                {
                    throw new ArgumentException();
                }

                return false;
            }

            source.MoveNext();
            return true;
        }

        private static AstProperty ParseProperty(IEnumerator<string> tokenizer)
        {
            string name, type, initializer = null, doc = null;
            bool skipNullCheck;

            if (Accept(tokenizer, "///", false) && !Current(tokenizer, out doc))
            {
                return null;
            }

            if (!Accept(tokenizer, "+", false))
            {
                return null;
            }

            Current(tokenizer, out name);
            Accept(tokenizer, ":", true);
            Current(tokenizer, out type);

            skipNullCheck = Accept(tokenizer, "?", false);

            if (Accept(tokenizer, "@", false))
            {
                if (!Current(tokenizer, out initializer))
                {
                    return null;
                }

                initializer = initializer.Trim();
            }

            return new AstProperty() { Name = name, SkipNullCheck = skipNullCheck, Type = type, Initialization = initializer, Documentation = doc };
        }

        private static string ParseIdentity(IEnumerator<string> source)
        {
            var result = source.Current;
            if (result == null)
            {
                throw new ArgumentException();
            }
            source.MoveNext();
            return result;
        }

        private static AstDefault ParseDefault(IEnumerator<string> source)
        {
            if (!Accept(source, "^", false))
            {
                return null;
            }

            string name;

            Current(source, out name);

            Accept(source, "=", true);

            var list = ParseList(source, ",", ParseIdentity);

            return new AstDefault() { Name = name, Values = list };
        }

        private static IList<T> ParseList<T>(
            IEnumerator<string> source,
            string delimeter,
            Func<IEnumerator<string>, T> parser)
        {
            List<T> result = new List<T>();

            do
            {
                T current = parser(source);

                if (current == null)
                {
                    break;
                }

                result.Add(current);
            } while (delimeter == null || Accept(source, delimeter, false));

            return result;
        }

        private static AstClass ParseClass(IEnumerator<string> tokenizer)
        {
            string name, parent = null, doc = null;
            bool isAbstract = false;

            if (Accept(tokenizer, "///", false) && !Current(tokenizer, out doc))
            {
                return null;
            }

            if (!Accept(tokenizer, "-", false))
            {
                return null;
            }

            Current(tokenizer, out name);

            if (Accept(tokenizer, ":", false) && !Current(tokenizer, out parent))
            {
                return null;
            }

            isAbstract = Accept(tokenizer, "abstract", false);

            var defaults = ParseList(tokenizer, null, ParseDefault);

            if (defaults.Count > 0)
            {
                Accept(tokenizer, ";", true);
            }

            var properties = ParseList(tokenizer, ";", ParseProperty);

            return new AstClass()
            {
                Name = name,
                Parent = parent,
                IsAbstract = isAbstract,
                Properties = properties,
                Defaults = defaults,
                Documentation = doc
            };
        }

        private static string[] EOS = new string[] { null };
        private static Regex LONG_PATTERN = new Regex(@"(@\s+[^;\n\r]+|///\s+[^\n\r]+)");

        private static IEnumerable<IEnumerable<string>> ParseComplex(string source)
        {
            int pos = 0;

            for (var match = LONG_PATTERN.Match(source); match.Success; match = match.NextMatch())
            {
                if (match.Index > pos)
                {
                    yield return source.Substring(pos, match.Index - pos).Split(" \r\t\n".ToCharArray(), StringSplitOptions.RemoveEmptyEntries);
                }

                yield return match.Value.Split(" \r\t\n".ToCharArray(), 2, StringSplitOptions.RemoveEmptyEntries);

                pos = match.Index + match.Length;
            }

            if (pos < source.Length)
            {
                yield return source.Substring(pos).Split(" \r\t\n".ToCharArray(), StringSplitOptions.RemoveEmptyEntries);
            }
        }

        private static IEnumerable<string> MergeDocumentation(this IEnumerable<string> source)
        {
            string last = string.Empty;

            var enumerator = source.GetEnumerator();

            while (enumerator.MoveNext())
            {
                var element = enumerator.Current;

                if ("///".Equals(element) && enumerator.MoveNext())
                {
                    last += enumerator.Current + Environment.NewLine;
                    continue;
                }
                else if (!string.IsNullOrWhiteSpace(last))
                {
                    yield return "///";
                    yield return last;
                    last = string.Empty;
                }

                yield return element;
            }

            if (!string.IsNullOrWhiteSpace(last))
            {
                throw new ArgumentException();
            }
        }

        public static IList<AstClass> Parse(string source)
        {
            var tokenEnumerator = ParseComplex(source).SelectMany(e => e).Concat(EOS).MergeDocumentation().GetEnumerator();

            if (!tokenEnumerator.MoveNext())
            {
                throw new ArgumentException();
            }

            return new List<AstClass>(ParseList(tokenEnumerator, null, ParseClass));
        }
    }
}
