﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Text;

namespace libdyn.generator.Lexer
{
    using state = System.Int32;

    [DebuggerDisplay("{ToString(),nq}")]
    public sealed class Dfa
    {
        private readonly state i;
        private readonly state[] a;
        private readonly state[] s;
        private readonly Transition[] t;

        public Dfa(state initial, IEnumerable<state> finals, IEnumerable<Transition> transitions)
        {
            Guard.EnsureNotNull("finals", finals);
            Guard.EnsureNotNull("transitions", transitions);

            var t_array = transitions.ToArray();

            ValidateTransitions(t_array);

            this.i = initial;
            this.a = finals.ToArray();
            this.s = Utilities.ToEnumerable(this.i).Concat(a).Concat(transitions.Select(t => t.Current)).Concat(transitions.Select(t => t.Next)).Distinct().ToArray();
            this.t = t_array;
        }

        private static void ValidateTransitions(Transition[] transitions)
        {
            foreach (var fails in transitions.GroupBy(e => e.Current + "_" + e.Next, e => e.Input).Where(e => e.Distinct().Count() != e.Count()))
            {
                throw new Exception("Oops");
            }
        }

        public state Initial
        {
            get
            {
                return i;
            }
        }

        public ICollection<state> Accepting
        {
            get
            {
                return new ReadOnlyCollection<state>(a);
            }
        }

        public IEnumerable<Transition> Transitions
        {
            get
            {
                return new ReadOnlyCollection<Transition>(t);
            }
        }

        public ICollection<state> States
        {
            get
            {
                return new ReadOnlyCollection<state>(s);
            }
        }

        public static Dfa operator !(Dfa value)
        {
            return new Dfa(value.Initial, value.States.Except(value.Accepting), value.Transitions);
        }

        public Nfa ToNondeterministicFiniteAutomata()
        {
            var max = States.Max() + 1;

            int INITIAL = max++;
            int FINAL = max++;

            return new Nfa(
                INITIAL, FINAL,
               Utilities.ToEnumerable(new Transition(INITIAL, Nfa.EPSILON, this.Initial))
               .Concat(this.Accepting.Select(s => new Transition(s, Nfa.EPSILON, FINAL)))
               .Concat(this.Transitions));
        }

        private static ISet<state> Reachable(IEnumerable<Transition> transitions, state start)
        {
            ISet<state> visited = new HashSet<state>();
            var pending = new Stack<state>();

            pending.Push(start);
            visited.Add(start);

            while (pending.Count > 0)
            {
                var current = pending.Pop();

                foreach (var t in transitions.Where(e => e.Current == current))
                {
                    if (!visited.Contains(t.Next))
                    {
                        pending.Push(t.Next);
                        visited.Add(t.Next);
                    }
                }
            }

            return visited;
        }

        // at least working
        public Dfa Minimize()
        {
            Func<Transition, Transition> reverseTransition = e => new Transition(e.Next, e.Input, e.Current);

            Func<Dfa, Nfa> reverseDfa = d =>
            {
                int final = d.States.Max() + 1;

                return new Nfa(
                    final, d.Initial,
                    d.Transitions.Select(reverseTransition).Concat(d.Accepting.Select(e => reverseTransition(new Transition(e, Nfa.EPSILON, final)))));
            };

            Func<Nfa, Dfa> nfa2dfa = n => n.ToDeterministic();

            return nfa2dfa(reverseDfa(nfa2dfa(reverseDfa(this))));
        }

        public override string ToString()
        {
            var builder = new StringBuilder();

            builder.AppendFormat("{0,8}   {1,8} -> {2,8}\n", String.Empty, String.Empty, this.Initial);

            foreach (var t in Transitions)
            {
                builder.AppendLine(t.ToString());
            }

            foreach (var f in Accepting)
            {
                builder.AppendFormat("{0,8}   {1,8} -> {2,8}\n", f, String.Empty, String.Empty);
            }

            return builder.ToString();
        }

        private static readonly Dfa EMPTY = new Dfa(0, Enumerable.Empty<state>(), Transition.Empty);

        public static Dfa Empty
        {
            get
            {
                return EMPTY;
            }
        }
    }
}
