﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;

namespace libdyn.generator.Lexer
{
    using input = System.Char;
    using state = System.Int32;

    public static class DynLexerGenerator
    {
        public static Nfa Single(char c)
        {
            return c;
        }

        public static Nfa Range(char from, char to)
        {
            const state INTITIAL = 0;
            const state FINAL = 1;

            var transitions = new List<Transition>();

            int nextFreeState = FINAL + 1;

            for (var c = from; c <= to; ++c)
            {
                var ei = nextFreeState++;
                transitions.Add(new Transition(INTITIAL, Nfa.EPSILON, ei));
                transitions.Add(new Transition(ei, c, FINAL));
            }

            return new Nfa(INTITIAL, FINAL, transitions);
        }

        public static Nfa Not(char c)
        {
            return Single((char)(c | Transition.NOT_FLAG));
        }

        private static Nfa Only(params String[] source)
        {
            const state INTITIAL = 0;
            const state FINAL = 1;

            var transitions = new List<Transition>();

            int nextFreeState = FINAL + 1;

            foreach (var e in source)
            {
                var ei = nextFreeState++;
                transitions.Add(new Transition(INTITIAL, Nfa.EPSILON, ei));

                var last = ei;

                foreach (var c in e)
                {
                    var ci = nextFreeState++;
                    transitions.Add(new Transition(last, (input)c, ci));

                    last = ci;
                }

                transitions.Add(new Transition(last, Nfa.EPSILON, FINAL));
            }

            return new Nfa(INTITIAL, FINAL, transitions);
        }

        [DebuggerDisplay("{ToString(),nq}")]
        private struct MicroLexer
        {
            public Dfa Automata;
            public String Name;

            public override string ToString()
            {
                var builder = new StringBuilder();

                builder.Append(Name).Append(':');

                builder.AppendLine().AppendLine(Automata.ToString());

                return builder.ToString();
            }
        }

        public static ILookup<int, TElement> EquivalenceClasses<TElement>(IEnumerable<IEnumerable<TElement>> source, out int counter)
        {
            Guard.EnsureNotNull("source", source);

            var sourceList = source.ToList();
            var input = new HashSet<TElement>(source.SelectMany(e => e).Distinct());

            var result = new Dictionary<TElement, int>();

            var equivalence = new HashSet<TElement>();

            counter = 0;

            while (input.Count > 0)
            {
                var current = input.First();
                input.Remove(current);

                equivalence.Clear();

                bool? state = null;

                foreach (var set in sourceList)
                {
                    bool inSet = set.Contains(current);

                    if (state == null)
                    {
                        state = inSet;
                        equivalence.UnionWith(input);
                    }

                    if (inSet)
                    {
                        equivalence.IntersectWith(set);
                    }
                    else
                    {
                        equivalence.ExceptWith(set);
                    }

                    if (equivalence.Count == 0) // only self
                    {
                        break;
                    }
                }

                result.Add(current, counter);

                foreach (var e in equivalence)
                {
                    input.Remove(e);
                    result.Add(e, counter);
                }

                ++counter;
            }

            return result.ToLookup(e => e.Value, e => e.Key);
        }

        private const String INVALID_LEXER_NAME = "Invalid";
        private const String INITIAL_LEXER_NAME = "Initial";

        private const byte UNDEFINED = byte.MaxValue;

        private const byte IEQC_INVALID = 1;
        private const byte IEQC_EOS = 0;
        private const byte IEQC_SHIFT = 0 + 1 + 1; // + Invalid + End of Stream

        private const byte STATE_INITIAL = 0;
        private const byte STATE_INVALID = 1;

        private const byte TRANSITION_TOKEN_COMPLETE_FLAG = 0x80;

        private static void GenerateLexer(IEnumerable<MicroLexer> source, TextWriter debug, TextWriter code)
        {
            Guard.EnsureNotNull("source", source);
            Guard.EnsureNotNull("debug", debug);
            Guard.EnsureNotNull("code", code);

            IList<MicroLexer> compressed;

            int ieqc_count, states_count;

            {
                /* generate list of all micro lexer including special invalid and initital */
                var sources = new List<MicroLexer>();

                sources.Add(new MicroLexer()
                {
                    Automata = Dfa.Empty,
                    Name = INITIAL_LEXER_NAME,
                });
                sources.Add(new MicroLexer()
                {
                    Automata = Dfa.Empty,
                    Name = INVALID_LEXER_NAME,
                });
                sources.AddRange(source);

                debug.WriteLine("List of micro lexer:");
                foreach (var e in sources)
                {
                    debug.WriteLine(e.Name + ":");
                    debug.WriteLine(DfaExtensions.ToStringCombined(e.Automata, DfaExtensions.CHARACTER_COMBINER));
                }
                debug.WriteLine();

                var state_input_mapping = sources.SelectMany(
                    e => e.Automata.Transitions.Where(t => (t.Input & Transition.NOT_FLAG) != Transition.NOT_FLAG).Select(
                        t => Tuple.Create(e.Name + "_" + t.Current + "_" + t.Next, t.Input)))
                        .ToLookup(e => e.Item1, e => e.Item2);

                debug.WriteLine("Input by classes:");
                foreach (var grouping in state_input_mapping)
                {
                    debug.WriteLine("{0,32}  :  {1}", grouping.Key, DfaExtensions.CHARACTER_COMBINER(grouping));
                }

                var input_equiv_mapping = AddDefaultEquivalenceClasses(EquivalenceClasses<char>(state_input_mapping, out ieqc_count), ref ieqc_count);

                debug.WriteLine("Equivalence Classes:");
                foreach (var grouping in input_equiv_mapping)
                {
                    debug.WriteLine("{0,32}  :  {1}", grouping.Key, DfaExtensions.CHARACTER_COMBINER(grouping));
                }
                debug.WriteLine();

                WriteInputEquivGroups(input_equiv_mapping, debug, code, ieqc_count);

                compressed = CompressLexers(sources, input_equiv_mapping, out states_count);

                debug.WriteLine("Compressed:");
                foreach (var e in compressed)
                {
                    debug.WriteLine(e.Name + ":");
                    debug.WriteLine(DfaExtensions.ToStringCombined(e.Automata, DfaExtensions.INT_COMBINER));
                }
                debug.WriteLine();
            }
            byte[] stt;

            BuildStateTransitionTable(compressed, debug, code, ieqc_count, states_count, out stt);

            WriteStateLexerTokenKindTable(compressed, debug, code);

            DebugValidateStateTransitionTable(compressed, stt);
        }

        private static ILookup<state, input> AddDefaultEquivalenceClasses(ILookup<state, input> source, ref int ieqc_count)
        {
            var invalids = new List<input>();

            // do not need to test 0 because line above
            for (int i = 1; i < 128; ++i)
            {
                if (!source.Any(e => e.Contains((char)i)))
                {
                    invalids.Add((char)i);
                }
            }

            ieqc_count += IEQC_SHIFT;

            // because IEnumerable<IGrouping<TKey, TElement>> => ILookup<TKey, TElement> is not easy
            return Utilities.ToEnumerable('\0').Select(e => Tuple.Create((input)e, (int)IEQC_EOS))
                .Concat(invalids.Select(e => Tuple.Create((input)e, (int)IEQC_INVALID)))
                .Concat(source.SelectMany(e => e, (g, e) => Tuple.Create(e, g.Key + IEQC_SHIFT)))
                .ToLookup(e => e.Item2, e => e.Item1);
        }

        private static void DebugValidateStateTransitionTable(IEnumerable<MicroLexer> source, byte[] table)
        {
        }

        private static void WriteParseMethods(MicroLexer lexer, TextWriter code, StringBuilder builder)
        {
            code.WriteLine("   token = Parse{0}(span, builder); return true;", lexer.Name);
            builder.AppendFormat("private static LexerToken token Parse{0}(Span span, StringBuilder builder)\n", lexer.Name);
            builder.AppendLine("{");
            builder.AppendLine("  throw new NotImplementedException();");
            builder.AppendLine("}");
            builder.AppendLine();
        }

        private static void WriteStateLexerTokenKindTable(IEnumerable<MicroLexer> source, TextWriter debug, TextWriter code)
        {
            const string CASE_TEMPLATE = "case 0x{0:x02}:\n";

            Guard.EnsureNotNull("source", source);
            Guard.EnsureNotNull("debug", debug);

            code.WriteLine("// switch statement inside lexer main loop ");
            code.WriteLine("// after applying state transition table and");
            code.WriteLine("// accepting flag is set");
            code.WriteLine("switch(sn)");
            code.WriteLine("{");

            var builder = new StringBuilder();
            var invalidBuilder = new StringBuilder();

            invalidBuilder.AppendFormat(CASE_TEMPLATE, STATE_INVALID);

            foreach (var element in source)
            {
                if (StringComparer.Ordinal.Equals(INVALID_LEXER_NAME, element.Name))
                {
                    continue;
                }

                var a = element.Automata;

                bool has_states = false;
                foreach (var state in a.States)
                {
                    if (state == a.Initial)
                    {
                        continue;
                    }

                    if (a.Accepting.Contains(state))
                    {
                        code.Write(CASE_TEMPLATE, state);
                        has_states = true;
                    }
                    else
                    {
                        invalidBuilder.AppendFormat(CASE_TEMPLATE, state);
                    }
                }

                if (has_states)
                {
                    WriteParseMethods(element, code, builder);
                }
            }

            // invalid cases

            code.Write(invalidBuilder);
            WriteParseMethods(source.Single(e => StringComparer.Ordinal.Equals(INVALID_LEXER_NAME, e.Name)), code, builder);

            code.WriteLine("}\n");

            code.WriteLine(builder);
        }

        private static void BuildStateTransitionTable(IEnumerable<MicroLexer> source, TextWriter debug, TextWriter code, int ieqc_count, int states_count, out byte[] stt)
        {
            Guard.EnsureNotNull("source", source);
            Guard.EnsureNotNull("debug", debug);
            Guard.EnsureNotNull("code", code);

            code.WriteLine("const byte TRANSITION_TOKEN_COMPLETE_FLAG = 0x{0:X2};", TRANSITION_TOKEN_COMPLETE_FLAG);
            code.WriteLine("const byte STATE_INITIAL = 0x{0:X2};", STATE_INITIAL);
            code.WriteLine("const byte STATE_INVALID = 0x{0:X2};", STATE_INVALID);
            code.WriteLine();

            stt = new byte[ieqc_count * states_count];
            stt.Fill(UNDEFINED);
            // assert states count && iecq count

            // eos transition
            stt.SafeSet(STATE_INITIAL * ieqc_count + IEQC_EOS, UNDEFINED, (byte)(TRANSITION_TOKEN_COMPLETE_FLAG | STATE_INITIAL));
            stt.SafeSet(STATE_INVALID * ieqc_count + IEQC_INVALID, UNDEFINED, (byte)(0 | IEQC_INVALID));

            foreach (var element in source)
            {
                var dfa = element.Automata;

                foreach (var t in dfa.Transitions)
                {
                    stt.SafeSet(t.Current * ieqc_count + t.Input, UNDEFINED, (byte)(0 | t.Next));
                }

                // non accepting state
                foreach (var s in dfa.States.Except(dfa.Accepting).Where(e => e != STATE_INITIAL).Where(e => e != STATE_INVALID))
                {
                    foreach (var j in Enumerable.Range(0, ieqc_count).Where(e => e != IEQC_EOS))
                    {
                        var index = s * ieqc_count + j;

                        if (stt[index] == UNDEFINED)
                        {
                            stt.SafeSet(index, UNDEFINED, (byte)(0 | STATE_INVALID));
                        }
                    }
                }
            }

            // fill initial states with invalid transitions
            for (int j = 0; j < ieqc_count; ++j)
            {
                var index = STATE_INITIAL * ieqc_count + j;

                if (stt[index] == UNDEFINED)
                {
                    stt.SafeSet(index, UNDEFINED, (byte)(STATE_INVALID));
                }
            }

            // copy initial transitions to each transition
            for (int i = 1; i < states_count; ++i)
            {
                for (int j = 0; j < ieqc_count; ++j)
                {
                    var index = i * ieqc_count + j;
                    var copyFromIndex = STATE_INITIAL * ieqc_count + j;

                    if (stt[index] == UNDEFINED)
                    {
                        stt.SafeSet(index, UNDEFINED, (byte)(TRANSITION_TOKEN_COMPLETE_FLAG | stt[copyFromIndex]));
                    }
                }
            }

            code.WriteLine("private const byte STT_STATES = {0};\n", states_count);
            code.WriteLine("private static readonly byte[] stt = new byte[IEQ_COUNT * STT_STATES]");
            code.WriteLine("{");
            code.WriteArray(stt, ieqc_count, e => string.Format("0x{0:x2}", e));
            code.WriteLine("};\n");
        }

        private static void WriteInputEquivGroups(ILookup<int, input> source, TextWriter debug, TextWriter code, int ieqc_count)
        {
            Guard.EnsureNotNull("source", source);

            byte[] ieqt = new byte[128]; // 0 - 127

            Utilities.Fill(ieqt, UNDEFINED);

            foreach (var grouping in source)
            {
                foreach (var element in grouping)
                {
                    ieqt.SafeSet(element, UNDEFINED, (byte)grouping.Key);
                }
            }

            if (ieqt.Any(e => e == UNDEFINED))
            {
                throw new ArgumentException("Oops");
            }

            const int COLUMNS = 16;

            code.WriteLine("// input equivalence groups:");
            code.WriteLine("private const byte IEQ_COUNT = {0};", ieqc_count); // + eos + inv
            code.WriteLine("private static readonly byte[] ieqt = new byte[128]");
            code.WriteLine("{");
            code.WriteArray(ieqt, COLUMNS, e => string.Format("0x{0:x2}", e));
            code.WriteLine("};\n");
        }

        private static IList<MicroLexer> CompressLexers(List<MicroLexer> source, ILookup<int, input> mapping, out int states_count)
        {
            states_count = 2;

            Guard.EnsureNotNull("source", source);
            Guard.EnsureNotNull("mapping", mapping);

            var result = new List<MicroLexer>();

            var reverseMapping = mapping.SelectMany(e => e, (g, e) => Tuple.Create(e, g.Key)).ToDictionary(e => e.Item1, e => e.Item2);

            foreach (var element in source)
            {
                var dfa = element.Automata;
                var initial = dfa.Initial;

                if (dfa.Transitions.Any(e => e.Next == initial))
                {
                    throw new ArgumentOutOfRangeException("Cannot handle initial state beeing reachable.");
                }

                var uniqueStates = new Dictionary<state, state>();

                uniqueStates.Add(initial, 0);

                foreach (var s in dfa.States.Where(e => e != initial))
                {
                    uniqueStates.Add(s, states_count++);
                }

                var transitions = new List<Transition>(dfa.Transitions
                        .Where(e => (e.Input & Transition.NOT_FLAG) == 0)
                        .Select(e => new Transition(uniqueStates[e.Current], (char)reverseMapping[e.Input], uniqueStates[e.Next])).Distinct());

                // handling negated transitions
                foreach (var grouping in dfa.Transitions.Where(e => (e.Input & Transition.NOT_FLAG) == Transition.NOT_FLAG).GroupBy(e => Tuple.Create(e.Current, e.Next)))
                {
                    var current = uniqueStates[grouping.Key.Item1];
                    var next = uniqueStates[grouping.Key.Item2];

                    foreach (var t in grouping)
                    {
                        var input = reverseMapping[(char)(t.Input & ~Transition.NOT_FLAG)];

                        transitions.AddRange(
                            mapping.Select(e => (input)e.Key)
                            .Where(e => e != input && e != IEQC_EOS)
                            .Except<input>(transitions.Where(e => e.Current == current && e.Next == next).Select(e => e.Input))
                            .Select(e => new Transition(current, e, next)).ToArray());
                    }
                }

                result.Add(
                    new MicroLexer()
                    {
                        Name = element.Name,
                        Automata = new Dfa(
                            0,
                            dfa.Accepting.Select(e => uniqueStates[e]),
                            transitions)
                    });
            }

            return result;
        }

        public static void MainLexer()
        {
            var comments = '#' * Not('\n').Star() * '\n';
            var letterRange = Range('a', 'z') | Range('A', 'Z') | '_';
            var identifiers = (letterRange) * (letterRange | Range('0', '9')).Star();

            // we allow 01... because we can
            var numbers = Range('0', '9').Plus();

            // \xfd any char except \"
            var strings = '"' * Not('"').Star() * '"';

            var symbols = Only("(", ")", "[", "]", "{", "}", "@", ".", ",", ";", "?", "+", "-", "*", "/", "%", "=", "&&", "||", "<", ">", "<=", ">=", "==", "===", "!", ":", "%", "'");

            var whitespaces = Only(" ", "\n", "\r", "\t").Plus(); //!!! string

            // var invalid = //!!! string

            Func<Nfa, string, MicroLexer> k =
                (a, s) => new MicroLexer() { Automata = a.ToDeterministic().Minimize(), Name = s };

            GenerateLexer(
                Utilities.ToEnumerable(k(comments, "Comments"),
                k(identifiers, "Identifiers"),
                k(numbers, "Numbers"),
                k(strings, "Strings"),
                k(symbols, "Symbols"),
                k(whitespaces, "Whitespace")),
                Console.Out,
                Console.Out);
        }
    }
}
