﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Text;

namespace libdyn.generator.Lexer
{
    using input = System.Char;
    using state = System.Int32;

    internal class StateSetComparer : IEqualityComparer<ISet<state>>
    {
        #region IEqualityComparer<ISet<int>> Members

        public bool Equals(ISet<state> x, ISet<state> y)
        {
            return x.SetEquals(y);
        }

        public int GetHashCode(ISet<state> obj)
        {
            int result = 1;

            foreach (var element in obj)
            {
                result = (result << 5 + result) ^ element;
            }

            return result;
        }

        #endregion IEqualityComparer<ISet<int>> Members
    }

    [DebuggerDisplay("{ToString(),nq}")]
    public sealed class Nfa
    {
        private readonly state ini;
        private readonly state acc;

        public state Initial
        {
            get
            {
                return ini;
            }
        }

        public state Accepting
        {
            get
            {
                return acc;
            }
        }

        public const input EPSILON = default(input);

        private static readonly Nfa EMPTY = new Nfa(0, 1, new Transition[0]);

        public static Nfa Empty { get { return EMPTY; } }

        private readonly Transition[] trans;
        private readonly state[] states;
        internal readonly state freeRange;

        public Nfa(int initial, int final, IEnumerable<Transition> transitions)
        {
            Guard.EnsureNotNull("transitions", transitions);

            this.ini = initial;
            this.acc = final;

            trans = transitions.ToArray();
            states = Utilities.ToEnumerable(initial, final).Concat(trans.Select(t => t.Current)).Concat(trans.Select(t => t.Next)).Distinct().ToArray();
            freeRange = states.Max() + 1;
        }

        public IReadOnlyCollection<Transition> Transitions
        {
            get
            {
                return new ReadOnlyCollection<Transition>(trans);
            }
        }

        public IReadOnlyCollection<state> States
        {
            get
            {
                return new ReadOnlyCollection<state>(states);
            }
        }

        // concat
        public static Nfa operator *(Nfa left, Nfa right)
        {
            var rightOffset = left.freeRange;

            return new Nfa(
                left.Initial,
                right.Accepting + rightOffset,
                Utilities.ToEnumerable(new Transition(left.Accepting, EPSILON, rightOffset + right.Initial))
                .Concat(left.Transitions)
                .Concat(right.Transitions.Select(t => new Transition(rightOffset + t.Current, t.Input, rightOffset + t.Next))));
        }

        // or
        public static Nfa operator |(Nfa left, Nfa right)
        {
            var leftOffset = 0;
            var rightOffset = left.freeRange;
            var mappingNew = rightOffset + right.freeRange;

            var initial = mappingNew;
            var final = initial + 1;

            return new Nfa(initial, final,
                Utilities.ToEnumerable(
                    new Transition(initial, EPSILON, leftOffset + left.Initial),
                    new Transition(initial, EPSILON, rightOffset + right.Initial),
                    new Transition(leftOffset + left.Accepting, EPSILON, final),
                    new Transition(rightOffset + right.Accepting, EPSILON, final))
                    .Concat(left.Transitions).Concat(right.Transitions.Select(t => new Transition(rightOffset + t.Current, t.Input, rightOffset + t.Next))));
        }

        public Nfa Star()
        {
            var newInitial = this.freeRange;
            var newFinal = newInitial + 1;

            return new Nfa(newInitial, newFinal, Utilities.ToEnumerable(
                new Transition(this.Accepting, EPSILON, this.Initial),
                new Transition(newInitial, EPSILON, this.Initial),
                new Transition(this.Accepting, EPSILON, newFinal),
                new Transition(newInitial, EPSILON, newFinal)
                ).Concat(this.Transitions));
        }

        public Nfa Plus()
        {
            var newInitial = this.freeRange;
            var newFinal = newInitial + 1;

            return new Nfa(newInitial, newFinal, Utilities.ToEnumerable(
                new Transition(newInitial, EPSILON, this.Initial),
                new Transition(this.Accepting, EPSILON, newFinal),
                new Transition(this.Accepting, EPSILON, this.Initial)
                ).Concat(this.Transitions));
        }

        public static implicit operator Nfa(char value)
        {
            return new Nfa(0, 1, Utilities.ToEnumerable(new Transition(0, value, 1)));
        }

        public static Nfa operator !(Nfa value)
        {
            return (!value.ToDeterministic().Minimize()).ToNondeterministicFiniteAutomata();
        }

        public static ISet<state> Move(ICollection<Transition> transitions, ISet<state> current, input value)
        {
            return new HashSet<state>(current.SelectMany(e => transitions.Where(t => e == t.Current && t.Input == value)).Select(e => e.Next));
        }

        // subset construction
        public Dfa ToDeterministic()
        {
            var pending = new Stack<ISet<state>>();

            state nfaInitial = Initial;
            state nfaFinal = Accepting;

            ISet<state> dfaFinals = new HashSet<state>();

            IDictionary<ISet<state>, state> mapping = new Dictionary<ISet<state>, state>(new StateSetComparer());

            state dfaInitial;

            var epsilons = this.trans.Where(e => e.Input == EPSILON).ToArray();
            var input = this.trans.Select(e => e.Input).Distinct().Where(e => e != EPSILON).ToArray();

            // initial
            {
                var firstState = new HashSet<state>();
                firstState.Add(nfaInitial);

                var first = EpsilonClosure(epsilons, firstState);

                mapping.Add(first, mapping.Count);

                pending.Push(first);

                dfaInitial = mapping[first];
            }

            var transitions = new List<Transition>();

            while (pending.Count > 0)
            {
                var current = pending.Pop();

                if (current.Contains(nfaFinal))
                {
                    dfaFinals.Add(mapping[current]);
                }

                foreach (var currentInput in input)
                {
                    ISet<state> next = EpsilonClosure(epsilons, Move(this.trans, current, currentInput));

                    if (next.Count == 0)
                    {
                        continue;
                    }

                    if (!mapping.ContainsKey(next))
                    {
                        pending.Push(next);
                        mapping.Add(next, mapping.Count);
                    }

                    transitions.Add(new Transition(mapping[current], currentInput, mapping[next]));
                }
            }

            return new Dfa(dfaInitial, dfaFinals, transitions);
        }

        private static ISet<state> EpsilonClosure(ICollection<Transition> epsilonTransitions, ISet<state> start)
        {
            ISet<state> visited = new HashSet<state>(start);
            var pending = new Stack<state>(start);

            while (pending.Count > 0)
            {
                var current = pending.Pop();

                foreach (var t in epsilonTransitions.Where(e => e.Current == current))
                {
                    if (!visited.Contains(t.Next))
                    {
                        pending.Push(t.Next);
                        visited.Add(t.Next);
                    }
                }
            }

            return visited;
        }

        public override string ToString()
        {
            var builder = new StringBuilder();

            builder.AppendFormat("{0,8}   {1,8} -> {2,8}\n", String.Empty, String.Empty, this.Initial);

            foreach (var t in Transitions)
            {
                builder.AppendLine(t.ToString());
            }

            builder.AppendFormat("{0,8}   {1,8} -> {2,8}\n", this.Accepting, String.Empty, String.Empty);

            return builder.ToString();
        }
    }
}
