﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace libdyn.generator.Lexer
{
    using state = System.Int32;

    public static class Utilities
    {
        public static IEnumerable<T> ToEnumerable<T>(params T[] source)
        {
            return source;
        }

        public static void Fill<T>(this IList<T> source, T value)
        {
            Guard.EnsureNotNull("source", source);

            for (int i = 0, max = source.Count; i < max; ++i)
            {
                source[i] = value;
            }
        }

        public static Func<Transition, Transition> Remap(Func<state, state> mapping)
        {
            return t => new Transition(mapping(t.Current), t.Input, mapping(t.Next));
        }

        public static string EscapeInput(this char value)
        {
            switch (value)
            {
                case '\0':
                    return "\u03B5";

                case '\t':
                    return "\\t";

                case '\r':
                    return "\\r";

                case '\n':
                    return "\\n";

                case ' ':
                    return "\\s";
            }

            if (value < 32)
            {
                return string.Format("\\u{0:X2}", (int)value);
            }
            else if ((value & Transition.NOT_FLAG) == Transition.NOT_FLAG)
            {
                return string.Format("!{0}", EscapeInput((char)(value & ~Transition.NOT_FLAG)));
            }
            else if (value > 255)
            {
                return string.Format("\\u{0:X4}", (int)value);
            }

            return "" + value;
        }

        public static Func<T, string> ToStringNullSafe<T>()
        {
            return o => o != null ? o.ToString() : "null";
        }

        public static void Write<TKey, TValue>(this TextWriter output, ILookup<TKey, TValue> source, Func<TKey, string> keyPrinter = null,
            Func<TValue, string> valuePrinter = null)
        {
            Guard.EnsureNotNull("source", source);
            Guard.EnsureNotNull("output", output);

            var kp = keyPrinter ?? ToStringNullSafe<TKey>();
            var vp = valuePrinter ?? ToStringNullSafe<TValue>();

            foreach (var group in source)
            {
                output.WriteLine(kp(group.Key));

                output.Write('\t');

                bool tail = false;

                foreach (var element in group)
                {
                    if (tail)
                    {
                        output.Write(", ");
                    }

                    output.Write(vp(element));
                    tail = true;
                }

                output.WriteLine();
            }
        }

        public static void WriteArray<T>(this TextWriter writer, IList<T> array, int columns, Func<T, string> formatter)
        {
            int counter = 0, length = array.Count;

            Func<T, string> f = formatter ?? (e => e.ToString());

            if (length > 0)
            {
                do
                {
                    if ((counter % columns) == 0)
                    {
                        writer.Write("    ");
                    }

                    writer.Write(f(array[counter]));

                    if (counter + 1 < length)
                    {
                        writer.Write(',');
                    }

                    if (counter % columns < columns - 1)
                    {
                        writer.Write(' ');
                    }
                    else
                    {
                        writer.WriteLine();
                    }
                }
                while (++counter < length);
            }
        }

        public static void SafeSet<TElement>(this IList<TElement> source, int index, TElement expected, TElement value)
        {
            Guard.EnsureNotNull("source", source);

            if (index < 0 || index >= source.Count)
            {
                throw new ArgumentOutOfRangeException("source");
            }

            TElement old_value = source[index];

            var cmp = EqualityComparer<TElement>.Default;

            if (!cmp.Equals(old_value, expected) && !cmp.Equals(old_value, value))
            {
                throw new Exception("Value already set");
            }

            source[index] = value;
        }

        public static void SafeSet<TElement>(this IList<TElement> source, int index, TElement value)
        {
            SafeSet<TElement>(source, index, default(TElement), value);
        }
    }
}
