﻿using Dyn.Ast;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using Xunit;

namespace tests.Ast
{
    public static class DynAstUnitTests
    {
        private static void TestLexerParser<T>(AstFuzzer fuzzer, Random rnd, Func<T, string> formatter)
        {
            const string FILENAME = "Random";

            var node = fuzzer.Fuzzy<T>(rnd);
            var source = formatter(node);
            DynLexer lexer = new DynLexer(FILENAME, source);
            Assert.All(lexer, e => Assert.NotEqual(SyntaxKind.InvalidKind, e.Kind));
        }

        [Theory]
        [InlineData(64)]
        [InlineData(96)]
        [InlineData(128)]
        public static void RandomEqualityTest(int count)
        {
            TestRandomEquality<DynProgramClassNode>(DynFuzzer.Programs, count, AstExtensions.Normal, DynNodeFormatter.ToSource);
            TestRandomEquality<DynProgramMethodDeclarationNode>(DynFuzzer.Programs, count, AstExtensions.Normal, DynNodeFormatter.ToSource);
            TestRandomEquality<DynProgramStatementNode>(DynFuzzer.Programs, count, AstExtensions.Normal, DynNodeFormatter.ToSource);
            TestRandomEquality<DynAssertionExpressionNode>(DynFuzzer.Assertions, count, AstExtensions.Normal, DynNodeFormatter.ToSource);
        }

        private static void TestRandomEquality<T>(AstFuzzer fuzzer, int count, IEqualityComparer<T> comparer, Func<T, string> formatter)
        {
            var sampleAst = new List<T>();
            var sampleSource = new List<string>();

            var rnd = new Random(count);

            for (int i = 0; i < count; ++i)
            {
                var sample = fuzzer.Fuzzy<T>(new Random(rnd.Next()));
                sampleAst.Add(sample);
                sampleSource.Add(formatter(sample));
            }

            var Ordinal = StringComparer.Ordinal;

            for (int i = 0; i < count; ++i)
                for (int j = 0; j < count; ++j)
                {
                    var src = Ordinal.Equals(sampleSource[i], sampleSource[j]);
                    var ast = comparer.Equals(sampleAst[i], sampleAst[j]);

                    Assert.True(src == ast, string.Format("{0} vs. {1}:\n{2}\n{3}", i, j, sampleAst[i], sampleAst[j]));
                }
        }

        [Theory]
        [InlineData(1024)]
        [InlineData(2048)]
        public static void RandomLexerTest(int count)
        {
            for (int i = 0; i < count; ++i)
            {
                TestLexerParser<DynProgramClassNode>(DynFuzzer.Programs, new Random(i), DynNodeFormatter.ToSource);
            }

            for (int i = 0; i < count; ++i)
            {
                TestLexerParser<DynAssertionExpressionNode>(DynFuzzer.Assertions, new Random(i), DynNodeFormatter.ToSource);
            }

            for (int i = 0; i < count; ++i)
            {
                TestLexerParser<DynProgramExpressionNode>(DynFuzzer.Programs, new Random(i), DynNodeFormatter.ToSource);
            }
        }
    }
}
