﻿using Dyn.Ast;
using System;
using System.Text;

namespace tests.Ast
{
    using Dyn.Lexer;
    using System.Linq;
    using constructor = Func<Random, int, object>;

    public static class DynFuzzer
    {
        private static readonly string IDENTIFIER_TAIL =
            "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";

        private static readonly string IDENTIFIER_HEAD =
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

        private static string FuzzyIdentifier(Random r, int depth)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(IDENTIFIER_HEAD[r.Next(IDENTIFIER_HEAD.Length - 1)]);

            for (int i = 0, max = 1 + r.Next(3); i < max; ++i)
            {
                builder.Append(IDENTIFIER_TAIL[r.Next(IDENTIFIER_TAIL.Length - 1)]);
            }

            return builder.ToString();
        }

        public static constructor Identifier()
        {
            return (r, d) => FuzzyIdentifier(r, 0);
        }

        public static int FuzzyNumeric(Random r, int depth)
        {
            return r.Next(20);
        }

        public static string FuzzyString(Random r, int depth)
        {
            StringBuilder builder = new StringBuilder();

            int len = r.Next(4) + 2 + r.Next(4);

            // visible ASCII range after " and before ^DEL
            for (int i = 0; i < len; ++i)
            {
                builder.Append((char)r.Next('"' + 1, 0x7F - 1));
            }

            return builder.ToString();
        }

        private static int Cubic(Random r, int maximum)
        {
            return (int)(maximum * r.NextDouble() * r.NextDouble() * r.NextDouble());
        }

        private static DynProgramExpressionNode[] DEFAULT_PROG_VARS = Dyn.Utilities.AsArray<DynProgramExpressionNode>(
                AstExtensions.Programs.False,
                AstExtensions.Programs.True,
                AstExtensions.Programs.Zero,
                new DynProgramVariableNode(Span.Empty, DynProgramStatementNode.NodeKind.Variable, "r"),
                new DynProgramVariableNode(Span.Empty, DynProgramStatementNode.NodeKind.Variable, "self"),
                AstExtensions.Programs.Null
           );

        private static DynAssertionExpressionNode[] DEFAULT_ASRT_VARS = Dyn.Utilities.AsArray(
                AstExtensions.Assertions.False,
                AstExtensions.Assertions.True,
                AstExtensions.Assertions.Null,
                AstExtensions.Assertions.Self,
                AstExtensions.Assertions.r
            );

        private const int MAX_DEPTH = 200;

        private static readonly AstFuzzer _P = CreateProgramFuzzer();
        private static readonly AstFuzzer _A = CreateAssertionFuzzer();
        private static readonly AstFuzzer _BA = CreateBooleanAssertionFuzzer();
        private static readonly AstFuzzer _NA = CreateNumericAssertionFuzzer();
        private static readonly AstFuzzer _SA = CreateStringAssertionFuzzer();
        private static readonly AstFuzzer _OA = CreateObjectAssertionFuzzer();

        private static readonly object _ = null;

        private static AstFuzzer CreateObjectAssertionFuzzer()
        {
            var defaultObjects = Dyn.Utilities.AsArray(
                AstExtensions.Assertions.Null,
                AstExtensions.Assertions.Self);

            // r/ImGoingToHellForThis
            var inheritance = new AstFuzzerExtensions.InheritanceBuilder<DynAssertionExpressionNode>(() => _OA)
            {
                { (DynAssertionIfThenElseNode) _, 1 },
                { (DynAssertionInstanceVariableNode) _, 4 },
                { (DynAssertionVariableNode) _, 3 },
            }.ToFunc();

            return new AstFuzzer()
            {
                { inheritance },

                { (r, d) => new DynAssertionIfThenElseNode(
                    Span.Empty,
                    _BA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                    _OA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                    _OA.Fuzzy<DynAssertionExpressionNode>(r, d + 1))
                },

                { (r, d) => new DynAssertionInstanceVariableNode(
                    Span.Empty,
                    _OA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                    FuzzyIdentifier(r, d))
                },

                { (r, d) =>
                    r.Next(2) == 0 ?
                    new DynAssertionVariableNode(
                        Span.Empty,
                        DynAssertionExpressionNode.NodeKind.ProgramVariable,
                        FuzzyIdentifier(r, d)
                    ) :
                    defaultObjects.Random(r)
                }
            };
        }

        private static AstFuzzer CreateBooleanAssertionFuzzer()
        {
            var inheritance = new AstFuzzerExtensions.InheritanceBuilder<DynAssertionExpressionNode>(() => _BA)
            {
                { (DynAssertionBinaryNode) _, 6 },
                { (DynAssertionUnaryNode) _, 7 },
                { (DynAssertionIfThenElseNode) _, 3 },
                { (DynAssertionQuantifierNode) _, 4 },
                { (DynAssertionTypeCheckNode) _, 5 },
                { (DynAssertionVariableNode) _, 6 }
            }.ToFunc();

            var subType = Dyn.Utilities.AsArray(
                _SA,
                _OA,
                _OA,
                _NA,
                _NA,
                _NA,
                _BA,
                _BA,
                _BA,
                _BA);

            var l0 = Dyn.Utilities.AsArray(
                DynAssertionExpressionNode.NodeKind.AssertionVariable,
                DynAssertionExpressionNode.NodeKind.ProgramVariable);

            var quantifierVariableType = Dyn.Utilities.AsArray(
                "object",
                "boolean",
                "numeric",
                "string");

            var defaultBoolean = Dyn.Utilities.AsArray(
                AstExtensions.Assertions.Null,
                AstExtensions.Assertions.Self);

            var booleanOperator = Dyn.Utilities.AsArray(
                DynAssertionExpressionNode.NodeKind.And,
                DynAssertionExpressionNode.NodeKind.Equality,
                DynAssertionExpressionNode.NodeKind.Implies,
                DynAssertionExpressionNode.NodeKind.Or);

            return new AstFuzzer()
            {
                // operators which are working on booleans
                { (r, d) => booleanOperator.Random(r) },

                { inheritance },

                { (r, d) => {
                    var fuzzer = subType.Random(r);

                    return new DynAssertionBinaryNode(
                        Span.Empty,
                        fuzzer.Fuzzy<DynAssertionExpressionNode.NodeKind>(r, d + 1),
                        fuzzer.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        fuzzer.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ); } },

                { (r, d) => new DynAssertionUnaryNode(
                        Span.Empty,
                        _BA.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) },

                { (r, d) => {
                    var fuzzer = subType.Random(r);

                    return new DynAssertionIfThenElseNode(
                        Span.Empty,
                        _BA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        fuzzer.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        fuzzer.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ); } },

                { (r, d) => new DynAssertionQuantifierNode(
                        Span.Empty,
                        l0.Random(r),
                        FuzzyIdentifier(r, d),
                        quantifierVariableType.Random(r),
                        _BA.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                    ) },

                { (r, d) => new DynAssertionTypeCheckNode(
                        Span.Empty,
                        _OA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        AstFuzzerExtensions.List(Cubic(r, 6) + 1, r, d, FuzzyIdentifier)
                    ) },

                { (r, d) => defaultBoolean.Random(r) }
            };
        }

        private static AstFuzzer CreateNumericAssertionFuzzer()
        {
            var inheritance = new AstFuzzerExtensions.InheritanceBuilder<DynAssertionExpressionNode>(() => _NA)
            {
                { (DynAssertionNumericNode) _, 9 },
                { (DynAssertionIfThenElseNode) _, 3 },
                { (DynAssertionBinaryNode) _, 3 }
            }.ToFunc();

            var bop = Dyn.Utilities.AsArray(
                DynAssertionExpressionNode.NodeKind.Addition,
                DynAssertionExpressionNode.NodeKind.Subtraction,
                DynAssertionExpressionNode.NodeKind.Multiplication,
                DynAssertionExpressionNode.NodeKind.Division,
                DynAssertionExpressionNode.NodeKind.Modulus);

            var numericOperators = Dyn.Utilities.AsArray(
                DynAssertionExpressionNode.NodeKind.Equality,
                DynAssertionExpressionNode.NodeKind.GreaterThan,
                DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual,
                DynAssertionExpressionNode.NodeKind.LessThan,
                DynAssertionExpressionNode.NodeKind.LessThanOrEqual,
                DynAssertionExpressionNode.NodeKind.Addition,
                DynAssertionExpressionNode.NodeKind.Subtraction,
                DynAssertionExpressionNode.NodeKind.Multiplication,
                DynAssertionExpressionNode.NodeKind.Division,
                DynAssertionExpressionNode.NodeKind.Modulus);

            return new AstFuzzer()
            {
                // operators which are working on numerics
                { (r, d) => numericOperators.Random(r) },

                { inheritance },

                { (r, d) => new DynAssertionNumericNode(Span.Empty, FuzzyNumeric(r, d)) },

                { (r, d) => new DynAssertionIfThenElseNode(Span.Empty,
                    _BA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                    _NA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                    _NA.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) },

                { (r, d) => new DynAssertionBinaryNode(
                        Span.Empty,
                        bop.Random(r),
                        _NA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        _NA.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) },
            };
        }

        private static AstFuzzer CreateStringAssertionFuzzer()
        {
            var inheritance = new AstFuzzerExtensions.InheritanceBuilder<DynAssertionExpressionNode>(() => _SA)
            {
                { (DynAssertionStringNode) _, 5 },
                { (DynAssertionIfThenElseNode) _, 3 },
            }.ToFunc();

            var stringOperators = Dyn.Utilities.AsArray(
                DynAssertionExpressionNode.NodeKind.Addition,
                DynAssertionExpressionNode.NodeKind.Subtraction,
                DynAssertionExpressionNode.NodeKind.Multiplication,
                DynAssertionExpressionNode.NodeKind.Division,
                DynAssertionExpressionNode.NodeKind.Modulus);

            return new AstFuzzer()
            {
                // operators which are working on strings
                { (r, d) => stringOperators.Random(r) },

                { inheritance },

                { (r, d) => new DynAssertionStringNode(Span.Empty, FuzzyString(r, d)) },

                { (r, d) => new DynAssertionIfThenElseNode(Span.Empty,
                    _BA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                    _SA.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                    _SA.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) }
            };
        }

        private static AstFuzzer CreateAssertionFuzzer()
        {
            var inheritance = new AstFuzzerExtensions.InheritanceBuilder<DynAssertionExpressionNode>(() => _A) {
                    { (DynAssertionPredicateNode) _, 3 },
                    { (DynAssertionBinaryNode) _, 10 },
                    { (DynAssertionUnaryNode) _, 7 },
                    { (DynAssertionIfThenElseNode) _, 3 },
                    { (DynAssertionQuantifierNode) _, 4 },
                    { (DynAssertionTypeCheckNode) _, 5 },
                    { (DynAssertionNumericNode) _, 20 },
                    { (DynAssertionStringNode) _, 5 },
                    { (DynAssertionInstanceVariableNode) _, 10 },
                    { (DynAssertionVariableNode) _, 10 }
            }.ToFunc();

            var ac_nodes = Enum.GetValues(
                    typeof(DynAssertionExpressionNode.NodeKind))
                    .Cast<DynAssertionExpressionNode.NodeKind>()
                    .ToList();

            var l1 = Dyn.Utilities.AsArray(
                    DynAssertionExpressionNode.NodeKind.Existential,
                    DynAssertionExpressionNode.NodeKind.Universal);

            var l2 = Dyn.Utilities.AsArray(
                    DynAssertionExpressionNode.NodeKind.Or,
                    DynAssertionExpressionNode.NodeKind.Implies,
                    DynAssertionExpressionNode.NodeKind.Addition,
                    DynAssertionExpressionNode.NodeKind.Subtraction,
                    DynAssertionExpressionNode.NodeKind.Multiplication,
                    DynAssertionExpressionNode.NodeKind.Division,
                    DynAssertionExpressionNode.NodeKind.Modulus,
                    DynAssertionExpressionNode.NodeKind.GreaterThan,
                    DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual,
                    DynAssertionExpressionNode.NodeKind.LessThan,
                    DynAssertionExpressionNode.NodeKind.LessThanOrEqual,
                    DynAssertionExpressionNode.NodeKind.Equality);

            var l3 = Dyn.Utilities.AsArray(
                    DynAssertionExpressionNode.NodeKind.AssertionVariable,
                    DynAssertionExpressionNode.NodeKind.ProgramVariable);

            return new AstFuzzer()
            {
                { inheritance },

                { (r, d) => new DynAssertionPredicateNode(
                        Span.Empty,
                        FuzzyIdentifier(r, d),
                        AstFuzzerExtensions.List(Cubic(r, 4), r, d, _A.Fuzzy<DynAssertionExpressionNode>)
                ) },

                { (r, d) => new DynAssertionIfThenElseNode(
                        Span.Empty,
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) },

                { (r, d) => new DynAssertionQuantifierNode(
                        Span.Empty,
                        l1.Random(r),
                        FuzzyIdentifier(r, d),
                        FuzzyIdentifier(r, d),
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) },

                { (r, d) => new DynAssertionBinaryNode(
                        Span.Empty,
                        l2.Random(r),
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) },

                { (r, d) => new DynAssertionUnaryNode(
                        Span.Empty,
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) },

                { (r, d) => new DynAssertionTypeCheckNode(
                        Span.Empty,
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        AstFuzzerExtensions.List(Cubic(r, 6) + 1, r, d, FuzzyIdentifier)
                ) },

                { (r, d) => new DynAssertionInstanceVariableNode(
                        Span.Empty,
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        FuzzyIdentifier(r, d)
                ) },

                { (r, d) => new DynAssertionVariableNode(
                        Span.Empty,
                        l3.Random(r),
                        FuzzyIdentifier(r, d)
                ) },

                { (r, d) => new DynAssertionNumericNode(Span.Empty, FuzzyNumeric(r, d)) },

                { (r, d) => new DynAssertionStringNode(Span.Empty, FuzzyString(r, d)) },
            };
        }

        private static AstFuzzer CreateProgramFuzzer()
        {
            var inheritance = new AstFuzzerExtensions.InheritanceBuilder<DynProgramExpressionNode>(() => _P) {
                {(DynProgramMethodCallNode) _, 3},
                {(DynProgramConstructorCallNode) _, 2},
                {(DynProgramAssignmentNode) _, 7},
                {(DynProgramIfNode) _, 3},
                {(DynProgramWhileNode) _, 1},
                {(DynProgramVariableNode) _, 10}
            }.ToFunc();

            var l0 = Dyn.Utilities.AsArray(
                    DynProgramStatementNode.NodeKind.Ensures,
                    DynProgramStatementNode.NodeKind.Requires,
                    DynProgramStatementNode.NodeKind.Invariant);

            var l1 = Dyn.Utilities.AsArray(
                    DynProgramStatementNode.NodeKind.Assert,
                    DynProgramStatementNode.NodeKind.ByTheWay);

            var l2 = Dyn.Utilities.AsArray(
                    DynProgramStatementNode.NodeKind.Variable,
                    DynProgramStatementNode.NodeKind.InstanceVariable);

            return new AstFuzzer()
            {
                { (r, d) => new DynHoareNode(
                        Span.Empty,
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        AstFuzzerExtensions.List(Cubic(r, 4) + 2, r, d, _P.Fuzzy<DynProgramStatementNode>),
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1)
                ) },

                { (r, d) => new DynProgramClassNode(
                        Span.Empty,
                        FuzzyIdentifier(r, d),
                        FuzzyIdentifier(r, d),
                        AstFuzzerExtensions.List(Cubic(r, 2) + 1, r, d, _P.Fuzzy<DynProgramRenameDeclarationNode>),
                        AstFuzzerExtensions.List(Cubic(r, 4) + 1, r, d, _P.Fuzzy<DynProgramMethodDeclarationNode>),
                        AstFuzzerExtensions.List(Cubic(r, 10) + 1, r, d, (r2, d2) => new DynProgramContractDeclarationNode(Span.Empty,
                                l0.Random(r),
                                _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1))
                )) },

                { (r, d) => new DynProgramRenameDeclarationNode(
                        Span.Empty,
                        FuzzyIdentifier(r, d),
                        FuzzyIdentifier(r, d)) },

                { (r, d) => new DynProgramMethodDeclarationNode(
                        Span.Empty,
                        FuzzyIdentifier(r, d),
                        AstFuzzerExtensions.List(Cubic(r, 4), r, d, FuzzyIdentifier),
                        AstFuzzerExtensions.List(Cubic(r, 16) + 1, r, d, _P.Fuzzy<DynProgramStatementNode>),
                        AstFuzzerExtensions.List(Cubic(r, 8) + 1, r, d, (r2, d2) => new DynProgramContractDeclarationNode(
                                Span.Empty,
                                l0.Random(r),
                                _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1))
                )) },

                { (r, d) => r.Next(3) == 0 ?
                        new DynProgramContractDeclarationNode(
                                Span.Empty,
                                l1.Random(r),
                                _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1)) :
                        (DynProgramStatementNode) _P.Fuzzy<DynProgramExpressionNode>(r, d + 1)
                },

                { (r, d) => d > MAX_DEPTH ? DEFAULT_PROG_VARS.Random(r) : inheritance(r, d + 1) },

                { (r, d) => new DynProgramMethodCallNode(
                        Span.Empty,
                        r.Next(8) == 0 ? _P.Fuzzy<DynProgramExpressionNode>(r, d + 1) : AstExtensions.Programs.Global,
                        FuzzyIdentifier(r, d),
                        AstFuzzerExtensions.List(Cubic(r, 4), r, d, _P.Fuzzy<DynProgramExpressionNode>)
                ) },

                { (r, d) => new DynProgramConstructorCallNode(
                        Span.Empty,
                        FuzzyIdentifier(r, d),
                        AstFuzzerExtensions.List(Cubic(r, 4), r, d, _P.Fuzzy<DynProgramExpressionNode>)
                ) },

                { (r, d) => new DynProgramAssignmentNode(
                        Span.Empty,
                        _P.Fuzzy<DynProgramVariableNode>(r, d + 1),
                        _P.Fuzzy<DynProgramExpressionNode>(r, d + 1)
                ) },

                { (r, d) => new DynProgramIfNode(
                        Span.Empty,
                        _P.Fuzzy<DynProgramExpressionNode>(r, d + 1),
                        AstFuzzerExtensions.List(Cubic(r, 4), r, d, _P.Fuzzy<DynProgramStatementNode>),
                        r.Next(3) == 0 ? AstFuzzerExtensions.List(1 + Cubic(r, 4), r, d, _P.Fuzzy<DynProgramStatementNode>) : null
                ) },

                { (r, d) => new DynProgramWhileNode(
                        Span.Empty,
                        _P.Fuzzy<DynProgramExpressionNode>(r, d + 1),
                        _A.Fuzzy<DynAssertionExpressionNode>(r, d + 1),
                        AstFuzzerExtensions.List(1 + Cubic(r, 8), r, d, _P.Fuzzy<DynProgramStatementNode>)
                ) },

                { (r, d) => new DynProgramVariableNode(
                        Span.Empty,
                        l2.Random(r),
                        FuzzyIdentifier(r, d)
                ) },
            };
        }

        public static AstFuzzer Programs { get { return _P; } }
        public static AstFuzzer Assertions { get { return _A; } }
        public static AstFuzzer BooleanAssertion { get { return _BA; } }
        public static AstFuzzer NumericAssertion { get { return _NA; } }
        public static AstFuzzer ObjectAssertion { get { return _OA; } }
        public static AstFuzzer StringAssertion { get { return _SA; } }
    }
}
