﻿using Dyn.Console.CommandLine;
using System;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using Xunit;

namespace Dyn.Tests.Console.CommandLine
{
    public static class CommandLineLexerUnitTests
    {
        // Simple Expected format:
        //
        // START := (SEGMENT)*
        // SEGMENT := MODIFIER + anything but modifiers
        // MODIFIER := [ˢˡᵖᵃ]
        //
        // ˢ starts a short option
        // ˡ starts a long option
        // ² starts a parameter
        // ³ starts an argument
        private static readonly Regex LEXER =
            new Regex(
                @"[ˢˡᵖᵃ][^ˢˡᵖᵃ]+",
                RegexOptions.Compiled |
                RegexOptions.Singleline |
                RegexOptions.CultureInvariant);

        /*[Fact]
    public static void SelfTest()
    {
        Assert.Equal(
            @"a§b§c§""d dd ddd""§e§f§""g \""  gg   ggg""".Split("§".ToArray()),
            SplitOnUnquotedWhitespace(@"a b c ""d dd ddd"" e  f   ""g \""  gg   ggg"""),
            StringComparer.Ordinal);
        Assert.Equal(
            @"a§b§c§d§e§f".Split("§".ToArray()),
            SplitOnUnquotedWhitespace("a\nb\rc\n\rd\te\tf\r\n"),
            StringComparer.Ordinal);
    }*/
        [System.Diagnostics.DebuggerStepThrough]
        public static string[] SplitWithFirstLineContent(string data)
        {
            string head, tail;

            using (var reader = new StringReader(data.Replace("\r", string.Empty)))
            {
                head = reader.ReadLine();
                tail = reader.ReadToEnd();

                if (string.IsNullOrEmpty(head))
                {
                    return new string[0] { };
                }

                if (string.IsNullOrEmpty(tail))
                {
                    return head.Substring(1).Split(head[0]);
                }

                return tail.Split(
                    new string[] { head + "\n" },
                    StringSplitOptions.RemoveEmptyEntries);
            }
        }

        public static void AssertCommands(string expected, string[] input)
        {
            char c;
            string s;

            var lexer = new CommandLineLexer(input);

            foreach (var element in LEXER.Matches(expected).
                    Cast<Match>().
                    Select(e => e.Value))
            {
                switch (element[0])
                {
                    case 'ˢ':
                        Assert.False(lexer.TryLong(out s));
                        Assert.True(lexer.TryShort(out c));
                        Assert.Equal(element[1], c);
                        break;
                    case 'ˡ':
                        Assert.False(lexer.TryShort(out c));
                        Assert.True(lexer.TryLong(out s));
                        Assert.Equal(element.Substring(1), s);
                        break;
                    case 'ᵖ':
                        // Assert.False(lexer.TryShort(out c));
                        // Assert.False(lexer.TryLong(out s));
                        Assert.True(lexer.TryParameter(out s));
                        Assert.Equal(element.Substring(1), s);
                        break;
                    case 'ᵃ':
                        Assert.False(lexer.TryShort(out c));
                        Assert.False(lexer.TryLong(out s));
                        Assert.True(lexer.TryArgument(out s));
                        Assert.Equal(element.Substring(1), s);
                        break;
                    default:
                        throw new Exception("Invalid modifier");
                }
            }

            Assert.True(lexer.IsCompleted);
            Assert.False(lexer.TryShort(out c));
            Assert.False(lexer.TryLong(out s));
            Assert.False(lexer.TryParameter(out s));
            Assert.False(lexer.TryArgument(out s));
        }

        [Theory]
        [InlineData("ˢa", "|-a")]
        [InlineData("ˢb", "|-b")]
        [InlineData("ˢaˢb", "|-ab")]
        [InlineData("ˡa", "|--a")]
        [InlineData("ˡb", "|--b")]
        [InlineData("ˡaˢbᵖcᵃd", "|--a|-b|c|d")]
        [InlineData("ˡaˢbᵖc d e fᵃd", "|--a|-b|c d e f|d")]
        [InlineData("ᵃ--aᵃ-bᵃc d e fᵖd", "|--|--a|-b|c d e f|d")]
        [InlineData("ᵖ--aᵖ-bᵖc d e fᵖd", "|--|--a|-b|c d e f|d")]
        [InlineData("ˢaᵖbˢaᵖbˢaᵖbˢaᵖb", "|-ab|-a:b|-a=b|-a|b")]
        [InlineData("ˡaᵖbˡaᵖbˡaᵖb", "|--a:b|--a=b|--a|b")]
        [InlineData("ˢaᵃ-b", "|-a|--|-b")]
        [InlineData("ˢaᵖ-b", "|-a|--|-b")]
        [InlineData("ˢaᵃ--b", "|-a|--|--b")]
        [InlineData("ˢaᵖ--b", "|-a|--|--b")]
        [InlineData("ˡaᵃ-b", "|--a|--|-b")]
        [InlineData("ˡaᵖ-b", "|--a|--|-b")]
        [InlineData("ˡaᵃ--b", "|--a|--|--b")]
        [InlineData("ˡaᵖ--b", "|--a|--|--b")]
        [InlineData("ᵖ--", "|--|--")]
        [InlineData("ᵃ--", "|--|--")]
        [InlineData("ˢaᵖb", "|-ab")]
        public static void Positive(string expected, string input)
        {
            AssertCommands(expected, SplitWithFirstLineContent(input));
        }

        [Theory]
        [InlineData("ˢa", "|")]
        [InlineData("ˡa", "|")]
        [InlineData("ᵖa", "|")]
        [InlineData("ᵃa", "|")]
        [InlineData("", "|a")]
        [InlineData("ˡaᵖb", "|--ab")]
        [InlineData("ˢaᵃb", "|-ab")]
        [InlineData("ˢaᵃb", "||-ab")]
        [InlineData("ᵃa", "||a")]
        [InlineData("ᵃaᵃb", "|a||b")]
        [InlineData("ˡaᵃb", "|--a:b")]
        [System.Diagnostics.DebuggerHidden]
        public static void Negative(string expected, string input)
        {
            Assert.ThrowsAny<Xunit.Sdk.XunitException>(
                () => AssertCommands(expected, SplitWithFirstLineContent(input)));
        }

        [Fact]
        [System.Diagnostics.DebuggerHidden]
        public static void NegativeEmpty()
        {
            var lexer = new CommandLineLexer(SplitWithFirstLineContent("|"));
            string s;
            Assert.True(lexer.TryArgument(out s));
            Assert.Equal(string.Empty, s);
            Assert.True(lexer.IsCompleted);
        }
    }
}
