﻿using System;
using System.Collections.Generic;
using System.Reflection;
using Xunit;
using tests;
using Dyn.Console.CommandLine;

namespace Dyn.Tests.Console.CommandLine
{
    public static class ReflectionInvokerUnitTests
    {
        public static IEnumerable<MemberInfo> Members<T>()
        {
            return typeof(T).GetTypeInfo().GetMembers(
                BindingFlags.DeclaredOnly |
                BindingFlags.DeclaredOnly |
                BindingFlags.Public |
                BindingFlags.NonPublic |
                BindingFlags.Instance |
                BindingFlags.Static);
        }

        public static IEnumerable<object[]> NEGATIVES = Members<NegativeExamples>().FiXunit();
        public static IEnumerable<object[]> POSITIVES = Members<PostiveExamples>().FiXunit();
        public static IEnumerable<object[]> INVOCATIONS = Members<PostiveExamples>().FiXunit();

        private class PostiveExamples
        {
            public string String;
            public int Int;
            public bool Bool;
            public int? NullableInt;
            public bool? NullableBool;
            public string Property { set { String = value; } }
            public void Method()
            { }
            public void Invocation(string a, int b, bool c, int? d, bool? e)
            {
                String = a;
                Int = b;
                Bool = c;
                NullableInt = d;
                NullableBool = e;
            }
        }

        [Theory]
        [MemberData("POSITIVES")]
        public static void Positive(MemberInfo info)
        {
            if (info.MemberType == MemberTypes.Constructor)
            {
                return;
            }

            Assert.NotNull(ReflectionInvoker.Create(info));
        }



        [Theory]
        [InlineData("|a|1|true|-1|false")]
        [InlineData("|b|2|1|-2|0")]
        [InlineData("|c|3|y|-3|n")]
        [InlineData("|d|4|yes|-4|no")]
        [InlineData("|e|5|t|-5|f")]
        [InlineData("|f|6|+|-6|-")]
        [InlineData("|g|7|1")]
        public static void PositiveInvocations(string line)
        {
            var info = typeof(PostiveExamples).GetTypeInfo().GetDeclaredMethod("Invocation");
            Assert.NotNull(info);

            var instance = new PostiveExamples();

            var lexer = new CommandLineLexer(
                CommandLineLexerUnitTests.SplitWithFirstLineContent(line));

            var invoker = ReflectionInvoker.Create(info);
            invoker(lexer, instance);
            Assert.NotNull(instance.String);
            Assert.NotEqual(instance.Int, instance.NullableInt);
            Assert.True(instance.Int >= 0);
            Assert.NotEqual(instance.Bool, instance.NullableBool);
            Assert.True(instance.Bool);

            if (instance.NullableInt.HasValue)
            {
                Assert.Equal(instance.Int, -instance.NullableInt);
            }
            if (instance.NullableBool.HasValue)
            {
                Assert.Equal(instance.Bool, !instance.NullableBool);
            }
        }

        [Theory]
        [InlineData("|invalid bool|0|true|0|invalid")]
        [InlineData("|invalid  int|0|true|invalid|false")]
        [InlineData("|not enough  |3")]
        [System.Diagnostics.DebuggerHidden]
        public static void NegativeInvocations(string input)
        {
            Assert.Throws<ArgumentParserException>(() => PositiveInvocations(input));
        }


        private class NegativeExamples
        {
            public event Action InvalidMember;
            public NegativeExamples()
            {
                InvalidMember();
                InvalidType = new NotSupportedException(ReadOnly1);
                PrivateMethod(InvalidType);
            }
            private class Nested
            {
            }

            public readonly string ReadOnly1 = "InvalidMember";
            public NotSupportedException InvalidType;

            private string ReadOnly2 { get { return null; } set { } }
            private string ReadOnly3 { get { return null; } }

            private void PrivateMethod() { }
            public void PrivateMethod(NotSupportedException invalidType)
            {
                throw invalidType;
            }
        }

        [Theory]
        [MemberData("NEGATIVES")]
        [System.Diagnostics.DebuggerHidden]
        public static void Negative(MemberInfo info)
        {
            Assert.Throws<OptionDefinitionException>(() => ReflectionInvoker.Create(info));
        }
    }
}
