﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using Dyn.Tests.Execution;
using Xunit;

namespace Dyn.Tests.Execution
{
    public static class Discoverer
    {
        public static IEnumerable<T> Flatten<T>(this IEnumerable<IEnumerable<T>> source)
        {
            foreach (var inner in source)
            {
                foreach (var element in inner)
                {
                    yield return element;
                }
            }
        }

        public static IEnumerable<T> Flatten<T>(params IEnumerable<T>[] source)
        {
            foreach (var inner in source)
            {
                foreach (var element in inner)
                {
                    yield return element;
                }
            }
        }

        public static IEnumerable<Test> Group(this IEnumerable<Test> source, string name)
        {
            Guard.EnsureNotNull("name", name);

            foreach (var element in source)
            {
                if (element.Action == null)
                {
                    throw new ArgumentNullException("Element is not valid.");
                }

                var oldValue = element.Hierarchy;

                var newValue = string.IsNullOrEmpty(oldValue) ?
                    name :
                    oldValue + "." + name;

                yield return new Test(
                    element.Action,
                    element.Method,
                    element.Attribute,
                    newValue);
            }
        }

        public static IEnumerable<Test> SearchAssembly<T>()
        {
            return typeof(T).SearchAssembly();
        }

        public static IEnumerable<Test> SearchAssembly(this Type input)
        {
            return input.GetTypeInfo().SearchAssembly();
        }

        public static IEnumerable<Test> SearchAssembly(this TypeInfo input)
        {
            return input.Assembly.SearchAssembly();
        }

        public static IEnumerable<Test> SearchAssembly(this Assembly input)
        {
            Guard.EnsureNotNull("input", input);

            return input.DefinedTypes.SelectMany(Search);
        }

        public static IEnumerable<Test> Search<T>()
        {
            return typeof(T).GetTypeInfo().Search();
        }

        public static IEnumerable<Test> Search(this Type input)
        {
            Guard.EnsureNotNull("input", input);

            return input.GetTypeInfo().Search();
        }

        public static IEnumerable<Test> Search(this TypeInfo input)
        {
            Guard.EnsureNotNull("input", input);

            return input.DeclaredMethods.SelectMany(Search);
        }

        private static IEnumerable<Test> Search(this MethodInfo input)
        {
            var fact = input.GetCustomAttribute<FactAttribute>(false);
            var theory = fact as TheoryAttribute;

            if(fact == null)
            {
                yield break;
            }

            if (theory == null)
            {
                yield return new Test(
                    () => input.Invoke(null, null),
                    input,
                    fact,
                    string.Empty);
                yield break;
            }

            foreach (var row in input
                .GetCustomAttributes<Xunit.Sdk.DataAttribute>()
                .SelectMany(e => e.GetData(input)))
            {
                yield return new Test(
                    () => input.Invoke(null, row),
                    input,
                    theory,
                    string.Empty);
            }
        }
    }
}
