﻿using Dyn;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Xunit;

namespace tests.Lexer
{
    public static class DynLexerUnitTests
    {
        public static SyntaxToken[] ALL_TOKENS = Tokens.KINDS.SelectMany(Tokens.GetSamples).ToArray();

        private static int IncompatibilityEquivalenceGroups(SyntaxKind kind)
        {
            // positive number represents: comptatible with itself
            // otherwise negative
            switch (kind)
            {
                case SyntaxKind.IdentifierKind:
                case SyntaxKind.NumericKind:
                case SyntaxKind.SelfKeyword:
                case SyntaxKind.NewKeyword:
                case SyntaxKind.IfKeyword:
                case SyntaxKind.ThenKeyword:
                case SyntaxKind.ElseKeyword:
                case SyntaxKind.EndKeyword:
                case SyntaxKind.WhileKeyword:
                case SyntaxKind.DoKeyword:
                case SyntaxKind.DoneKeyword:
                case SyntaxKind.MethodKeyword:
                case SyntaxKind.RenameKeyword:
                case SyntaxKind.NullKeyword:
                case SyntaxKind.ClassKeyword:
                case SyntaxKind.RequiresKeyword:
                case SyntaxKind.EnsuresKeyword:
                case SyntaxKind.ForAllKeyword:
                case SyntaxKind.ExistsKeyword:
                case SyntaxKind.InKeyword:
                case SyntaxKind.IsAKeyword:
                case SyntaxKind.TrueKeyword:
                case SyntaxKind.FalseKeyword:
                case SyntaxKind.AssertKeyword:
                case SyntaxKind.InvariantKeyword:
                case SyntaxKind.ImpliesOperator:
                case SyntaxKind.ByTheWayKeyword:
                    return -(int)SyntaxKind.IdentifierKind;

                case SyntaxKind.AssignmentOperator:
                case SyntaxKind.EqualityOperator:
                case SyntaxKind.IdentityOperator:
                case SyntaxKind.LessThanOperator:
                case SyntaxKind.GreaterThanOperator:
                case SyntaxKind.LessThanOrEqualOperator:
                case SyntaxKind.GreaterThanOrEqualOperator:
                    return -(int)SyntaxKind.AssignmentOperator;

                case SyntaxKind.AndOperator:
                case SyntaxKind.OrOperator:
                    return -(int)SyntaxKind.AndOperator;

                case SyntaxKind.AdditionOperator:
                case SyntaxKind.SubtractionOperator:
                case SyntaxKind.MultiplicationOperator:
                case SyntaxKind.DivisionOperator:
                case SyntaxKind.ModulosOperator:
                    return (int)kind;

                case SyntaxKind.ParenthesisLeftSymbol:
                case SyntaxKind.ParenthesisRightSymbol:
                case SyntaxKind.SquareBracketLeftSymbol:
                case SyntaxKind.SquareBracketRightSymbol:
                case SyntaxKind.CurlyBracketLeftSymbol:
                case SyntaxKind.CurlyBracketRightSymbol:
                case SyntaxKind.AtSymbol:
                case SyntaxKind.DotSymbol:
                case SyntaxKind.CommaSymbol:
                case SyntaxKind.SemicolonSymbol:
                case SyntaxKind.QuestionMarkSymbol:
                case SyntaxKind.ExclamationMarkSymbol:
                case SyntaxKind.ColonSymbol:
                case SyntaxKind.ApostropheSymbol:
                    return (int)kind;

                case SyntaxKind.EndOfStreamKind:
                    return -(int)kind;

                case SyntaxKind.InvalidKind:
                case SyntaxKind.StringKind:
                case SyntaxKind.WhitespaceKind:
                    return -(int)kind;

                case SyntaxKind.CommentKind:
                    return (int)kind;

                default:
                    throw Guard.EnumOutOfRangeException<SyntaxKind>("kind", (int)kind);
            }
        }

        public static bool AreCompatible(SyntaxKind current, SyntaxKind next)
        {
            if (current == SyntaxKind.EndOfStreamKind)
            {
                return false;
            }

            if (current == next)
            {
                return IncompatibilityEquivalenceGroups(current) > 0;
            }

            return IncompatibilityEquivalenceGroups(current) != IncompatibilityEquivalenceGroups(next);
        }

        private static void AssertLexer(String fileName, params SyntaxToken[] source)
        {
            var builder = new StringBuilder();

            foreach (var e in source)
            {
                builder.Append(e.ToSource());
            }

            using (var reader = new StringReader(builder.ToString()))
            {
                var expectedLength = source.Length;

                var actual = new DynLexer(fileName, reader).ToList();

                Assert.Equal(SyntaxKind.EndOfStreamKind, actual.Last().Kind);
                Assert.Equal(source, actual.Take(source.Length).ToArray());
            }
        }

        [Fact]
        public static void EmptyString()
        {
            const string FILE_NAME = "EmptyString";

            using (var reader = new StringReader(""))
            {
                Assert.Equal(new SyntaxToken[] { new SyntaxToken(SyntaxKind.EndOfStreamKind, new Span(FILE_NAME, 0, 0, 0, 0)) }, new DynLexer(FILE_NAME, reader));
            }
        }

        [Fact]
        public static void Single()
        {
            const string FILENAME = "Single";
            foreach (var t1 in ALL_TOKENS)
            {
                if (t1.Kind != SyntaxKind.EndOfStreamKind)
                {
                    AssertLexer(FILENAME, Tokens.ModifySpan(FILENAME, t1));
                }
            }
        }

        [Fact]
        public static void Double()
        {
            const string FILENAME = "Double";

            foreach (var t1 in ALL_TOKENS)
            {
                foreach (var t2 in ALL_TOKENS)
                {
                    if (AreCompatible(t1.Kind, t2.Kind))
                    {
                        AssertLexer(FILENAME, Tokens.ModifySpan(FILENAME, t1, t2));
                    }
                }
            }
        }

        [Fact(Skip = "Method became too slow, because there are too many combinations since Operator & Symbol & Keyword inlining.")]
        public static void Triple()
        {
            const string FILENAME = "Triple";

            foreach (var t1 in ALL_TOKENS)
            {
                foreach (var t2 in ALL_TOKENS)
                {
                    if (AreCompatible(t1.Kind, t2.Kind))
                    {
                        foreach (var t3 in ALL_TOKENS)
                        {
                            if (AreCompatible(t2.Kind, t3.Kind))
                            {
                                AssertLexer(FILENAME, Tokens.ModifySpan(FILENAME, t1, t2, t3));
                            }
                        }
                    }
                }
            }
        }

        public static void KeywordsUpperCase()
        {
            foreach (var element in Tokens.KINDS.Where(e => Enum.GetName(typeof(SyntaxKind), e).EndsWith("Keyword")))
            {
                var token = new SyntaxToken(element, new Span(string.Empty, 0, 0, 0, Enum.GetName(typeof(SyntaxKind), element).Length));

                var input = token.ToSource().ToUpperInvariant();

                var l = new DynLexer(input).ToList();

                Assert.True(l.Count == 1);

                Assert.Equal(new Span(string.Empty, 0, 0, 0, input.Length), l[0].Span);
                Assert.NotEqual(element, l[0].Kind);
                Assert.NotEqual(token.ToSource(), l[0].ToSource());
                Assert.Equal(input, l[0].ToSource());
            }
        }

        [Theory]
        [InlineData(8192)]
        [InlineData(16384)]
        [InlineData(32768)]
        public static void Random(int count)
        {
            const string FILENAME = "Random";
            AssertLexer(FILENAME, Tokens.ModifySpan(FILENAME, GenerateTokens(FILENAME, count, count, AreCompatible).ToArray()));
        }

        public static IEnumerable<SyntaxToken> GenerateTokens(string fileName, int seed, int count, Func<SyntaxKind, SyntaxKind, bool> followUp)
        {
            var rnd = new Random(seed);

            var prev = Tokens.KINDS.Random(rnd);

            var normalKinds = Tokens.KINDS.Where(e => e != SyntaxKind.EndOfStreamKind).ToArray();

            var sane = normalKinds.ToDictionary(e => e, e => normalKinds.Where(f => followUp(e, f)).ToArray());

            for (int i = 0; i < count; ++i)
            {
                var current = sane[prev].Random(rnd);

                Assert.True(AreCompatible(prev, current));

                var result = Tokens.GetSamples(current).ToArray().Random(rnd);

                yield return result;

                prev = current;
            }
        }

        [Fact]
        public static void HelpSpecial()
        {
            const string FILENAME = "HelpSpecial";
            AssertLexer(FILENAME, Tokens.ModifySpan(FILENAME,
                    new SyntaxToken(SyntaxKind.ParenthesisLeftSymbol, default(Span)),
                    new SyntaxToken(SyntaxKind.ParenthesisRightSymbol, default(Span))));
        }
    }
}
