﻿using Dyn;
using Dyn.Lexer;
using System.Collections.Generic;
using System.IO;
using System.Text;
using Xunit;

namespace tests.Lexer
{
    public static class SourceReaderUnitTests
    {
        private static readonly string[] FILENAMES = new string[] { string.Empty, " ", "Word", "This is a sentence!" };

        public static IEnumerable<object[]> XunitLongString(int count)
        {
            yield return new object[1] { GetLongString(count) };
        }

        public static string GetLongString(int count)
        {
            const string INGRIENTS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";

            Guard.EnsureNonnegative("count", count);

            if (count == 0)
            {
                return "";
            }

            if (count < INGRIENTS.Length)
            {
                return INGRIENTS.Substring(0, count);
            }
            if (count == INGRIENTS.Length)
            {
                return INGRIENTS;
            }

            var builder = new StringBuilder(INGRIENTS);

            while (builder.Length * 2 <= count)
            {
                builder.Append(builder);
            }

            if (builder.Length < count)
            {
                builder.Append(builder.ToString(), 0, count - builder.Length);
            }

            return builder.ToString();
        }

        [Theory]
        [InlineData("")]
        [InlineData(" ")]
        [InlineData(".~|<")]
        [InlineData(" A somewhat log filename without a path seperator ")]
        public static void FileNames(string fileNames)
        {
            using (var reader = new StringReader(string.Empty))
            {
                foreach (var fn in FILENAMES)
                {
                    var srh = new SourceReaderHelper(fn, reader);

                    Assert.Equal(fn, srh.FileName);

                    Assert.False(srh.TryNext());

                    Assert.Equal(fn, srh.FileName);
                }
            }
        }

        [Theory]
        [MemberData("XunitLongString", 0)]
        [MemberData("XunitLongString", 1000)]
        [MemberData("XunitLongString", 10000)]
        [MemberData("XunitLongString", 100000)]
        public static void TestPosition(string content)
        {
            const string FILENAME = "TestPosition";

            using (var reader = new StringReader(content))
            {
                var srh = new SourceReaderHelper(FILENAME, reader);

                Assert.True(srh.BeforeContent);

                for (int i = 0; i < content.Length; ++i)
                {
                    Assert.True(srh.TryNext());

                    Assert.Equal(content[i], srh.Current);
                    Assert.Equal(new Span(FILENAME, i, 0, i, 0), srh.CreateSpan());
                    Assert.True(srh.AtContent);
                }

                Assert.False(srh.TryNext());
                Assert.True(srh.AfterContent);
            }
        }

        [Fact]
        public static void EmptyNewLines()
        {
            const string FILENAME = "EmptyNewLines";

            string content = new string('\n', 4000);

            using (var reader = new StringReader(content))
            {
                var srh = new SourceReaderHelper(FILENAME, reader);

                for (int i = 0; i < content.Length; ++i)
                {
                    Assert.True(srh.TryNext());

                    Assert.Equal(content[i], srh.Current);
                    Assert.Equal(new Span(FILENAME, i, i, 0, 0), srh.CreateSpan());
                }

                Assert.False(srh.TryNext());
            }
        }

        [Fact]
        public static void RectangleNewLines()
        {
            const int LINES = 256;
            const int LINE_LENGTH = 128;

            var builder = new StringBuilder(LINES * LINE_LENGTH);

            var line = GetLongString(LINE_LENGTH - 1) + "\n";

            for (int i = 0; i < LINES; ++i)
            {
                builder.Append(line);
            }

            var srh = new SourceReaderHelper("Ractangle", new StringReader(builder.ToString()));

            for (int l = 0; l < LINES; ++l)
            {
                for (int c = 0; c < LINE_LENGTH - 1; ++c)
                {
                    Assert.True(srh.TryNext());

                    Assert.NotEqual(srh.Current, '\n');

                    Assert.Equal(srh.CreateSpan(), new Span("Ractangle", l * LINE_LENGTH + c, l, c, 0));
                }

                Assert.True(srh.TryNext());
                Assert.Equal(srh.Current, '\n');
                Assert.Equal(srh.CreateSpan(), new Span("Ractangle", (l + 1) * LINE_LENGTH - 1, l, LINE_LENGTH - 1, 0));
            }

            Assert.False(srh.TryNext());
        }

        [Theory]
        [InlineData("test\r\n\tHere\tit\ncomes!", 21, 2, 6)]
        [InlineData("", 0, 0, 0)]
        [InlineData("$\n", 2, 1, 0)]
        public static void SpecialTests(string content, int pos, int line, int column)
        {
            const string FILENAME = "Special";

            using (var reader = new StringReader(content))
            {
                var srh = new SourceReaderHelper(FILENAME, reader);

                Span s = srh.CreateSpan();
                while (srh.TryNext())
                {
                    s = srh.CreateSpan();
                }

                Assert.True(srh.AfterContent);
                Assert.Equal(new Span(FILENAME, pos, line, column, 0), srh.CreateSpan());
            }
        }
    }
}
