﻿using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Linq;
using Xunit;

namespace tests.Lexer
{
    public static class SyntaxTokenUnitTests
    {
        public static readonly SyntaxKind[] KINDS = new SyntaxKind[] {
            // End of stream
            SyntaxKind.EndOfStreamKind,

            // Kinds with string value attached
            SyntaxKind.InvalidKind,
            SyntaxKind.WhitespaceKind,
            SyntaxKind.IdentifierKind,
            SyntaxKind.StringKind,
            SyntaxKind.CommentKind,

            // Kind with numeric value attached
            SyntaxKind.NumericKind,

            // Operators
            SyntaxKind.AssignmentOperator,
            SyntaxKind.EqualityOperator,
            SyntaxKind.IdentityOperator,
            SyntaxKind.AdditionOperator,
            SyntaxKind.SubtractionOperator,
            SyntaxKind.MultiplicationOperator,
            SyntaxKind.DivisionOperator,
            SyntaxKind.ModulosOperator,
            SyntaxKind.LessThanOperator,
            SyntaxKind.GreaterThanOperator,
            SyntaxKind.LessThanOrEqualOperator,
            SyntaxKind.GreaterThanOrEqualOperator,
            SyntaxKind.AndOperator,
            SyntaxKind.OrOperator,
            SyntaxKind.ImpliesOperator,

            // Symbols
            SyntaxKind.ParenthesisLeftSymbol,
            SyntaxKind.ParenthesisRightSymbol,
            SyntaxKind.SquareBracketLeftSymbol,
            SyntaxKind.SquareBracketRightSymbol,
            SyntaxKind.CurlyBracketLeftSymbol,
            SyntaxKind.CurlyBracketRightSymbol,
            SyntaxKind.AtSymbol,
            SyntaxKind.DotSymbol,
            SyntaxKind.CommaSymbol,
            SyntaxKind.SemicolonSymbol,
            SyntaxKind.QuestionMarkSymbol,
            SyntaxKind.ExclamationMarkSymbol,
            SyntaxKind.ColonSymbol,

            // Keywords
            SyntaxKind.SelfKeyword,
            SyntaxKind.NewKeyword,
            SyntaxKind.IfKeyword,
            SyntaxKind.ThenKeyword,
            SyntaxKind.ElseKeyword,
            SyntaxKind.EndKeyword,
            SyntaxKind.WhileKeyword,
            SyntaxKind.DoKeyword,
            SyntaxKind.DoneKeyword,
            SyntaxKind.MethodKeyword,
            SyntaxKind.RenameKeyword,
            SyntaxKind.NullKeyword,
            SyntaxKind.ClassKeyword,
            SyntaxKind.FalseKeyword,
            SyntaxKind.TrueKeyword,
            SyntaxKind.IsAKeyword,

            // Assertion language
            SyntaxKind.RequiresKeyword,
            SyntaxKind.EnsuresKeyword,
            SyntaxKind.InvariantKeyword,
            SyntaxKind.AssertKeyword,
            SyntaxKind.ByTheWayKeyword,

            SyntaxKind.ForAllKeyword,
            SyntaxKind.ExistsKeyword,
            SyntaxKind.InKeyword,
            SyntaxKind.ApostropheSymbol,
        };

        public static readonly IEnumerable<object[]> TEST_KINDS = KINDS.Select(e => new object[] { e }).ToArray();

        [Fact]
        public static void SelfTest()
        {
            AssertExtensions.SetEqual(Utils.GetAllValuesEnum<SyntaxKind>(), KINDS, e => Enum.GetName(typeof(SyntaxKind), e));
        }

        [Fact]
        public static void EmptyConstructorIsEos()
        {
            var empty = new SyntaxToken();

            Assert.Equal(SyntaxKind.EndOfStreamKind, empty.Kind);
            Assert.Equal("", empty.ToSource());
            Assert.Equal(" (0:0)", empty.ToString());
            Assert.Throws<InvalidOperationException>(() => empty.Number);
            Assert.Throws<InvalidOperationException>(() => empty.String);
        }

        [Fact(Skip = "Noisy during debug mode")]
        public static void StringConstructorThrows()
        {
            Assert.Throws<ArgumentNullException>(() => new SyntaxToken(SyntaxKind.StringKind, new Span(), null));
        }

        [Theory]
        [MemberData("TEST_KINDS")]
        public static void Simple(SyntaxKind kind)
        {
            const string SPAN_FORMATTER = " ({0})";
            Func<Span, string> sf = e => string.Format(SPAN_FORMATTER, e);

            Span SPAN = default(Span);

            switch (kind)
            {
                case SyntaxKind.InvalidKind:
                case SyntaxKind.WhitespaceKind:
                case SyntaxKind.IdentifierKind:
                case SyntaxKind.StringKind:
                case SyntaxKind.CommentKind:
                    {
                        Assert.Throws<InvalidOperationException>(() => new SyntaxToken(kind, SPAN, 0));
                        foreach (var n in Tokens.GetStringSamples(kind))
                        {
                            var span = new Span(string.Empty, 0, 0, 0, n.Length);
                            var current = new SyntaxToken(kind, span, n);

                            Assert.Equal(span, current.Span);
                            Assert.Equal(kind, current.Kind);
                            Assert.Equal(n, current.ToSource());
                            Assert.Equal(kind.ToString() + ": " + n.ToString() + sf(span), current.ToString());
                            Assert.Equal(n, current.String);
                            Assert.Throws<InvalidOperationException>(() => current.Number);
                        }
                        break;
                    }

                case SyntaxKind.NumericKind:
                    {
                        Assert.Throws<InvalidOperationException>(() => new SyntaxToken(kind, SPAN, ""));
                        foreach (var n in Tokens.GetIntSamples(kind))
                        {
                            var span = new Span(string.Empty, 0, 0, 0, n.ToString().Length);
                            var current = new SyntaxToken(kind, span, n);

                            Assert.Equal(span, current.Span);
                            Assert.Equal(kind, current.Kind);
                            Assert.Equal(n.ToString(), current.ToSource());
                            Assert.Equal(n.ToString() + sf(span), current.ToString());

                            Assert.Equal(n, current.Number);
                            Assert.Throws<InvalidOperationException>(() => current.String);
                        }

                        break;
                    }
                default:
                    {
                        Assert.Throws<InvalidOperationException>(() => new SyntaxToken(kind, SPAN, 0));
                        Assert.Throws<InvalidOperationException>(() => new SyntaxToken(kind, SPAN, ""));

                        var span = new Span(string.Empty, 0, 0, 0, 13);
                        var current = new SyntaxToken(kind, span);

                        Assert.Equal(kind, current.Kind);

                        // different test
                        Assert.Equal(span, current.Span);
                        Assert.EndsWith(sf(span), current.ToString());
                        Assert.DoesNotContain(kind.ToString(), current.ToString());
                        Assert.Throws<InvalidOperationException>(() => current.Number);
                        Assert.Throws<InvalidOperationException>(() => current.String);
                        break;
                    }
            }
        }

        [Fact]
        public static void ToSourceTests()
        {
            Func<string, SyntaxKind, KeyValuePair<string, SyntaxKind>> f = (k, v) => new KeyValuePair<string, SyntaxKind>(k, v);

            var postfix = " (" + default(Span) + ")";

            var kvp = new KeyValuePair<string, SyntaxKind>[] {
                f("",SyntaxKind.EndOfStreamKind),
                f("=",SyntaxKind.AssignmentOperator),
                f("==",SyntaxKind.EqualityOperator),
                f("===",SyntaxKind.IdentityOperator),
                f("+",SyntaxKind.AdditionOperator),
                f("-",SyntaxKind.SubtractionOperator),
                f("*",SyntaxKind.MultiplicationOperator),
                f("/",SyntaxKind.DivisionOperator),
                f("<",SyntaxKind.LessThanOperator),
                f(">",SyntaxKind.GreaterThanOperator),
                f("<=",SyntaxKind.LessThanOrEqualOperator),
                f(">=",SyntaxKind.GreaterThanOrEqualOperator),
                f("&&",SyntaxKind.AndOperator),
                f("||",SyntaxKind.OrOperator),
                f("(",SyntaxKind.ParenthesisLeftSymbol),
                f(")",SyntaxKind.ParenthesisRightSymbol),
                f("[",SyntaxKind.SquareBracketLeftSymbol),
                f("]",SyntaxKind.SquareBracketRightSymbol),
                f("{",SyntaxKind.CurlyBracketLeftSymbol),
                f("}",SyntaxKind.CurlyBracketRightSymbol),
                f("@",SyntaxKind.AtSymbol),
                f(".",SyntaxKind.DotSymbol),
                f(",",SyntaxKind.CommaSymbol),
                f(";",SyntaxKind.SemicolonSymbol),
                f("?",SyntaxKind.QuestionMarkSymbol),
                f("!",SyntaxKind.ExclamationMarkSymbol),
                f(":",SyntaxKind.ColonSymbol),
                f("'",SyntaxKind.ApostropheSymbol),
                f("%",SyntaxKind.ModulosOperator),
                f("self",SyntaxKind.SelfKeyword),
                f("new",SyntaxKind.NewKeyword),
                f("if",SyntaxKind.IfKeyword),
                f("then",SyntaxKind.ThenKeyword),
                f("else",SyntaxKind.ElseKeyword),
                f("end",SyntaxKind.EndKeyword),
                f("while",SyntaxKind.WhileKeyword),
                f("do",SyntaxKind.DoKeyword),
                f("done",SyntaxKind.DoneKeyword),
                f("method",SyntaxKind.MethodKeyword),
                f("rename",SyntaxKind.RenameKeyword),
                f("null",SyntaxKind.NullKeyword),
                f("class",SyntaxKind.ClassKeyword),
                f("requires",SyntaxKind.RequiresKeyword),
                f("ensures",SyntaxKind.EnsuresKeyword),
                f("forall",SyntaxKind.ForAllKeyword),
                f("exists",SyntaxKind.ExistsKeyword),
                f("in",SyntaxKind.InKeyword),
                f("is_a",SyntaxKind.IsAKeyword),
                f("true",SyntaxKind.TrueKeyword),
                f("false",SyntaxKind.FalseKeyword),
            };

            Assert.All(kvp, (e) => KINDS.Contains(e.Value));

            Assert.All(kvp, (e) => Assert.Equal(e.Key, new SyntaxToken(e.Value, default(Span)).ToSource()));
            Assert.All(kvp, (e) => Assert.Equal(e.Key + postfix, new SyntaxToken(e.Value, default(Span)).ToString()));
        }
    }
}
