﻿using Dyn;
using Dyn.Lexer;
using System.Collections.Generic;
using System.Linq;

namespace tests
{
    internal static class Tokens
    {
        public static readonly int[] NUMBERS = new int[] { 0, 1, 10, 11, 2, 4, 8 };

        public static readonly string[] INVALIDS = new string[] { "§", "^", @"\", "~", "üäö" };

        public static readonly string[] IDENTIFIERS = new string[] { "something", "some_identifier", "identifier", "tmp", "foobar", "long_class_name", "CamelCaseIdentifier", "Mixed_Camel_With_Snake" };

        public static readonly string[] WHITESPACES = new string[] { " ", "  ", "   ", "    ", "\t", "\n", "\n\r", "\r\n" };

        public static readonly string[] STRINGS = new string[] { "\"\"", "\" \"", "\"String\"", "\"42\"", "\"&$|~\"", "\"Badger\"", "\"\n\t\r\t\r\"", '"' + @"
//        " + '"' };

        public static readonly string[] COMMENTS = new string[] { "#\n", "##\n", "### \n", "# In a galaxy ...\n", @"# Placeholder
", "# &%$~|\n" };

        public static readonly SyntaxKind[] KINDS = Utils.GetAllValuesEnum<SyntaxKind>();

        public static void CountLineAndColumn(string value, ref int line, ref int column)
        {
            foreach (var c in value)
            {
                if (c == '\n')
                {
                    line++;
                    column = 0;
                }
                else
                {
                    column++;
                }
            }
        }

        public static SyntaxToken[] ModifySpan(string fileName, params SyntaxToken[] tokens)
        {
            SyntaxToken[] result = (SyntaxToken[])tokens.Clone();

            int line = 0, column = 0, pos = 0;

            for (int i = 0; i < result.Length; ++i)
            {
                var currentSource = result[i].ToSource();

                var currentLength = currentSource.Length;

                result[i] = ReplaceSpan(result[i], new Span(fileName, pos, line, column, currentLength));

                Tokens.CountLineAndColumn(currentSource, ref line, ref column);

                pos += currentLength;
            }

            return result;
        }

        private static SyntaxToken ReplaceSpan(SyntaxToken source, Span span)
        {
            switch (source.Kind)
            {
                case SyntaxKind.InvalidKind:
                case SyntaxKind.WhitespaceKind:
                case SyntaxKind.IdentifierKind:
                case SyntaxKind.StringKind:
                case SyntaxKind.CommentKind:
                    return new SyntaxToken(source.Kind, span, source.String);

                case SyntaxKind.NumericKind:
                    return new SyntaxToken(source.Kind, span, source.Number);

                default:
                    return new SyntaxToken(source.Kind, span);
            }
        }

        public static IEnumerable<SyntaxToken> GetSamples(SyntaxKind kind)
        {
            switch (kind)
            {
                case SyntaxKind.InvalidKind:
                    return INVALIDS.Select(e => new SyntaxToken(kind, default(Span), e));

                case SyntaxKind.WhitespaceKind:
                    return WHITESPACES.Select(e => new SyntaxToken(kind, default(Span), e));

                case SyntaxKind.IdentifierKind:
                    return IDENTIFIERS.Select(e => new SyntaxToken(kind, default(Span), e));

                case SyntaxKind.StringKind:
                    return STRINGS.Select(e => new SyntaxToken(kind, default(Span), e));

                case SyntaxKind.CommentKind:
                    return COMMENTS.Select(e => new SyntaxToken(kind, default(Span), e));

                case SyntaxKind.NumericKind:
                    return NUMBERS.Select(e => new SyntaxToken(kind, default(Span), e));

                case SyntaxKind.EndOfStreamKind:
                case SyntaxKind.SelfKeyword:
                case SyntaxKind.NewKeyword:
                case SyntaxKind.IfKeyword:
                case SyntaxKind.ThenKeyword:
                case SyntaxKind.ElseKeyword:
                case SyntaxKind.EndKeyword:
                case SyntaxKind.WhileKeyword:
                case SyntaxKind.DoKeyword:
                case SyntaxKind.DoneKeyword:
                case SyntaxKind.MethodKeyword:
                case SyntaxKind.RenameKeyword:
                case SyntaxKind.NullKeyword:
                case SyntaxKind.ClassKeyword:
                case SyntaxKind.RequiresKeyword:
                case SyntaxKind.EnsuresKeyword:
                case SyntaxKind.ForAllKeyword:
                case SyntaxKind.ExistsKeyword:
                case SyntaxKind.ByTheWayKeyword:
                case SyntaxKind.InKeyword:
                case SyntaxKind.IsAKeyword:
                case SyntaxKind.TrueKeyword:
                case SyntaxKind.FalseKeyword:
                case SyntaxKind.InvariantKeyword:
                case SyntaxKind.AssertKeyword:
                case SyntaxKind.AssignmentOperator:
                case SyntaxKind.EqualityOperator:
                case SyntaxKind.IdentityOperator:
                case SyntaxKind.AdditionOperator:
                case SyntaxKind.SubtractionOperator:
                case SyntaxKind.MultiplicationOperator:
                case SyntaxKind.DivisionOperator:
                case SyntaxKind.ModulosOperator:
                case SyntaxKind.LessThanOperator:
                case SyntaxKind.GreaterThanOperator:
                case SyntaxKind.LessThanOrEqualOperator:
                case SyntaxKind.GreaterThanOrEqualOperator:
                case SyntaxKind.AndOperator:
                case SyntaxKind.OrOperator:
                case SyntaxKind.ImpliesOperator:
                case SyntaxKind.ParenthesisLeftSymbol:
                case SyntaxKind.ParenthesisRightSymbol:
                case SyntaxKind.SquareBracketLeftSymbol:
                case SyntaxKind.SquareBracketRightSymbol:
                case SyntaxKind.CurlyBracketLeftSymbol:
                case SyntaxKind.CurlyBracketRightSymbol:
                case SyntaxKind.AtSymbol:
                case SyntaxKind.DotSymbol:
                case SyntaxKind.CommaSymbol:
                case SyntaxKind.SemicolonSymbol:
                case SyntaxKind.QuestionMarkSymbol:
                case SyntaxKind.ExclamationMarkSymbol:
                case SyntaxKind.ColonSymbol:
                case SyntaxKind.ApostropheSymbol:
                    return new SyntaxToken[] { new SyntaxToken(kind, default(Span)) };

                default:
                    throw Guard.EnumOutOfRangeException<SyntaxKind>("kind", (int)kind);
            }
        }

        public static IEnumerable<int> GetIntSamples(SyntaxKind kind)
        {
            switch (kind)
            {
                case SyntaxKind.NumericKind:
                    return NUMBERS;

                default:
                    throw Guard.EnumOutOfRangeException<SyntaxKind>("kind", (int)kind);
            }
        }

        public static IEnumerable<string> GetStringSamples(SyntaxKind kind)
        {
            switch (kind)
            {
                case SyntaxKind.InvalidKind:
                    return INVALIDS;

                case SyntaxKind.WhitespaceKind:
                    return WHITESPACES;

                case SyntaxKind.IdentifierKind:
                    return IDENTIFIERS;

                case SyntaxKind.StringKind:
                    return STRINGS;

                case SyntaxKind.CommentKind:
                    return COMMENTS;

                default:
                    throw Guard.EnumOutOfRangeException<SyntaxKind>("kind", (int)kind);
            }
        }
    }
}
