﻿using Dyn.Ast;
using Dyn.Lexer;
using Dyn.Parser;
using System;
using System.Collections.Generic;
using System.Linq;
using tests.Ast;
using Xunit;

namespace tests.Parser
{
    public static class DynParserUnitTests
    {
        [Theory]
        [InlineData("class a {}")]
        [InlineData("class a < b {}")]
        [InlineData("class a < b { rename c d method e (f, g, h, i, j) { }}")]
        public static void TestClass(string source)
        {
            Assert.NotNull(source.Tokenize().Parse(true, p => p.ParseClass()));
        }

        public static IEnumerable<T> Fuzzy<T>(AstFuzzer fuzzer, int count, int seed)
        {
            Random r = new Random(seed);

            for (int i = 0; i < count; ++i)
            {
                yield return fuzzer.Fuzzy<T>(new Random(r.Next()));
            }
        }

        [Theory]
        [InlineData(100)]
        [InlineData(200)]
        public static void Random(int count)
        {
            Assert.All(Fuzzy<DynProgramClassNode>(DynFuzzer.Programs, count, -count), element =>
                    AssertPrinterLexerParser(
                             element,
                             DynNodeFormatter.ToSource,
                             p => p.ParseClass(),
                             AstExtensions.IgnoreLocation));
            Assert.All(Fuzzy<DynProgramExpressionNode>(DynFuzzer.Programs, count, -count), element =>
                    AssertPrinterLexerParser(
                             element,
                             DynNodeFormatter.ToSource,
                             p => p.ParseProgramExpression(),
                             AstExtensions.IgnoreLocation));
            Assert.All(Fuzzy<DynAssertionExpressionNode>(DynFuzzer.Assertions, count, -count), element =>
                    AssertPrinterLexerParser(
                             element,
                             DynNodeFormatter.ToSource,
                             p => p.ParseAssertionExpression(),
                             AstExtensions.IgnoreLocation));
        }

        public static void AssertPrinterLexerParser<T>(T node, Func<T, string> formatter, Func<DynParser, T> parser, IEqualityComparer<T> comparer)
        {
            var nodeSource = formatter(node);
            var nodeSourceToken = nodeSource.Tokenize();
            var nodeSourceTokenNode = nodeSourceToken.Parse(true, parser);
            var nodeSourceTokenNodeSource = formatter(nodeSourceTokenNode);
            var nodeSourceTokenNodeSourceToken = nodeSourceTokenNodeSource.Tokenize();

            AssertExtensions.TokensEqual(nodeSourceToken, nodeSourceTokenNodeSourceToken);
            Assert.Equal(nodeSource.NormalizeWhitespace(), nodeSourceTokenNodeSource.NormalizeWhitespace());
            Assert.Equal(node, nodeSourceTokenNode, comparer);
        }

        public static void AssertLexerParserPrinter<T>(IEnumerable<SyntaxToken> sourceToken, Func<DynParser, T> parser, IEqualityComparer<T> comparer, Func<T, string> formatter)
        {
            var sourceTokenNode = sourceToken.Parse(true, parser);
            var sourceTokenNodeSource = formatter(sourceTokenNode);
            var sourceTokenNodeSourceToken = sourceTokenNodeSource.Tokenize();
            var sourceTokenNodeSourceTokenNode = sourceTokenNodeSourceToken.Parse(true, parser);

            AssertExtensions.TokensEqual(sourceToken, sourceTokenNodeSourceToken);
            Assert.Equal(sourceTokenNode, sourceTokenNodeSourceTokenNode, comparer);
        }

        private class EnumerableEqualityComparer<T> : IEqualityComparer<IEnumerable<T>>
        {
            private readonly IEqualityComparer<T> _inner;

            public EnumerableEqualityComparer(IEqualityComparer<T> inner)
            {
                _inner = inner;
            }

            public bool Equals(IEnumerable<T> x, IEnumerable<T> y)
            {
                return x.SequenceEqual(y, _inner);
            }

            public int GetHashCode(IEnumerable<T> obj)
            {
                return obj.Aggregate(13, (result, e) => result * 127 + _inner.GetHashCode(e));
            }
        }

        [Theory]
        [MemberData("Valid", MemberType = typeof(SampleHelpers))]
        public static void SampleTests(Sample sample)
        {
            AssertLexerParserPrinter(
                sample.Tokenize().TransformCommentToAssertion().AddWhileInvariant().Simplify().ToList(),
                p => p.ParseAll(q => q.ParseClass()).ToList(),
                new EnumerableEqualityComparer<DynProgramClassNode>(AstExtensions.IgnoreLocation),
                (IEnumerable<DynProgramClassNode> e) => e.FormatAll(DynNodeFormatter.ToSource));
        }
    }
}
