﻿using Dyn.Ast;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using Xunit;

namespace tests.Parser
{
    // "(a - b) - c"  == "a - b - c" != "a - (b - c)"
    // "a == (b == c)" == "a == b == c" ...
    // "a = (b = c)" == "a = b = c" ...

    public static class PrinterUnitTests
    {
        private static readonly DynAssertionExpressionNode r = new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.AssertionVariable, "r");
        private static readonly Span e = Span.Empty;

        private static readonly DynAssertionExpressionNode n1 = new DynAssertionNumericNode(e, 1);
        private static readonly DynAssertionExpressionNode n2 = new DynAssertionNumericNode(e, 2);
        private static readonly DynAssertionExpressionNode n3 = new DynAssertionNumericNode(e, 3);

        private static readonly DynProgramVariableNode a = new DynProgramVariableNode(e, DynProgramStatementNode.NodeKind.Variable, "a");
        private static readonly DynProgramVariableNode b = new DynProgramVariableNode(e, DynProgramStatementNode.NodeKind.Variable, "b");
        private static readonly DynProgramVariableNode c = new DynProgramVariableNode(e, DynProgramStatementNode.NodeKind.Variable, "c");
        private static readonly DynProgramVariableNode d = new DynProgramVariableNode(e, DynProgramStatementNode.NodeKind.Variable, "d");

        [System.Diagnostics.DebuggerHidden]
        public static DynProgramExpressionNode And(this DynProgramExpressionNode first, DynProgramExpressionNode second)
        {
            return new DynProgramMethodCallNode(e, first, OperatorTranslation.KindToMethod(SyntaxKind.AndOperator), Dyn.Utilities.AsArray(second));
        }

        [System.Diagnostics.DebuggerHidden]
        public static DynAssertionExpressionNode Neg(this DynAssertionExpressionNode first)
        {
            return new DynAssertionUnaryNode(e, first);
        }

        [System.Diagnostics.DebuggerHidden]
        public static DynAssertionInstanceVariableNode InstanceVariable(this DynAssertionExpressionNode expression, string name)
        {
            return new DynAssertionInstanceVariableNode(e, expression, name);
        }

        [System.Diagnostics.DebuggerHidden]
        public static DynAssertionExpressionNode And(this DynAssertionExpressionNode first, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(e, DynAssertionExpressionNode.NodeKind.And, first, second);
        }

        [System.Diagnostics.DebuggerHidden]
        public static DynProgramExpressionNode Or(this DynProgramExpressionNode first, DynProgramExpressionNode second)
        {
            return new DynProgramMethodCallNode(e, first, OperatorTranslation.KindToMethod(SyntaxKind.OrOperator), Dyn.Utilities.AsArray(second));
        }

        [System.Diagnostics.DebuggerHidden]
        public static DynAssertionExpressionNode Or(this DynAssertionExpressionNode first, DynAssertionExpressionNode second)
        {
            return new DynAssertionBinaryNode(e, DynAssertionExpressionNode.NodeKind.Or, first, second);
        }

        public static DynProgramExpressionNode Add(this DynProgramExpressionNode first, DynProgramExpressionNode second)
        {
            return new DynProgramMethodCallNode(e, first, OperatorTranslation.KindToMethod(SyntaxKind.AdditionOperator), Dyn.Utilities.AsArray(second));
        }

        public static DynProgramExpressionNode Sub(this DynProgramExpressionNode first, DynProgramExpressionNode second)
        {
            return new DynProgramMethodCallNode(e, first, OperatorTranslation.KindToMethod(SyntaxKind.SubtractionOperator), Dyn.Utilities.AsArray(second));
        }

        public static DynProgramExpressionNode Eq(this DynProgramExpressionNode first, DynProgramExpressionNode second)
        {
            return new DynProgramMethodCallNode(e, first, OperatorTranslation.KindToMethod(SyntaxKind.EqualityOperator), Dyn.Utilities.AsArray(second));
        }

        public static DynProgramExpressionNode Assign(this DynProgramVariableNode variable, DynProgramExpressionNode expr)
        {
            return new DynProgramAssignmentNode(e, variable, expr);
        }

        private static IEqualityComparer<string> CMP = StringComparer.Ordinal;

        [Fact]
        public static void InstanceVariableNegation()
        {
            var r = new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.ProgramVariable, "r");
            var a = r.Neg().InstanceVariable("r");
            var b = r.InstanceVariable("r").Neg();

            Assert.NotEqual(a, b, AstExtensions.IgnoreLocation);

            Assert.Equal("!r.@r", a.ToSource());
            Assert.Equal("!(r.@r)", b.ToSource());

            Assert.Equal("!(!r.@r)", "!((!r).@r)".Tokenize().Parse(true, p => p.ParseAssertionExpression()).ToSource());
            Assert.Equal("!(!r.@r)", "!(!r.@r)".Tokenize().Parse(true, p => p.ParseAssertionExpression()).ToSource());
            Assert.Equal("!(!r.@r)", "!((!r).@r)".Tokenize().Parse(true, p => p.ParseAssertionExpression()).ToSource());
            Assert.Equal("!!(r.@r)", "!(!(r.@r))".Tokenize().Parse(true, p => p.ParseAssertionExpression()).ToSource());

            Assert.Equal("!!(r.@r) == 1", "!(!(r.@r)) == 1".Tokenize().Parse(true, p => p.ParseAssertionExpression()).ToSource());
            Assert.Equal("!15.@C2M", "!15.@C2M".Tokenize().Parse(true, p => p.ParseAssertionExpression()).ToSource());
        }

        [Fact]
        public static void Difficult()
        {
            Func<string, string, KeyValuePair<string, string>> kvp = (k, v) => new KeyValuePair<string, string>(k, v);
            Assert.All(Dyn.Utilities.AsArray(
                kvp("a && b && c && d", a.And(b).And(c).And(d).ToSource()),
                kvp("a && (b && (c && d))", a.And(b.And(c.And(d))).ToSource()),
                kvp("'r && 'r || ('r && 'r)", (r.And(r)).Or(r.And(r)).ToSource()),
                kvp("'r && ('r && 'r)", r.And(r.And(r)).ToSource()),
                kvp("a && b || (c && d)", a.And(b).Or(c.And(d)).ToSource()),
                kvp("a && (b || (c && d))", a.And(b.Or(c.And(d))).ToSource())),
                e => Assert.Equal(e.Key, e.Value));
        }

        [Fact]
        public static void Simple()
        {
            var r = new DynAssertionVariableNode(Span.Empty, DynAssertionExpressionNode.NodeKind.ProgramVariable, "r");
            Func<string, string, KeyValuePair<string, string>> kvp = (k, v) => new KeyValuePair<string, string>(k, v);
            Assert.All(Dyn.Utilities.AsArray(
                kvp("r && r && r", r.And(r).And(r).ToSource()),
                kvp("r == 1 + 2 * 3", r.Equality(e, n1.Addition(e, n2.Multiplication(e, n3))).ToSource()),
                kvp("r == (1 + 2) * 3", r.Equality(e, n1.Addition(e, n2).Multiplication(e, n3)).ToSource()),
                kvp("r == 1 * (2 + 3)", r.Equality(e, n1.Multiplication(e, n2.Addition(e, n3))).ToSource()),
                kvp("a = b = c = d", a.Assign(b.Assign(c.Assign(d))).ToSource()),
                kvp("a && (b && (c && d))", a.And(b.And(c.And(d))).ToSource()),
                kvp("a && b && c && d", a.And(b).And(c).And(d).ToSource()),
                kvp("a && (b && (c && d))", a.And(b.And(c.And(d))).ToSource()),
                kvp("a + (b + (c + d))", a.Add(b.Add(c.Add(d))).ToSource()),
                kvp("a + b + c + d", a.Add(b).Add(c).Add(d).ToSource()),
                kvp("a + b == c + d", (a.Add(b)).Eq(c.Add(d)).ToSource()),
                kvp("a + (b - c)", a.Add(b.Sub(c)).ToSource()),
                kvp("a + b - c", a.Add(b).Sub(c).ToSource())),
                f => Assert.Equal(f.Key, f.Value));
        }

        [Fact]
        public static void Negation()
        {
            var x = "(!a).@b".Tokenize().Parse(true, p => p.ParseAssertionExpression());
            var y = "!(a.@b)".Tokenize().Parse(true, p => p.ParseAssertionExpression());
            Assert.NotEqual(x, y, AstExtensions.IgnoreLocation);
        }

        [Theory]
        // structural
        [InlineData("a = b = c = d")]
        [InlineData("a + b + c + d")]
        [InlineData("while a invariant true do b done")]
        [InlineData("while while a invariant true do b done invariant true do while c invariant true do d done done")]
        [InlineData("while a() invariant true do b(); c(@d) done")]
        [InlineData("while a invariant true do @b done.c()")]
        [InlineData("a")]
        [InlineData("@a")]
        [InlineData("@a.b().c().d(e(), f().g())")]
        [InlineData("if if a then null; null end then null else null; null end")]
        [InlineData("if if a then b end then if a then b end else if a then b end end")]
        [InlineData("if if a then b else c end then if a then b else c end else if a then b else c end end")]
        [InlineData("new a().b()")]
        [InlineData("new a(b()).c().d()")]

        // sugar
        [InlineData("while i < 5 invariant true do 0 done")]
        [InlineData("r == 1 + 2 * 3")]
        [InlineData("r == (1 + 2) * 3")]
        [InlineData("r == 1 * (2 + 3)")]
        [InlineData("a + b == c + d")]
        [InlineData("\"Hello World\"")]
        [InlineData("\"\"")]
        [InlineData("0")]
        [InlineData("1")]
        [InlineData("10")]
        [InlineData("100")]
        [InlineData("true")]
        [InlineData("false")]

        // mixed
        [InlineData("if if 1 then 2 end then 3 else 4 end")]
        [InlineData("1 + while a invariant true do b done")]
        [InlineData("while a invariant true do b done + 1")]
        [InlineData("a(while b invariant true do c done)")]
        [InlineData("while 3 * 4 invariant true do c done")]
        [InlineData("1 + a()")]
        [InlineData("2 * b()")]
        [InlineData("3 && c()")]
        [InlineData("4 || d()")]
        [InlineData("5 || d()")]
        [InlineData("1 + @a")]
        [InlineData("1 + @a.b()")]
        [InlineData("1 + (b = 3)")]
        [InlineData("c = 1 + (b = 3)")]
        [InlineData("assert !(a.@b)")]
        [InlineData("assert a || b && (c || d)")]
        [InlineData("assert a && b || (c && d)")]
        public static void Complex(string expected)
        {
            var actualNode = expected.Tokenize().Parse(true, p => p.ParseProgramStatement());
            AssertExtensions.SyntaxEqual(expected, actualNode.ToSource());
            Assert.Equal(expected.NormalizeWhitespace(), actualNode.ToSource().NormalizeWhitespace());
        }

        [Theory]
        [InlineData("a")]
        [InlineData("a = a")]
        [InlineData("a = 1 && 1")]
        [InlineData("a = 1 || 1")]
        [InlineData("a = 1 is_a? 1")]
        [InlineData("a = 1 + 1")]
        [InlineData("a = 1 - 1")]
        [InlineData("a = 1 * 1")]
        [InlineData("a = 1 / 1")]
        [InlineData("a = 1 % 1")]
        [InlineData("a = 1 > 1")]
        [InlineData("a = 1 >= 1")]
        [InlineData("a = 1 < 1")]
        [InlineData("a = 1 <= 1")]
        [InlineData("a = 1 == 1")]
        [InlineData("a = 1 === 1")]
        public static void OperatorPrecedence(string expected)
        {
            var actualNode = expected.Tokenize().Parse(true, p => p.ParseProgramExpression());
            Assert.Equal(expected, actualNode.ToSource(), CMP);
            AssertExtensions.SyntaxEqual(expected, actualNode.ToSource());
        }

        private static string Replace(this string source, string replacement, int start, int length)
        {
            var builder = new System.Text.StringBuilder(source);
            builder.Remove(start, length);
            builder.Insert(start, replacement);
            return builder.ToString();
        }

        private const int FIRST_HOLE = 0;
        private const int LAST_HOLE = 1;

        [Fact]
        public static void TestCalculateParenthesisRequired()
        {
            const int
                l0 = LEFT_ASSOC | 0,
                l1 = LEFT_ASSOC | 1,
                r0 = RIGHT_ASSOC | 0,
                r1 = RIGHT_ASSOC | 1;

            // primary > secondary => boring
            TestCalculateParenthesisRequiredCore(l0, 0, l1, false);
            TestCalculateParenthesisRequiredCore(l0, 0, r1, false);
            TestCalculateParenthesisRequiredCore(r0, 0, l1, false);
            TestCalculateParenthesisRequiredCore(r0, 0, r1, false);
            TestCalculateParenthesisRequiredCore(l0, 1, l1, false);
            TestCalculateParenthesisRequiredCore(l0, 1, r1, false);
            TestCalculateParenthesisRequiredCore(r0, 1, l1, false);
            TestCalculateParenthesisRequiredCore(r0, 1, r1, false);

            // primary < secondary => parenthesis
            TestCalculateParenthesisRequiredCore(l1, 0, l0, true);
            TestCalculateParenthesisRequiredCore(l1, 0, r0, true);
            TestCalculateParenthesisRequiredCore(r1, 0, l0, true);
            TestCalculateParenthesisRequiredCore(r1, 0, r0, true);
            TestCalculateParenthesisRequiredCore(l1, 1, l0, true);
            TestCalculateParenthesisRequiredCore(l1, 1, r0, true);
            TestCalculateParenthesisRequiredCore(r1, 1, l0, true);
            TestCalculateParenthesisRequiredCore(r1, 1, r0, true);

            // now the interesting stuff
            // primary == secondary

            // (a && a) && a => false
            // a && (a && a) => true
            // (a == a) == a => true
            // a == (a == a) => false
            TestCalculateParenthesisRequiredCore(l0, 0, l0, false);
            TestCalculateParenthesisRequiredCore(l0, 1, l0, true);
            TestCalculateParenthesisRequiredCore(r0, 0, r0, true);
            TestCalculateParenthesisRequiredCore(r0, 1, r0, false);
        }

        private static void TestCalculateParenthesisRequiredCore(int primary, int index, int secondary, bool expectedRequiresParenthesis)
        {
            Assert.Equal(expectedRequiresParenthesis, CalculateParenthesisRequired(primary, index, secondary));
        }

        private static bool CalculateParenthesisRequired(int primary, int index, int secondary)
        {
            if ((primary & 0xFF) < (secondary & 0xFF))
            {
                return false;
            }
            else if ((primary & 0xFF) > (secondary & 0xFF))
            {
                return true;
            }

            var assoc = primary & ~0xFF; // assert assoc == secondary & ~0xFF

            switch (assoc)
            {
                case LEFT_ASSOC:
                    return index == LAST_HOLE;

                case RIGHT_ASSOC:
                    return index == FIRST_HOLE;

                default:
                    throw new Exception();
            }
        }

        private const int LEFT_ASSOC = 0x200;
        private const int RIGHT_ASSOC = 0x400;

        public static IEnumerable<object[]> DoubleSource(string data, string holeVariable)
        {
            Regex regex = new Regex(Regex.Escape(holeVariable));
            var expressions = new Dictionary<string, int>();

            int index = 0;
            foreach (var line in data.Split("\r\n".ToCharArray(), StringSplitOptions.RemoveEmptyEntries))
            {
                int associativity;
                switch (line[0])
                {
                    case 'L': associativity = LEFT_ASSOC; break;
                    case 'R': associativity = RIGHT_ASSOC; break;
                    default: throw new Exception();
                }

                foreach (var expr in line.Substring(2).Split(",".ToCharArray(), StringSplitOptions.RemoveEmptyEntries))
                {
                    expressions.Add(expr.Trim(), index + associativity);
                }
                index++;
            }

            foreach (var primary in expressions)
                foreach (var secondary in expressions)
                {
                    // Regex needed to access each hole separately
                    foreach (var match in regex.Matches(primary.Key).Cast<Match>())
                    {
                        int hole;

                        if (match.Index == 0)
                        {
                            hole = FIRST_HOLE;
                        }
                        else if (match.Index + match.Length == primary.Key.Length)
                        {
                            hole = LAST_HOLE;
                        }
                        else
                        {
                            hole = 2;
                        }
                        var replacement = secondary.Key;
                        var exprWithout = primary.Key.Replace(replacement, match.Index, match.Length);
                        var exprWith = primary.Key.Replace("(" + replacement + ")", match.Index, match.Length);

                        // expect error because both sub expressions have the same precedence and are not-associative
                        yield return new object[] { exprWithout, exprWith, CalculateParenthesisRequired(primary.Value, hole, secondary.Value) };
                    }
                }
        }

        // N: if y then y end, while y do y done

        [Theory]
        [MemberData("DoubleSource", @"
R: y = x
L: x && x, x || b
R: x == x, x === x
L: x < x, x <= x, x > x, x >= x
L: x is_a? y
L: x + x, x - x
L: x * x, x / x, x % x
L: x.y()", "x")] // euqls are right acco
        public static void DoubleProgram(string without, string with, bool requiresParenthesis)
        {
            TestDouble(without, with, p => p.ParseProgramExpression(), DynNodeFormatter.ToSource, requiresParenthesis);
        }

        // N: if y then y else y end
        [Theory]
        [MemberData("DoubleSource", @"
R: exists 'y : y. blank, forall 'y : y. blank
L: blank && blank, blank || blank, blank implies blank
R: blank == blank
L: blank < blank, blank <= blank, blank > blank, blank >= blank
L: blank + blank, blank - blank
L: blank * blank, blank / blank, blank % blank
L: blank.@y
L: blank in {y}
R: !blank", "blank")]// N: y(y, y)
        public static void DoubleAssertion(string without, string with, bool requiresParenthesis)
        {
            TestDouble(without, with, p => p.ParseAssertionExpression(), DynNodeFormatter.ToSource, requiresParenthesis);
        }

        private static T TestDoubleCore<T>(string expected, string source, Func<Dyn.Parser.DynParser, T> parser, Func<T, string> formatter)
        {
            T sourceNode = source.Tokenize().Parse(true, parser);
            string sourceNodeSource = formatter(sourceNode);
            Assert.Equal(expected.NormalizeWhitespace(), sourceNodeSource.NormalizeWhitespace());
            AssertExtensions.SyntaxEqual(expected, sourceNodeSource);
            return sourceNode;
        }

        private static void TestDouble<T>(string without, string with, Func<Dyn.Parser.DynParser, T> parser, Func<T, string> formatter, bool requiresParenthesis)
        {
            T a = default(T), b = default(T);
            Exception ex1 = null, ex2 = null;

            try
            {
                a = without.Tokenize().Parse(true, parser);
            }
            catch (Exception e)
            {
                ex1 = e;
            }

            try
            {
                b = with.Tokenize().Parse(true, parser);
            }
            catch (Exception e)
            {
                ex2 = e;
            }

            if (requiresParenthesis)
            {
                Assert.Equal(with, formatter(b));

                // Assert.NotEqual(formatter(a), formatter(b));
                // Assert.NotEqual(a, b, AstExtensions.IgnoreLocation.As<IEqualityComparer<T>>());
            }
            else
            {
                Assert.Equal(without, formatter(a));
                Assert.Equal(without, formatter(b));

                Assert.Null(ex1);
                Assert.Null(ex2);
                Assert.Equal(formatter(a), formatter(b));

                var ieq = AstExtensions.IgnoreLocation as IEqualityComparer<T>;
                Assert.NotNull(ieq);
                Assert.Equal(a, b, ieq);
            }
        }
    }
}
