﻿using Dyn;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;

namespace tests
{
    public static class AssertExtensions
    {
        public static void SetEqual<T>(IEnumerable<T> expected, IEnumerable<T> actual, Func<T, string> formatter)
        {
            Guard.EnsureNotNull("expected", expected);
            Guard.EnsureNotNull("actual", actual);
            Guard.EnsureNotNull("formatter", formatter);

            var expectedSet = new HashSet<T>(expected);
            var actualSet = new HashSet<T>(actual);

            var builder = new StringBuilder();

            if (!expectedSet.SetEquals(actualSet))
            {
                if (!expectedSet.IsSubsetOf(actualSet))
                {
                    builder.Append("Elements in 'expected' but not in 'actual':\n\t");
                    builder.AppendLine(string.Join(", ", expectedSet.Except(actualSet).Select(formatter)));
                }
                if (!actualSet.IsSubsetOf(expectedSet))
                {
                    builder.Append("Elements in 'actual' but not in 'expected':\n\t");
                    builder.AppendLine(string.Join(", ", actualSet.Except(expectedSet).Select(formatter)));
                }

                Assert.True(false, builder.ToString());
            }

            Assert.True(true);
        }

        public static void ListEqual<T>(IList<T> expected, IList<T> actual, IEqualityComparer<T> comparer, Func<T, string> formatter)
        {
            Guard.EnsureNotNull("expected", expected);
            Guard.EnsureNotNull("actual", actual);
            Guard.EnsureNotNull("formatter", formatter);

            if (expected.SequenceEqual(actual, comparer))
            {
                return;
            }

            var builder = new StringBuilder();

            int e = 0, a = 0;

            foreach (var diff in LineDiff.Diff(expected, actual, comparer))
            {
                switch (diff.Type)
                {
                    case LineDiff.DiffSectionType.Copy:
                        {
                            builder.Append("= ");
                            for (int i = 0; i < diff.Length; ++i)
                            {
                                builder.Append(' ').Append(formatter(expected[e]));
                                e++; a++;
                            }
                            builder.AppendLine();
                        }
                        break;

                    case LineDiff.DiffSectionType.Delete:
                        {
                            builder.Append("- ");
                            for (int i = 0; i < diff.Length; ++i)
                            {
                                builder.Append(' ').Append(formatter(expected[e]));
                                e++;
                            }
                            builder.AppendLine();
                        }
                        break;

                    case LineDiff.DiffSectionType.Insert:
                        {
                            builder.Append("+ ");
                            for (int i = 0; i < diff.Length; ++i)
                            {
                                builder.Append(' ').Append(formatter(actual[a]));
                                a++;
                            }
                            builder.AppendLine();
                        }
                        break;
                }
            }

            Assert.True(false, builder.ToString());
        }

        public static void SyntaxEqual(string expected, string actual)
        {
            TokensEqual(expected.Tokenize(), actual.Tokenize());
        }

        private sealed class SyntaxTokenComparer : IEqualityComparer<SyntaxToken>
        {
            public static readonly IEqualityComparer<SyntaxToken> Comparer = new SyntaxTokenComparer();

            private SyntaxTokenComparer()
            { }

            public bool Equals(SyntaxToken x, SyntaxToken y)
            {
                return x.EqualsIgnoreSpan(y);
            }

            public int GetHashCode(SyntaxToken obj)
            {
                return obj.GetHashCode();
            }
        }

        public static void TokensEqual(IEnumerable<SyntaxToken> expected, IEnumerable<SyntaxToken> actual)
        {
            ListEqual(
                expected.Simplify().ToList(),
                actual.Simplify().ToList(),
                SyntaxTokenComparer.Comparer,
                e => e.ToSource());
        }

        public static void AssertEqualIgnoreOrder<T>(IEnumerable<T> expected, IEnumerable<T> actual,
    IEqualityComparer<T> comparer)
        {
            var set = new HashSet<T>(expected);

            foreach (var element in actual)
            {
                // more elements in actual than in set
                if (set.Count == 0)
                {
                    throw new Xunit.Sdk.EqualException(expected, actual);
                }

                // element in actual but not in excepted
                if (!set.Remove(element))
                {
                    throw new Xunit.Sdk.EqualException(expected, actual);
                }
            }

            // more elements in set than in actual
            if (set.Count != 0)
            {
                throw new Xunit.Sdk.EqualException(expected, actual);
            }
        }
    }
}
