﻿using Dyn;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;

namespace tests
{
    public static class LineDiff
    {
        [System.Diagnostics.DebuggerStepThrough]
        [DebuggerDisplay("{Offset}:{Count}")]
        public struct ListSegment<T>
        {
            private IList<T> _list;

            private int _offset;

            private int _count;

            public IList<T> Array
            {
                get
                {
                    return this._list;
                }
            }

            public int Offset
            {
                get
                {
                    return this._offset;
                }
            }

            public int Count
            {
                get
                {
                    return this._count;
                }
            }

            public ListSegment(IList<T> _list)
            {
                if (_list == null)
                {
                    throw new ArgumentNullException("array");
                }
                this._list = _list;
                this._offset = 0;
                this._count = _list.Count;
            }

            public ListSegment(IList<T> list, int offset, int count)
            {
                if (list == null)
                {
                    throw new ArgumentNullException("array");
                }
                if (offset < 0)
                {
                    throw new ArgumentOutOfRangeException("offset");
                }
                if (count < 0)
                {
                    throw new ArgumentOutOfRangeException("count");
                }
                if (list.Count - offset < count)
                {
                    throw new ArgumentException();
                }
                this._list = list;
                this._offset = offset;
                this._count = count;
            }

            public override int GetHashCode()
            {
                unchecked
                {
                    int slr = 5, sll = 32 - 5;

                    int result = this._list.GetHashCode();

                    result = ((result << slr) | (result >> sll)) ^ _offset;
                    result = ((result << slr) | (result >> sll)) ^ _count;

                    return result;
                }
            }

            public override bool Equals(object obj)
            {
                return obj is ListSegment<T> && this.Equals((ListSegment<T>)obj);
            }

            public bool Equals(ListSegment<T> obj)
            {
                return obj._list == this._list && obj._offset == this._offset && obj._count == this._count;
            }

            public static bool operator ==(ListSegment<T> a, ListSegment<T> b)
            {
                return a.Equals(b);
            }

            public static bool operator !=(ListSegment<T> a, ListSegment<T> b)
            {
                return !a.Equals(b);
            }
        }

        public enum DiffSectionType
        {
            Copy,
            Insert,
            Delete
        }

        [System.Diagnostics.DebuggerStepThrough]
        public struct DiffSection
        {
            private readonly DiffSectionType _type;
            private readonly int _length;

            public DiffSection(DiffSectionType type, int length)
            {
                _type = type;
                _length = length;
            }

            public DiffSectionType Type { get { return _type; } }

            public int Length { get { return _length; } }

            public override string ToString()
            {
                return string.Format("{0} {1}", Type, Length);
            }
        }

        private static IEnumerable<string> Lines(TextReader source)
        {
            string line;

            while ((line = source.ReadLine()) != null)
            {
                yield return line;
            }
        }

        public static void PrintLineDiff(Stream a, Stream b)
        {
            using (StreamReader ar = new StreamReader(a), br = new StreamReader(a))
            {
                long apos = a.Position, bpos = b.Position;

                var ah = Lines(ar).Select(e => e.GetHashCode()).ToArray();
                var bh = Lines(br).Select(e => e.GetHashCode()).ToArray();

                var diffs = Diff(ah, bh, EqualityComparer<int>.Default);

                ar.DiscardBufferedData();
                br.DiscardBufferedData();
                a.Position = apos;
                b.Position = bpos;
            }
        }

        public static IEnumerable<DiffSection> Diff<T>(IList<T> a, IList<T> b, IEqualityComparer<T> comparer)
        {
            Guard.EnsureNotNull("a", a);
            Guard.EnsureNotNull("b", b);
            Guard.EnsureNotNull("comparer", comparer);

            return Diff(new ListSegment<T>(a), new ListSegment<T>(b), comparer);
        }

        private static IEnumerable<DiffSection> Diff<T>(ListSegment<T> a, ListSegment<T> b, IEqualityComparer<T> comparer)
        {
            ListSegment<T> a2, b2;
            int count;

            if (!TryFindLongestCommonSubSequence(a, b, comparer, out a2, out b2, out count))
            {
                if (a2.Count > 0)
                {
                    yield return new DiffSection(DiffSectionType.Delete, a.Count);
                }
                if (b2.Count > 0)
                {
                    yield return new DiffSection(DiffSectionType.Insert, b.Count);
                }
                yield break;
            }

            if (count == a.Count && count == b.Count)
            {
                yield return new DiffSection(DiffSectionType.Copy, count);
                yield break;
            }

            // deal with the section before
            var sectionsBefore = Diff(
                new ListSegment<T>(a.Array, a.Offset, a2.Offset - a.Offset),
                new ListSegment<T>(b.Array, b.Offset, b2.Offset - b.Offset), comparer);

            foreach (var section in sectionsBefore)
                yield return section;

            // output the copy operation
            yield return new DiffSection(
                DiffSectionType.Copy,
                count);

            // deal with the section after
            var sectionsAfter = Diff(
                new ListSegment<T>(a.Array, a2.Offset + count, a.Offset + a.Count - a2.Offset - a2.Count),
                new ListSegment<T>(b.Array, b2.Offset + count, b.Offset + b.Count - b2.Offset - b2.Count), comparer);

            foreach (var section in sectionsAfter)
                yield return section;

            yield break;
        }

        internal static bool TryFindLongestCommonSubSequence<T>(
            ListSegment<T> a, ListSegment<T> b, IEqualityComparer<T> comparer, out ListSegment<T> a2, out ListSegment<T> b2, out int count)
        {
            a2 = a;
            b2 = b;
            count = 0;

            for (int i = a.Offset, ic = a.Count; ic > 0; --ic, ++i)
            {
                for (int j = b.Offset, jc = b.Count; jc > 0 && ic > 0; --jc, ++j)
                {
                    if (comparer.Equals(a.Array[i], b.Array[j]))
                    {
                        int currentCount = CountEqual(
                            new ListSegment<T>(a.Array, i, ic),
                            new ListSegment<T>(b.Array, j, jc),
                            comparer);

                        // Is longer than what we already have --> record new LCS
                        if (currentCount > count)
                        {
                            a2 = new ListSegment<T>(a.Array, i, currentCount);
                            b2 = new ListSegment<T>(b.Array, j, currentCount);
                            count = currentCount;
                        }

                        --ic;
                        ++i; // move i forward too
                    }
                }
            }

            return count > 0;
        }

        internal static int CountEqual<T>(ListSegment<T> a, ListSegment<T> b, IEqualityComparer<T> qualityComparer)
        {
            int result = 0;

            int i = a.Offset, ic = a.Count, j = b.Offset, jc = b.Count;

            while (ic > 0 && jc > 0 && qualityComparer.Equals(a.Array[i], b.Array[j]))
            {
                --ic; ++i; --jc; ++j;
                result++;
            }

            return result;
        }
    }
}
