﻿using Dyn;
using Dyn.Diagnostics;
using Dyn.Lexer;
using Dyn.Parser;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using Xunit;

namespace tests
{
    public static class Magic
    {
        private static IList<SyntaxToken> TokenizeCore(DynLexer lexer)
        {
            var result = lexer.ToList();

            Assert.NotEmpty(result);
            Assert.All(result, e => Assert.True(e.Kind != SyntaxKind.InvalidKind));

            return result;
        }

        public static IEnumerable<SyntaxToken> Tokenize(this string source)
        {
            Assert.NotNull(source);
            return TokenizeCore(new DynLexer(source));
        }

        public static IEnumerable<SyntaxToken> Tokenize(this System.IO.TextReader source)
        {
            Assert.NotNull(source);
            return TokenizeCore(new DynLexer(source));
        }

        public static IEnumerable<SyntaxToken> Tokenize(this Sample source)
        {
            using (var reader = source.ToReader())
            {
                return TokenizeCore(new DynLexer(reader));
            }
        }

        public static IEnumerable<SyntaxToken> Tokenize(this Sample source, string file)
        {
            using (var reader = source.ToReader())
            {
                return TokenizeCore(new DynLexer(file, reader));
            }
        }

        public static T Parse<T>(this IEnumerable<SyntaxToken> source, bool expectEndOfStream, Func<DynParser, T> func)
        {
            Assert.NotNull(source);
            Assert.NotNull(func);
            var parser = new DynParser(source.GetEnumerator());
            parser.DiagnosticInfo += ThrowOnDiagnostic;
            var result = func(parser);
            Assert.NotNull(result);
            Assert.True(expectEndOfStream || parser.IsEndOfStream);
            return result;
        }

        public static IEnumerable<T> ParseAll<T>(this IEnumerable<SyntaxToken> source, bool expectEndOfStream, Func<DynParser, T> func)
        {
            Assert.NotNull(source);
            Assert.NotNull(func);
            var parser = new DynParser(source.GetEnumerator());
            parser.DiagnosticInfo += ThrowOnDiagnostic;

            T result;

            while ((result = func(parser)) != null)
            {
                Assert.NotNull(result);
                yield return result;
            }

            Assert.True(expectEndOfStream || parser.IsEndOfStream);
        }

        public static IEnumerable<T> ParseAll<T>(this DynParser parser, Func<DynParser, T> func)
        {
            Assert.NotNull(parser);
            Assert.NotNull(func);

            T result;

            while ((result = func(parser)) != null)
            {
                Assert.NotNull(result);
                yield return result;
            }
        }

        public static string FormatAll<T>(this IEnumerable<T> source, Func<T, string> func)
        {
            Assert.NotNull(source);
            Assert.NotNull(func);
            return source.Aggregate(new StringBuilder(), (builder, node) => builder.Append(func(node))).ToString();
        }

        public static IEnumerable<SyntaxToken> Simplify(this IEnumerable<SyntaxToken> source)
        {
            return source.Where(e => !SkipKind(e.Kind));
        }

        public static bool SkipKind(SyntaxKind kind)
        {
            switch (kind)
            {
                case SyntaxKind.WhitespaceKind:
                case SyntaxKind.CommentKind:
                    return true;

                default:
                    return false;
            }
        }

        public sealed class MagicException : Exception
        {
            public MagicException()
            {
            }

            public MagicException(string message) : base(message)
            {
            }

            public MagicException(string message, Exception inner) : base(message, inner)
            {
            }
        }

        public static void ThrowOnDiagnostic(object sender, DiagnosticEventArgs e)
        {
            throw new MagicException(e.Info.ToString());
        }

        public static void ThrowOnError(object sender, DiagnosticEventArgs e)
        {
            if (e.Info.Level == DiagnosticSeverity.Error)
            {
                throw new MagicException(e.Info.ToString());
            }
        }

        private static Regex SHORT_WHITESPACE = new Regex(@"\s+", RegexOptions.Multiline);

        public static string NormalizeWhitespace(this string source)
        {
            return SHORT_WHITESPACE.Replace(source, " ");
        }
    }
}
